// (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
// Heinrich Heine Universitaet Duesseldorf
// This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

#include <map>
#include <string>

#include <sicstus/sicstus.h>
#include "counter_glue.h"

using namespace std;

map<SP_atom, SP_integer> counters;
SP_term_ref exception_term;

void create_exception_message(SP_atom name) {
  exception_term =  SP_new_term_ref();
  string message = "The counter <<";
  message.append(SP_string_from_atom(name));
  message.append(">> does not exist");
  SP_put_atom(exception_term, SP_atom_from_string(message.c_str()));
}

extern "C" {
  void new_counter(SP_atom name) {
    // currently, declaring a counter twice does not lead to failure!
    counters[name] = 0;
  }

  SP_integer get_counter(SP_atom name) {
    // query can not work if the counter does not exist
    // sicstus sees an exception in this case
    if(counters.count(name) == 1) {
      return counters[name];
    } else {
      create_exception_message(name);
      SP_raise_exception(exception_term);
      return SP_ERROR;
    }
  }

  SP_integer inc_counter(SP_atom name) {
    // query can not work if the counter does not exist
    // sicstus sees an exception in this case
    if(counters.count(name) == 1) {
      counters[name] = counters[name] + 1;
      return counters[name];
    } else {
      create_exception_message(name);
      SP_raise_exception(exception_term);
      return SP_ERROR;
    }
  }

  void reset_counter(SP_atom name) {
    //counters[name] = 0; return; // this was the old behaviour; comment in if you want an extension compatible with older probclis
    // resetting a counter that has not been defined yet
    // might be a sign of an error
    // thus, send failure back to sicstus
    if(counters.count(name) == 1) {
      counters[name] = 0;
    } else {
      create_exception_message(name);
      SP_raise_exception(exception_term);
    }
  }

  void set_counter(SP_atom name, SP_integer val) {
    // setting a counter that has not been defined yet
    // might be a sing of an error
    // thus, send failure back to sicstus
    if(counters.count(name) == 1) {
      counters[name] = val;
    } else {
      create_exception_message(name);
      SP_raise_exception(exception_term);
    }
  }
}
