
#include "bitoperations.h"

static unsigned int bits_per_bitfield;
static unsigned int *usage_bitmasks = 0;

void bitoperations_init() {
  bitfield x = 0;
  x = ~x;
  bits_per_bitfield=0;
  while (x) {
    bits_per_bitfield++;
    x >>= 1;
  }
  usage_bitmasks = (unsigned int*) malloc(sizeof(unsigned int) * bits_per_bitfield);
  unsigned int mask = 0;
  unsigned int i;
  for (i=0; i<bits_per_bitfield; i++) {
    usage_bitmasks[i] = mask;
    mask <<= 1;
    mask |= 1;
  }
}

void bitoperations_delete() {
    if(usage_bitmasks) free(usage_bitmasks);
}

static int get_bitfield_size(unsigned int size) {
  return ((size+bits_per_bitfield-1) / bits_per_bitfield);
}

size_t get_bitfield_memsize(unsigned int size) {
  return sizeof(bitfield) * get_bitfield_size(size);
}

void setbit(bitfield *b, int pos, int value) {
  bitfield mask = 1 << (pos % bits_per_bitfield);
  b = b + (pos/bits_per_bitfield);
  if (value) {
    *b |= mask;
  } else {
    *b &= ~mask;
  }
}

void setbit_by_long(bitfield *b, int pos, unsigned long value, int size) {
  int i,rest;
  bitfield mask_bitfield = 1 << (pos % bits_per_bitfield);
  b = b + (pos/bits_per_bitfield);
  rest = bits_per_bitfield - (pos % bits_per_bitfield) - 1;
  bitfield mask_long = 1;
  for (i=0; i<size; i++) {
    if (value & mask_long) {
      *b |= mask_bitfield;
    } else {
      *b &= ~mask_bitfield;
    }
    mask_long <<= 1;
    if (rest > 0) {
      rest--;
      mask_bitfield <<= 1;
    } else {
      rest = bits_per_bitfield - 1;
      b++;
      mask_bitfield = 1;
    }
  }
}

int getbit(const bitfield *b, int pos) {
  b = b + (pos/bits_per_bitfield);
  bitfield mask = 1 << (pos % bits_per_bitfield);
  return (*b & mask) ? ~0 : 0;
}

void copy_bitfield(const bitfield *src,
                   bitfield *dst,
                   int offset,
                   int size) {
  int i,rest;
  src = src + (offset/bits_per_bitfield);
  dst = dst + (offset/bits_per_bitfield);
  rest = bits_per_bitfield - (offset % bits_per_bitfield) -1;
  bitfield mask = 1 << (offset % bits_per_bitfield);
  for (i=0; i<size; i++) {
    *dst = (*dst & ~mask) | (*src & mask);
    if (rest==0) {
      rest = bits_per_bitfield - 1;
      src++;
      dst++;
      mask = 1;
    } else {
      rest--;
      mask <<= 1;
    }
  }
}

void insert_bitfield(const bitfield *src,
                     bitfield *dst,
                     int offset,
                     int size) {
  int i;
  for (i=0; i<size; i++)
    setbit(dst, offset+i, getbit(src, i));
}

void add_set_bits(const bitfield *src, bitfield *dst, unsigned int size) {
  int i,max;
  max = get_bitfield_size(size);
  for (i=0; i<max; i++) {
    dst[i] |= src[i];
  }
}

void clear_bitfield(bitfield *b, unsigned int size) {
  int i,max;
  max = get_bitfield_size(size);
  for (i=0; i<max; i++) {
    b[i] = 0;
  }
}

void fshow_bitfield(FILE *stream, const bitfield *b, unsigned int offset, unsigned int size) {
  int i;
  for (i=0; i<size; i++) {
    fputc(getbit(b,offset + i) ? '1' : '0', stream);
  }
}

void show_bitfield(const bitfield *b, unsigned int size) {
  fshow_bitfield(stdout, b, 0, size);
}

void clone_bitfield(const bitfield *src, bitfield *dst, unsigned int size) {
  int i,nrfields = get_bitfield_size(size);
  for (i=0; i<nrfields; i++) dst[i] = src[i];
}

int equal_bitfields(const bitfield *a, const bitfield *b, unsigned int size) {
  int i,full = (size/bits_per_bitfield);
  for (i=0; i<full; i++) {
    if (a[i] != b[i]) return 0;
  }
  int rest = size % bits_per_bitfield;
  if (rest != 0
      && ((a[full] & usage_bitmasks[rest]) 
          != (b[full] & usage_bitmasks[rest]))) {
    return 0;
  }
  return 1;
}

int is_sub_bitfield(const bitfield *sub, const bitfield *super, unsigned int size) {
  int i,full = (size/bits_per_bitfield);
  for (i=0; i<full; i++) {
    if (sub[i] & ~super[i]) {
      return 0;
    }
  }
  int rest = size % bits_per_bitfield;
  if (rest != 0
      && ( (sub[full] & ~super[full]) & usage_bitmasks[rest]) ) {
    return 0;
  }
  return 1;  
}

int is_null_bitfield(const bitfield *b, unsigned int size) {
  int i,full = (size/bits_per_bitfield);
  for (i=0; i<full; i++) {
    if (b[i] != 0) return 0;
  }
  int rest = size % bits_per_bitfield;
  if (rest!=0
      && ((b[full] & usage_bitmasks[rest]) != 0)) {
    return 0;
  }
  return 1;
}
