//
//  loop_fairness.c
//  
//
//  Created by Ivaylo Dobrikov on 24/09/14.
//
//

#include "loop_fairness.h"


SP_term_ref create_loop_nodes_list(ltlarray *bf_items, unsigned int solution, int size) {
    loopitem *item;
    unsigned int index = solution;
    
    ltlatom *atom;
    ltlnode *node;
    SP_term_ref ref,list,tail;
    unsigned int i;
    tail = SP_new_term_ref();
    list = tail;
    if (size > 0) {
        item = (loopitem*) get_array_element(bf_items, solution);
        index = item->previous;
        for (i=0; i<size; i++) {
            item = (loopitem*) get_array_element(bf_items,index);
            atom = item->atom;
            node = atom->node;
            if (node->type == int_node) {
                ref = SP_new_term_ref();
                SP_put_integer(ref, node->id.id);
            } else {
                ref = node->id.ref;
            }
            list = SP_new_term_ref();
            SP_cons_list(list,ref,tail);
            tail = list;
            index = item->previous;
        }
    } else {
        fprintf(stderr, "No atom in the loop.");
        exit(1);
    }
    return list;
}

int is_fair_loop(ltl_formula *f, ltlarray *bf_items, unsigned int solution, int size) {

    SP_term_ref loop_node_ids = create_loop_nodes_list(bf_items,solution,size);

    // In addition to guaranteeing that the SCC of the tarjan search is fair,
    // we also need to assure that the found loop is also fair in regard to the fairness conditions
    tarjandata t;
    init_tarjan_data(&t,f);
    
    tarjandata tres;
    int i,tres_scc_size;
    init_tarjan_data(&tres, f);

    loopitem *item;
    unsigned int index = solution;

    item = (loopitem*) get_array_element(bf_items, solution);
    index = item->previous;
    ltlatom* atom;
    for (i=0; i<size; i++) {
        item = (loopitem*) get_array_element(bf_items, index);
        atom = item->atom;
        push_tarjan_stack(&t,atom,0,1);
        push_to_stack(&t,atom,1);
        index = item->previous;
    }
    
    int result = is_fair_scc_check(&t, size, loop_node_ids, &tres, &tres_scc_size,1);
    delete_tarjan_data(&t);
    delete_tarjan_data(&tres);
    
    return result;
}
