
#ifndef NOSICSTUS
#include "sicstus/sicstus.h"
#endif

#include "ltlmem.h"
#include "stdio.h"

static int open_mallocs = 0;

void print_number_of_open_mallocs(const char *msg) {
    printf("%s. Rest of residual mallocs: %d\n",msg,open_mallocs);
}

void init_mallocs_counter() {
    open_mallocs = 0;
}

void *ltl_malloc(size_t size) {
  void *ptr = 0;
  if (size>0) {
    open_mallocs++;
#ifndef NOSICSTUS
    ptr = SP_malloc(size);
#else
    ptr = malloc(size);
#endif
  }
  return ptr;
}

void ltl_free(void *ptr) {
  if (ptr) {
    open_mallocs--;
#ifndef NOSICSTUS  
    SP_free(ptr);
#else
    free(ptr);
#endif
  }
}

void *ltl_realloc(void *ptr, size_t size) {
  if (ptr == 0) {
    /* Workaround for bug in SP_realloc(): */
    return ltl_malloc(size);
  } else {
#ifndef NOSICSTUS
    return SP_realloc(ptr, size);
#else
    return realloc(ptr, size);
#endif
  }
}

void ltl_check_memory() {
  if (open_mallocs>0) {
    fprintf(stderr, "WARNING: memory leak detected. %i free() calls missing.\n", open_mallocs);
  }
}
