% (c) 2009-2010 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(myheap, [myheap_init/0,
  myheap_reset/0, myheap_add_node/2, myheap_add_node_at_front/1, myheap_add_node_at_end/1,
  myheap_add_node_random/1,
  myheap_size/1, myheap_isempty/0, 
  myheap_top_id/1, myheap_top_weight/1, 
  myheap_pop/1,
  
  myheap_max_weight/1,
  myheap_pop_max/1,
  myheap_portray/0,
  
  mycounter_set/1, mycounter_dec/1,
  test/0]).


/* the interface to a heap global variable stored and maintained in C */

foreign(myheap_initialize,myheap_initialize). % initializes the data structure
foreign(myheap_reset,myheap_reset). % resets the heap to the empty heap; frees memory
foreign(myheap_add_node,myheap_add_node(+integer,+integer)).  % ID, Weight
foreign(myheap_add_node_at_front,myheap_add_node_at_front(+integer)).  % ID
foreign(myheap_add_node_at_end,myheap_add_node_at_end(+integer)).  % ID
foreign(myheap_add_node_random,myheap_add_node_random(+integer)).  % ID
foreign(myheap_size,myheap_size([-integer])).
foreign(myheap_empty,myheap_empty([-integer])). %return 1 if empty, 0 if non-empty
foreign(myheap_top_id,myheap_top_id([-integer])).
foreign(myheap_top_weight,myheap_top_weight([-integer])).
foreign(myheap_max_weight,myheap_max_weight([-integer])).
foreign(myheap_pop,myheap_pop([-integer])). % pops the ID with minimum Weight, -1 if empty
foreign(myheap_pop_max,myheap_pop_max([-integer])). % pops the ID with maximum Weight, -1 if empty
foreign(myheap_portray,myheap_portray). % pops the ID with maximum Weight, -1 if empty

foreign(mycounter_set,mycounter_set(+integer)). % set a counter to some value
foreign(mycounter_dec,mycounter_dec([-integer])). % decrement and return value

% to do: also implement a pop from end to replace the not_all_z/1 relation
% for mixed depth-first, breadth-first traversal

foreign_resource(myheap,
  [
   myheap_reset
  ,myheap_initialize
  ,myheap_add_node
  ,myheap_add_node_at_front
  ,myheap_add_node_at_end
  ,myheap_add_node_random
  ,myheap_size
  ,myheap_empty
  ,myheap_top_id
  ,myheap_top_weight
  ,myheap_max_weight
  ,myheap_pop
  ,myheap_pop_max
  ,myheap_portray
  
  ,mycounter_set
  ,mycounter_dec
  ]).
  
  
  :- dynamic loaded/0.

myheap_init :- loadfr, print(initializing_myheap),nl, myheap_initialize.

loadfr :-
  (loaded -> true
   ; (assert(loaded), 
      %assert_dir, 
      %assert(user:library_directory(prob_home('.'))),
      print(loading_foreign_resource(myheap)),nl,
      load_foreign_resource(library(myheap)))
   ).

assert_dir :- (user:library_directory('.') -> true ; assert(user:library_directory('.'))).

myheap_isempty :- myheap_empty(X),X=1.

test :-  assert(user:library_directory('.')),
         myheap_init,
         mycounter_set(10),
         mycounter_dec(X1), X1==9,
         mycounter_dec(X2), X2==8,
         myheap_add_node(22,2),
         myheap_add_node(33,3),
         myheap_add_node(221,2),
         myheap_add_node_at_front(11),
         myheap_add_node_at_front(5),
         myheap_add_node_random(12345),
         myheap_add_node_random(6789),
         myheap_portray,
         myheap_top_weight(W),
         myheap_pop(X),
         print(popped_pl(X:W)),nl, flush_output(user),
         myheap_portray,
         myheap_size(Sz),
         print(size(Sz)),nl, flush_output(user),
         myheap_max_weight(W2),
         myheap_pop_max(Y),
         print(popped_max(Y:W2)),nl, flush_output(user),
         myheap_portray,
         myheap_size(Sz2),
         print(size(Sz2)),nl, flush_output(user),
         myheap_pop(Y2),
         print(popped_pl(Y2)),nl, flush_output(user),
         myheap_portray,
         myheap_empty(Y3),
         print(empty_pl(Y3)),nl, flush_output(user),
         myheap_top_weight(W3),
         print(W3),nl.
   
 bench :- bench(100000).
 bench(X) :- myheap_init,
         statistics(runtime,_),
         add_nodes(X),
         pop, 
         assert_nodes(X),
         pop_nodes.
         
 pop :- myheap_pop(P), P \= -1,!, pop.
 pop :- print_time(pop).
 
 add_nodes(X) :- X>0, !,myheap_add_node_at_front(X), X1 is X-1, add_nodes(X1).
 add_nodes(_) :- print_time(add_nodes).
 
 :- dynamic o1/1, o2/1.
 assert_nodes(X) :- X>0, !,asserta(o1(X)),assertz(o2(X)), X1 is X-1, assert_nodes(X1).
 assert_nodes(_) :- print_time(assert_nodes).
 
 pop_nodes :- retract(o1(X)), !, retract(o2(X)), pop_nodes.
 pop_nodes :- print_time(pop_nodes).
 print_time(T) :- statistics(runtime,[_,T2]), print(T), print(' : '), print(T2), print(' ms'),nl.