#include <sys/stat.h>
#include <stdlib.h>

#include "../include/zmq.h"
#include "../include/czmq.h"
#include "shm_cln.h"
#include "mapper.h"
#include "hashtrie.h"

#include "hashrecovery.h"


int receive_hashes(int numWorkers, zctx_t *ctx, int port_start) {
    void *recover_socket = zsocket_new(ctx, ZMQ_REP);
    int rc = zsocket_bind(recover_socket, "tcp://*:%d", port_start + 8);
    assert(rc == 5008);
    int empty_workers = 0;
    int i;
    for (i = 0; i < numWorkers; i++) {
        puts("trying to recv msg");
        zmsg_t *msg = zmsg_recv(recover_socket);
        puts("received msg");
        assert(msg);
        if (zmsg_size(msg) == 1 && zframe_size(zmsg_first(msg)) < 20) {
            empty_workers++;
            puts("empty worker");
            zstr_send(recover_socket, "ok");
            zmsg_destroy(&msg);
            continue;
        }
        zframe_t *frame;
        while ((frame = zmsg_pop(msg)) != NULL) {
            char *digest = zframe_strdup(frame);
            trie_put_unchecked(digest);
            free(digest);
            zframe_destroy(&frame);
        }
        zmsg_destroy(&msg);
        zstr_send(recover_socket, "ok");
        puts("sent answer");
    }
    zsocket_destroy(ctx, recover_socket);
    recover_socket = NULL;
    return empty_workers;
}

void read_hashes(mmap_env *env, size_t flen) {
    size_t pos = 0;
    flen = ((size_t) (flen / 21)) * 21; // if there is an incomplete digest, we don't want it
    printf("assuming %ld hashes\n", flen/21);

    while (pos < flen) {
        char *digest = get_in_mmap(env, pos, 21, 0);
        assert(digest[0] == 1);
        trie_put(digest);
        pos += 21;
    }
    printf("recovered %d hashes\n", trie_get_checked_states_count()); 

    close(env->fd);
}

void recover_hashes(char *dirname, zctx_t *ctx, char *skip_file) {
    //init_shm(ctx);

    DIR *dir = opendir(dirname);
    assert(dir);
    struct dirent *ent;
    printf("skipping file: %s\n", skip_file);


    while ((ent = readdir(dir)) != NULL) {
        char *fname = ent->d_name;
        if (strncmp(fname, "zmqh-", 5) == 0) {
            int len = strlen(fname) + strlen(dirname) + 2;
            char *buf = malloc(len);
            memset(buf, 0, len);

            strcat(buf, dirname);
            if (dirname[strlen(dirname) -1] != '/') {
                strcat(buf, "/");
            }
            strcat(buf, fname);

            if (0 == strcmp(skip_file, buf)) {
                free(buf);
                continue;
            }

            struct stat st;
            stat(buf, &st);
            size_t flen = st.st_size;

            printf("recovering: %s\n", buf);
            int fd = open(buf, O_RDWR);
            assert(fd >= 0);

            mmap_env *env = mapper_init(buf);
            read_hashes(env, flen);
            free(env);

            free(buf);
        }
    }
    closedir(dir);


}



//int main(int argc, char *argv[]) {
//    zctx_t *ctx = zctx_new();
//    recover_hashes("/var/tmp/", ctx);
//    teardown_shm();
//    zctx_destroy(&ctx);
//    return 0;
//}
