#include <assert.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hashtrie.h"
#include "shm_common.h"

// TODO: fix this maybe
#define root 0



void print_readable_digest(unsigned char *key, char *dest) {
	sprintf(dest, "%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",key[0],key[1],key[2],key[3],key[4],key[5],key[6],key[7],key[8],key[9],key[10],key[11],key[12],key[13],key[14],key[15],key[16],key[17],key[18],key[19]);
}


int calculate_pos(char *key, int depth) {
    // pattern repeats after 8 times
    int offset = depth / 8;
    int moddepth = depth % 8;

    char *offi = key + 5 * offset;

    uint64_t a = 0;
    int i;
    for (i = 0; i < 5; i++) {
       a += (uint64_t) (offi[i] * pow(2, 8*i));
    }
    uint64_t c = offi[0] + offi[1] * pow(2, 8) + offi[2] * pow(2, 16)
               + offi[3] * pow(2, 24) + offi[4] * pow(2, 32);
    assert(c == a);

    uint64_t b = a >> (moddepth * 5);
    int x = b & 31;
    return x;
}


int is_digest(tNode *node, int pos) {
    uint32_t bits = node->is_digest;
    int bit_at = (bits >> (31 - pos)) & 1;
    return bit_at;
}



unsigned int skip_first_layer(char key[20], int *index) {
    *index = calculate_pos(key, 0);
    unsigned int id = (lookup_node(root))->ptrs[*index];
    return id;
}


int trie_get_checked_states_count(void) {
    struct hash_meta_infos x;
    get_meta_infos(&x);
    return x.checked_states;
}

int trie_count_elements(void) {
    struct hash_meta_infos x;
    get_meta_infos(&x);
    return x.pos_hashes;
}

char *trie_iterate(unsigned int *blocklen) {
    struct hash_meta_infos x;
    get_meta_infos(&x);
    lock_digests_read();
    char *res = get_shm_digests(x.pos_hashes);
    release_digests_read();
    *blocklen = x.pos_hashes;
    return res;
}

int get_max_depth(void) {
    struct hash_meta_infos x;
    get_meta_infos(&x);
    return x.max_depth;
}

static int h_teardown(tNode *node) {
    int i, sum = 0;
    for (i = 0; i < 32; i++) {
        if (node->ptrs[i] == (unsigned int) -1)
            continue;
        if (is_digest(node, i)) {
            sum++;
        } else {
            sum += h_teardown(lookup_node(node->ptrs[i]));
        }
    }
    return sum;
}

int teardown_hashtrie(void) {
    int res = h_teardown(lookup_node(root));
    return res;
}



static int lookup(unsigned int id, char *key, int depth, const int branchpos) {
    tNode *node = lookup_node(id);
    int pos = calculate_pos(key, depth);
    if (!is_digest(node, pos)) {
        return lookup(node->ptrs[pos], key, depth+1, branchpos);
    } else {
        lock_nodes_read(branchpos);
        if (is_digest(node, pos)) {
            if (node->ptrs[pos] == (unsigned int) -1) {
                release_nodes_read(branchpos);
                return 0;
            } else {
                char *ref = lookup_digest(node->ptrs[pos]);
                release_nodes_read(branchpos);
                if (memcmp(key, ref + 1, 20) != 0) {
                    return 0;
                } else if (ref[0]) {
                    return 1;
                }
                return -1;
            }
        } else {
            release_nodes_read(branchpos);
            return lookup(node->ptrs[pos], key, depth+1, branchpos);
        }
    }
}

int trie_contains(char key[20]) {
    int branch;
    unsigned int node = skip_first_layer(key, &branch);
    int res = lookup(node, key, 1, branch);
    return res;
}

int trie_contains_processed(char key[20]) {
    return trie_contains(key) == 1;
}


static unsigned int transform_with(unsigned int digest_id, int checked, int newdepth);

static void init_node(struct shm_node *res) {
    alloc_node(res);
    memset(res->nodeptr->ptrs, 0xFF, 32 * sizeof(unsigned int));
    res->nodeptr->is_digest = (uint32_t) -1;
}

static void set_digest(tNode *node, int pos) {
    uint32_t bits = node->is_digest;
    uint32_t newbit = 1 << (31 - pos);
    node->is_digest = bits | newbit;
}

static void set_node(tNode *node, int pos) {
    uint32_t bits = node->is_digest;
    uint32_t newbit = 1 << (31 - pos);
    node->is_digest = bits & ~newbit;
}

static unsigned int insert(char *digest, int checked) {
    struct shm_digest x;
    lock_digests_write();
    alloc_digest(&x);
    char *target = x.ptr;
    memcpy(target+1, digest, 20);
    release_digests_write();
    target[0] = checked;
    if (checked) {
        inc_checked_states();
    }
    return x.id;
}


static void add(unsigned int id, char key[20], int checked, int depth, const int branchpos) {
    tNode *node = lookup_node(id);
    assert(depth < 31);
    int pos = calculate_pos(key, depth);

    if (!is_digest(node, pos)) {
        add(node->ptrs[pos], key, checked, depth+1, branchpos);
    } else {
        lock_nodes_write(branchpos);
        if (is_digest(node, pos)) {
            if (node->ptrs[pos] == (unsigned int) -1) {
                node->ptrs[pos] = insert(key, checked);
                release_nodes_write(branchpos);
                //set_digest(node, pos);
                update_max_depth(depth);
            } else {
                char *digest = lookup_digest(node->ptrs[pos]);
                if (memcmp(digest + 1, key, 20) == 0) {
                    release_nodes_write(branchpos);
                    if (checked && !(digest[0]))
                        inc_checked_states();
                    digest[0] |= checked;
                } else {
                    unsigned int succ = transform_with(node->ptrs[pos], checked, depth+1);
                    node = lookup_node(id); // memory block may be moved; recalculate the pointer is necessary
                    node->ptrs[pos] = succ;
                    set_node(node, pos);
                    release_nodes_write(branchpos);
                    add(node->ptrs[pos], key, checked, depth+1, branchpos);
                }
            }
        } else {
            release_nodes_write(branchpos);
            add(node->ptrs[pos], key, checked, depth+1, branchpos);
        }
    }
}

// MAYMOVE shm node structure
static unsigned int transform_with(unsigned int digest_id, int checked, int newdepth) {
    struct shm_node r;
    init_node(&r);
    tNode *node = r.nodeptr;
    char *digest = lookup_digest(digest_id);

    int pos = calculate_pos(digest+1, newdepth);
    assert(node->ptrs[pos] == (unsigned int) -1);
    node->ptrs[pos] = digest_id;
    set_digest(node, pos);

    return r.id;
}

void trie_put(char digest[21]) {
    assert(digest[0] == 0 || digest[0] == 1);
    int branch;
    unsigned int node = skip_first_layer(digest+1, &branch);
    add(node, digest+1, digest[0], 1, branch);
}

void trie_put_checked(char digest[20]) {
    int branch;
    unsigned int node = skip_first_layer(digest, &branch);
    add(node, digest, 1, 1, branch);
}

void trie_put_unchecked(char digest[20]) {
    int branch;
    unsigned int node = skip_first_layer(digest, &branch);
    add(node, digest, 0, 1, branch);
}
