#include <sys/mman.h>
#include <sys/stat.h>

#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

#include "mapper.h"

static size_t page_size;

static char *create_mapping(mmap_env *env, size_t pos, size_t len, int force_update) {
    assert(pos != (size_t) -1);
    if (!force_update && env->current_start < pos && env->current_end > pos+len) {
        //msync(map_start, size, MS_SYNC);
        return env->map_start;
    }

    if (env->map_start != NULL) {
        int rc = munmap(env->map_start, env->size);
        assert(rc == 0);
    }

    off_t aligned_pos = ((off_t) (pos / page_size)) * page_size; // page we need
    size_t aligned_offset = pos % page_size; // how many bytes there are at the beginning of the page in front of the term

    while (env->size < len + aligned_offset) {
        env->size = 2*env->size;
    }

    char *addr = mmap(NULL, env->size, PROT_READ | PROT_WRITE, MAP_SHARED, env->fd, aligned_pos);
    assert(addr != (void*) -1);

    env->current_start = aligned_pos;
    env->current_end = aligned_pos + env->size;
    assert(env->current_start + len <= env->current_end);
    env->map_start = addr;

    return addr;
}

char *get_in_mmap(mmap_env *env, size_t pos, size_t len, int force_update) {
    char *addr = create_mapping(env, pos, len, force_update);
    size_t delta = pos - env->current_start;
    return addr + delta;
}

mmap_env *mapper_init(char *filename) {
    page_size = sysconf(_SC_PAGE_SIZE);

    struct flock lock;
    lock.l_type = F_WRLCK;
    lock.l_whence = SEEK_SET;
    lock.l_start = 0;
    lock.l_len = 0;

    mmap_env *env = (mmap_env*) malloc(sizeof(mmap_env));
    env->filename = filename;
    // TODO: O_FLAGS
    env->fd = open(filename, O_RDWR | O_CREAT /* | O_EXCL*/, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
    assert(env->fd);

    int rc = fcntl(env->fd, F_SETLKW, &lock);
    assert(rc != -1);

    env->size = page_size;
    env->current_start = 0;
    env->current_end = 0;
    env->map_start = NULL;
    printf("size: %ld\n", env->size);

    return env;
}

