#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <semaphore.h>

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>

#include "../include/zmq.h"
#include "../include/czmq.h"

#include "shm_cln.h"
#include "shm_common.h"
#include "hashtrie.h"


static struct semap_env *open_semaphores(void) {
    struct semap_env *r = (struct semap_env *) malloc(sizeof(struct semap_env));
    r->semap_meta   = sem_open("/hashes-semap_infos", 0);
    r->semap_hashes = sem_open("/hashes-semap_digests", 0);
    r->semap_hashwrite = sem_open("/hashes-semap_digestswrite", 0);
    assert(r->semap_meta   != SEM_FAILED);
    assert(r->semap_hashes != SEM_FAILED);
    assert(r->semap_hashwrite != SEM_FAILED);
    int i;
    for (i = 0; i < 32; i++) {
        int len = strlen("/hashes-semap_nodes") + 3; // "31" + zero termination
        char name[len];
        sprintf(name, "/hashes-semap_nodes%d", i);
        r->semap_nodes[i] = sem_open(name, 0);
        assert(r->semap_nodes[i] != SEM_FAILED);

        len = strlen("/hashes-semap_meta") + 3; // "31" + zero termination
        char name2[len];
        sprintf(name2, "/hashes-semap_meta%d", i);
        r->semap_multiwrite[i] = sem_open(name2, 0);
        assert(r->semap_multiwrite[i] != SEM_FAILED);
    }
    return r;
}

void teardown_shm(void) {
    int rc = sem_close(semaphores->semap_meta);
    assert(rc == 0);
    rc = sem_close(semaphores->semap_hashes);
    assert(rc == 0);
    rc = sem_close(semaphores->semap_hashwrite);
    assert(rc == 0);
    int i;
    for (i = 0; i < 32; i++) {
        rc = sem_close(semaphores->semap_nodes[i]);
        assert(rc == 0);
        rc = sem_close(semaphores->semap_multiwrite[i]);
        assert(rc == 0);
    }
    free(semaphores);
    shm_cleanup(0);
    #ifdef BLOCKS
    printf("%u readblocks\n", blocks_read);
    #endif
}


void init_shm(zctx_t *ctx) {
    void *zocket = zsocket_new(ctx, ZMQ_REQ);
    zsocket_connect(zocket, "ipc:///tmp/zmq-syncer");
    zstr_send(zocket, "hello");

    char *rep = zstr_recv(zocket);
    assert(strcmp(rep, "go") == 0);
    free(rep);
    zsocket_destroy(ctx, zocket);
            
    struct semap_env *senv = open_semaphores();
    init_shm_common(senv);
}


