//#include <sys/mman.h>
//#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>
//#include <fcntl.h>
#include <semaphore.h>

#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <math.h>
#include <string.h>
#include "shm_common.h"

#ifdef BLOCKS
unsigned int blocks_read = 0;
#endif

struct semap_env *semaphores = NULL;
struct shm_env *shm = NULL;

#define DIGEST_KEY_START    4000
#define NODES_KEY_START     5000
const int last_meta_information = 4999;
int next_digest_key = DIGEST_KEY_START;
int next_nodes_key = NODES_KEY_START;

#define ELEMENTS_PER_BLOCK      (10000000) //(1e7)
#define SHM_BLOCK_LEN_DIGEST    (ELEMENTS_PER_BLOCK * 21)
#define SHM_BLOCK_LEN_NODES     (ELEMENTS_PER_BLOCK * sizeof(tNode))


enum segment_types {
    DIGESTS,
    NODES,
    META
};

#include <stdio.h>
struct shm_block *get_shm(enum segment_types t) {
    key_t key;
    size_t size = sizeof(struct hash_meta_infos);
    switch(t) {
        case META:
            key = last_meta_information; break;
        case DIGESTS:
            key = next_digest_key++;
            size = SHM_BLOCK_LEN_DIGEST;
            assert(next_digest_key <= last_meta_information); break;
        case NODES:
            key = next_nodes_key++;
            size = SHM_BLOCK_LEN_NODES;
            assert(next_nodes_key <= NODES_KEY_START + 1000); break;
    }
    
    int shmid = shmget(key, size, IPC_CREAT | 0666);
    assert(shmid >= 0);

    void *ptr = shmat(shmid, NULL, 0);
    assert(ptr != (void*) -1);

    struct shm_block *block = (struct shm_block*) malloc(sizeof(struct shm_block));
    block->ptr = ptr;
    block->shmid = shmid;

    return block;
}

#include <errno.h>

void shm_cleanup(int server) {
    int rc = shmdt(shm->shm_meta_infos->ptr);
    assert(rc == 0);
    if (server) {
        rc = shmctl(shm->shm_meta_infos->shmid, IPC_RMID, NULL);
        assert(rc != -1);
    }
    free(shm->shm_meta_infos);
    
    int i;
    for (i = next_digest_key - 1 - DIGEST_KEY_START; i >= 0; i--) {
        rc = shmdt(shm->shm_digests[i]->ptr);
        assert(rc == 0);
        if (server) {
            rc = shmctl(shm->shm_digests[i]->shmid, IPC_RMID, NULL);
            assert(rc != -1);
        }
        free(shm->shm_digests[i]);
    }
      
    for (i = next_nodes_key - 1 - NODES_KEY_START; i >= 0; i--) {
        rc = shmdt(shm->shm_nodes[i]->ptr);
        assert(rc == 0);
        if (server) {
            rc = shmctl(shm->shm_nodes[i]->shmid, IPC_RMID, NULL);
            assert(rc != -1);
        }
        free(shm->shm_nodes[i]);
    }
    free(shm);
}

void get_pos_tuple(const unsigned int input, size_t *block, size_t *offset) {
    *block  = input / ELEMENTS_PER_BLOCK;
    *offset = input % ELEMENTS_PER_BLOCK;
}

void init_shm_common(struct semap_env *env) {
    semaphores = env;
    int i;
    for (i = 0; i < 32; i++) {
        lock_nodes_write(i);
        release_nodes_write(i);
    }
    struct shm_env *res = (struct shm_env *) malloc(sizeof(struct shm_env));
    res->shm_meta_infos = get_shm(META);
    memset(res->shm_nodes, 0, 999 * sizeof(struct shm_block*));
    res->shm_nodes[0] = get_shm(NODES);
    memset(res->shm_digests, 0, 999 * sizeof(struct shm_block*));
    res->shm_digests[0] = get_shm(DIGESTS);

    shm = res;
}

inline void lock_digests_read(void) {
    sem_wait(semaphores->semap_hashes);
}

inline void release_digests_read(void) {
    sem_post(semaphores->semap_hashes);
}


inline void lock_infos(void) {
    sem_wait(semaphores->semap_meta);
}

inline void release_infos(void) {
    sem_post(semaphores->semap_meta);
}

void get_meta_infos(struct hash_meta_infos *res) {
    lock_infos();
    *res = *((struct hash_meta_infos*) shm->shm_meta_infos->ptr);
    release_infos();
}

char *get_shm_digests(size_t amount) {
    char *buf = calloc(21, amount);
    size_t pos = 0;
    int i;
    for (i = DIGEST_KEY_START; i < next_digest_key; i++) {
        size_t blocklen = SHM_BLOCK_LEN_DIGEST;
        if (amount < blocklen) {
            blocklen = amount;
        }
        memcpy(buf + pos * 21, shm->shm_digests[i - DIGEST_KEY_START]->ptr, blocklen);
        amount = amount - blocklen;
        if (amount <= 0)
            break;
    }
    return buf;
}


tNode *lookup_node(unsigned int id) {
    size_t block, offset;
    get_pos_tuple(id, &block, &offset);
    while (block >= next_nodes_key - NODES_KEY_START) {
        int pos = next_nodes_key - NODES_KEY_START;
        shm->shm_nodes[pos] = get_shm(NODES);
    } 
    return ((tNode*) shm->shm_nodes[block]->ptr) + offset; // pointer arithmetic does its magic
}

char *lookup_digest(unsigned int id) {
    size_t block, offset;
    get_pos_tuple(id, &block, &offset);
    while (block >= next_digest_key - DIGEST_KEY_START) {
        int pos = next_digest_key - DIGEST_KEY_START;
        shm->shm_digests[pos] = get_shm(DIGESTS);
    } 
    return ((char*) shm->shm_digests[block]->ptr) + (21 * offset);
}

inline void lock_nodes_read(const int pos) {
#ifdef BLOCKS
    int rc = sem_trywait(semaphores->semap_nodes[pos]);
    if (rc == 0)
        return;
    blocks_read++;
#endif
    sem_wait(semaphores->semap_nodes[pos]);
}


inline void release_nodes_read(const int pos) {
    sem_post(semaphores->semap_nodes[pos]);
}

void inc_checked_states(void) {
    lock_infos();
    ((struct hash_meta_infos*) (shm->shm_meta_infos->ptr))->checked_states++;
    release_infos();
}


void update_max_depth(int depth) {
    lock_infos();
    int max = ((struct hash_meta_infos*) shm->shm_meta_infos->ptr)->max_depth;
    if (max < depth)
        ((struct hash_meta_infos*) shm->shm_meta_infos->ptr)->max_depth = depth;
    release_infos();
}




void alloc_node(struct shm_node *x) {
    lock_infos();
    unsigned int res = ((struct hash_meta_infos*) shm->shm_meta_infos->ptr)->pos_nodes;
    ((struct hash_meta_infos*) shm->shm_meta_infos->ptr)->pos_nodes++;
    release_infos();


    x->nodeptr = lookup_node(res);
    x->id = res;
}

void alloc_digest(struct shm_digest *r) {
    lock_infos();
    unsigned int res = ((struct hash_meta_infos*) shm->shm_meta_infos->ptr)->pos_hashes;
    ((struct hash_meta_infos*) shm->shm_meta_infos->ptr)->pos_hashes++;
    release_infos();

    r->ptr = lookup_digest(res);
    r->id = res;
}


inline void lock_nodes_write(const int pos) {
    sem_wait(semaphores->semap_multiwrite[pos]);
    int i;
    for (i = 0; i < SEMAP_NODES_VAL; i++) {
        sem_wait(semaphores->semap_nodes[pos]);
    }
    sem_post(semaphores->semap_multiwrite[pos]);
}

inline void lock_digests_write(void) {
    sem_wait(semaphores->semap_hashwrite);
    int i;
    for (i = 0; i < SEMAP_DIGESTS_VAL; i++) {
        sem_wait(semaphores->semap_hashes);
    }
    sem_post(semaphores->semap_hashwrite);
}

inline void release_nodes_write(const int pos) {
    int i;
    for (i = 0; i < SEMAP_NODES_VAL; i++) {
        sem_post(semaphores->semap_nodes[pos]);
    }
}

inline void release_digests_write(void) {
    int i;
    for (i = 0; i < SEMAP_DIGESTS_VAL; i++) {
        sem_post(semaphores->semap_hashes);
    }
}

