#ifndef SHM_COMMON_H_INCLUDED
#define SHM_COMMON_H_INCLUDED

#include "hashtrie.h"
#include <semaphore.h>

#define INITIAL_SIZE    10000
#define SEMAP_DIGESTS_VAL   10
#define SEMAP_NODES_VAL     5

#ifdef BLOCKS
unsigned int blocks_read;
#endif

struct hash_meta_infos {
    volatile unsigned int pos_hashes; // equals known states
    volatile unsigned int pos_nodes;
    volatile unsigned int checked_states;
    volatile unsigned int max_depth;
};

struct shm_digest {
    char *ptr;
    unsigned int id;
};

struct shm_node {
    tNode *nodeptr;
    unsigned int id;
};


struct semap_env {
    sem_t *semap_meta;
    sem_t *semap_hashes;
    sem_t *semap_hashwrite;
    sem_t *semap_nodes[32];
    sem_t *semap_multiwrite[32];
};

struct shm_block {
    int shmid;
    void *ptr;
};

struct shm_env {
    // FIXME: this is me being lazy.
    // 999 pointers are /probably/ enough
    // this is the upper limit restricted by arbitrarily chosen key limits
    // I assume we have the 8K memory for 1k pointers.
    struct shm_block *shm_meta_infos;
    struct shm_block *shm_digests[999];
    struct shm_block *shm_nodes[999];
};

struct semap_env *semaphores;
struct shm_env *shm;

void init_shm_common(struct semap_env *env);
char *get_shm_digests(size_t amount);
void get_meta_infos(struct hash_meta_infos *res);
void lock_digests_read(void);
void release_digests_read(void);
void lock_nodes_read(const int pos);
void release_nodes_read(const int pos);
void lock_infos(void);
void release_infos(void);
tNode *lookup_node(unsigned int id);
char *lookup_digest(unsigned int id);
void alloc_node(struct shm_node *x);
void alloc_digest(struct shm_digest *r);
void inc_checked_states(void);
void update_max_depth(int depth);
void shm_cleanup(int server);

void lock_nodes_write(const int pos);
void lock_digests_write(void);
void release_nodes_write(const int pos);
void release_digests_write(void);

#endif
