#include <stdlib.h>
#include <assert.h>
#include <pthread.h>

#include "mctrl.h"
#include "mbeat.h"
#include "mbalancer.h"
#include "mprint.h"

#define TIMEOUT_CONSTANT    3 // in seconds

static int *worker_alive = NULL;
static volatile int worker_alive_len = 0;
static void *local_printer;

static pthread_mutex_t beat_lock; // for extending worker_alive

static pthread_t ecg;

static void *check_heartbeat(void *arg);

void init_mbeat(zctx_t *ctx) {
    local_printer = zsocket_new(ctx, ZMQ_PUSH);
    zsocket_connect(local_printer, "inproc://printer");
    int rc = pthread_mutex_init(&beat_lock, NULL);
    assert(rc == 0);
    pthread_create(&ecg, NULL, check_heartbeat, NULL);
}

void teardown_mbeat(zctx_t *ctx) {
    pthread_cancel(ecg);
    pthread_mutex_destroy(&beat_lock);
    assert(worker_alive);
    free(worker_alive);
    zsocket_disconnect(local_printer, "inproc://printer");
    zsocket_destroy(ctx, local_printer);
}

void set_worker_alive(int id) {
    if (id > worker_alive_len) {
        pthread_mutex_lock(&beat_lock);
        worker_alive = realloc(worker_alive, (id + 1) * sizeof(int));
        pthread_mutex_unlock(&beat_lock);
        int i;
        for (i = worker_alive_len + 1; i <= id; i++) {
            worker_alive[i] = 5;
        }
        worker_alive_len = id;
    } else {
        worker_alive[id] = 5;
    }
}

static int system_alive(void) {
    int i;
    pthread_mutex_lock(&beat_lock);
    for (i = 1; i <= worker_alive_len; i++) {
        if (get_q_size(i) >= 0) { // do not touch uninitialised workers
            worker_alive[i]--;
            //printf("i: %d, val: %d\n", i, worker_alive[i]);
            if (worker_alive[i] <= 0) {
                DEBUG_PRINT(DEBUG_NORMAL,,,":worker-died %d", i);
                pthread_mutex_unlock(&beat_lock);
                return 0;
            }
        }
    }
    pthread_mutex_unlock(&beat_lock);
    return 1;
}


static void *check_heartbeat(void *arg) {
    while(1) {
        if (system_alive()) {
            sleep(TIMEOUT_CONSTANT);
        } else {
//            abort_run();
            sleep(1); // wait so that the debug message may get to the printer
            assert(0);
        }
    }
}
