#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include "../../../include/zmq.h"
#include "../../../include/czmq.h"

#include "../../../common/hashtrie.h"

#include "mid.h"
#include "mbalancer.h"
#include "mprint.h"

#define NOT_INITIALIZED -1

static int next_worker_id = 1;
static void *id_response = NULL;
static char *model = NULL;
static int model_len;
static void *local_printer;

extern volatile int running;

static void *h_joiner(void *arg) {
    while (running) {
        DEBUG_PRINT(DEBUG_VERBOSE,,,":enter :h_joiner");
        zsocket_poll(id_response, -1); // block until input is availabe

        char *string = zstr_recv (id_response);
        assert(string != NULL);
        if (strcmp(string, "worker") == 0) {
            zmsg_t *msg =  zmsg_new();
            zmsg_addstr(msg, "%d", next_worker_id);
            zmsg_addmem(msg, model, model_len);
            zmsg_send(&msg, id_response);
            zmsg_destroy(&msg);
            next_worker_id++;
        } else if (strcmp(string, "hasher") == 0) {
            zmsg_t *msg = zmsg_new();
            unsigned int blocklen;
            char *hashes = trie_iterate(&blocklen);
            zmsg_addmem(msg, hashes, blocklen);
            zmsg_send(&msg, id_response);
            zmsg_destroy(&msg);
            free(hashes);
        } else {
            zstr_send(id_response, "SRSLY?");
        }
        DEBUG_PRINT(DEBUG_VERBOSE,,,":exit :h_joiner");
    }
    pthread_exit(NULL);
}


void start_h_joiner(pthread_t *idthread) {
    pthread_create(idthread, NULL, h_joiner, NULL);
}


void init_id_rep(zctx_t *ctx, zmq_pollitem_t *poller, int port_start) {
    local_printer = zsocket_new(ctx, ZMQ_PUSH);
    zsocket_connect(local_printer, "inproc://printer");
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :init_id_rep");
    assert(id_response == NULL);
    id_response = zsocket_new (ctx, ZMQ_REP);
    zsocket_bind (id_response, "tcp://*:%d", port_start + 5);

    *poller = (zmq_pollitem_t) { id_response, 0, ZMQ_POLLIN };
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :init_id_rep");
}

void teardown_id_rep(zctx_t *ctx) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :teardown_id_rep");
    assert(id_response != NULL);
    zsocket_destroy(ctx, id_response);
    id_response = NULL;
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :teardown_id_rep");
    zsocket_disconnect(local_printer, "inproc://printer");
    zsocket_destroy(ctx, local_printer);
}

void set_model(char *term, int model_size) {
    model = (char*) malloc(model_size);
    assert(model);
    memcpy(model, term, model_size);
    model_len = model_size;
}

