//#include <sys/mman.h>
//#include <sys/stat.h>
//#include <fcntl.h>
#include <semaphore.h>

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "../include/zmq.h"
#include "../include/czmq.h"

#include "../common/shm_common.h"
#include "../common/hashtrie.h"
#include "../common/memusage.h"


void *hash_zocket = NULL;
void *ctrl_zocket = NULL;
void *sync_workers = NULL;
char *semap_names[32];
char *semap_multiwrite_names[32];

int write_hash(zloop_t *loop, zmq_pollitem_t *poller, void *arg) {
    int count = 0;
    // FIXME: magic constant 1000
    // hash insertion has top priority
    // polling termination or adding a new worker may take a few milliseconds
    // possible enhancement: heartbeating with the master, master sends number of current workers
    while (count < 1000 && zsocket_poll(hash_zocket, 0) == true) {
        zmsg_t *msg = zmsg_recv(hash_zocket);
        zframe_t *frame;
        while ((frame = zmsg_pop(msg)) != NULL) {
            assert(21 == zframe_size(frame));
            char *digest = zframe_strdup(frame);
            trie_put(digest);
            free(digest);
            zframe_destroy(&frame);
        }
        zmsg_destroy(&msg);
        count++;
    }
    return 0;
}

int signal_go(zloop_t *loop, zmq_pollitem_t *poller, void *arg) {
    zmsg_t *msg = zmsg_recv(sync_workers);
    zmsg_destroy(&msg);
    zstr_send(sync_workers, "go");
    return 0;
}

int handle_ctrl(zloop_t *loop, zmq_pollitem_t *poller, void *arg) {
    char *cmd = zstr_recv(ctrl_zocket);
    if (strcmp(cmd, "TERM") == 0) {
        free(cmd);
        return -1;
    } else if (strcmp(cmd, "MEMORY") == 0) {
        printf("memory allocated: %ld\n", memory_usage_in_kb());
    }
    free(cmd);
    // TODO: invalid message
    return 0;
}


void init_semaphore_names(void) {
    int i;
    for (i = 0; i < 32; i++) {
        int len = strlen("/hashes-semap_nodes") + 3; // "31" + zero termination
        semap_names[i] = (char*) malloc(len);
        sprintf(semap_names[i], "/hashes-semap_nodes%d", i);
        len = strlen("/hashes-semap_meta") + 3;
        semap_multiwrite_names[i] = (char*) malloc(len);
        sprintf(semap_multiwrite_names[i], "/hashes-semap_meta%d", i);
    }
}

void rmsem(void) {
    sem_unlink("/hashes-semap_infos");
    sem_unlink("/hashes-semap_digests");
    sem_unlink("/hashes-semap_digestswrite");
    int i;
    for (i = 0; i < 32; i++) {
        sem_unlink(semap_names[i]);
        sem_unlink(semap_multiwrite_names[i]);
    }
}

struct semap_env *create_semaphores(void) {
    init_semaphore_names();

    rmsem();

    struct semap_env *r = (struct semap_env *) malloc(sizeof(struct semap_env));
    r->semap_meta   = sem_open("/hashes-semap_infos", O_CREAT | O_EXCL, 0666, 1);
    r->semap_hashes = sem_open("/hashes-semap_digests", O_CREAT | O_EXCL, 0666, SEMAP_DIGESTS_VAL);
    r->semap_hashwrite = sem_open("/hashes-semap_digestswrite", O_CREAT | O_EXCL, 0666, 1);
    assert(r->semap_meta   != SEM_FAILED);
    assert(r->semap_hashes != SEM_FAILED);
    assert(r->semap_hashwrite != SEM_FAILED);
    int i;
    for (i = 0; i < 32; i++) {
        r->semap_nodes[i] = sem_open(semap_names[i], O_CREAT | O_EXCL, 0666, SEMAP_NODES_VAL);
        assert(r->semap_nodes[i] != SEM_FAILED);
        r->semap_multiwrite[i] = sem_open(semap_multiwrite_names[i], O_CREAT | O_EXCL, 0666, 1);
        assert(r->semap_multiwrite[i] != SEM_FAILED);
    }

    return r;
}

void init_trie(void) {
    sem_wait(semaphores->semap_meta);
    memset(shm->shm_meta_infos->ptr, 0, sizeof(struct hash_meta_infos));

//  this pointer should be initialized already
//  tNode *shm_nodes_ptr = (tNode*) mmap(NULL, 33 * sizeof(tNode), PROT_WRITE, MAP_SHARED, fd_nodes, 0);

    int i;
    for (i = 0; i < 32; i++) {
        // init node
        tNode *node = ((tNode*) (shm->shm_nodes[0]->ptr)) + i + 1;
        memset(node->ptrs, 0xFF, 32 * sizeof(unsigned int));
        memset(&(node->is_digest), 0xFF, sizeof(uint32_t));
        ((struct hash_meta_infos*) (shm->shm_meta_infos->ptr))->pos_nodes++;

        ((tNode*) (shm->shm_nodes[0]->ptr))->ptrs[i] = i + 1; // set successors in root node
    }
    ((struct hash_meta_infos*) shm->shm_meta_infos->ptr)->pos_nodes++;
    memset(&(((tNode*) shm->shm_nodes[0]->ptr)->is_digest), 0, sizeof(uint32_t));
    sem_post(semaphores->semap_meta);
}

void put_hashes(char *hashes, int len) {
    int i;
    for (i = 0; i < len; i += 21) {
        trie_put(hashes + i);
    }
}

int main(int argc, char *argv[]) {
    assert(argc == 4);
    char *host = argv[1];
    int port = atoi(argv[2]);
    int masternode = atoi(argv[3]);

    struct semap_env *senv = create_semaphores();
    init_shm_common(senv);
    init_trie();

    zctx_t *ctx = zctx_new();

    hash_zocket = zsocket_new(ctx, ZMQ_SUB);
    int rc = zsocket_connect(hash_zocket, "tcp://%s:%d", host, port);
    assert(rc == 0);
    char *filter = "";
    zmq_setsockopt(hash_zocket, ZMQ_SUBSCRIBE, filter, strlen(filter));
    ctrl_zocket = zsocket_new(ctx, ZMQ_SUB);
    rc = zsocket_connect(ctrl_zocket, "tcp://%s:%d", host, port + 7);
    assert(rc == 0);
    zmq_setsockopt(ctrl_zocket, ZMQ_SUBSCRIBE, filter, strlen(filter));
    printf("Connected.\n");

    sync_workers = zsocket_new(ctx, ZMQ_REP);
    rc = zsocket_bind(sync_workers, "ipc:///tmp/zmq-syncer");
    assert(rc == 0);

    if (!masternode) {
        void *join_zocket = zsocket_new(ctx, ZMQ_REQ);
        rc = zsocket_connect(join_zocket, "tcp://%s:%d", host, port + 5);
        assert(rc == 0);

        zstr_send(join_zocket, "hasher");
        zmsg_t *msg = zmsg_recv(join_zocket);
        zframe_t *f = zmsg_pop(msg);
        assert(f);
        int len = zframe_size(f);
        char *hashes = zframe_strdup(f);
        put_hashes(hashes, len);
        free(hashes);
        zframe_destroy(&f);
        zmsg_destroy(&msg);

        zsocket_destroy(ctx, join_zocket);
    }
    
    zmq_pollitem_t hash_poller = (zmq_pollitem_t) {hash_zocket, 0, ZMQ_POLLIN};
    zmq_pollitem_t ctrl_poller = (zmq_pollitem_t) {ctrl_zocket, 0, ZMQ_POLLIN};
    zmq_pollitem_t sync_poller = (zmq_pollitem_t) {sync_workers, 0, ZMQ_POLLIN};

    zloop_t *reactor = zloop_new();
    
    zloop_poller(reactor, &hash_poller, write_hash, NULL);
    zloop_poller(reactor, &sync_poller, signal_go, NULL);
    zloop_poller(reactor, &ctrl_poller, handle_ctrl, NULL);
    zloop_start(reactor);
    zloop_destroy(&reactor);


    zsocket_destroy(ctx, hash_zocket);
    zctx_destroy(&ctx);
    
    // cleanup
    rc = sem_close(semaphores->semap_meta);
    assert(rc == 0);
    rc = sem_close(semaphores->semap_hashes);
    assert(rc == 0);
    rc = sem_close(semaphores->semap_hashwrite);
    assert(rc == 0);
    int i;
    for (i = 0; i < 32; i++) {
        rc = sem_close(semaphores->semap_nodes[i]);
        assert(rc == 0);
        rc = sem_close(semaphores->semap_multiwrite[i]);
        assert(rc == 0);
    }
    rc = sem_unlink("/hashes-semap_infos");
    assert(rc == 0);
    rc = sem_unlink("/hashes-semap_digests");
    assert(rc == 0);
    rc = sem_unlink("/hashes-semap_digestswrite");
    assert(rc == 0);
    for (i = 0; i < 32; i++) {
        rc = sem_unlink(semap_names[i]);
        rc = sem_unlink(semap_multiwrite_names[i]);
        assert(rc == 0);
        free(semap_names[i]);
        free(semap_multiwrite_names[i]);
    }
    shm_cleanup(1);

    return 0;
}


