#include "../include/zmq.h"
#include "../include/czmq.h"

#include "CuTest.h"
#include "../master/mhash.h"
#include "../common/hashtrie.h"

extern zctx_t *ctx;
static void *tmp_socket_send;
static void *tmp_socket_recv;
extern int multiple_calculations;

extern int port;

static void init(void) {
    tmp_socket_send = zsocket_new(ctx, ZMQ_PUSH);
    zsocket_connect(tmp_socket_send, "tcp://localhost:%d", port+1);

    tmp_socket_recv = zsocket_new(ctx, ZMQ_SUB);
    char *filter = "";
    zmq_setsockopt (tmp_socket_recv, ZMQ_SUBSCRIBE, filter, strlen (filter));
    zsocket_connect(tmp_socket_recv, "tcp://localhost:%d", port);

    zmq_pollitem_t ignored;
    sleep(1);
    init_hash_module(ctx, &ignored, 3, port);

    port += 2;

}

static void send_message_no_results(void) {
    zmsg_t *msg = zmsg_new();
    zmsg_addstr(msg, "Results:");
    zmsg_addstr(msg, "Hashes:");
    zmsg_addmem(msg, "\0Fake hash 0123456789", 21);
    zmsg_addmem(msg, "\0Fake hash 1111111111", 21);
    zmsg_addmem(msg, "\0Fake hash 2222222222", 21);
    zmsg_addmem(msg, "\1Fake hash 3333333333", 21);
    zmsg_send(&msg, tmp_socket_send);
    zmsg_destroy(&msg);
}

static void recv_message_no_results(CuTest *tc) {
    int rc = zsocket_poll(tmp_socket_recv, 5000);
    CuAssertTrue(tc, rc);
    zmsg_t *msg = zmsg_recv(tmp_socket_recv);
    CuAssertIntEquals(tc, 4, zmsg_size(msg));

    char *hash;
    hash = zmsg_popstr(msg);
    CuAssertMemEquals(tc, "\0Fake hash 0123456789", hash, 21);
    free(hash);

    hash = zmsg_popstr(msg);
    CuAssertMemEquals(tc, "\0Fake hash 1111111111", hash, 21);
    free(hash);

    hash = zmsg_popstr(msg);
    CuAssertMemEquals(tc, "\0Fake hash 2222222222", hash, 21);
    free(hash);

    hash = zmsg_popstr(msg);
    CuAssertMemEquals(tc, "\1Fake hash 3333333333", hash, 21);
    free(hash);

    zmsg_destroy(&msg);
}


static void send_message_with_results(void) {
    zmsg_t *msg = zmsg_new();
    zmsg_addstr(msg, "Results:");
    char *result = "DAI'm not really checking states, shh";
    zmsg_addmem(msg, result, strlen(result)+1);
    zmsg_addstr(msg, "Hashes:");
    zmsg_addmem(msg, "\0Fake hash 0123456789", 21);
    zmsg_send(&msg, tmp_socket_send);
    zmsg_destroy(&msg);
}

static void recv_message_with_results(CuTest *tc) {
    int rc = zsocket_poll(tmp_socket_recv, 5000);
    CuAssertTrue(tc, rc);
    zmsg_t *msg = zmsg_recv(tmp_socket_recv);
    CuAssertIntEquals(tc, 1, zmsg_size(msg));

    char *hash;
    hash = zmsg_popstr(msg);
    CuAssertMemEquals(tc, "\0Fake hash 0123456789", hash, 21);
    free(hash);

    zmsg_destroy(&msg);
}

static void send_message_all_states_checked(void) {
    zmsg_t *msg = zmsg_new();
    zmsg_addstr(msg, "Results:");
    zmsg_addstr(msg, "Hashes:");
    zmsg_addmem(msg, "\1Fake hash 0123456789", 21);
    zmsg_addmem(msg, "\1Fake hash 1111111111", 21);
    zmsg_send(&msg, tmp_socket_send);
    zmsg_destroy(&msg);
}

static void recv_message_all_states_checked(CuTest *tc) {
    int rc = zsocket_poll(tmp_socket_recv, 5000);
    CuAssertTrue(tc, rc);
    zmsg_t *msg = zmsg_recv(tmp_socket_recv);
    CuAssertIntEquals(tc, 2, zmsg_size(msg));

    char *hash;
    hash = zmsg_popstr(msg);
    CuAssertMemEquals(tc, "\1Fake hash 0123456789", hash, 21);
    free(hash);
    hash = zmsg_popstr(msg);
    CuAssertMemEquals(tc, "\1Fake hash 1111111111", hash, 21);
    free(hash);

    zmsg_destroy(&msg);
}

static void teardown(void) {
    zsocket_destroy(ctx, tmp_socket_send);
    zsocket_destroy(ctx, tmp_socket_recv);
    teardown_hash_module(ctx);
}


void Test_h_mhash_no_results(CuTest *tc) {
    init();
    send_message_no_results();
    int rc = h_mhash(NULL, NULL, NULL);
    CuAssertIntEquals(tc, 0, rc);
    CuAssertIntEquals(tc, 1, trie_get_checked_states_count());
    CuAssertIntEquals(tc, 4, trie_count_elements());
    recv_message_no_results(tc);
    teardown();
}

void Test_h_mhash_with_results(CuTest *tc) {
    init();
    send_message_with_results();
    int rc = h_mhash(NULL, NULL, NULL);
    CuAssertIntEquals(tc, 0, trie_get_checked_states_count());
    CuAssertIntEquals(tc, 1, trie_count_elements());
    recv_message_with_results(tc);
    CuAssertIntEquals(tc, 0, rc);
    teardown();
}

void send_multiple_calculations(void) {
    zmsg_t *msg = zmsg_new();
    zmsg_addstr(msg, "Results:");
    zmsg_addstr(msg, "Hashes:");
    zmsg_addmem(msg, "\1Fake hash 1111111111", 21);
    zmsg_addmem(msg, "\1Fake hash 1111111111", 21);
    zmsg_send(&msg, tmp_socket_send);
    zmsg_destroy(&msg);
}

void Test_multiple_processed_counter(CuTest *tc) {
    init();
    send_multiple_calculations();
    CuAssertIntEquals(tc, 0, multiple_calculations);

    int rc = h_mhash(NULL, NULL, NULL);
    CuAssertIntEquals(tc, -1, rc); // all states processed
    CuAssertIntEquals(tc, 1, multiple_calculations);
    send_multiple_calculations();

    rc = h_mhash(NULL, NULL, NULL);
    CuAssertIntEquals(tc, -1, rc); 
    CuAssertIntEquals(tc, 3, multiple_calculations);

    rc = zsocket_poll(tmp_socket_recv, 5000);
    CuAssertTrue(tc, rc);
    zmsg_t *foo = zmsg_recv(tmp_socket_recv); // get messages off sockets
    zmsg_destroy(&foo);
    rc = zsocket_poll(tmp_socket_recv, 5000);
    CuAssertTrue(tc, rc);
    zmsg_t *bar = zmsg_recv(tmp_socket_recv);
    zmsg_destroy(&bar);

    teardown();
}

void Test_h_mhash_all_states_checked(CuTest *tc) {
    init();
    send_message_all_states_checked();
    int rc = h_mhash(NULL, NULL, NULL);
    CuAssertIntEquals(tc, 2, trie_get_checked_states_count());
    CuAssertIntEquals(tc, 2, trie_count_elements());
    recv_message_all_states_checked(tc);
    CuAssertIntEquals(tc, -1, rc);
    teardown();
}
