#include "../include/zmq.h"
#include "../include/czmq.h"

#include "CuTest.h"
#include "../worker/whash_recv.h"
#include "../common/hashtrie.h"

extern zctx_t *ctx;
static void *tmp_socket;

extern int port;

static void init(void) {
    zmq_pollitem_t ignored;
    init_recv_hashes(ctx, &ignored, "localhost", port);
    init_hashtrie();
    port++;
}

static void send_message(void) {
    tmp_socket = zsocket_new(ctx, ZMQ_PUB);
    zsocket_bind (tmp_socket, "tcp://*:%d", port-1); // port is inc'd in init
    sleep(1);

    zmsg_t *msg = zmsg_new();
    char *fakedigest = "\1foobartest1234567890";
    char *fakesucc   = "\0foobartest9876543210";
    zmsg_addmem(msg, fakedigest, 21);
    zmsg_addmem(msg, fakesucc,   21);
    zmsg_send(&msg, tmp_socket);
    zmsg_destroy(&msg);
    sleep(1);
}

static void teardown(void) {
    zsocket_destroy(ctx, tmp_socket);
    teardown_recv_hashes(ctx);
    teardown_hashtrie();
}

void Test_h_hash(CuTest *tc) {
    init();
    send_message();

    CuAssertFalse (tc, trie_contains_processed("foobartest1234567890"));
    CuAssertFalse(tc, trie_contains_processed("foobartest9876543210"));
    CuAssertFalse (tc, trie_contains("foobartest1234567890"));
    CuAssertFalse (tc, trie_contains("foobartest9876543210"));

    CuAssertIntEquals(tc, 0, trie_get_checked_states_count());
    CuAssertIntEquals(tc, 0, trie_count_elements());

    h_hash(NULL, NULL, NULL);

    CuAssertTrue (tc, trie_contains_processed("foobartest1234567890"));
    CuAssertFalse(tc, trie_contains_processed("foobartest9876543210"));
    CuAssertTrue (tc, trie_contains("foobartest1234567890"));
    CuAssertTrue (tc, trie_contains("foobartest9876543210"));

    CuAssertIntEquals(tc, 1, trie_get_checked_states_count());
    CuAssertIntEquals(tc, 2, trie_count_elements());

    teardown();
}

CuSuite *WHashGetSuite() {
    CuSuite *suite = CuSuiteNew();
    SUITE_ADD_TEST(suite, Test_h_hash);
    return suite;
}
