#include <stdlib.h>
#include <string.h>

#include "../../../include/zmq.h"
#include "../../../include/czmq.h"
#include "../../../common/ring.h"
#include "../../../common/queue.h"

#include "wdebug.h"
#include "whash_recv.h"
#include "wbeat.h"

extern wQueue *local_queue;

/* this queue buffers workpackages until the master acknowledges
 * via a hash package, that the state is checked */
static wQueue *durable_q = NULL;

static void *hash_recv = NULL;

void init_whash_recv(zctx_t *ctx, zmq_pollitem_t *poller, char *masteraddr, int port) {
    assert(hash_recv == NULL);
    hash_recv = zsocket_new(ctx, ZMQ_SUB);
    char *filter = "";
    zmq_setsockopt (hash_recv, ZMQ_SUBSCRIBE, filter, strlen (filter));
    zsocket_connect(hash_recv,"tcp://%s:%i", masteraddr, port);
    durable_q = init_queue(0);

    *poller = (zmq_pollitem_t) { hash_recv, 0, ZMQ_POLLIN };
}

void teardown_whash_recv(zctx_t *ctx) {
    assert(hash_recv != NULL);
    zsocket_destroy(ctx, hash_recv);
    hash_recv = NULL;
    teardown_queue(&durable_q);
}

static void clean_up_after_ack(tCell *t) {
    ring_free(local_queue->env, t); // TODO: maybe write into the other file.
    free(t->term);
    free(t->digest);
    free(t);
}

static void consume_hash_msg(void) {
    zmsg_t *msg = zmsg_recv(hash_recv);
    //set_master_alive();
    zframe_t *frame;
    while (!is_empty(durable_q) && (frame = zmsg_pop(msg)) != NULL) {
        assert(21 == zframe_size(frame));
        char *data = zframe_strdup(frame);
        if (data[0]) {
            if (0 == memcmp(data+1, durable_q->first->digest, 20)) {
                assert(durable_q->first->term);
                tCell *cell = dequeue(durable_q);
                clean_up_after_ack(cell);
            }
        }
        free(data);
        zframe_destroy(&frame);
    }
    zmsg_destroy(&msg);
}

void clear_durable_queue(void) {
    DEBUG_PRINT(DEBUG_NORMAL,,,":enter :clear-durable-queue");
    while (!is_empty(durable_q)) {
        consume_hash_msg();
    }
    DEBUG_PRINT(DEBUG_NORMAL,,,":exit :clear-durable-queue");
}

int h_whash(zloop_t *loop, zmq_pollitem_t *poller, void *arg) {
    int i = 0;
    while (zsocket_poll(hash_recv, 0) == true) {
        if (i++ % 200 == 0) {
            DEBUG_PRINT(DEBUG_NORMAL,,,":in :h_whash :iterations %d", i);
        }
        consume_hash_msg();
    }
    return 0;
}

void enqueue_durable(tCell *cell) {
    enqueue_cell(durable_q, cell);
}
