#include <assert.h>
#include <stdlib.h>

#include "../../../include/zmq.h"
#include "../../../include/czmq.h"

#include "../../../common/queue.h"
#include "../../../common/hashtrie.h"

#include "wstats.h"
#include "wdebug.h"

static void *send_stats = NULL;
static int id;

static char full_results = 0;
static double prolog_time;
static struct timeval prolog_starttime;
static long int global_starttime;
static int run = 0;

extern wQueue *local_queue;

void init_send_stats(zctx_t *ctx, char *masteraddr, char *workerid, int port) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :init_send_starts");
    sscanf(workerid, "%d", &id);

    assert(send_stats == NULL);
    send_stats = zsocket_new(ctx, ZMQ_PUSH);
    zsocket_connect(send_stats,"tcp://%s:%i", masteraddr, port+4);

    struct timeval tmpgt;
    gettimeofday(&tmpgt, NULL);
    global_starttime = ((int)tmpgt.tv_sec * 1000) + (int)(tmpgt.tv_usec/1000);
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :init_send_starts");
}

void teardown_send_stats(zctx_t *ctx) {
    assert(send_stats);
    zsocket_destroy(ctx, send_stats);
    send_stats = NULL;
    run = 0;
}

void send_statistics(char *digest) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,
                char d[41]; print_readable_digest((unsigned char*)digest, d);,,
                ":enter :send_statistics :digest :%s", d);
    zmsg_t *msg = zmsg_new();
    //zmsg_addstr(msg, "%d %d", id, local_queue->c);
    zmsg_addstr(msg, "%d", id);
    zmsg_addstr(msg, "%d", local_queue->c);

    if (full_results) {
        zmsg_addmem(msg, digest, 20);
        zmsg_addstr(msg, "%.3lf", prolog_time);
        struct timeval tmpgt;
        gettimeofday(&tmpgt, NULL);
        int timestamp = ((int)tmpgt.tv_sec * 1000) + (int)(tmpgt.tv_usec/1000);
        zmsg_addstr(msg, "%d", timestamp - global_starttime);
        zmsg_addstr(msg, "%d", run++);
    }

    zmsg_send(&msg, send_stats);
    zmsg_destroy(&msg);
    prolog_time = 0;
    full_results = 0;
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :send_statistics");
}

void start_timer() {
    full_results = 1;
    gettimeofday(&prolog_starttime, NULL);
}

void stop_timer() {
    struct timeval tv;
    gettimeofday(&tv, NULL);
    suseconds_t time1 = prolog_starttime.tv_usec;
    time_t   sectime1 = prolog_starttime.tv_sec;
    suseconds_t time2 = tv.tv_usec;
    time_t   sectime2 = tv.tv_sec;
    prolog_time = (double) (time2 - time1)/1000 + (double)(sectime2 - sectime1) * 1000;    
}

int send_heartbeat(zloop_t *loop, int timer_id, void *args) {
    if (is_empty(local_queue)) {
        zmsg_t *msg = zmsg_new();
        zmsg_addstr(msg, "%d", id);
        zmsg_send(&msg, send_stats);
        zmsg_destroy(&msg);
    }
    return 0;
}
