#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sicstus/sicstus.h>

#include "../../../include/zmq.h"
#include "../../../include/czmq.h"

#include "../../../common/queue.h"
#include "../../../common/hashtrie.h"
#include "../../../common/sha.h"
#include "../../../common/ring.h"

#include "wwork.h"
#include "whash_send.h"
#include "whash_recv.h"
#include "wstats.h"
#include "wtok.h"
#include "wdebug.h"

static void work_hard (void);
static void clean_up_after_work(tCell *t);
static void check_state(char *term);

static SP_pred_ref pred_zmq_check;

extern wQueue *local_queue;


static void *send_tick = NULL;
static void *recv_tick = NULL;

void tick(void) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :tick");
    zstr_send(send_tick, "TOCK");
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :tick");
}


void init_tok(zctx_t *ctx, zmq_pollitem_t *poller, SP_pred_ref check_pred_ref) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :init_tok");
    send_tick = zsocket_new(ctx, ZMQ_PUSH);
    recv_tick = zsocket_new(ctx, ZMQ_PULL);

    int tickport = zsocket_bind(recv_tick, "tcp://*:*");
    char prot[22]; 
    sprintf(prot,"tcp://localhost:%i",tickport); 
    
    zsocket_connect(send_tick, prot);
    *poller = (zmq_pollitem_t) { recv_tick, 0, ZMQ_POLLIN };
    pred_zmq_check = check_pred_ref;
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :init_tok");
}


void teardown_tok(zctx_t *ctx) {
    zsocket_destroy(ctx, send_tick);
    send_tick = NULL;
    zsocket_destroy(ctx, recv_tick);
    recv_tick = NULL;
}



int h_tick (zloop_t *loop, zmq_pollitem_t *poller, void *arg) {
    zmsg_t *msg = zmsg_recv(recv_tick);
    assert(msg != NULL);
    zmsg_destroy(&msg);
    work_hard(); 
    return 0;
}


static void clean_up_already_processed(tCell *t) {
    ring_free(local_queue->env, t);
    free(t->term);
    free(t->digest);
    free(t);
}

static void work_hard (void) {
    if (is_empty(local_queue)) {
        return;
    }

    tCell *t = dequeue(local_queue);

    if (!trie_contains_processed(t->digest)) {
        init_successors(); // successors and results

        check_state(t->term);


        trie_put_checked(t->digest);

        send_successors(t);
        enqueue_durable(t);
    } else {
        clean_up_already_processed(t);
    }
    send_statistics(t->digest);
    tick(); 

}


static void check_state(char *term) {
    SP_pred_ref pred = pred_zmq_check;
    SP_term_ref val = SP_new_term_ref();

    SP_put_address(val, term);

    start_timer();
    int status = SP_query_cut_fail(pred, val);
    stop_timer();
    assert(status == SP_SUCCESS);
}

void put_successor(char *term, int len,
                   char *digest, SP_term_ref ignored) {

    if (!trie_contains(digest)) {
        char *termi = (char*) malloc(len);
        memcpy(termi,term,len);

        enqueue(local_queue, termi, len, digest);
        add_queued_digest(digest);
        trie_put_unchecked(digest);
    } else {
        free(digest);
    }
}

char *shachar(char *term, int len) {
    char *digest = (char*) calloc(20, sizeof(char));
    assert(digest);
    sha1(term, len, digest);
    return digest;
}
