% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(units_alias, [
                         unit_alias/2,
                         init_user_defined_alias/0
                       ]).

:- use_module(probsrc(pragmas)).
:- use_module(probsrc(error_manager)).

:- use_module(unit_parser).

:- use_module(probsrc(module_information)).
:- module_info(group,plugin_units).
:- module_info(description,'Units Plugin: Short aliases for commonly used units.').

init_user_defined_alias :-
    findall(alias(AliasName,Definition),
            global_pragma(unit_alias,[AliasName|Definition]),
            List),
    init_user_defined_alias2(List).
    
init_user_defined_alias2([]).
init_user_defined_alias2([alias(AliasName,Definition)|T]) :-
    unit_args_to_list(Definition,List),
    ((\+ unit_alias(AliasName,List2), List2 \= List) -> add_error(incorrect_unit_definition, 'Unit Alias used twice: ', AliasName) ;  true),
    assert(unit_alias(AliasName,List)),
    init_user_defined_alias2(T).
    

:- dynamic unit_alias/2.

% lengths
unit_alias(mm, [[-3,m,1]]).
unit_alias(cm, [[-2,m,1]]).
unit_alias(dm, [[-1,m,1]]).
unit_alias(m, [[0,m,1]]).
unit_alias(km, [[3,m,1]]).

% railway-specific units
unit_alias(hm, [[0,hm,1]]). % half-meter

% volume
unit_alias(l, [[-3,m,3]]).

% time
unit_alias(s, [[0,s,1]]).
unit_alias(mins, [[0,mins,1]]).
unit_alias(h, [[0,h,1]]).
unit_alias(d, [[0,d,1]]).

% speed
unit_alias(mps, [[0,m,1],[0,s,-1]]).
unit_alias(kmh, [[3,m,1],[0,h,-1]]).

% temperature
unit_alias(degC, [[0,degC,1]]).
unit_alias('K', [[0,'K',1]]).
