% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


:- module(b_compiler,[b_compile/5]).

:- use_module(library(lists)).

:- use_module(self_check).
:- use_module(bsyntaxtree).
:- use_module(error_manager).
:- use_module(btypechecker,[fasttype/2]).
:- use_module(b_interpreter).
:- use_module(custom_explicit_sets).

:- use_module(module_information,[module_info/2]).
:- module_info(group,interpreter).
:- module_info(description,'This module compiles set comprehensions into closures; making them independent of the state of the B machine.').

/* compile boolean expression into a closure where the local state and the global
 state has been incorporated, but any parameter is left in the closure */

:- public test_bexpr/3, test_bexpr2/3.
test_bexpr(Expr,[bind(cc,int(1)),bind(nn,int(2))], [bind(db,[(int(1),int(2))])]) :-
    fasttype( +conjunct( +conjunct( +member( global('Name')<<identifier(nn), +identifier('Name')),
                                    +member( global('Code')<<identifier(cc), +identifier('Code'))),
                         +not_member( global('Name')<<identifier(nn),
                                      +domain(set(couple(global('Name'),global('Code')))<<identifier(db)))), Expr).

test_bexpr2(Expr,[],[]) :-
    fasttype( +implication( +equal( seq(any)<<identifier(ss), +empty_sequence),
                            +equal( set(any)<<identifier(res), +empty_set)), Expr).

:- assert_must_succeed(( b_compiler:test_bexpr(E,L,S),
                         b_compiler:b_compile(E,[nn],L,S,_R)  )).
:- assert_must_succeed(( b_compiler:test_bexpr2(E,L,S),
                         b_compiler:b_compile(E,[ss,res],L,S,_R)  )).
 
%:- assert_pre(b_compiler:b_compile_boolean_expression(E,Parameters,LS,S,_), 
%                  (type_check(E,boolean_expression),type_check(Parameters,list(variable_id)),
%                   type_check(LS,store),type_check(S,store))).
%:- assert_post(b_compiler:b_compile_boolean_expression(_,_,_,_,E), type_check(E,boolean_expression)).

/* probably not worthwhile: 
b_compile(TExpr,Parameters,LS,S,NTExpr) :- get_texpr_info(TExpr,Info),
   (memberchk(compiled(Parameters),Info) -> NTExpr = TExpr  ,print(already_compiled(Parameters)),nl, translate:print_bexpr(TExpr),nl
     ; b_compile0(TExpr,Parameters,LS,S,b(E,T,I)), translate:print(compiling(Parameters)),nl, translate:print_bexpr(TExpr),nl,
       NTExpr = b(E,T,[compiled(Parameters)|I])).
*/

b_compile(TExpr,Parameters,LS,S,NTExpr) :- check_is_id_list(Parameters),
   %  print('compile : '),translate:print_bexpr(TExpr),nl,  statistics(runtime,[Start,_]),%%
   if(b_compile1(TExpr,Parameters,LS,S,NTExpr,eval,_FullyKnown),true,
     (add_internal_error('Compilation failed: ',Parameters),
       NTExpr = TExpr)
     ). %, check_infos(TExpr,NTExpr).
   %, print('compiled: '), translate:print_bexpr(NTExpr), print(' '),statistics(runtime,[Stop,_]), T is Stop-Start, print(T), print(' ms  '),nl, bsyntaxtree:check_ast(NTExpr). % , print(NTExpr),nl. check_ast

/*
check_infos(Old,New) :- bsyntaxtree:get_texpr_info(Old,IO),
    bsyntaxtree:get_texpr_info(New,IN),
    member(X,IO), \+ member(X,IN),
    bsyntaxtree:important_info(X),
    print(check_infos(Old,New)),nl,
    print('Not copied: '), print(X),fail.
check_infos(_,_). */

:- use_module(bsyntaxtree,[always_well_defined/1]).
%:- use_module(hit_profiler,[add_profile_hit/1]).
b_compile1(TExpr,Parameters,LS,S,ResultTExpr,Eval,FullyKnown1) :-
    TExpr = b(Expr,Type,Infos),
    NTExpr = b(NewUntypedExpr,Type,NewInfos),
    %remove_bt(TExpr,Expr,NewUntypedExpr,NTExpr), % remove top-level Type Information
    %hit_profiler:add_profile_hit(Expr),
    b_compile2(Expr,Parameters,LS,S,NewUntypedExpr,Eval,FullyKnown),
    update_infos(NewUntypedExpr,Infos,Parameters,NewInfos),
    % print('  compiled : '),translate:print_bexpr(Expr),nl,
    % print(res(FullyKnown,Eval,NTExpr)),nl,
    ((FullyKnown=true, evaluate_this(Eval,NTExpr))
     ->  %  print('eval : '),translate:print_bexpr(NTExpr),nl, %
        ((nonvar(NewUntypedExpr),
          b_interpreter:b_compute_expression(NTExpr,LS,S,ResultValue))
          -> %print(result(NTExpr,ResultValue)),nl,
             %remove_bt(TExpr,Expr,value(ResultValue),ResultTExpr)
             get_texpr_type(TExpr,TT),ResultTExpr = b(value(ResultValue),TT,[]),
             FullyKnown1 = FullyKnown
              %, print(compiled_value(ResultTExpr)),nl
          ;  translate:translate_bexpression_with_limit(NTExpr,TTS),
             add_internal_error('Undefined Value: ',b_compile1(TTS,LS,S)),
             ResultTExpr=NTExpr, FullyKnown1=false
        )
       ;    % ((FullyKnown=true, NTExpr\=b(value(_),_,_)) -> format('not eval: ~w (~w)~n',[NTExpr,Eval]) ; true),
            ResultTExpr=NTExpr, 
            (ResultTExpr = b(value(_),_,_) -> FullyKnown1=FullyKnown ; FullyKnown1=false)
     ).

:- use_module(library(ordsets)).
update_infos(forall(EParas,LHS,RHS),Infos,Parameters,NewInfos) :- 
  % forall also has used_ids field as it may call exists in negative context
  conjunct_predicates([LHS,RHS],Cond),
  update_infos_aux(EParas,Cond,Infos,Parameters,NewInfos),!.
update_infos(exists(EParas,Cond),Infos,Parameters,NewInfos) :-
  update_infos_aux(EParas,Cond,Infos,Parameters,NewInfos),!.
update_infos(_,I,_,I).

update_infos_aux(EParas,Cond,Infos,Parameters,NewInfos) :- 
  select_used_ids(UsedIds,Infos,I1,EParas,Cond),
  UsedIds \= [],!,
  % remove those used identifiers which will be compiled into the predicate
  sort(Parameters,SParas),
  ord_intersection(UsedIds,SParas,NewUsedIds), 
  %print(update(Parameters,UsedIds,I1,new(NewUsedIds))),nl,
  NewInfos = [used_ids(NewUsedIds)|I1].

select_used_ids(UsedIds,Infos,I1,_,_) :- select(used_ids(UsedIds),Infos,I1),!.
  % check_used_ids_info(Parameters,Condition,UsedIds,b_compiler)
select_used_ids(UsedIds,Infos,Infos,_Parameters,Condition) :-
   %add_error(bcompiler,'Expected information of used identifiers in exists operation information : ',Parameters:Infos), %% missing info can happen due to simplifcation rules below !
   bsyntaxtree:find_identifier_uses(Condition, [], UsedIds). % ,print(used(UsedIds,in_exist(Parameters))),nl.
 % TO DO: avoid re-computing used-ids; we should refactor b_compile2 to return new Info field
 
%is_pred(b(_,pred,_)).


evaluate_this(eval,NTExpr) :- !, worth_evaluating(NTExpr).
evaluate_this(_,NTExpr) :- always_wd_and_efficient(NTExpr).


always_wd_and_efficient(b(X,_,_Info)) :- wd_and_efficient(X).

% things which are very quick to compute
wd_and_efficient(sequence_extension(_)) :- !.
wd_and_efficient(integer_set(_)) :- !. % will be converted into value(global_set(_))
%% wd_and_efficient(value(_)) :- !. % is already computed
% the following will reduce the size of the representation; usually a good thing;
% we assume only avl_sets apprear here; inner set comprehensions are never computed into symbolic form by b_compile (see known_value below):
wd_and_efficient(range(_)) :- !.
wd_and_efficient(domain(_)) :- !.
wd_and_efficient(domain_restriction(_,_)) :- !.
wd_and_efficient(domain_subtraction(_,_)) :- !.
wd_and_efficient(range_restriction(_,_)) :- !.
wd_and_efficient(range_subtraction(_,_)) :- !.
wd_and_efficient(intersection(_,_)) :- !.
wd_and_efficient(set_subtraction(_,_)) :- !.
wd_and_efficient(image(_,_)) :- !.
% what about union: usually the union of two avl_sets will be smaller than keeping them separate
% some other efficient operators:
wd_and_efficient(X) :- worth_it_int(X),!.
wd_and_efficient(X) :- worth_it_other(X).

check_wd(NTExpr) :- bsyntaxtree:always_well_defined(NTExpr). % ,print(wd(NTExpr)),nl.
  %(bsyntaxtree:always_well_defined(NTExpr) -> true ; print(not_guaranteed_wd(NTExpr)),nl).

worth_evaluating(b(X,Type,Info)) :- worth_it_type(Type,X,Info).
  % (worth_it_type(Type,X,Info) -> true ;  X=value(_) -> true ; print(not_evaluating(X)),nl,fail).
worth_it_type(integer,X,Info) :- worth_it_int_wd(X),!, check_wd(b(X,integer,Info)).
worth_it_type(T,function(A,B),Info) :- !,
 (check_wd(b(function(A,B),T,Info)) -> true
   ; fail). %print('not_precompiling_function '),translate:print_bexpr(A), print(' @ '), translate:print_bexpr(B),nl,fail).
worth_it_type(integer,X,_) :- worth_it_int(X),!.
worth_it_type(T,X,Info) :- worth_it_wd(X),!, check_wd(b(X,T,Info)).
worth_it_type(set(_),X,_) :- worth_it_set(X),!.
worth_it_type(seq(_),X,_) :- worth_it_set(X),!.
worth_it_type(_,X,_) :- worth_it_other(X).

% TO DO: we should use the predicate has_wd_condition from b_ast_cleanup

% integer operations that could generate WD-errors
worth_it_int_wd(div(_,_)).
worth_it_int_wd(max(_)).
worth_it_int_wd(min(_)).
worth_it_int_wd(mod(_,_)).
worth_it_int_wd(power_of(_,_)).

% other operations that could generate WD-errors
worth_it_wd(first(_)).
worth_it_wd(last(_)).
worth_it_wd(tail(_)).
worth_it_wd(front(_)).
worth_it_wd(restrict_front(_,_)).
worth_it_wd(restrict_tail(_,_)).
worth_it_wd(rel_iterate(_,_)). %% could be expensive
worth_it_wd(general_intersection(_)).
worth_it_wd(general_concat(_)).

% to do: check other operators that could be not well-defined !

worth_it_int(unary_minus(_)).
worth_it_int(add(_,_)).
worth_it_int(minus(_,_)).
worth_it_int(multiplication(_,_)).
worth_it_int(max_int).
worth_it_int(min_int).
worth_it_int(card(_)).
worth_it_int(size(_)).

%worth_it_set(empty_set). % treated below
%worth_it_set(empty_sequence). % treated below
worth_it_set(closure(_)).  %% ?? need to be more careful here; can be expensive --> need to keep track which parts need definitely to be evaluated
worth_it_set(integer_set(_)). % will be converted into value(global_set(_))
worth_it_set(interval(_,_)).
worth_it_set(reflexive_closure(_)).
worth_it_set(reverse(_)).
worth_it_set(domain(_)). worth_it_set(range(_)).
worth_it_set(union(_,_)). worth_it_set(intersection(_,_)).  worth_it_set(set_subtraction(_,_)).
worth_it_set(image(_,_)). worth_it_set(composition(_,_)).
worth_it_set(domain_restriction(_,_)). worth_it_set(domain_subtraction(_,_)).
worth_it_set(range_restriction(_,_)). worth_it_set(range_subtraction(_,_)).
worth_it_set(direct_product(_,_)). worth_it_set(parallel_product(_,_)).
worth_it_set(iteration(_,_)).
worth_it_set(set_extension(_)).
worth_it_set(sequence_extension(_)).
worth_it_set(rev(_)).
worth_it_set(concat(_,_)).
worth_it_set(seq(_)). % will be kept symbolic anyway
worth_it_set(seq1(_)). % will be kept symbolic anyway
worth_it_set(iseq(_)). % will be kept symbolic anyway
worth_it_set(iseq1(_)). % will be kept symbolic anyway
worth_it_set(perm(_)). % will be kept symbolic anyway
worth_it_set(pow_subset(_)). % will be kept symbolic anyway
worth_it_set(pow1_subset(_)). % will be kept symbolic anyway
worth_it_set(fin_subset(_)). % will be kept symbolic anyway
worth_it_set(fin1_subset(_)). % will be kept symbolic anyway
worth_it_set(general_union(_)).
worth_it_set(first_of_pair(_)).  % can also produce set
worth_it_set(second_of_pair(_)). % can also produce set
worth_it_set(cartesian_product(_,_)). % will usually be kept symbolic


worth_it_other(boolean_true).
worth_it_other(boolean_false).
worth_it_other(string(_)).
worth_it_other(first_of_pair(_)).
worth_it_other(second_of_pair(_)).
worth_it_other(couple(_,_)).
worth_it_other(record_field(_,_)).
%worth_it_other(value(_)). % does not have to be evaluted; already a value

/*
% To do: distinguish which parts definitely need to be evaluated
%dont_eval_subexpressions(member).  % special membership code can be used
%dont_eval_subexpressions(not_member).  % special membership code can be used
dont_eval_subexpressions(disjunct(_,_)). % it may not be necessary to eval RHS
dont_eval_subexpressions(implication(_,_)).  % it may not be necessary to eval RHS
*/

:- use_module(kernel_objects,[equal_object/3]).
:- use_module(kernel_mappings,[binary_boolean_operator/3]).
b_compile2(Exp,Parameters,LS,S,NExpr,Eval,FullyKnown) :- var(Exp), !,
    add_internal_error('Variable Expression: ',b_compile2(Exp,Parameters,LS,S,NExpr,Eval,FullyKnown)),
    NExpr=Exp, FullyKnown=false.
b_compile2(empty_set,_,_LS,_S,NExpr,_Eval,FullyKnown) :- !, NExpr=value([]), FullyKnown=true.
b_compile2(empty_sequence,_,_LS,_S,NExpr,_Eval,FullyKnown) :- !, NExpr=value([]), FullyKnown=true.
b_compile2(value(Val),_Parameters,_LS,_S,NExpr,_Eval,FullyKnown) :- !,
    NExpr=value(Val),
    (known_value(Val) -> % ground(Val) -> known_value ?
     FullyKnown=true ; FullyKnown=false).
b_compile2(integer(Val),_Parameters,_LS,_S,NExpr,_Eval,FullyKnown) :- !,
    NExpr=value(int(Val)), (number(Val) -> FullyKnown=true ; FullyKnown=false).
b_compile2(lazy_lookup_pred(Id),Parameters,LS,S,NExpr,_Eval,FullyKnown) :- !,
    ( memberchk(Id,Parameters) ->  %print(lazy_lookup_pred_id_as_parameter(Id,Parameters)),nl,
        NExpr = lazy_lookup_pred(Id), FullyKnown=false
    ; otherwise -> store:lookup_value_for_existing_id(Id,LS,S,LazyValue),
                   LazyValue = (pred_true,Value),
                   NExpr = value(Value), (known_value(Value) -> FullyKnown=true ; FullyKnown=false)
    ).
b_compile2(lazy_lookup_expr(Id),Parameters,LS,S,NExpr,_Eval,FullyKnown) :- !,
    ( memberchk(Id,Parameters) -> %print(lazy_lookup_expr_id_as_parameter(Id,Parameters)),nl,
        NExpr = lazy_lookup_expr(Id), FullyKnown=false
    ; otherwise -> store:lookup_value_for_existing_id(Id,LS,S,LazyValue),
                   LazyValue = (pred_true,Value),
                   NExpr = value(Value), (known_value(Value) -> FullyKnown=true ; FullyKnown=false)
    ).
b_compile2(identifier(Id),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    ( memberchk(Id,Parameters) ->
        NExpr = identifier(Id), FullyKnown=false
    ; b_interpreter:lookup_value_in_store_and_global_sets(Id,LS,S,Value) ->
        NExpr = value(Value), (known_value(Value) -> FullyKnown=true ; FullyKnown=false)
    ; add_internal_error('Compilation of identifier failed: ',
                          b_compile2(identifier(Id),Parameters,LS,S,NExpr,Eval,FullyKnown)), 
       NExpr = identifier(Id), FullyKnown=false
    ).
% treat a few common cases, more efficiently than syntaxtransformation
b_compile2(equal(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    FullyKnown=false, % predicates never evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,FullyKnownA),
    b_compile1(B,Parameters,LS,S,NExprB,Eval,FullyKnownB),
    ((FullyKnownA=true, FullyKnownB=true,
      NExprA = b(value(VA),_,_), NExprB = b(value(VB),_,_))
      ->  %print(computing_eq(VA,VB)),nl,
          (kernel_objects:equal_object(VA,VB,b_compile2_1) -> NExpr = truth ; NExpr = falsity)
          %, print(result(NExpr)),nl
       ;  generate_equality(NExprA,NExprB,NExpr)
    ).
b_compile2(not_equal(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    FullyKnown=false, % predicates never evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,FullyKnownA),
    b_compile1(B,Parameters,LS,S,NExprB,Eval,FullyKnownB),
    ((FullyKnownA=true, FullyKnownB=true,
      NExprA = b(value(VA),_,_), NExprB = b(value(VB),_,_))
      ->  %print(computing_neq(VA,VB)),nl,
          (kernel_objects:equal_object(VA,VB,b_compile2_2) -> NExpr = falsity ; NExpr = truth)
          %, print(result(NExpr)),nl
       ;  NExpr = not_equal(NExprA,NExprB)
    ).
b_compile2(member(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    FullyKnown=false, % predicates never evaluated in b_compile1
    b_compile1(B,Parameters,LS,S,NExprB,Eval,_),
    ((NExprB = b(value(X),_,_),X==[], always_well_defined(A)) -> NExpr = falsity
     ;  b_compile1(A,Parameters,LS,S,NExprA,Eval,_),
        % check if we can decide membership (e.g., 1:{1})
        (quick_test_membership(NExprA,NExprB,PRes)
          -> convert_pred_res(PRes,NExpr)
          % TO DO: maybe optimize x: {y|P(y)}  --> P(x) ; by commenting in code below
          % However, ensure test 273 succeeds and closure equality correctly detected (e.g., for BV16 = {bt|bt : BIT_VECTOR & bv_size(bt) = 16})
          % ; (get_texpr_id(NExprA,IDA),
          %    get_comprehension_set(NExprB,IDB,Pred), get_texpr_info(NExprB,Info),
          %     print(rename(IDB,IDA,Info)),nl,
          %    bsyntaxtree:rename_bt(Pred,[rename(IDB,IDA)],PredA)) -> get_texpr_expr(PredA,NExpr)
           ; NExpr = member(NExprA,NExprB))
         %, print('compile: '),print(NExpr),nl
    ).
% TO DO: also optimize things like: (ev1 |-> 0) |-> eg : #221:{((1|->0)|->3),((2|->0)|->0),...,((233|->0)|->218),((234|->0)|->228)}

b_compile2(not_member(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    FullyKnown=false, % predicates never evaluated in b_compile1
    b_compile1(B,Parameters,LS,S,NExprB,Eval,_),
    ((NExprB = b(value(X),_,_),X==[], always_well_defined(A)) -> NExpr = truth
     ;  b_compile1(A,Parameters,LS,S,NExprA,Eval,_),
        % check if we can decide membership
        (quick_test_membership(NExprA,NExprB,PRes)
          -> convert_neg_pred_res(PRes,NExpr) ; NExpr = not_member(NExprA,NExprB))
          % TO DO: maybe optimize x/: {y|P(y)}  --> not(P(x))
        %,print('compile: '),print(NExpr),nl
    ).
b_compile2(function(Fun,B),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    (is_extension_function(Fun) -> Eval1=false % extension function applications treated lazily
     % TO DO: first evaluate B, if known then only evaluate relevant part of extension function
     ; Eval1=Eval),
    b_compile1(Fun,Parameters,LS,S,NExprA,Eval1,FullyKnown1),
    combine_fully_known(FullyKnown1,FullyKnown2,FullyKnown12),
    b_compile1(B,Parameters,LS,S,NExprB,Eval,FullyKnown2),
    %tools_printing:print_term_summary(compiled_function(FullyKnown1,FullyKnown2,NExprA,NExprB)),nl,
    ((FullyKnown2=true, preferences:preference(find_abort_values,false),
      % check if we have enough information to apply a partially specified function as a list
      % e.g. if the function is [(int(1),A),(int(2),B)] and the argument is int(2): we can apply the function without knowing A or B;  can have a dramatic importance when expanding unversal quantifiers !
      can_apply_partially_specified_function(NExprA,NExprB,ResultValue))
      -> NExpr = value(ResultValue),
         (FullyKnown12=true -> FullyKnown=true
           ; known_value(ResultValue) -> FullyKnown=true
           ; FullyKnown = false)
      ;  NExpr = function(NExprA,NExprB), FullyKnown = FullyKnown12).

b_compile2(conjunct(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    FullyKnown=false, % predicates never automatically evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,_),
    (is_falsity(NExprA) -> NExpr = falsity
     ;
    % TO DO: propagate static information from A to B, e.g. if A :: x=10  -> add x=10 to LS/Parameters
    % also: try and detect unsatisfiable predicates beforehand
      b_compile1(B,Parameters,LS,S,NExprB,Eval,_),
      (is_falsity(NExprB) -> NExpr = falsity
       ; is_truth(NExprB) -> get_texpr_expr(NExprA,NExpr)
       ; is_truth(NExprA) -> get_texpr_expr(NExprB,NExpr)
       ; NExpr = conjunct(NExprA,NExprB))
    ).
b_compile2(disjunct(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    FullyKnown=false, % predicates never automatically evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,_),
    (is_truth(NExprA) -> NExpr = truth
     ; b_compile1(B,Parameters,LS,S,NExprB,false,_), % does not have to be executed in a successful branch: set Eval to false to avoid computing expensive expressions which may not be needed
       (is_truth(NExprB) -> NExpr = truth
       ; is_falsity(NExprA) -> get_texpr_expr(NExprB,NExpr)
       ; is_falsity(NExprB) -> get_texpr_expr(NExprA,NExpr)
       ; NExpr = disjunct(NExprA,NExprB))
    ).
b_compile2(implication(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    FullyKnown=false, % predicates never automatically evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,_),
    (is_falsity(NExprA) -> NExpr = truth
     ; is_truth(NExprA) -> b_compile1(B,Parameters,LS,S,NExprB,false,_), get_texpr_expr(NExprB,NExpr)
     ; NExpr = implication(NExprA,NExprB),
       b_compile1(B,Parameters,LS,S,NExprB,false,_) % does not have to be executed in a succesful branch: set Eval to false to avoid computing expensive expressions which may not be needed
    ).
b_compile2(negation(A),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    FullyKnown=false, % predicates never automatically evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,_),
    ( get_negated_expr(NExprA,NegatedA) -> NExpr = NegatedA
      ; NExpr = negation(NExprA)
    ).

/* b_compile2(comprehension_set(CParas,Body),Parameters,LS,S,NExpr,_Eval,FullyKnown) :-
     sort(Parameters,SParas),
     b_ast_cleanup:not_occurs_in_predicate(SParas,Body), %can only be applied if Body does not reference Parameters
     !,
    b_generate_inner_closure_if_necessary(Parameters,CParas,Body,LS,S,Result), % will itself call b_compile
    NExpr = value(Result),
    (ground(Result) -> FullyKnown=true ; FullyKnown=false). */
b_compile2(exists(ExistsPara,Pred),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    FullyKnown=false, % predicates never automatically evaluated in b_compile1
    get_texpr_ids(ExistsPara,AtomicIds),
    append(Parameters,AtomicIds,NParameters),
    b_compile1(Pred,NParameters,LS,S,NPred,Eval,_FullyKnown1),
    (is_falsity(NPred) -> NExpr = falsity
     ; is_truth(NPred) -> NExpr = truth
     ; NExpr = exists(ExistsPara,NPred)).
b_compile2(forall(ForallPara,A,B),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    FullyKnown=false, % predicates never automatically evaluated in b_compile1
    get_texpr_ids(ForallPara,AtomicIds),
    append(Parameters,AtomicIds,NParameters),
    b_compile1(A,NParameters,LS,S,NA,Eval,_FullyKnownA),
    (is_falsity(NA) -> NExpr = truth
     ; b_compile1(B,NParameters,LS,S,NB,Eval,_FullyKnownB),
       (is_truth(NB) -> NExpr = truth
         ; NExpr = forall(ForallPara,NA,NB))
    ).
% some special cases to avoid calling syntaxtransformation
b_compile2(record_field(A,FieldName),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    b_compile1(A,Parameters,LS,S,NExprA,Eval,FullyKnown),
    NExpr = record_field(NExprA,FieldName).
b_compile2(couple(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    NExpr = couple(NExprA,NExprB),
    b_compile1(A,Parameters,LS,S,NExprA,Eval,FullyKnown1),
    combine_fully_known(FullyKnown1,FullyKnown2,FullyKnown),
    b_compile1(B,Parameters,LS,S,NExprB,Eval,FullyKnown2).
b_compile2(image(A,B),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    NExpr = image(NExprA,NExprB),
    b_compile1(A,Parameters,LS,S,NExprA,Eval,FullyKnown1),
    combine_fully_known(FullyKnown1,FullyKnown2,FullyKnown),
    b_compile1(B,Parameters,LS,S,NExprB,Eval,FullyKnown2).
b_compile2(assign(LHS_List,RHS_List),Parameters,LS,S,NExpr,Eval,FullyKnown) :- !,
    %print(compile_assign(LHS_List)),nl,
    NExpr = assign(New_LHS_List,NewRHS_List), FullyKnown = false,
    maplist(b_compile_lhs(Parameters,LS,S,Eval),LHS_List,New_LHS_List),
    b_compile_l(RHS_List,Parameters,LS,S,NewRHS_List,Eval,_FullyKnown2).
b_compile2(assign_single_id(ID,B),Parameters,LS,S,assign_single_id(ID,NExprB),Eval,FullyKnown) :- !,
    FullyKnown = false,
    b_compile1(B,Parameters,LS,S,NExprB,Eval,_).
b_compile2(Expr,Parameters,LS,S,NExpr,Eval,FullyKnown) :-
    functor(Expr,BOP,2),
    kernel_mappings:binary_boolean_operator(BOP,_,_),!, % slightly more efficient than syntaxtransformation
    arg(1,Expr,A),
    FullyKnown=false, % predicates never evaluated in b_compile1
    b_compile1(A,Parameters,LS,S,NExprA,Eval,FullyKnownA),
    arg(2,Expr,B),
    b_compile1(B,Parameters,LS,S,NExprB,Eval,FullyKnownB),
    (eval_binary_bool(FullyKnownA,NExprA,FullyKnownB,NExprB,BOP,Result)
      -> NExpr = Result
       ; functor(NExpr,BOP,2), arg(1,NExpr,NExprA), arg(2,NExpr,NExprB)
    ).
b_compile2(Expr,Parameters,LS,S,NExpr,Eval,FullyKnown) :- % GENERAL CASE
    syntaxtransformation(Expr,Subs,Names,NSubs,NExpr),
    get_texpr_ids(Names,Ids), append(Parameters,Ids,NParameters),
    b_compile_l(Subs,NParameters,LS,S,NSubs,Eval,FullyKnown).

b_compile_l([],_,_,_,[],_Eval,true).
b_compile_l([Expr|ExprRest],P,LS,S,[NExpr|NExprRest],Eval,FullyKnown) :-
    b_compile1(Expr,P,LS,S,NExpr,Eval,FullyKnown1),
    combine_fully_known(FullyKnown1,FullyKnown2,FullyKnown),
    b_compile_l(ExprRest,P,LS,S,NExprRest,Eval,FullyKnown2).

/* comment in if we want to compile inner comprehension sets:
:- use_module(closures,[construct_closure_if_necessary/4]).
b_generate_inner_closure_if_necessary(OuterParameters,ClosureParas,Condition,LocalState,State,Result) :-
    split_names_and_types(ClosureParas,Names,Types),
    append(OuterParameters,Names,FullNames),
    b_compile(Condition,FullNames,LocalState,State,ClosurePred),
    print('INNER COMPILE: '), translate:print_bexpr(ClosurePred),nl,
    construct_closure_if_necessary(Names,Types,ClosurePred,Result).
    */


% the LHS contains l-values which should not be looked up in the environment
% however, we need to deal with things like f(i) := RHS and compile i
b_compile_lhs(Parameters,LS,S,Eval,TExpr,NewTExpr) :- 
    TExpr = b(Expr,Type,Info), NewTExpr = b(NewExpr,Type,Info),
    b_compile_lhs_aux(Expr,Parameters,LS,S,NewExpr,Eval).

b_compile_lhs_aux(function(A,B),Parameters,LS,S,function(NewA,NewB),Eval) :- !, 
    b_compile_lhs(Parameters,LS,S,Eval,A,NewA), % we can have nexted function calls f(e)(g) :=
    b_compile1(B,Parameters,LS,S,NewB,Eval,_).
b_compile_lhs_aux(E,_,_,_,E,_). % other LHS should be identifier -> just copy


eval_binary_bool(true,b(value(ValA),_,[]),true,b(value(ValB),_,[]),BOP,Result) :-
   eval_binary_aux(BOP,ValA,ValB,Result).
   %print(eval(BOP,ValA,ValB,Result)),nl,nl.
eval_binary_aux(less,int(A),int(B),Result) :-
   (A < B -> Result = truth ; Result = falsity).
eval_binary_aux(greater,int(A),int(B),Result) :-
   (A > B -> Result = truth ; Result = falsity).
eval_binary_aux(less_equal,int(A),int(B),Result) :-
   (A =< B -> Result = truth ; Result = falsity).
eval_binary_aux(greater_equal,int(A),int(B),Result) :-
   (A >= B -> Result = truth ; Result = falsity).
% TO DO: other operators are subset, strict_subset, ...

get_negated_expr(b(E,pred,_),Res) :- negate_expr_aux(E,Res).
negate_expr_aux(falsity,truth).
negate_expr_aux(truth,falsity).
negate_expr_aux(member(X,S),not_member(X,S)).
negate_expr_aux(not_member(X,S),member(X,S)).
negate_expr_aux(subset(X,S),not_subset(X,S)).
negate_expr_aux(not_subset(X,S),subset(X,S)).
negate_expr_aux(subset_strict(X,S),not_subset_strict(X,S)).
negate_expr_aux(not_subset_strict(X,S),subset_strict(X,S)).
negate_expr_aux(equal(X,S),not_equal(X,S)).
negate_expr_aux(not_equal(X,S),equal(X,S)).
negate_expr_aux(less(X,S),greater_equal(X,S)).
negate_expr_aux(greater_equal(X,S),less(X,S)).
negate_expr_aux(less_equal(X,S),greater(X,S)).
negate_expr_aux(greater(X,S),less_equal(X,S)).


combine_fully_known(true,A,A).
combine_fully_known(false,_,false).

is_extension_function(b(A,_,_Info)) :- is_extension_function_aux(A).
is_extension_function_aux(set_extension(_)).
is_extension_function_aux(sequence_extension(_)).

:- use_module(bsyntaxtree,[is_set_type/2]).
% check if we can apply an argument to a partially specified function (as a list)
%can_apply_partially_specified_function(Fun,X,_) :- print(fun(Fun)),nl,print(val(X)),nl,fail.
can_apply_partially_specified_function(b(value(VAL),TYPE,_), b(value(X),TA,_),ResultValue) :-
    is_set_type(TYPE,couple(TA,_TB)),
    lookup_result(VAL,X,ResultValue).
:- use_module(kernel_equality,[equality_objects/3]).
lookup_result(VAR,_,_) :- var(VAR),!,fail.
lookup_result([(HFrom,HTo)|T],X,Result) :-
    kernel_equality:equality_objects(HFrom,X,PredRes),  %print(compare(PredRes,HFrom,X)),nl,
    nonvar(PredRes), % only proceed if we have enough information to determine the result of the comparison
    (PredRes = pred_true -> Result = HTo % we have found the value; we assume there is no other value later
      % TO DO: if find_abort_values=true: look later in the list if for all other HFrom PredRes=pred_false
      ; lookup_result(T,X,Result)).

% get_comprehension_set(b(SET,_,_),ID,PRED) :- get_comprehension_set_aux(SET,ID,PRED).
% get_comprehension_set_aux(comprehension_set([TID],PRED),ID,PRED) :-
%   get_texpr_id(TID,ID).
% get_comprehension_set_aux(value(V),ID,PRED) :- nonvar(V), V = closure([ID],_,PRED).

quick_test_membership(b(value(VA),_,_),b(value(VB),_,_),Result) :-
   quick_test_membership_aux(VB,VA,Result).
quick_test_membership_aux(VAR,_,_) :- var(VAR),!,fail.
quick_test_membership_aux([],_,pred_false).
quick_test_membership_aux(avl_set(AVL),X,Result) :-
  custom_explicit_sets:quick_test_avl_membership(AVL,X,Result). % we could also check that in case X is not ground but partially instantiated, whether there is a possible match in AVL (if not Result = pred_false)

convert_pred_res(pred_false,falsity).
convert_pred_res(pred_true,truth).
convert_neg_pred_res(pred_true,falsity).
convert_neg_pred_res(pred_false,truth).

generate_equality(b(A,_,_),b(B,_,_),Res) :- generate_equality_aux(A,B,Res),!.
generate_equality(A,B,equal(A,B)).

generate_equality_aux(convert_bool(AA),value(PT),Res) :- PT==pred_true,get_texpr_expr(AA,TA),!,
   %print(simplify_bool(AA,PT)),nl,
   Res = TA.
generate_equality_aux(convert_bool(AA),value(PF),Res) :- PF==pred_false,!,
   %print(simplify_bool(AA,PF)),nl,
   Res = negation(AA).
generate_equality_aux(value(PT),convert_bool(AA),Res) :- PT==pred_true,get_texpr_expr(AA,TA),!,
   %print(simplify_bool(AA,PT)),nl,
   Res = TA.
generate_equality_aux(value(PF),convert_bool(AA),Res) :- PF==pred_false,!,
   %print(simplify_bool(AA,PF)),nl,
   Res = negation(AA).


known_value(X) :- nonvar(X), known_value2(X).
known_value2(global_set(GS)) :- nonvar(GS).
known_value2(freetype(GS)) :- nonvar(GS).
known_value2(closure(P,T,B)) :- !,
   custom_explicit_sets:is_interval_closure(P,T,B,Low,Up), number(Low), number(Up),
   Up < Low+1000.
   %Note: for other closures: they still may have to be computed; some of the computations in wd_and_efficient could be expensive for closures
known_value2(avl_set(_)). % already fully normalised
known_value2((X,Y)) :- known_value(X),known_value(Y).
known_value2(fd(A,B)) :- number(A),atomic(B).
known_value2(int(N)) :- number(N).
known_value2([]).
known_value2(pred_true).
known_value2(pred_false).
known_value2(string(S)) :- atomic(S).
known_value2([H|T]) :- known_value(H), known_value(T).
known_value2(record(Fields)) :- known_fields(Fields).

known_fields(X) :- var(X),!,fail.
known_fields([]).
known_fields([field(N,V)|T]) :- ground(N),known_value(V),known_fields(T).

%b_evaluate1(TExpr,Parameters,LS,S,ResultTExpr,FullyKnown) :- 

check_is_id_list([]).
check_is_id_list([H|T]) :-
   (H=b(identifier(_),_,_)
    -> add_error(b_compile,'Typed id list as argument: ',[H|T]) % use get_texpr_ids or maplist(def_get_texpr_id,P,PIDs)
    ; true).
