% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(b_enumerate, [
  b_type_values_in_store/3,
  b_enumerate_values_in_store/5,
  b_enumerate_values/2,
  b_enumerate_typed_values/3,
  %b_tighter_enumerate_values_in_store/4,
  b_tighter_enumerate_values/2, b_tighter_enumerate_single_value/4,
  b_trace_tighter_enumerate_values/2,
  
  b_extract_typedvalc/4,
  
  enumerate_integer_with_min_range_wf/4]).
  
  
:- use_module(self_check).
:- use_module(debug).
:- use_module(kernel_objects).
:- use_module(kernel_waitflags).
:- use_module(error_manager,[add_error/3]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,interpreter).
:- module_info(description,'This module takes care of enumerating B variables.').

  /* ------------------------------------------ */
    
	
/* can be used to apply the result of 	
	b_getVarTypes_from_boolean_expression to a store */
	
:- use_module(store,[lookup_value_for_existing_id/3]).
	
	
:- assert_pre(b_enumerate:b_type_values_in_store(V,T,S), 
                  (typechecker:type_check(V,list(variable_id)),
                   typechecker:type_check(T,list(basic_type_descriptor)),typechecker:type_check(S,store))).
:- assert_post(b_enumerate:b_type_values_in_store(_,_,_), true).

b_type_values_in_store([],[],_).
b_type_values_in_store([Variable|VT],[Type|TT],Store) :-
   b_type_value_in_store(Variable,Type,Store),
   b_type_values_in_store(VT,TT,Store).
	
b_type_value_in_store(Var,Type,Store) :- 
   (Var = 'Identifier'(VarID)
      -> print(b_type_value_in_store(Var,Type)),print(' variable not in proper form'),nl
       ; VarID = Var),
   store:lookup_value_for_existing_id(VarID,Store,Val),
   (Val==fail -> add_error(b_type_value_in_store,'Variable has not been given value: ',VarID)
              ; kernel_objects:basic_type(Val,Type)).




/* ------------------------------------------ */
    

:- assert_pre(b_enumerate:b_enumerate_values_in_store(V,T,_,S,_WF), 
                  (typechecker:type_check(V,list(variable_id)),
                   typechecker:type_check(T,list(basic_type_descriptor)),typechecker:type_check(S,store))).
:- assert_post(b_enumerate:b_enumerate_values_in_store(_,_,_,_,_), true).
	
b_enumerate_values_in_store(Variables,Types,Values,Store,WF) :-
    generate_typed_var_list(Variables,Types,Store,Values,TypedValues),
    b_enumerate_values(TypedValues,WF).
    
:- block generate_typed_var_list(-,?,?,?,?).
generate_typed_var_list([],[],_,[],[]).
generate_typed_var_list([VarID|VT],[Type|TT],Store,Values,Res) :-
    ((store:lookup_value_for_existing_id(VarID,Store,Val), Val \== fail)
     ->  Res = [typedval(Val,Type,VarID,trigger_true(VarID))|Rest], Values=[Val|ValT] % always trigger enumeration warning
     ;   add_error(b_enumerate_values_in_store,'Variable has not been given value: ',VarID),
          Res=Rest,Values=ValT
   ),
   generate_typed_var_list(VT,TT,Store,ValT,Rest).
 
b_enumerate_sorted_values([],_).
b_enumerate_sorted_values([typedvalc(Val,Type,VarID,EnumWarning,Card)|Rest],WF) :-
    get_wait_flag(Card,enum(Val,Type),WF,LWF),
    b_enumerate_sorted_value_and_continue(LWF,Val,Type,VarID,EnumWarning,Rest,WF).

:- block b_enumerate_sorted_value_and_continue(-,?,?,?,?,?,?).
b_enumerate_sorted_value_and_continue(_LWF,Val,Type,_VarID,EnumWarning,Rest,WF) :-
    %%  print(enum_sorted_value(_VarID,Val,Type)),nl, %%
    enum_basic_with_wf(Val,Type,EnumWarning,WF),
    b_enumerate_sorted_values(Rest,WF).


b_enumerate_values(TypedValues,WF) :-
    ~~mnf(sort_typed_values(TypedValues,[],SortedValues)),
   % print(sort_vals(SortedValues)),nl,
    b_enumerate_sorted_values(SortedValues,WF).
/* ------------------------------------------ */

b_enumerate_typed_values(Values,Types,WF) :-
   b_generate_typed_values(Values,Types,TypedVals),
   b_enumerate_values(TypedVals,WF).

% generated typed values trigger the enumeration warning
b_generate_typed_values([],[],[]).
b_generate_typed_values([Val|Vrest],[Type|Trest],[typedval(Val,Type,'$UNKNOWN_ID',trigger_true(unknown))|TVrest]) :-
   b_generate_typed_values(Vrest,Trest,TVrest).

/*
:- assert_pre(b_enumerate:b_tighter_enumerate_values_in_store(V,T,S,_WF), 
                  (typechecker:type_check(V,list(variable_id)),
                   typechecker:type_check(T,list(basic_type_descriptor)),typechecker:type_check(S,store))).
:- assert_post(b_enumerate:b_tighter_enumerate_values_in_store(_,_,_,_), true).
b_tighter_enumerate_values_in_store(Vars,Types,Store,WF) :-
    ~~mnf(b_extract_typedvalc(Vars,Types,Store,TypedVals)),
    b_tighter_enumerate_values(TypedVals,WF).
*/
   
b_extract_typedvalc([],[],_,[]).
b_extract_typedvalc([Variable|VT],[Type|TT],Store,Res) :-
   (Variable = '_lambda_result_' % TO DO: use lambda_result info field ?; Note the check is repeated below in sort_typed_values
    -> b_extract_typedvalc(VT,TT,Store,Res) % lambda result does not need to be enumerated; will be computed from rest
    % TO DO: should we also deal specially with _lambda_result_ below in b_tighter_enumerate_values;
    %  or even infer in ANY statements, ... which variables are simply defined by equalities from the others
    ;  Res=[ExtractedH|ET],
       b_extract_typedvalc_var(Variable,Type,Store,ExtractedH),
       b_extract_typedvalc(VT,TT,Store,ET)).
 
b_extract_typedvalc_var(Var,Type,Store,typedval(Val,Type,VarID,trigger_true(VarID))) :- 
   (Var = 'Identifier'(VarID)
      -> print(b_tighter_enumerate_value_in_store(Var,Type)),print(' variable not in proper form'),nl
       ; VarID = Var),
   ((store:lookup_value_for_existing_id(VarID,Store,Val), Val \== fail)
     ->  true
     ;   add_error(b_tighter_enumerate_value_in_store,'Variable has not been given value: ',VarID)
   ).

% update priority for certain types
% TO DO: we should probably group all infinite integer values together and take the variable with the smallest size

b_tighter_enumerate_sorted_values([],_WF).
b_tighter_enumerate_sorted_values([typedvalc(Val,Type,VarID,EnumWarning,Card)|Rest],WF) :- 
    (ground(Val)
      -> check_reached(Val,Type,VarID,Card,det,WF), %% <-----
         b_tighter_enumerate_sorted_values(Rest,WF) %% no need to enumerate this value
      ;  treat_next_typed_valc(Type,Val,VarID,EnumWarning,Card,Rest,WF)).

:- use_module(library(lists)).
treat_next_typed_valc(integer,Val,VarID,EnumWarning,_Card,Rest,WF) :-
         nonvar(Val), Val=int(FD), 
         clpfd:fd_size(FD,Size), % will return sup if no bounds + also works in non-CLPFD mode
         number(Size),!,
         treat_next_typed_valc_normal(integer,Val,VarID,EnumWarning,Size,Rest,WF). 
treat_next_typed_valc(integer,Val,VarID,EnumWarning,Card,Rest,WF) :- 
         !,
         kernel_waitflags:large_finite_priority(LP),
         treat_next_integer(first_iteration,[LP],Val,VarID,EnumWarning,Card,Rest,WF).  % TO DO: maybe do several iterations 1000,LP
         % however: test 1003 with plavis-TransData_SP_13.prob fails in this case; to do: investigate
treat_next_typed_valc(Type,Val,VarID,EnumWarning,Card,Rest,WF) :- % TO DO: also look for finite values in Rest if Type is infinite but not integer
         treat_next_typed_valc_normal(Type,Val,VarID,EnumWarning,Card,Rest,WF).

% find another typedvalc whose domain is now finite
/* select_finite_typed_valc(TC, ListOfTC, Rest) :- 
         select( TC, ListOfTC, Rest),
         finite_typed_valc(TC,_). */
finite_typed_valc(typedvalc(Val,Type,_,_,_),Size) :- %Type=integer,
         finite_value(Type,Val,Size),
         (Size=1 -> true ; \+ currently_useless_variable(Val)). % ,print(finite(Val,Size)),nl.

:- use_module(b_global_sets,[b_global_set_cardinality/2]).
finite_value(integer,Val,Size) :- nonvar(Val), Val=int(FD), clpfd:fd_size(FD,Size), number(Size).
finite_value(boolean,Val,Size) :- (ground(Val) -> Size = 1 ; Size =2).
finite_value(string(_),Val,Size) :- ground(Val), Size=1.
finite_value(global(G),Val,Size) :- (ground(Val) -> Size = 1 ; b_global_sets:b_global_set_cardinality(G,Size)).
finite_value(set(T),Val,Size) :- nonvar(Val), l_finite_value(T,Val,Size).
finite_value(couple(T1,T2),Val,Size) :- nonvar(Val), Val = (V1,V2),
    finite_value(T1,V1,Size1),
    finite_value(T2,V2,Size2),
    kernel_objects:safe_mul(Size1,Size2,Size), number(Size).
% TO DO: missing records/fields + freetypes         

l_finite_value(_Type,Val,_) :- var(Val),!,fail.
l_finite_value(_,avl_set(_),1).
l_finite_value(_Type,[],1).
l_finite_value(Type,[H|T],Size) :- l_finite_value(Type,T,SizeT), finite_value(Type,H,SizeH),
    kernel_objects:safe_mul(SizeT,SizeH,Size), number(Size).


% or alternatively find the typedvalc with the smallest domain
select_typed_valc_with_smallest_domain(SmallestDomainTC,[FirstTC|ListOfTC],Rest,Size) :-
        finite_typed_valc(FirstTC,FirstSize), !, % finite domain found, look for smaller one:
        select_smallest_aux(ListOfTC,FirstSize,FirstTC,SmallestDomainTC,Rest,Size).
select_typed_valc_with_smallest_domain(TC,[FirstTC|ListOfTC],[FirstTC|Rest],Size) :-
        % no finite domain found yet; look further
        select_typed_valc_with_smallest_domain(TC,ListOfTC,Rest,Size).

select_smallest_aux(ListOfTC,Size,CurSmallestDomainTC,SmallestTC,Rest,ResSize) :-  Size =< 2, !,
        SmallestTC = CurSmallestDomainTC, Rest = ListOfTC, ResSize = Size.
select_smallest_aux([],Size,SmallestDomainTC,SmallestDomainTC,[],Size).
select_smallest_aux([FirstTC|ListOfTC],CurMinSize,CurSmallestDomainTC, SmallestTC,[CurSmallestDomainTC|Rest],ResSize) :-
        finite_typed_valc(FirstTC,Size), Size < CurMinSize, !,
        select_smallest_aux(ListOfTC,Size,FirstTC,SmallestTC,Rest,ResSize).
select_smallest_aux([FirstTC|ListOfTC],CurMinSize,CurSmallestDomainTC,SmallestTC,[FirstTC|Rest],ResSize) :-
        select_smallest_aux(ListOfTC,CurMinSize,CurSmallestDomainTC,SmallestTC,Rest,ResSize).

/*
% find the typedvalc with the most constraints (currently used to select between infinite valcs)
typed_valc_num_constraints(typedvalc(Val,integer,_,_,_),Degree) :-
        nonvar(Val), Val=int(FD), clpfd:fd_degree(FD,Degree).

select_most_constrained_typed_valc(MostConstrainedTC,[FirstTC|ListOfTC],Rest) :-
        typed_valc_num_constraints(FirstTC,NumAttachedConstraints), !,
        select_most_constrained_aux(ListOfTC,NumAttachedConstraints,FirstTC,MostConstrainedTC,Rest).
select_most_constrained_aux([],_NumAttachedConstraints,MostConstrainedTC,MostConstrainedTC,[]).
select_most_constrained_aux([FirstTC|ListOfTC],CurMaxNumAttachedConstraints,CurMostConstrainedTC,MostConstrainedTC,[CurMostConstrainedTC|Rest]) :-
        typed_valc_num_constraints(FirstTC,NumAttachedConstraints), NumAttachedConstraints > CurMaxNumAttachedConstraints, !,
        select_most_constrained_aux(ListOfTC,NumAttachedConstraints,FirstTC,MostConstrainedTC,Rest).
select_most_constrained_aux([FirstTC|ListOfTC],CurMaxNumAttachedConstraints,CurMostConstrainedTC,MostConstrainedTC,[FirstTC|Rest]) :-
        select_most_constrained_aux(ListOfTC,CurMaxNumAttachedConstraints,CurMostConstrainedTC,MostConstrainedTC,Rest).
*/

% we try again and see if the domain is now finite
:- block treat_next_integer(-,?,?,?,?,?,?,?).
treat_next_integer(LWF,_,Val,VarID,EnumWarning,_Card,Rest,WF) :- LWF \== first_iteration, % print(treat(Val,VarID)),nl,
         nonvar(Val), Val=int(FD), 
         clpfd:fd_size(FD,Size), % will return sup if no bounds + also works in non-CLPFD mode
         %print(treat_next_integer(VarID,Val,Size,_Card,Rest)),nl,
         number(Size),!,
         (Size = 1 % for Size=1 no enumeration is necessary anyway
          -> b_tighter_enumerate_sorted_values(Rest,WF) % no enumeration required
          ;  treat_next_typed_valc_normal(integer,Val,VarID,EnumWarning,Size,Rest,WF)).
treat_next_integer(LWF,Prios,Val,VarID,EnumWarning,Card,OtherTCs,WF) :- % ideally we should move this above the clause with NextPrio; but then test 412 fails
         select_typed_valc_with_smallest_domain(TC, OtherTCs, Rest,Size), % see if we can find another typed value whose domain is now finite
         !,
         % print(prioritising(Size,TC,VarID,Card,Rest)),nl,
         CurTC = typedvalc(Val,integer,VarID,EnumWarning,Card),
         (Size == 1 -> % no enumeration required for this variable; just remove it from OtherTCs to avoid cluttering
                       treat_next_integer(LWF,Prios,Val,VarID,EnumWarning,Card,Rest,WF)
           ; b_tighter_enumerate_sorted_values([TC, CurTC | Rest],WF)).
treat_next_integer(_,[NextPrio|Prios],Val,VarID,EnumWarning,Card,Rest,WF) :- !, % for test 1161 this clause causes a considerable slowdown (VarID == dz1 -> trace ; true),
         % print(get_wait_flag_for(VarID,Val,NextPrio,Rest)),nl,
         get_wait_flag(NextPrio,treat_next_integer(VarID,Val),WF,LWF),
         treat_next_integer(LWF,Prios,Val,VarID,EnumWarning,Card,Rest,WF).
% commented out: the fd_degree seems to be 0 for all of them?
%treat_next_integer(_,_,Val,VarID,EnumWarning,Card,OtherTC,WF) :-
%         select_most_constrained_typed_valc(TC, OtherTC, Rest), % if we can not find a finite one, select the one with the most constraints (to limit other variables)
%         !,
%         %print(prioritising(TC,VarID,Rest)),nl,
%         CurTC = typedvalc(Val,integer,VarID,EnumWarning,Card),
%         b_tighter_enumerate_sorted_values([TC, CurTC | Rest],WF).
% the selection rules can only work if there is an element left
treat_next_integer(_,_,Val,VarID,EnumWarning,Card,[],WF) :-
         treat_next_typed_valc_normal(integer,Val,VarID,EnumWarning,Card,[],WF). % we have already delayed; no need to delay further
% one of the rules above should work - set up an error to notice if not
treat_next_integer(_,_,Val,VarID,EnumWarning,Card,[R|Rest],WF) :-
%         add_error(treat_next_integer, 'Both smallest domain selection and most constrained variable selection failed for variable with id', VarID),
         treat_next_typed_valc_normal(integer,Val,VarID,EnumWarning,Card,[R|Rest],WF). % we have already delayed


treat_next_typed_valc_normal(Type,Val,VarID,EnumWarning,Card,Rest,WF) :-
         get_wait_flag(Card,tighter_enum(VarID,Val,Type),WF,LWF),
         b_tighter_enumerate_sorted_value_and_continue(LWF,Val,Type,VarID,EnumWarning,Card,Rest,WF).

:- block b_tighter_enumerate_sorted_value_and_continue(-,?,?,?,?,?,?,?).
b_tighter_enumerate_sorted_value_and_continue(_LWF,Val,Type,VarID,EnumWarning,Card,Rest,WF) :- 
     % nl,(ground(Val) -> print(ground(Val,VarID,Rest)),nl, tools_printing:print_term_summary(Rest),nl
     %  ;  print(tight_enum_sorted_value(Val,Type,VarID,Rest)),nl), %%
    (( Rest \= [],currently_useless_variable(Val), %print(useless(Val,VarID)),nl,
       get_next_useful_variable(Rest,RVal,RType,RVarID,REnumWarning,RCard,RRest)
     ) 
      ->  %debug_println(9,skipping(Val,VarID,Card,new(RVal,RType,RVarID,RCard))),
         (RCard =< Card /* we can enumerate straight away */
           -> check_reached(RVal,RType,RVarID,RCard,WF), %% <------
              enum_tight_with_wf(RVal,RType,EnumWarning,WF),
              b_tighter_enumerate_sorted_values([typedvalc(Val,Type,VarID,EnumWarning,Card)|RRest],WF)
            ; b_tighter_enumerate_sorted_values([typedvalc(RVal,RType,RVarID,REnumWarning,RCard),
                                                 typedvalc(Val,Type,VarID,EnumWarning,Card)|RRest],WF)
         )
     ;   check_reached(Val,Type,VarID,Card,WF), %% <-----
         %% tools:print_bt_message(enumerate(VarID,Val)), %%
         enum_tight_with_wf(Val,Type,EnumWarning,WF),
         %% tools:print_bt_message(after_enumerate(VarID,Val)),
         %(VarID=='VS_NOERRORCOND' -> print(done_enum(VarID,Val)),nl ; true),
         b_tighter_enumerate_sorted_values(Rest,WF)
    ).

% --------------------

% add WF information; so that we can provide better enumeration warning messages
add_wf_context(trigger_true(Info),WF,trigger_true(enum_wf_context(WF,Info))).
add_wf_context(trigger_false(Info),WF,trigger_false(enum_wf_context(WF,Info))).
add_wf_context(trigger_throw(Info),WF,trigger_throw(enum_wf_context(WF,Info))).

% TO DO: check whether we should not call enumerate_basic_type_wf instead

:- use_module(kernel_objects,[enumerate_tight_type/3]).
enum_tight_with_wf(Val,Type,EnumWarning,WF) :- 
   (add_wf_context(EnumWarning,WF,NewWarning) -> enumerate_tight_type(Val,Type,NewWarning)
     ; enumerate_tight_type(Val,Type,EnumWarning)).

:- use_module(kernel_objects,[enumerate_basic_type/3]).
enum_basic_with_wf(Val,Type,EnumWarning,WF) :- 
   (add_wf_context(EnumWarning,WF,NewWarning) -> enumerate_basic_type(Val,Type,NewWarning)
     ; enumerate_basic_type(Val,Type,EnumWarning)).


% ---------------------

get_next_useful_variable([typedvalc(Val,Type,VarID,EnumWarning,Card)|Rest],RVal,RType,RVarID,REnumWarning,RCard,RRest) :-
  (currently_useless_variable(Val)
    ->  RRest=[typedvalc(Val,Type,VarID,EnumWarning,Card)|TRRest],
        get_next_useful_variable(Rest,RVal,RType,RVarID,REnumWarning,RCard,TRRest)
    ;   RRest=Rest, RVal=Val, RType=Type, RVarID=VarID, REnumWarning = EnumWarning, RCard=Card
  ).
  

/* purpose: check if an enumeration is useful or not and give priority to useful variables ? */
%:- use_module(library(clpfd),[fd_dom/2]).
currently_useless_variable(Val) :- var(Val),!,frozen(Val,Goal),  %print(enum(Val,Goal)),nl,
      enumeration_only_goal(Goal,Val).
currently_useless_variable(int(Val)) :- currently_useless_variable(Val).
currently_useless_variable(fd(Val,_)) :- currently_useless_variable(Val).
  %fd_dom(Val,Dom),print(uselessfd(Val,Dom)),nl.
enumeration_only_goal(true,_).
%enumeration_only_goal(fd_utils_clpfd:delayed_enum_fd(V,_,_,_),Val) :- V==Val.
enumeration_only_goal(kernel_objects:enumerate_int_wf(V,_,_),Val) :- V==Val.
enumeration_only_goal(clpfd:(X in _),Var) :- Var==X. % only goal is the enumerator itself


b_tighter_enumerate_single_value(Val,Type,VarID,WF) :-
    b_tighter_enumerate_values([typedval(Val,Type,VarID,trigger_true(VarID))],WF).

b_tighter_enumerate_values(TypedValues,WF) :-
    ~~mnf(sort_typed_values(TypedValues,[],SortedValues)),  % no longer required ??
    b_tighter_enumerate_sorted_values(SortedValues,WF).

:- if(debug:global_debug_flag).
:- dynamic reached/4.
my_spy('P_Env_Output_TargetSpeed_Speed').
check_reached(Val,Type,VarID,Card,WF) :- 
  (my_spy(VarID) -> print(' SPY POINT : '), print(VarID), print(' = '), print(Val),nl,
    portray_waitflags(WF), trace ; true),
  (ground(Val) -> check_reached(Val,Type,VarID,Card,det,WF)
     ; check_reached(Val,Type,VarID,Card,enum,WF)).
% findall(r(Card,VarID,Info),b_enumerate:reached(VarID,Card,Info),_A),sort(_A,_AS),reverse(_AS,RAS).
check_reached(Val,Type,VarID,Card,Info,WF) :- 
    (retract(reached(VarID,RR,Info)) -> true ; RR=0),
    R1 is RR+1,
    (RR mod 1000 < 1 -> (RR=0 -> print('REACHED : ') ; print(' AGAIN  : ')), print(Info), print(' '),
                 print(VarID), print(' : '), print(Card),
                 print(' : '), print(Type), print(' : '), print(Val), print(' :: '), print(RR), nl,
                 (my_spy(VarID)  %(RR>8000,RR<40001)
                  -> portray_waitflags(WF),trace ; portray_waitflags(WF))
                 %, (RR=1000 -> trace ; true)
           ; true),
    assert(reached(VarID,R1,Info)).
b_trace_tighter_enumerate_values(TypedValues,WF) :-
    ~~mnf(sort_typed_values(TypedValues,[],SortedValues)), 
     print('ENUMERATING VALUES:'),nl,print_values(SortedValues), %%
     nl, portray_waitflags(WF),nl,
    b_trace_tighter_enumerate_values2(SortedValues,WF),
    print('FINISHED SETTING UP ENUMERATION'),nl, portray_waitflags(WF),nl.
 b_trace_tighter_enumerate_values2(SortedValues,WF) :-   retractall(reached(_,_,_)),
    b_tighter_enumerate_sorted_values(SortedValues,WF).

print_values([]).
print_values([typedvalc(Val,Type,VarID,_EnumWarning,Card)|Rest]) :-
  print_value(VarID,Card,Type,Val),
  (ground(Val) -> true ; when(ground(Val),(nl,print('HAS BECOME GROUND: '),print_value(VarID,Card,Type,Val)))),
  print_values(Rest).
  
print_value(VarID,Card,Type,Val) :-
  print(' '),
  print(VarID), print(' :card='), print(Card), print(' ('), print(Type), print(') = '), print(Val),nl.
:- else.
check_reached(_Val,_Type,_VarID,_Card,_WF).
check_reached(_Val,_Type,_VarID,_Card,_Info,_WF).
b_trace_tighter_enumerate_values(TypedValues,WF) :- b_tighter_enumerate_values(TypedValues,WF).
:- endif.

% dummy call to register variable name with free Prolog variables:
% can be obtained using frozen()
:- public register_prolog_variable_name/3. % Debugging tool
:- block register_prolog_variable_name(-,?,?).
register_prolog_variable_name(int(Val),Name,Type) :- !, %print(register_int(Val,Name,Type)),nl,
   register_prolog_variable_name(Val,Name,Type).
register_prolog_variable_name(fd(Val,_),Name,Type) :- !,
   register_prolog_variable_name(Val,Name,Type).
register_prolog_variable_name(string(Val),Name,Type) :- !,
   register_prolog_variable_name(Val,Name,Type).
register_prolog_variable_name((A,B),Name,Type) :- !,
   register_prolog_variable_name(A,left(Name,mapto(B)),Type),
   register_prolog_variable_name(B,right(Name,mapfrom(A)),Type).
register_prolog_variable_name([A|T],Name,Type) :- !,
   register_prolog_variable_name(A,el(Name),Type),
   register_prolog_variable_name(T,Name,Type).
register_prolog_variable_name(_,_,_).
% TO DO: something that pretty prints the path information for the average user

% a tool to get variable name if it was registered
:- public get_prolog_variable_name/3. % Debugging tool
get_prolog_variable_name(X,Name,Type) :- var(X),
   frozen(X,Goal), print(goal(X,Goal)),nl,
   goal_member(Goal,b_enumerate:register_prolog_variable_name(_,ID,T)),
   Name=ID, Type=T.
goal_member(X,X).
goal_member((A,B),X) :- goal_member(A,X) ; goal_member(B,X).
  

sort_typed_values([],R,R).
sort_typed_values([typedval(Val,Type,VarID,EnumWarning)|Rest],SortedSoFar,Res) :-
   %% comment in next line to keep track of variable names for enumeration warnings, etc...
   %% register_prolog_variable_name(Val,VarID,Type), % dummy call to register variable name with free Prolog variables
   (VarID = '_lambda_result_' -> NewSorted=SortedSoFar
    ; ground(Val) -> NewSorted=SortedSoFar
    ;  get_max_cardinality_as_priority(Type,CardPrio),
       %% TO DO: maybe also take shape of Val into account; e.g. if cardinality of a set already fixed...
       insert_val(SortedSoFar,Val,Type,VarID,EnumWarning,CardPrio,NewSorted)
   ),
   sort_typed_values(Rest,NewSorted,Res).

get_max_cardinality_as_priority(Type,Prio) :-
  ~~mnf(kernel_objects:max_cardinality(Type,Card)),
  (Card \= inf -> Prio=Card
    ; inf_type_prio(Type,Prio)
  ).

% Try to give priority to simpler infinite types (such as integer over seq(_) ,...)
% TO DO: to a full-blown analysis ...
inf_type_prio(integer,Prio) :- !, integer_priority(Prio).
inf_type_prio(seq(X),Prio) :- !, inf_type_prio(X,P1), Prio is P1+11.
inf_type_prio(set(X),Prio) :- !, inf_type_prio(X,P1), Prio is P1+10.
%inf_type_prio(couple(_,_),Prio) :- !, TO DO check if both args are infinite or not ...
inf_type_prio(_,Prio) :- !, integer_priority(P), Prio is P+1.


%typedvalc = typedval + cardinality information
insert_val([],Val,Type,VarID,EnumWarning,Card,[typedvalc(Val,Type,VarID,EnumWarning,Card)]).
insert_val([TC|T], Val,Type, VarID, EnumWarning, Card, Res) :-
   TC = typedvalc(V1,_T1,_Var1,_W1,C1),
   ((%cardval_greater_than(C1,Card)
     Card \== inf, (C1=inf -> true ; C1>Card)
     ; Card=C1, var(V1), nonvar(Val)) % if card equal give priority to nonvariable values
     -> Res = [typedvalc(Val,Type,VarID,EnumWarning,Card),TC|T]
     ;  Res = [TC|RT],
        insert_val(T,Val,Type,VarID,EnumWarning,Card,RT)
   ).

%cardval_greater_than(X,Y) :- Y \== inf, (X=inf -> true ; X>Y).

% call to enumerate an integer ensuring that you go at least from MinFrom to MinTo
%:- block enumerate_integer_with_min_range_wf(-,?,?,?).
enumerate_integer_with_min_range_wf(int(X),MinFrom,MinTo,WF) :-
   integer_priority(Prio), P1 is Prio-1,
   get_wait_flag(P1,enum_int_min_range(X,MinFrom,MinTo),WF,LWF),
    % print(enum_int_min_range(X,MinFrom,MinTo)),nl, portray_waitflags(WF),
    %((MinFrom == -4, MinTo == 17) -> trace ; true),
   enum_int_min_range(X,MinFrom,MinTo,LWF).

:- use_module(clpfd_interface,[clpfd_domain/3]).
:- block enum_int_min_range(?,?,?,-).
enum_int_min_range(X,MinFrom,MinTo,_) :-
   %  print(trigger_enum_int_min_range(X,MinFrom,MinTo)),nl,
  (ground(X) -> true ;
    clpfd_domain(X,FDLow,FDUp),
    (number(FDLow) -> From = FDLow ; From=MinFrom),
    (number(FDUp) -> To = FDUp ; To=MinTo),
    % print(enumerate_int(X,From,To)),nl,
    kernel_objects:enum_warning('INTEGER',FDLow:FDUp,From:To),
    kernel_objects:enumerate_int(X,From,To)).  

