% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(b_interpreter_check,[imply_test_boolean_expression/6,
                               equiv_test_boolean_expression/6,
                               equiv_bidrectional_test_boolean_expression/7,
                               get_priority_of_boolean_expression/2, get_priority_of_boolean_expression2/2,
                               b_check_boolean_expression/5,b_check_boolean_expression/7,
                               
                               /* some lower-level propagation predicates */
                               imply_true/2,
                               
                               reify_closure_with_small_cardinality/5,
                               
                               register_predicate/6,

                               conjoin/6
          ]).

/* A version for checking the truth value of boolean-expressions,
 the result is instantiated with pred_true/pred_false as soon as the result is known */
/* Warning: does not cover all expressions */
/* It provides SMT-like performance, for predicates involving arithmetic comparisons (<,...),
   equality, disequality and membership/not_membership
   It is not restricted to CNF
*/

:- use_module(debug).
:- use_module(self_check).
:- use_module(error_manager).
:- use_module(b_interpreter,[b_compute_expression/5, b_not_test_boolean_expression/4, b_test_boolean_expression/4]).
:- use_module(kernel_waitflags).


:- use_module(kernel_objects,[less_than_direct/2, less_than_equal_direct/2]).

:- use_module(kernel_equality).

:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,interpreter).
:- module_info(description,'This module provides a reified interpreter for certain predicates.').

:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(partially_evaluate, true)).
:- use_module(partial_evaluator(pe)).
:- unfold(b_check_boolean_expression(_LHS,_LocalState,_State,_CWF,_PredRes)).
:- unfold(b_check_boolean_expression(_LHS,_LocalState,_State,_CWF,_PredRes,_Ai,_Aii)).
:- unfold(b_check_boolean_expression0(_WDE,_WDV,_Expr,_LS,_S,_WF,_Res,_Ai,_Ao)).
:- unfold(b_check_boolean_expression1(_EXPR,_LS,_S,_WFD,_Res,_Ai,_Ao)).
:- unfold(b_check_boolean_expression2(_Expr,_Infos,_LS,_S,_,_Res,_Ai,_Ao)).
:- unfold(negate_equal(A,B,_),
      ((ground(A); A \= b(boolean_false,boolean,_)),(ground(B); B \= b(boolean_false,boolean,_)))).
:- unfold(b_check_boolean_expression3_pos(_,_Info,_LocalState,_State,_WFD,_Res,_Ai,_Ao)).
:- unfold(b_check_boolean_expression3_neg(_,_Info,_LocalState,_State,_WFD,_Res,_Ai,_Ao)).
:- unfold(b_check_boolean_expression4(_Pred,_Infos,_LocalState,_State,_WFD,_Res)).
:- unfold(register_predicate_wfd(WFD,_Pred,Infos,_Res,_Reused,_Ai,_Ao),(WFD=wfwd(_,A,B),(A==B;A\=B,ground(Infos)))).
:- unfold(register_predicate_aux(Pred,_PredTruthVar,_Reused,_Ai,_Ao),nonvar(Pred)).
:- unfold(check_pred_truth_var(PredTruthVar),nonvar(PredTruthVar)).
:- execute(get_texpr_type(_,_)).
:- execute(always_well_defined(A),ground(A)).
:- unfold(b_wd_compute_expression(_Expr,_LocalState,_State,_Value,_)).
:- unfold(b_compute_expression_delay(_WDE,_WDV, _Expr,_LocalState,_State,_Value,_WF)).
:- execute(instiantiate_to_any_value(A,_E,_WF),ground(A)).
:- unfold(imply_test_boolean_expression(_PredRes1,_PredRes2, _RHS,_LocalState,_State,_WF)).
:- unfold(equiv_bidrectional_test_boolean_expression(_PredRes,_PredRes2,_LHS,_RHS,_LS,_S,_WF)).
:- unfold(equiv_test_boolean_expression(_PredRes,_PredRes2,_RHS,_LocalState,_State,_WF)).
:- unfold(arithmetic_op(Pred,_Op,_LHS,_RHS),nonvar(Pred)).
:- unfold(check_arithmetic_operator(Op,_LHValue,_RHValue,_Res),nonvar(Op)).
:- execute(get_priority_of_boolean_expression(A,_),ground(A)).
:- execute(get_priority_of_boolean_expression2(A,_),ground(A)).
:- execute(normalise_disjunct(E,_,_),ground(E)).
:- execute(normalise_disjunct2(E,_Info,_,_),ground(E)).
:- execute(construct_norm_disjunct(E,_R),ground(E)).
:- execute(construct_norm_disjunct2(E,_R,_I),ground(E)).
:- execute(empty_avl(_)).
:- unfold(norm_pred(A,_R),nonvar(A)).
:- execute(norm_pred_check(A,_R),nonvar(A)). % Dangerous
:- unfold(norm_expr(A,_R),nonvar(A)).
:- unfold(norm_expr2(A,_R),nonvar(A)).
:- unfold(add_predicate(_Pred,_Var,_AVL,_NewAVL)).
:- unfold(reuse_predicate(_Pred,_Var,_AVL)).
:- execute(avl_fetch(Pred,AVL,_Var), (ground(AVL),(AVL==empty ; ground(Pred)))).
:- execute(avl_store(Pred,AVL,_Var,_NewAVL), (ground(AVL),(AVL==empty ; ground(Pred)))).

:- forward_block(b_test_disjunction(?,?,?,?,?,?,?)) .
:- forward_block(b_enum_test_equivalence(?,?,?,?,?,?)) .
:- forward_block(block_test_implication(?,?,?,?,?,?)) .
:- forward_block(b_compute_expression_delay(?,?,?,?,?,?,?)) .
:- endif.

:- block imply_test_boolean_expression(-, ?,?,?,?,?). % TO DO: pass at least Ai to it
imply_test_boolean_expression(PredRes1,PredRes2, RHS,LocalState,State,WF) :- 
      (PredRes1=PredRes2
         -> % print('enforcing RHS '),print(PredRes), print(' : '), translate:print_bexpr(RHS),nl, %%
            % print(LocalState),nl,
            b_test_boolean_expression(RHS,LocalState,State,WF)
         ;  true %,print(not_checking_rhs(_PRes)), print(' : '), print_bexpr(_RHS),nl, print(_State),nl
      ).
         
:- block equiv_test_boolean_expression(-, ?,?,?,?,?).
equiv_test_boolean_expression(PredRes,PredRes, RHS,LocalState,State,WF) :- !,
         %print('enforcing RHS '),print(PredRes), print(' : '), print_bexpr(RHS),nl,
         b_test_boolean_expression(RHS,LocalState,State,WF).
equiv_test_boolean_expression(_PredRes,_, RHS,LocalState,State,WF) :- 
         %print('enforcing negation of RHS '),print(_PredRes), print(' : '), print_bexpr(RHS),nl,
         b_not_test_boolean_expression(RHS,LocalState,State,WF).

equiv_bidrectional_test_boolean_expression(PredResLHS,PredResRHS, _LHS,_RHS,_LocalState,_State,_WF) :- PredResLHS=PredResRHS.
/*
:- block equiv_bidrectional_test_boolean_expression(-,-, ?,?,?,?,?).
equiv_bidrectional_test_boolean_expression(PredResLHS,PredResRHS, LHS,RHS,LocalState,State,WF) :- 
   (  PredResLHS == pred_true -> b_test_boolean_expression(RHS,LocalState,State,WF)
    ;  PredResLHS == pred_false -> b_not_test_boolean_expression(RHS,LocalState,State,WF)
    ;  PredResRHS == pred_true -> b_test_boolean_expression(LHS,LocalState,State,WF)
    ;  PredResRHS == pred_false -> b_not_test_boolean_expression(LHS,LocalState,State,WF)
    ;  add_error_fail(equiv,'Illegal values: ',equiv_bidrectional_test_boolean_expression(PredResLHS,PredResRHS))
   ).
*/

get_priority_of_boolean_expression(priority(P),Prio) :- !, Prio=P.
get_priority_of_boolean_expression(b(Expr,_,_Infos),Prio) :- !,
   % try to estimate a priority for performing a case split upon a predicate
   %   i.e., forcing a predicate Expr to be true/false
	get_priority_of_boolean_expression2(Expr,Prio).
get_priority_of_boolean_expression(E,Prio) :- 
  add_internal_error('Boolean expression not properly wrapped: ',get_priority_of_boolean_expression(E,Prio)),
  get_priority_of_boolean_expression2(E,Prio).
	
:- use_module(bsyntaxtree).
get_priority_of_boolean_expression2(truth,0) :- !.
get_priority_of_boolean_expression2(falsity,0) :- !.
get_priority_of_boolean_expression2(_,R) :- !, R=3. % force SMT style case-splitting
/*
get_priority_of_boolean_expression2(negation(BExpr),Prio) :- !, 
        get_priority_of_boolean_expression(BExpr,NPrio), NPrio=Prio  %can we do better ?
get_priority_of_boolean_expression2(norm_conjunct(LHS,RHS),Prio) :- !, 
       get_priority_of_boolean_expression2(conjunct(LHS,RHS),Prio).
get_priority_of_boolean_expression2(conjunct(LHS,RHS),Prio) :- !, 
        get_priority_of_boolean_expression(LHS,LPrio),
        get_priority_of_boolean_expression(RHS,RPrio),
        Prio is min(LPrio,RPrio).
get_priority_of_boolean_expression2(disjunct(LHS,RHS),Prio) :- !, 
        get_priority_of_boolean_expression(LHS,LPrio),
        get_priority_of_boolean_expression(RHS,RPrio),
        Prio is LPrio+RPrio+2.
get_priority_of_boolean_expression2(equal(L,R),Prio) :- has_large_type(L),
   uses_ids(L,R),!,Prio=2.
 % TO DO: check if it binds a value
get_priority_of_boolean_expression2(not_equal(L,R),Prio) :- has_large_type(L),uses_ids(L,R),!,Prio=2.
get_priority_of_boolean_expression2(member(L,R),Prio) :- uses_ids(L,R), !, Prio=3.
get_priority_of_boolean_expression2(not_member(L,R),Prio) :- uses_ids(L,R), !, Prio=3.
get_priority_of_boolean_expression2(less(L,R),Prio) :- uses_ids(L,R),!,Prio=4.
get_priority_of_boolean_expression2(less_equal(L,R),Prio) :- uses_ids(L,R),!,Prio=4.
get_priority_of_boolean_expression2(greater(L,R),Prio) :- uses_ids(L,R),!,Prio=4.
get_priority_of_boolean_expression2(greater_equal(L,R),Prio) :- uses_ids(L,R),!,Prio=4.
get_priority_of_boolean_expression2(_,10). % magic constant; what should we put here
                                
uses_ids(L,R) :- (get_texpr_id(L,_);get_texpr_id(R,_)).                                                   
has_large_type(E) :- get_texpr_type(E,Type), large_type(Type).
% is it worth making a case split on this type
large_type(integer).
large_type(set(_)).
*/


count_number_of_conjuncts(b(Expr,_,_Infos),Prio) :- !,
	count_number_of_conjuncts2(Expr,Prio).
count_number_of_conjuncts2(conjunct(A,B),Nr) :- !, count_number_of_conjuncts(A,NA),
    count_number_of_conjuncts(B,NB), Nr is NA+NB.
count_number_of_conjuncts2(norm_conjunct(_,RHS),Res) :- length(RHS,Len),!, 
     Res is Len+1.
count_number_of_conjuncts2(_,1).

:- public count_number_of_disjuncts/2. %currently commented out above
count_number_of_disjuncts(b(Expr,_,_Infos),Prio) :- !,
	count_number_of_disjuncts2(Expr,Prio).
count_number_of_disjuncts2(disjunct(A,B),Nr) :- !, count_number_of_disjuncts(A,NA),
    count_number_of_disjuncts(B,NB), Nr is NA+NB.
count_number_of_disjuncts2(norm_disjunct(_,RHS),Res) :- length(RHS,Len),!, 
     Res is Len+1.
count_number_of_disjuncts2(_,1).



% we need to ensure that b_check_boolean_expression does not create a choice point on its own 

b_check_boolean_expression(B,LS,S,WF,Res) :- empty_avl(Ai),
  b_check_boolean_expression(B,LS,S,WF,Res,Ai,_).

b_check_boolean_expression(E,LS,S,WF,Res,Ai,Ao) :-
% add information about WD context: wfwd(WF_store, ExpectedVal, Val)
%  when Val becomes nonvar: if Val==ExpectedVal we need the value of E, otherwise it should be discarded
   b_check_boolean_expression1(E,LS,S,wfwd(WF,pred_true,pred_true),Res,Ai,Ao).

b_check_boolean_expression0(WDE,WDV,Expr,LS,S,WF,Res,Ai,Ao) :-
  b_check_boolean_expression1(Expr,LS,S,wfwd(WF,WDE,WDV),Res,Ai,Ao).
%  ((nonvar(WDV),WDE \= WDV) -> Ai=Ao, Res=pred_false
%   ;  b_check_boolean_expression1(Expr,LS,S,wfwd(WF,WDE,WDV),Res,Ai,Ao)).

b_check_boolean_expression1(b(Expr,_,Infos),LS,S,WFD,Res,Ai,Ao) :- WFD=wfwd(_WF,WDE,WDV),!,
    % print('check : '), translate:print_bexpr(b(Expr,pred,Infos)),nl,
    %when(nonvar(Res), (print(Res), print(' <--> '),translate:print_bexpr(b(Expr,pred,Infos)),nl)),
    ((nonvar(WDV),WDE \= WDV)
        -> Ai=Ao, (var(Res) -> Res=pred_false ; true)
	    ; b_check_boolean_expression2(Expr,Infos,LS,S,WFD,Res,Ai,Ao)).
b_check_boolean_expression1(b(Expr,_,Infos),LS,S,WFD,Res,Ai,Ao) :- !,
     add_error(b_check_boolean_expression1,'WFWD info missing: ',WFD),
	 b_check_boolean_expression2(Expr,Infos,LS,S,wfwd(WFD,true,true),Res,Ai,Ao).
b_check_boolean_expression1(E,LS,S,WFD,Res,Ai,Ao) :- 
     add_internal_error('Boolean expression not properly wrapped: ',b_check_boolean_expression1(E,LS,S,WFD,Res,Ai,Ao)),
     b_check_boolean_expression2(E,[],LS,S,WFD,Res,Ai,Ao).

% normalise conjunction into flat list of conjuncts
normalise_conjunct(b(E,_,Info)) --> normalise_conjunct2(E,Info).
normalise_conjunct2(conjunct(A,B),_) --> !,normalise_conjunct(A),normalise_conjunct(B).
normalise_conjunct2(F,Info) --> [b(F,pred,Info)].

construct_norm_conjunct(A,b(B,pred,Info)) :-  construct_norm_conjunct2(A,B,Info).
construct_norm_conjunct2([],truth,[]).
construct_norm_conjunct2([H|T],Res,Info) :-
  (T==[] -> H=b(Res,pred,Info) ; Res=norm_conjunct(H,T),Info=[]).
% TO DO: build up member(contains_wd_condition,Infos) 


% normalise disjunction into flat list of disjuncts
normalise_disjunct(b(E,_,Info)) --> normalise_disjunct2(E,Info).
normalise_disjunct2(disjunct(A,B),_) --> !,normalise_disjunct(A),normalise_disjunct(B).
normalise_disjunct2(F,Info) --> [b(F,pred,Info)].

construct_norm_disjunct(A,b(B,pred,Info)) :-  construct_norm_disjunct2(A,B,Info).
construct_norm_disjunct2([],falsity,[]).
construct_norm_disjunct2([H|T],Res,Info) :-
  (T==[] -> H=b(Res,pred,Info) ; Res=norm_disjunct(H,T),Info=[]).

can_negate_expression(b(Expr,pred,I),b(NExpr,pred,I)) :- can_negate2(Expr,NExpr).
can_negate2(equal(A,B),not_equal(A,B)).
can_negate2(not_equal(A,B),equal(A,B)).
can_negate2(member(A,B),not_member(A,B)).
can_negate2(not_member(A,B),member(A,B)).
can_negate2(subset(A,B),not_subset(A,B)).
can_negate2(not_subset(A,B),subset(A,B)).
can_negate2(subset_strict(A,B),not_subset_strict(A,B)).
can_negate2(not_subset_strict(A,B),subset_strict(A,B)).
can_negate2(greater_equal(A,B),less(A,B)).
can_negate2(less(A,B),greater_equal(A,B)).
can_negate2(less_equal(A,B),greater(A,B)).
can_negate2(greater(A,B),less_equal(A,B)).

b_check_boolean_expression2(truth,_,_,_,_WFD,Res,Ai,Ao) :- !,Res=pred_true, Ai=Ao.
b_check_boolean_expression2(falsity,_,_,_,_WFD,Res,Ai,Ao) :- !,Res=pred_false, Ai=Ao.
b_check_boolean_expression2(negation(BExpr),_,LocalState,State,WFD,Res,Ai,Ao) :- !,
        (can_negate_expression(BExpr,NBExpr)
          -> /* avoid introducing negate propagator; maybe not necessary */
             b_check_boolean_expression1(NBExpr,LocalState,State,WFD,Res,Ai,Ao)
          ;  negate(NR,Res),
             b_check_boolean_expression1(BExpr,LocalState,State,WFD,NR,Ai,Ao)).
b_check_boolean_expression2(conjunct(LHS,RHS),CInfo,LocalState,State,WFD,Res,Ai,Ao) :- !,
        normalise_conjunct2(conjunct(LHS,RHS),CInfo,NormRes,[]),
        construct_norm_conjunct2(NormRes,NC,Info), % print(normalised(NC)),nl,
        b_check_boolean_expression2(NC,Info,LocalState,State,WFD,Res,Ai,Ao).
b_check_boolean_expression2(norm_conjunct(LHS,RHS),_,LocalState,State,wfwd(WF,WDE,WDV),Res,Ai,Ao) :- !,
        construct_norm_conjunct(RHS,NC),
        conjoin(LR,RR,Res,LHS,NC,WF),
        b_check_boolean_expression1(LHS,LocalState,State,wfwd(WF,WDE,WDV),LR,Ai,Aii),
        % when(nonvar(LR),(print(LR), print(' && : '), translate:print_bexpr(LHS),nl)),
        propagagate_wfwd(WDE,WDV,GuardFlag,LR,pred_false), % if WDE/=WDV then set GuardFlag to pred_false; indicating to RHS that it is not needed also
        b_check_boolean_expression0(pred_true,GuardFlag,NC,LocalState,State,WF,RR,Aii,Ao).
b_check_boolean_expression2(implication(LHS,RHS),_,LocalState,State,wfwd(WF,WDE,WDV),Res,Ai,Ao) :- !,  impl(LR,RR,Res),
        b_check_boolean_expression1(LHS,LocalState,State,wfwd(WF,WDE,WDV),LR,Ai,Aii),
        % when(nonvar(LR),(LR=pred_true -> print(' => : '), translate:print_bexpr(RHS),nl ; true)),
        propagagate_wfwd(WDE,WDV,GuardFlag,LR,pred_false),
        b_check_boolean_expression0(pred_true,GuardFlag,RHS,LocalState,State,WF,RR,Aii,Ao).
b_check_boolean_expression2(equivalence(LHS,RHS),_,LocalState,State,WFD,Res,Ai,Ao) :- !,  equiv(LR,RR,Res),
        b_check_boolean_expression1(LHS,LocalState,State,WFD,LR,Ai,Aii),
        b_check_boolean_expression1(RHS,LocalState,State,WFD,RR,Aii,Ao).
b_check_boolean_expression2(disjunct(LHS,RHS),CInfo,LocalState,State,WFD,Res,Ai,Ao) :- !,
        normalise_disjunct2(disjunct(LHS,RHS),CInfo,NormRes,[]),
        construct_norm_disjunct2(NormRes,NC,Info),  % print(normalised_disj(NC)),nl,
        b_check_boolean_expression2(NC,Info,LocalState,State,WFD,Res,Ai,Ao).
b_check_boolean_expression2(norm_disjunct(LHS,RHS),_,LocalState,State,wfwd(WF,WDE,WDV),Res,Ai,Ao) :- !, 
        construct_norm_disjunct(RHS,NC), 
        disjoin(LR,RR,Res,LHS,NC,WF),
        %print(disjoin(WDE,WDV)), translate:print_bexpr(LHS),nl,
        b_check_boolean_expression1(LHS,LocalState,State,wfwd(WF,WDE,WDV),LR,Ai,Aii),
        propagagate_wfwd(WDE,WDV,GuardFlag,LR,pred_true),
        b_check_boolean_expression0(pred_false,GuardFlag,NC,LocalState,State,WF,RR,Aii,Ao).
b_check_boolean_expression2(let_predicate(Ids,AssignmentExprs,Pred),_Info,LocalState,State,wfwd(WF,WDE,WDV),Res,Ai,Ao) :-
    %print(check_let_predicate(Ids)),nl,
	!,b_interpreter:set_up_localstate_for_let(Ids,AssignmentExprs,LocalState,State,LetState,WF),
	b_check_boolean_expression1(Pred,LetState,State,wfwd(WF,WDE,WDV),Res,Ai,Ao).
% TO DO: compile forall, exists + setup choice point if expansion fails + remove compile calls in b_interpreter
b_check_boolean_expression2(forall(Parameters,LHS,RHS),Info,LocalState,State,WFD,NR,Ai,Ao) :-
    % could be generalised to take into consideration domain as restricted by LHS
        small_quantifier_cardinality(Parameters,LHS,LHS1,LHSRest),!, Ao=Ai,
          % print(expanding_check_forall(_Card)),nl,
        expand_quantifier(Parameters,LHS1,List), %print(List),nl,
        Body = b(implication(LHSRest,RHS),pred,Info),% translate:print_bexpr(Body),nl,
        WFD=wfwd(WF,_,_),
        check_expanded_forall_quantifier(List,Body, LocalState, State,WF,WFD,NR).
b_check_boolean_expression2(exists(Parameters,Body),_Info,LocalState,State,WFD,NR,Ai,Ao) :-
    % could be generalised to take into consideration domain as restricted by Body
        small_quantifier_cardinality(Parameters,Body,LHS,RHS),
           % print(expanding_check_exists(_Card,Parameters)),nl, % portray_waitflags(WF),
         !, Ao=Ai,
        expand_quantifier(Parameters,LHS,List),  
         % print(List),nl, translate:print_bexpr(Body),nl,nl,
        WFD=wfwd(WF,_,_),
        check_expanded_exists_quantifier(List,RHS, LocalState, State,WF,WFD,NR).
b_check_boolean_expression2(lazy_let_pred(Id,AssignmentExpr,Pred),_I,LocalState,State,wfwd(WF,WDE,WDV),Res,Ai,Ao) :- !,
	store:set_up_localstate([Id],[(Trigger,IdValue)],LocalState,LetState),
	% print(setting_up_lazy_let_pred(Id)),nl,
	b_interpreter:lazy_compute_expression(Trigger,AssignmentExpr,LocalState,State,IdValue,WF,Ai),
	b_check_boolean_expression1(Pred,LetState,State,wfwd(WF,WDE,WDV),Res,Ai,Ao).
b_check_boolean_expression2(lazy_lookup_pred(Id),_Info,LocalState,_State,_WFD,Res,Ai,Ao) :- !, Ai=Ao,
        % print(lazy_lookup_pred_check(Id,LocalState)),nl,
        store:lookup_value_for_existing_id(Id,LocalState,(Trigger,Value)), % should normally only occur in LocalState; value introduced by lazy_let
        (Trigger,Value) = (pred_true,Res). % ,print(res(Res)),nl. % force evaluation
b_check_boolean_expression2(value(V),_Info,_LocalState,_State,_WFD,Res,Ai,Ao) :- !,  % this can occur when lazy_lookup_pred gets compiled by b_compiler
        Res=V,Ai=Ao.

b_check_boolean_expression2(not_equal(LHS,RHS),Info,LocalState,State,WFD,Res,Ai,Ao) :- !,
       (negate_equal(LHS,RHS,LHS1)
          -> RHS1=b(boolean_true,boolean,[]),  /* X/=FALSE equivalent to X=TRUE */
             b_check_boolean_expression3_pos(equal(LHS1,RHS1),Info,LocalState,State,WFD,Res,Ai,Ao)
          ;  b_check_boolean_expression3_neg(equal(LHS,RHS),Info,LocalState,State,WFD,Res,Ai,Ao)
        ).
b_check_boolean_expression2(equal(LHS,RHS),Info,LocalState,State,WFD,Res,Ai,Ao) :- !,
       (negate_equal(LHS,RHS,LHS1)
          -> RHS1=b(boolean_true,boolean,[]),  /* X/=FALSE equivalent to X=TRUE */
             b_check_boolean_expression3_neg(equal(LHS1,RHS1),Info,LocalState,State,WFD,Res,Ai,Ao)
          ;  b_check_boolean_expression3_pos(equal(LHS,RHS),Info,LocalState,State,WFD,Res,Ai,Ao)
        ).
b_check_boolean_expression2(not_member(LHS,RHS),Info,LocalState,State,WFD,Res,Ai,Ao) :- !,
        b_check_boolean_expression3_neg(member(LHS,RHS),Info,LocalState,State,WFD,Res,Ai,Ao).
b_check_boolean_expression2(not_subset(LHS,RHS),Info,LocalState,State,WFD,Res,Ai,Ao) :- !,
        b_check_boolean_expression3_neg(subset(LHS,RHS),Info,LocalState,State,WFD,Res,Ai,Ao).
b_check_boolean_expression2(not_subset_strict(LHS,RHS),Info,LocalState,State,WFD,Res,Ai,Ao) :- !,
        b_check_boolean_expression3_neg(subset_strict(LHS,RHS),Info,LocalState,State,WFD,Res,Ai,Ao).
b_check_boolean_expression2(greater(LHS,RHS),Info,LocalState,State,WFD,Res,Ai,Ao) :- !, 
        b_check_boolean_expression3_pos(less(RHS,LHS),Info,LocalState,State,WFD,Res,Ai,Ao).
b_check_boolean_expression2(greater_equal(LHS,RHS),Info,LocalState,State,WFD,Res,Ai,Ao) :- !, 
        b_check_boolean_expression3_neg(less(LHS,RHS),Info,LocalState,State,WFD,Res,Ai,Ao).
b_check_boolean_expression2(less_equal(LHS,RHS),Info,LocalState,State,WFD,Res,Ai,Ao) :- !,
        b_check_boolean_expression3_neg(less(RHS,LHS),Info,LocalState,State,WFD,Res,Ai,Ao).
/* not yet ready for production: 
b_check_boolean_expression2(exists(Parameters,Body),Infos,LocalState,State,WFD,Res,Ai,Ao) :-
        print(reifying_exists(Parameters)),nl, Ao=Ai,
        b_interpreter:set_up_typed_localstate(Parameters,_FreshOutputVars,TypedVals,LocalState,NewLocalState,positive),
        b_check_boolean_expression(Body,NewLocalState,State,WF,BodyRes), !, % to do: eliminate all expressions in Ai which are no longer valid
        translate:print_bexpr(Body),nl,
        b_interpreter:get_exists_used_ids(Parameters,Body,Infos,Used),
        b_interpreter:lookup_values_if_vars(Used,LocalState,State,WaitVars), % TO DO: we could use b_compiler instead
        %kernel_waitflags:portray_waitflags(WF),
        b_propagate_exists(Res,BodyRes,WaitVars,Parameters,TypedVals,WF). */
b_check_boolean_expression2(Pred,Infos,LocalState,State,WFD,Res,Ai,Ao) :- 
        b_check_boolean_expression3_pos(Pred,Infos,LocalState,State,WFD,Res,Ai,Ao).



negate_equal(b(boolean_false,boolean,_),E,E).
negate_equal(E,b(boolean_false,boolean,_),E).

b_check_boolean_expression3_pos(Pred,Infos,LocalState,State,WFD,Res,Ai,Ao) :- 
        %print(register_predicate(WFD,Infos,Pred)),nl,
        register_predicate_wfd(WFD,Pred,Infos,Res,Reused,Ai,Ao),
        (Reused=true
           ->  true % , print(reused_pred(WFD,Infos)),nl, translate:print_bexpr(Pred),nl
           ;  b_check_boolean_expression4(Pred,Infos,LocalState,State,WFD,Res)
        ).

b_check_boolean_expression3_neg(Pred,Infos,LocalState,State,WFD,NegRes,Ai,Ao) :- 
        %print(register_neg_predicate(WFD,Infos,Pred)),nl,
        register_predicate_wfd(WFD,Pred,Infos,Res,Reused,Ai,Ao), negate(Res,NegRes),
        (Reused=true
           ->  true
           ;   b_check_boolean_expression4(Pred,Infos,LocalState,State,WFD,Res)
        ).

% Register a new predicate in the AVL tree if either forced or non-WD condition inside
register_predicate_wfd(_,_Pred,_Infos,_,false,Ai,Ao) :- 
   preferences:preference(use_common_subexpression_elimination,true),
   preferences:preference(use_common_subexpression_also_for_predicates,true),!, % CSE already statically detects these ( all of the time !?)
   Ai=Ao.
register_predicate_wfd(wfwd(_,A,B),Pred,Infos,PredTruthVar,Reused,Ai,Ao) :-
   (A==B -> register_predicate_aux(Pred,PredTruthVar,Reused,Ai,Ao) % it will be evaluated; we can share it 
    ; nonmember(contains_wd_condition,Infos) -> register_predicate_aux(Pred,PredTruthVar,Reused,Ai,Ao)
    ;  Ai=Ao, Reused=false % not guaranteed to be evaluated: WD-condition + not in forced WFD context
      % ,print('NOT REGISTERING: '), translate:print_bexpr(Pred),nl
    ). 
% Register a new predicate in the AVL tree
register_predicate(Pred,_Infos,PredTruthVar,Reused,Ai,Ao) :-
    register_predicate_aux(Pred,PredTruthVar,Reused,Ai,Ao).
register_predicate_aux(Pred,PredTruthVar,Reused,Ai,Ao) :- check_pred_truth_var(PredTruthVar),
   norm_pred_check(Pred,NPred), % We could store this information in the info field computed by ast_cleanup ?
   (%too_simple(NPred) -> Reused=false, Ao=Ai ; %% even for simple equalities it actually pays off !
     reuse_predicate(NPred,Var,Ai)
           -> %% print(reusing(NPred,Var)),nl, %%
              PredTruthVar=Var,Ao=Ai, Reused=true
           ;  add_predicate(NPred,PredTruthVar,Ai,Ao), Reused=false
              %, portray_avl(Ao),nl
   ).



check_pred_truth_var(X) :- var(X),!.
check_pred_truth_var(pred_true) :- !.
check_pred_truth_var(pred_false) :- !.
check_pred_truth_var(X) :- add_internal_error('Illegal Predicate Truth Value: ',check_pred_truth_var(X)).

:- use_module(b_enumerate, [b_tighter_enumerate_values/2]).
expand_quantifier(Parameters,Pred,ListOfNewLocalStates) :-
 % at the moment LS,State not really needed; only necessary if non-compiled Pred can be used
 % also: feeding in any non-bound variables in LocalState or State would cause problems in findall !
    findall(ParLocalState,
             (b_interpreter:set_up_typed_localstate(Parameters,_ParaValues,ParamTypedVals,[],ParLocalState,all_solutions),
              kernel_waitflags:init_wait_flags(WF),
              b_interpreter:b_test_boolean_expression(Pred,[],ParLocalState,WF),
              b_tighter_enumerate_values(ParamTypedVals,WF),
              kernel_waitflags:ground_wait_flags(WF)),
            ListOfNewLocalStates).

check_expanded_forall_quantifier([], _Body, _LS, _State,_WF,_WFD,Res) :- 
     Res=pred_true.
check_expanded_forall_quantifier([LS1|TLS], Body, LS, State,WF,WFD,Res) :- 
     conjoin(Res1,TRes,Res,Body,Body,WF),
     empty_avl(InnerAi),
     append(LS1,LS,InnerLS),
     % Note: we do not need to guard against wd-definition from other instances inside a quantified expression
     %  either all conjuncts can be evaluated or none
     % print(expand_forall(WFD)), translate:print_bexpr(Body),nl,
     b_check_boolean_expression1(Body,InnerLS,State,WFD,Res1,InnerAi,_Aii),
     %instantiate_wfwd_result(WDE,WDV,Res1),
     check_expanded_forall_quantifier(TLS,Body,LS,State,WF,WFD,TRes).

/*
:- block instantiate_wfwd_result(?,-,-).
% instantiate a boolean variable in case it is no longer needed and not set by something else
instantiate_wfwd_result(WDE,WDV,Res) :-
     (nonvar(Res) -> true
      ; WDE==WDV -> true
      ; Res = pred_false). */

check_expanded_exists_quantifier([], _Body, _LS, _State,_WF,_WFD,Res) :- 
     Res=pred_false.
check_expanded_exists_quantifier([LS1|TLS], Body, LS, State,WF,WFD,Res) :- 
     disjoin(Res1,TRes,Res,Body,Body,WF),
     empty_avl(InnerAi),
     append(LS1,LS,InnerLS),
     b_check_boolean_expression1(Body,InnerLS,State,WFD,Res1,InnerAi,_Aii),
     % Note: we do not need to guard against wd-definition from other instances inside a quantified expression
     check_expanded_exists_quantifier(TLS,Body,LS,State,WF,WFD,TRes).
     

   :- use_module(library(lists),[maplist/4]).
% try and convert a closure into a list of 0/1 variables for each potential element
reify_closure_with_small_cardinality(P,T,Body, WF,ReifiedList) :- %print(try(P)),nl,
    maplist(create_typed_id,P,T,Parameters),
    small_quantifier_cardinality(Parameters,Body,LHS,RHS,350,25000), % TO DO: how to choose these parameters ?
    % for card({x|x:1..n & x mod 3=0})=c & n=24000 -> 340 ms with reification; 320 ms without; n=74000 : 1020 ms without, 1160 with reification
    % but there is a big difference for card({x|x:1..n & x mod 3=0 & x<10}) with n=74000 : 0 ms without reification, 1580 with;  n=500: 20 ms with reification; n=250: 10 ms with reification
    %print(clos_card(Card,LHS,RHS)),nl,
    expand_quantifier(Parameters,LHS,List),
    %print(list(List)),nl,
    WFD=wfwd(WF,pred_true,pred_true), % is this ok ??
    reifiy_list(List,RHS,WFD,ReifiedList).

reifiy_list([], _Body,_WFD,[]).
reifiy_list([LS1|TLS], Body,WFD,[Res_01|TRes]) :- 
     empty_avl(InnerAi),
     b_check_boolean_expression1(Body,LS1,[],WFD,Res1_pred,InnerAi,_Aii),
     %print(reified(Res1_pred,LS1)),nl,
     prop_pred_01(Res1_pred,Res_01),
     % Note: we do not need to guard against wd-definition from other instances inside a quantified expression
     reifiy_list(TLS,Body,WFD,TRes).



:- use_module(library(lists),[select/3]).

% check if Body produces a small cardinality for the paramters Par
% if yes: the predicates constraining Par are put into LHS, the rest into RHS
% also: LHS must ensure that ground values are produced for Par and that we can enumerate with a separate WF (in expand_quantifier)
small_quantifier_cardinality(Par,Body,LHS,RHS) :-
   small_quantifier_cardinality(Par,Body,LHS,RHS,10,35).
small_quantifier_cardinality(Par,Body,LHS,RHS,NormalLimit,SMTLimit) :-
   conjunction_to_list(Body,BodyList),
   small_quantifier_cardinality_aux(Par,BodyList,_UpBoundOnSize,LLHS,LRHS,NormalLimit,SMTLimit),
   conjunct_predicates(LLHS,LHS),
   conjunct_predicates(LRHS,RHS).

% do not rely on size for anything: it is just an upper bound on the size; the actual size could be smaller
small_quantifier_cardinality_aux([],Body,Size,LHS,RHS,_,_) :- !,
  LHS=[],RHS=Body,Size=1.
small_quantifier_cardinality_aux(Parameters,[LHS|RHS],FullSize,FullLHS,RestRHS,NormalLimit,SMTLimit) :- 
  LHS = b(member(SID,b(MemRHS,_,_)),pred,_),
  ((select(ID,Parameters,RestParameters),
   same_id(SID,ID,_)) -> true),
  is_small_set(MemRHS,Size,NormalLimit,SMTLimit),!,
  small_quantifier_cardinality_aux(RestParameters,RHS,RestSize,RestLHS,RestRHS,NormalLimit,SMTLimit),
  FullSize is Size*RestSize,
  FullLHS = [LHS|RestLHS],
  is_small_size(FullSize,NormalLimit,SMTLimit).
small_quantifier_cardinality_aux(Parameters,[H|Rest],FullSize,RestLHS,[H|RestRHS],NormalLimit,SMTLimit) :- !,
   % skip the conjunct H
   small_quantifier_cardinality_aux(Parameters,Rest,FullSize,RestLHS,RestRHS,NormalLimit,SMTLimit).
small_quantifier_cardinality_aux(Parameters,Body,ParCard,[],Body,NormalLimit,SMTLimit) :- 
  b_interpreter:parameter_list_cardinality(Parameters,ParCard),
  is_small_size(ParCard,NormalLimit,SMTLimit).

:- use_module(custom_explicit_sets,[efficient_card_for_set/3]).

%is_small_set(S,_,LocalState,State) :- print(is_small_set(S,LocalState,State)),nl,fail.

is_small_set(Val,Size,NormalLimit,SMTLimit) :-
    get_small_set_size(Val,Size),
    is_small_size(Size,NormalLimit,SMTLimit).

get_small_set_size(value(S),Size) :- !,
%print(value(S)),nl,
    ground(S),  % otherwise we could have S=[X] and expand_quantifier will erroneously create multiple solutions for parameter=X
    efficient_card_for_set(S,Size,C),
    call(C).
get_small_set_size(interval(From,To),Size) :- !, 
    custom_explicit_sets:is_interval_set(interval(From,To),Low,Up),
    number(Low), number(Up),
    (Low > Up -> Size = 1 ; Size is 1+Up-Low).
% we provide dom/ran here explicitly as this is often used {i|i:dom(f) ...}
get_small_set_size(domain(b(Val,_,_)),Size) :- !, get_small_domain_set_size(Val,Size). % this is only an upper bound on the size !! 
get_small_set_size(range(b(Val,_,_)),Size) :- !, get_small_set_size(Val,Size). % this is only an upper bound on the size !! 
%get_small_set_size(set_extension(L),Size,_,_,_) :- !, length(L,Size), is_small_size(Size). what if elements itself notknown ??
%get_small_set_size(sequence_extension(L),Size,_,_,_) :- !, length(L,Size), is_small_size(Size). ditto
%get_small_set_size(interval...
%get_small_set_size(domain(value([....]))...
% for this to work efficiently one should call b_compiler:compile on the predicates before sending them
% to b_interpreter_check; otherwise the sets will not yet be in value(_) form

% TO DO: the same for range or find more principled solution
get_small_domain_set_size(value([H|T]),Size) :- !, 
    efficient_card_for_set([H|T],Size,C),
    ground_domain([H|T]), % rather than requiring ground of entire list; we only require ground for domain (see test 1272)
    call(C).
get_small_domain_set_size(Val,Size) :- get_small_set_size(Val,Size).
ground_domain(V) :- var(V),!,fail.
ground_domain([]).
ground_domain([H|T]) :- nonvar(H), H=(D,_), ground(D), ground_domain(T).

% maybe in other cases try b_compiler:compile ?
%is_small_set(EXPR,Size,Parameters,LocalState,State) :-
%     print('check is_small_set: '),translate:print_bexpr(EXPR),nl,
%     get_texpr_id_names(Parameters,PN),
%     print(parameters(PN)),nl,
%     b_compiler:b_compile(b(EXPR,pred,[]),PN,LocalState,State,CLHS),
%     print('compiled: '), print(CLHS),nl,
%     CLHS = value(S),efficient_card_for_set(S,Size,C),call(C), is_small_size(Size).

is_small_size(Size,NormalLimit,SMTLimit) :- Size \= inf,
   (Size < NormalLimit -> true
      ; Size < SMTLimit -> preferences:preference(use_smt_mode,true)).
    


b_wd_compute_expressions([], _, _, [],_WFD).
b_wd_compute_expressions([EXPRsHd|EXPRsTl],LocalState,State,[ValHd|ValTl],WFD) :-
  b_wd_compute_expression(EXPRsHd,LocalState,State,ValHd,WFD),
  b_wd_compute_expressions(EXPRsTl,LocalState,State,ValTl,WFD). 
    
b_wd_compute_expression(Expr,LocalState,State,Value,wfwd(WF,WDE,WDV)) :-
       (nonvar(WDV)
         -> (WDE==WDV
             -> b_compute_expression(Expr,LocalState,State,Value,WF)
              ; instiantiate_to_any_value(Value,Expr,WF))
         ;  always_well_defined(Expr) % should we use always_well_defined_or_disprover_mode ?
              ->   % print('ALWAYS WD: '), translate:print_bexpr(Expr),nl, %
                 b_compute_expression(Expr,LocalState,State,Value,WF)
     %    ;  \+ preferences:preference(use_smt_mode,_____false) %,\+ may_instantiate(Expr)
     %        ->
     %     % Necessary for tests 323, 324 : but dangerous inside set comprehensions ,... !!
     %       print('NOT DELAYING WD in SMT MODE: '), translate:print_bexpr(Expr),nl,
     %        b_compute_expression(Expr,LocalState,State,Value,WF)
         ;   % print('DELAYING DUE TO WD: '), translate:print_bexpr(Expr),nl, %%
            b_compute_expression_delay(WDE,WDV, Expr,LocalState,State,Value,WF)
            % TO DO: allow further partial instantiation
         ).

/*
% what if function is used inside Expr ??
may_instantiate(b(E,_,_Infos)) :- !,may_instantiate_aux(E).
% Warning: function application can force a variable to be bound to non-empty set !
may_instantiate_aux(function(_,_)).
*/

:- block b_compute_expression_delay(?,-, ?,?,?,?,?).
b_compute_expression_delay(WDE,WDV,Expr,LocalState,State,Value,WF) :-
    (WDE==WDV
      ->  % print('WD Evaluation: '), print(WDE),print(' =?= '), print(WDV), print(' '),translate:print_bexpr(Expr),nl,
         b_compute_expression(Expr,LocalState,State,Value,WF)
         % ,print(' Value: '), print(Value),nl
      ;  %print(instiantiate_to_any_value(Value,Expr)),nl,
         instiantiate_to_any_value(Value,Expr,WF) % does not matter anyway; but there can be pending co-routines :-(
    ).

% WARNING: the variable could be used in another context, where it is relevant !

:- use_module(kernel_freetypes,[get_freeval_type/3]).

%instiantiate_to_any_value(V,_) :- print(inst(V)),nl,nonvar(V),!.
instiantiate_to_any_value(V,_,_) :- ground(V),!.
instiantiate_to_any_value(V,b(_B,TYPE,_I),WF) :-
   get_enumeration_finished_wait_flag(WF,EWF),
   when((nonvar(EWF);ground(V)),inst1(TYPE,V)). % , print(inst2(TYPE,V)),translate:print_bexpr(b(_B,TYPE,_I)),nl.
:- use_module(b_global_sets,[b_fd_type/3]).
inst1(T,V) :- if(inst2(T,V),true, (nl,print('*** instantiate_to_any_value failed: '), print(T:V),nl,nl)).
%inst2(_,V) :- ground(V),!. 
%inst2(Type,V) :- print(inst2(Type,V)),nl,trace,fail.
inst2(boolean,V) :- !, inst3(V,pred_true).
inst2(integer,V) :- !, V=int(X),inst3(X,0).
inst2(set(_),V) :- !,  inst3(V,[]). % TO DO: what if V already partially instantiated
inst2(seq(_),V) :- !,  inst3(V,[]). % TO DO: what if V already partially instantiated
inst2(couple(TA,TB),V) :- !, V=(A,B), inst2(TA,A), inst2(TB,B).
inst2(string,V) :- !, V=string(X), inst3(X,'').
inst2(global(T),V) :- !, V=fd(X,T), %(T=='MODES' -> trace ; true),
  b_global_sets:b_fd_type(T,LowBnd,_UpBnd),
  inst3(X,LowBnd).
inst2(record(_F),_V) :- !, true. % TO DO
inst2(freetype(FT),V) :- !, V=freeval(FT,CASE,Val),
  (nonvar(CASE) -> (get_freeval_type(FT,CASE,Type) -> inst1(Type,Val)) % ensure no backtracking
    ; nonvar(Val) -> print((nonvar_value(V))),nl, get_freeval_type(FT,CASE,Type), inst1(Type,Val)
    ; (get_freeval_type(FT,C,Type) -> CASE=C, inst1(Type,Val))
  ).
inst2(X,V) :- print(cannot_inst2(X,V)),nl.

inst3(X,V) :- (var(X) -> X=V ; true).

:- block propagagate_wfwd(-,?,?,-,?), propagagate_wfwd(?,-,?,-,?).
% propagate expected and actual value guarding left with actual predicate value obtained for left
propagagate_wfwd(WDE,WDV,Res,F1,F2) :-
  (F1==F2 -> Res=F1  % then value of WDE does not matter at all; Res always the same
  ; propagagate_wfwd2(WDE,WDV,Res,F1,F2)).

:- block propagagate_wfwd2(-,?,?,?,?), propagagate_wfwd2(?,-,?,?,?).
propagagate_wfwd2(WDE,WDV,Res,F1,F2) :- (WDE==WDV -> Res=F1 ; Res=F2).

:- use_module(translate,[print_bexpr/1]).
:- use_module(external_functions,[call_external_predicate/6]).
:- use_module(preferences).
b_check_boolean_expression4(equal(LHS,RHS),_,LocalState,State,WFD,EqRes) :- !,   
        b_wd_compute_expression(LHS,LocalState,State,LHValue,WFD),
        b_wd_compute_expression(RHS,LocalState,State,RHValue,WFD),
        get_texpr_type(LHS,Type),
        %translate:translate_bexpression(b(equal(LHS,RHS),pred,[]),SSS),preq(EqRes,SSS),
        % was translate_eq(Eq,Res),
        %when(nonvar(EqRes),(print(EqRes),print(' : '),translate:print_bexpr(b(equal(LHS,RHS),pred,[])),nl)),
        %print(equality(Type,LHValue,RHValue,WFD)), translate:print_bexpr(b(equal(LHS,RHS),pred,[])),nl,
        equality_objects_with_type(Type,LHValue,RHValue,EqRes).
%:- block preq(-,?).
%preq(Res,Expr) :- print('Equality determined '), print(Res), print(' '),print(Expr),nl.
% we may need to improve not_member for closure to invert symbolic operators
% we may want to add things like partial_function_check, subset_check, ...
b_check_boolean_expression4(member(LHS,RHS),_,LocalState,State,WFD,Res) :- 
        %member_check_should_be_reified(LHS,RHS), % no longer need this: symbolic operators will be kept as closures if large ?!
        !, 
        get_texpr_expr(RHS,ERHS),
  
  b_check_member_expression(ERHS,RHS,LHS,LocalState,State,WFD,Res).
% TO DO: add member(,pow_subset, fin_subset) --> subset_test
b_check_boolean_expression4(subset(LHS,RHS),_,LocalState,State,WFD,Res) :- !, WFD=wfwd(WF,_,_),
        b_wd_compute_expression(LHS,LocalState,State,LHValue,WFD),
        b_wd_compute_expression(RHS,LocalState,State,RHValue,WFD),
        % print(subset_check),print(' '),translate:print_bexpr(b(subset(LHS,RHS),pred,[])),nl, %%
        subset_test(LHValue,RHValue,Res,WF).
b_check_boolean_expression4(subset_strict(LHS,RHS),_,LocalState,State,WFD,Res) :- !, WFD=wfwd(WF,_,_),
        b_wd_compute_expression(LHS,LocalState,State,LHValue,WFD),
        b_wd_compute_expression(RHS,LocalState,State,RHValue,WFD),
        % print(subset_strict_check),print(' '),translate:print_bexpr(b(subset_strict(LHS,RHS),pred,[])),nl, %%
        subset_strict_test(LHValue,RHValue,Res,WF).
b_check_boolean_expression4(external_pred_call(FunName,Args),Info,LocalState,State,WFD,Res) :- 
        !, WFD=wfwd(WF,_,_),
        b_wd_compute_expressions(Args,  LocalState,State, EvaluatedArgs, WFD),  
        %print(external_pred_check),print(FunName),nl,%%
        external_functions:call_external_predicate(FunName,Args,EvaluatedArgs,Res,Info,WF).
b_check_boolean_expression4(freetype_case(FT,IsCase,Expr),_Infos,LocalState,State,WFD,Res) :- !,
        b_wd_compute_expression(Expr,LocalState,State,freeval(FT,ActualCase,_A),WFD),
        %print(eq_atomic(IsCase,ActualCase,FT,_A,Res,WFD)),nl,translate:print_bexpr(Expr),nl,
        eq_atomic(IsCase,ActualCase,freeval_case,Res).
b_check_boolean_expression4(Pred,_,LocalState,State,WFD,Res) :- 
        arithmetic_op(Pred,Op,LHS,RHS),!,  
        % print('checking arith: '), print(op(Op,Res)),nl,translate:print_bexpr(Pred),nl,   %
        b_wd_compute_expression(LHS,LocalState,State,int(LHValue),WFD),
        % kernel_objects:basic_type2(integer,int(LHValue)), %
        b_wd_compute_expression(RHS,LocalState,State,int(RHValue),WFD),
        % kernel_objects:basic_type2(integer,int(RHValue)), %
        check_arithmetic_operator(Op,LHValue,RHValue,Res). %
% TO DO ???: use idea of b_artihmetic_expression to avoid intermediate CLPFD variables
% TO DO: add other operators,  ...
/* use_smt_mode = full is never set; at some point we should enable the following clause by default
b_check_boolean_expression4(Pred,Infos,LocalState,State,wfwd(WF,WDE,WDV),Res) :-
         preferences:preference(use_smt_mode,full), %% comment out to enable check testing of complicated predicates inside;
         % causes slowdowns of cbtc/actions_cbtc.mch; TO DO: investigate why
        b_check_boolean_expression4_delay(WDE,WDV,Pred,Infos,LocalState,State,WF,Res).
*/

% TO DO: add member(,pow_subset, fin_subset) --> subset_test
% we could distribute RHS=union -> disjunction, LHS=intersection -> conjunction ,... ?
%b_check_member_expression(EHRS,_RHS,_LHS,_LocalState,_State,_WFD,_Res) :-
%        print('member : '), print(EHRS),nl,fail.
%b_check_member_expression(union(A,B),LHS,LocalState,State,WFD,Res) :-
b_check_member_expression(pow_subset(RHS),_,LHS,LocalState,State,WFD,Res) :- !,
        b_check_boolean_expression4(subset(LHS,RHS),[],LocalState,State,WFD,Res).
b_check_member_expression(_,RHS,LHS,LocalState,State,WFD,Res) :- WFD=wfwd(WF,_,_),
        b_wd_compute_expression(LHS,LocalState,State,LHValue,WFD),
        b_wd_compute_expression(RHS,LocalState,State,RHValue,WFD),
        %translate_mem(Mem,Res),
        % print(mem_check),print(' '),translate:print_bexpr(b(member(LHS,RHS),pred,[])),nl,
        % print(mem_test(RHValue,LHValue,Res)),nl, trace,
        membership_test_wf(RHValue,LHValue,Res,WF),
        force_membership_test(Res,LHValue,RHValue,WF).

/*
:- block b_check_boolean_expression4_delay(?,-,?,?,?,?,?,?).
b_check_boolean_expression4_delay(WDE,WDV,_,_,_,_,_,Res) :- WDE \= WDV,!,
        Res=pred_false. % does not matter
b_check_boolean_expression4_delay(_WDE,_WDV,Pred,Infos,LocalState,State,WF,Res) :- 
        % we currently have not yet implemented a way to check the Pred; wait until Result is known
        (preferences:preference(use_smt_mode,false)
          -> 
             get_last_wait_flag(b_check_test_boolean_expression,WF,WF2)
            %Prio=10000,get_wait_flag(Prio,b_check_test_boolean_expression,WF,WF2)
          ; get_binary_choice_wait_flag(b_check_test_boolean_expression,WF,WF2)
        ),
        (debug:debug_mode(on) -> print(' Check Testing: '),translate:print_bexpr(Pred),nl ; true),
        b_check_test_boolean_expression(Res,WF2,b(Pred,pred,Infos),LocalState,State,WF).
        
:- block b_check_test_boolean_expression(-,-,?,?,?,?).
b_check_test_boolean_expression(pred_true,_,Pred,LocalState,State,WF) :-
      b_test_boolean_expression(Pred,LocalState,State,WF).
b_check_test_boolean_expression(pred_false,_,Pred,LocalState,State,WF) :-
      b_not_test_boolean_expression(Pred,LocalState,State,WF).
 */   

/*
:- use_module(kernel_mappings).
member_check_should_be_reified(_,b(RHS,_,_)) :- functor(RHS,BOP,Arity), print(check(BOP,Arity)),nl,!.
member_check_should_be_reified(_,_) :-  \+ preferences:preference(use_smt_mode,false),!.
member_check_should_be_reified(_LHS,b(RHS,_,_)) :- functor(RHS,BOP,Arity), 
  % check if we have optimized treatments available, for which we do not yet have reified versions
        (Arity=1 -> \+ kernel_mappings:unary_in_boolean_type(BOP,_)
         ; Arity=2 -> \+ kernel_mappings:binary_in_boolean_type(BOP,_)
         ; Arity=0 -> \+  kernel_mappings:cst_in_boolean_type(BOP,_)
         ; true).
*/
      

:- block force_membership_test(-,?,?,?).
% currently required for ensuring that following fails:
% kernel_objects:union(closure(['_zzzz_unit_tests'],[integer],b(member(b(identifier('_zzzz_unit_tests'),integer,[generated]),b(value([int(3),int(4)]),set(integer),[])),pred,[])),closure(['_zzzz_unit_tests'],[integer],b(member(b(identifier('_zzzz_unit_tests'),integer,[generated]),b(value([int(2),int(1)]),set(integer),[])),pred,[])),[int(1),int(3),int(2)])
% Reason: membership_test does not on its own enumerate
force_membership_test(pred_true,X,Set,WF) :-  % print(forcing_element_true(X,Set)),nl,%
     (ground(X) -> true
      ; kernel_objects:check_element_of_wf(X,Set,WF)
        )
     %, tools:bt_trace(force_membership)
     %, print(forced(X)),nl
     .
force_membership_test(pred_false,_X,_Set,_WF).


:- use_module(library(avl)).
reuse_predicate(Pred,Var,AVL) :- avl_fetch(Pred,AVL,Var). %pred_var(Var)).
add_predicate(Pred,Var,AVL,NewAVL) :- avl_store(Pred,AVL,Var,NewAVL). %pred_var(Var),NewAVL).

% normalies a typed predicate by removing position information and ordering commutative operators in a canonical way

norm_pred_check(B,Res) :- 
  (norm_pred(B,Res) -> true 
   ; bget_functor(B,F,N),print(norm_pred_failed(F/N)),nl, translate:print_bexpr(B),nl, % print(B),nl,
     Res=B).
bget_functor(b(B,_,_),F,N) :- functor(B,F,N).
bget_functor(B,F,N) :- functor(B,F,N).

norm_pred(X,Res) :- var(X),!,Res=X.
norm_pred(b(B,_,_),Res) :- !, norm_pred(B,Res).
norm_pred(conjunct(A,B),Res) :- !, norm_pred(A,AA), norm_pred(B,BB),
   (BB @< AA -> Res = conjunct(AA,BB) ; Res= conjunct(BB,AA)).
norm_pred(disjunct(A,B),Res) :- !, norm_pred(A,AA), norm_pred(B,BB),
   (BB @< AA -> Res = disjunct(AA,BB) ; Res= disjunct(BB,AA)).
norm_pred(equal(A,B),Res) :- !,norm_expr(A,AA),norm_expr(B,BB),
   (BB @< AA -> Res = equal(AA,BB) ; Res= equal(BB,AA)).
norm_pred(equivalence(A,B),Res) :- !, norm_pred(A,AA), norm_pred(B,BB),
   (BB @< AA -> Res = equivalence(AA,BB) ; Res= equivalence(BB,AA)).
norm_pred(exists(A,B),Res) :- !, Res=exists(AA,BB),l_norm_expr(A,AA), norm_pred(B,BB).
norm_pred(forall(A,B,C),Res) :- !, Res=forall(AA,BB,CC),l_norm_expr(A,AA), norm_pred(B,BB), norm_pred(C,CC).
norm_pred(greater(A,B),less(BB,AA)) :- !,norm_expr(A,AA),norm_expr(B,BB).
norm_pred(greater_equal(A,B),less_equal(BB,AA)) :- !,norm_expr(A,AA),norm_expr(B,BB).
norm_pred(implication(A,B),Res) :- !, Res=implication(AA,BB),norm_pred(A,AA), norm_pred(B,BB).
norm_pred(less(A,B),less(AA,BB)) :- !,norm_expr(A,AA),norm_expr(B,BB).
norm_pred(less_equal(A,B),less_equal(AA,BB)) :- !,norm_expr(A,AA),norm_expr(B,BB).
norm_pred(let_predicate(A,B,C),Res) :- !,
   Res=let_predicate(AA,BB,CC),
   l_norm_expr(A,AA), l_norm_expr(B,BB),norm_pred(C,CC).
norm_pred(lazy_let_pred(A,B,C),Res) :-  !,
   Res = lazy_let_pred(AA,BB,CC),
   norm_expr(A,AA),
   norm_pred_or_expr(B,BB),norm_pred(C,CC).
norm_pred(lazy_lookup_pred(A),Res) :- !, Res = lazy_lookup_pred(A).
norm_pred(member(A,B),member(AA,BB)) :- !,norm_expr(A,AA),norm_expr(B,BB).
norm_pred(negation(A),Res) :- !, Res=negation(AA), norm_pred(A,AA).
norm_pred(not_equal(A,B),Res) :- !,norm_expr(A,AA),norm_expr(B,BB),
   (BB @< AA -> Res = not_equal(AA,BB) ; Res= not_equal(BB,AA)).
norm_pred(not_member(A,B),not_member(AA,BB)) :- !,norm_expr(A,AA),norm_expr(B,BB).
norm_pred(not_subset(A,B),not_subset(AA,BB)) :- !,norm_expr(A,AA),norm_expr(B,BB).
norm_pred(not_subset_strict(A,B),not_subset_strict(AA,BB)) :- !,norm_expr(A,AA),norm_expr(B,BB).
norm_pred(subset(A,B),subset(AA,BB)) :- !,norm_expr(A,AA),norm_expr(B,BB).
norm_pred(subset_strict(A,B),subset_strict(AA,BB)) :- !,norm_expr(A,AA),norm_expr(B,BB).
norm_pred(freetype_case(Type,Case,A),freetype_case(Type,Case,AA)) :- !,norm_expr(A,AA).
norm_pred(X,X). % :- print(norm_pred(X)),nl.

norm_pred_or_expr(b(B,pred,_),Res) :- norm_pred(B,Res).
norm_pred_or_expr(B,Res) :- norm_expr(B,Res).

:- public norm_expr_check/2.
norm_expr_check(X,Res) :- var(X),!,Res=X.
norm_expr_check(b(B,_,_),Res) :- !, norm_expr_check2(B,Res).
norm_expr_check(X,X).

norm_expr_check2(B,Res) :- 
  (norm_expr2(B,Res) -> true 
   ; functor(B,F,N),print(norm_expr2_failed(F/N)),nl,
     Res=B).

norm_expr(X,Res) :- var(X),!,Res=X.
norm_expr(b(B,_,_),Res) :- !, norm_expr2(B,Res). 
                              %norm_expr_check2(B,Res). %% comment in to obtain details about failed normalisation for expressions
norm_expr(X,X).

norm_expr2(assertion_expression(Cond,E,Expr),assertion_expression(AA,E,BB)) :- norm_pred(Cond,AA),norm_expr(Expr,BB).
norm_expr2(add(A,B),Res) :- norm_expr(A,AA), norm_expr(B,BB), (BB @< AA -> Res = add(AA,BB) ; Res= add(BB,AA)).
norm_expr2(bag_items(A),bag_items(AA)) :-  norm_expr(A,AA).
norm_expr2(boolean_false,boolean_false).
norm_expr2(boolean_true,boolean_true).
norm_expr2(bool_set,bool_set).
norm_expr2(card(A),card(AA)) :-  norm_expr(A,AA).
norm_expr2(cartesian_product(A,B),cartesian_product(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(closure(A),closure(AA)) :-  norm_expr(A,AA).
norm_expr2(closure1(A),closure1(AA)) :-  norm_expr(A,AA).
norm_expr2(compaction(A),compaction(AA)) :-  norm_expr(A,AA).
norm_expr2(composition(A,B),composition(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(comprehension_set(A,B),comprehension_set(AA,BB)) :- norm_expr(A,AA), norm_pred(B,BB).
norm_expr2(concat(A,B),concat(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(convert_bool(A),convert_bool(AA)) :- norm_pred_check(A,AA).
norm_expr2(couple(A,B),couple(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(direct_product(A,B),direct_product(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(div(A,B),div(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(domain_restriction(A,B),domain_restriction(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(domain_subtraction(A,B),domain_subtraction(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(domain(A),domain(AA)) :- norm_expr(A,AA).
norm_expr2(empty_sequence,empty_sequence).
norm_expr2(empty_set,empty_set).
norm_expr2(event_b_identity,event_b_identity).
norm_expr2(external_function_call(A,B),external_function_call(A,BB)) :- l_norm_expr(B,BB).
norm_expr2(fin_subset(A),fin_subset(AA)) :- norm_expr(A,AA).
norm_expr2(fin1_subset(A),fin1_subset(AA)) :- norm_expr(A,AA). 
norm_expr2(first(A),first(AA)) :- norm_expr(A,AA). 
norm_expr2(first_of_pair(A),first_of_pair(AA)) :- norm_expr(A,AA).
norm_expr2(front(A),front(AA)) :- norm_expr(A,AA). 
norm_expr2(function(A,B),function(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(general_concat(A),general_concat(AA)) :- norm_expr(A,AA).
norm_expr2(general_intersection(A),general_intersection(AA)) :- norm_expr(A,AA).
norm_expr2(general_product(A,B,C),Res) :- !, Res=general_sum(AA,BB,CC),l_norm_expr(A,AA), norm_pred(B,BB), norm_expr(C,CC).
norm_expr2(general_sum(A,B,C),Res) :- !, Res=general_sum(AA,BB,CC),l_norm_expr(A,AA), norm_pred(B,BB), norm_expr(C,CC).
norm_expr2(general_union(A),general_union(AA)) :- norm_expr(A,AA).
norm_expr2(identifier(A),'$'(A)). % need wrapper to avoid confusion with other terms !
norm_expr2(identity(A),identity(AA)) :- norm_expr(A,AA).
norm_expr2(if_then_else(P,A,B),if_then_else(PP,AA,BB)) :- norm_pred(P,PP),norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(image(A,B),image(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(insert_front(A,B),insert_front(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(insert_tail(A,B),insert_tail(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(integer_set(A),A).
norm_expr2(integer(A),A).
norm_expr2(intersection(A,B),Res) :- norm_expr(A,AA), norm_expr(B,BB), (BB @< AA -> Res = union(AA,BB) ; Res= intersection(BB,AA)).
norm_expr2(interval(A,B),interval(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(iseq(A),iseq(AA)) :- norm_expr(A,AA).
norm_expr2(iseq1(A),iseq1(AA)) :- norm_expr(A,AA).
norm_expr2(iteration(A,B),iteration(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(last(A),last(AA)) :- norm_expr(A,AA). 
norm_expr2(lazy_let_expr(A,B,C),lazy_let_expr(AA,BB,CC)) :-
     norm_expr(A,AA),norm_pred_or_expr(B,BB),norm_expr(C,CC).
norm_expr2(lazy_lookup_expr(A),lazy_lookup_expr(A)) :- !.
norm_expr2(let_expression(A,B,C),let_expression(AA,BB,CC)) :- l_norm_expr(A,AA), l_norm_expr(B,BB),norm_pred(C,CC).
norm_expr2(max(A),max(AA)) :- norm_expr(A,AA). 
norm_expr2(max_int,max_int).
norm_expr2(min(A),min(AA)) :- norm_expr(A,AA). 
norm_expr2(min_int,min_int).
norm_expr2(minus(A,B),minus(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(modulo(A,B),modulo(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(mu(A),mu(AA)) :-  norm_expr(A,AA).
norm_expr2(multiplication(A,B),Res) :- norm_expr(A,AA), norm_expr(B,BB), (BB @< AA -> Res = multiplication(AA,BB) ; Res= multiplication(BB,AA)).
norm_expr2(overwrite(A,B),overwrite(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(parallel_product(A,B),parallel_product(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(partial_bijection(A,B),partial_bijection(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(partial_function(A,B),partial_function(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(partial_injection(A,B),partial_injection(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(partial_surjection(A,B),partial_surjection(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(perm(A),perm(AA)) :- norm_expr(A,AA). 
norm_expr2(power_of(A,B),power_of(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(pow_subset(A),pow_subset(AA)) :- norm_expr(A,AA). 
norm_expr2(pow1_subset(A),pow1_subset(AA)) :- norm_expr(A,AA). 
norm_expr2(range_restriction(A,B),range_restriction(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(range_subtraction(A,B),range_subtraction(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(range(A),range(AA)) :- norm_expr(A,AA).
norm_expr2(rec(A),rec(A)).  % should we do norm_fields here ?
norm_expr2(record_field(A,B),record_field(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(relations(A,B),relations(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(restrict_front(A,B),restrict_front(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(restrict_tail(A,B),restrict_tail(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(reflexive_closure(A),reflexive_closure(AA)) :- norm_expr(A,AA). 
norm_expr2(rev(A),rev(AA)) :- norm_expr(A,AA). 
norm_expr2(reverse(A),reverse(AA)) :- norm_expr(A,AA).
norm_expr2(second_of_pair(A),second_of_pair(AA)) :- norm_expr(A,AA).
norm_expr2(seq(A),seq(AA)) :- norm_expr(A,AA).
norm_expr2(seq1(A),seq1(AA)) :- norm_expr(A,AA).
norm_expr2(sequence_extension(L),sequence_extension(NL)) :- l_norm_expr(L,NL).
norm_expr2(set_extension(L),set_extension(NL)) :- l_norm_expr(L,NL).
norm_expr2(set_subtraction(A,B),set_subtraction(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(size(A),size(AA)) :- norm_expr(A,AA).
norm_expr2(string(A),string(A)). % need wrapper to avoid confusion with other terms !
norm_expr2(string_set,string_set).
norm_expr2(struct(A),struct(A)).  % should we do norm_fields here ?
norm_expr2(surjection_relation(A,B),surjection_relation(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(tail(A),tail(AA)) :- norm_expr(A,AA). 
norm_expr2(total_bijection(A,B),total_bijection(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(total_function(A,B),total_function(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(total_injection(A,B),total_injection(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(total_relation(A,B),total_relation(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(total_surjection(A,B),total_surjection(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).
norm_expr2(total_surjection_relation(A,B),total_surjection_relation(AA,BB)) :- norm_expr(A,AA), norm_expr(B,BB).

norm_expr2(unary_minus(A),unary_minus(AA)) :- norm_expr(A,AA).
norm_expr2(union(A,B),Res) :- norm_expr(A,AA), norm_expr(B,BB), (BB @< AA -> Res = union(AA,BB) ; Res= union(BB,AA)).
norm_expr2(value(A),value(A)).

norm_expr2(freetype_set(T),freetype_set(T)).
norm_expr2(freetype_constructor(FT,Case,A), freetype_constructor(FT,Case,AA)) :- norm_expr(A,AA).
norm_expr2(freetype_destructor(FT,Case,A),freetype_destructor(FT,Case,AA)) :- norm_expr(A,AA).
norm_expr2(recursive_let(A,B),recursive_let(AA,BB)) :- norm_expr(A,AA),norm_expr(B,BB).

l_norm_expr([],[]).
l_norm_expr([H|T],[NH|NT]) :- norm_expr(H,NH), l_norm_expr(T,NT).


arithmetic_op(less(LHS,RHS),'<',LHS,RHS).
arithmetic_op(less_equal(LHS,RHS),'<=',LHS,RHS).
arithmetic_op(greater(LHS,RHS),'<',RHS,LHS).
arithmetic_op(greater_equal(LHS,RHS),'<=',RHS,LHS).

:- use_module(clpfd_interface).
check_arithmetic_operator('<',X,Y,Res) :- check_less(X,Y,Res),
  (nonvar(Res) -> true
   ; clpfd_interface:post_constraint( (X#<Y) #<=> R01), prop_pred_01(Res,R01)).
check_arithmetic_operator('<=',X,Y,Res) :- check_less_than_equal(X,Y,Res),
  (nonvar(Res) -> true
   ; clpfd_interface:post_constraint( (X#=<Y) #<=> R01), prop_pred_01(Res,R01)).

:- block prop_pred_01(-,-).
% prop_pred_01(A,B) :- print(prop(A,B)),nl,fail.
prop_pred_01(pred_true,1).
prop_pred_01(pred_false,0).

:- block check_less(-,?,-), check_less(?,-,-).
check_less(X,Y,Res) :- nonvar(Res),!, /* truth value known: enforce it */
    % print(forcing_less(X,Y,Res)),nl, %%
   (Res=pred_true -> less_than_direct(X,Y) ; less_than_equal_direct(Y,X)).
check_less(X,Y,Res) :-   % print(check_less(X,Y)),nl,%%
    X < Y,!,/* we could call safe_less_than(X,Y), */
    Res=pred_true.
check_less(_,_,pred_false).
:- block check_less_than_equal(-,?,-), check_less_than_equal(?,-,-).
check_less_than_equal(X,Y,Res) :- nonvar(Res),!, /* truth value known: enforce it */
   %% print(forcing_less_than_equal(X,Y,Res)),nl,%%
   (Res=pred_true -> less_than_equal_direct(X,Y) ; less_than_direct(Y,X)).
check_less_than_equal(X,Y,Res) :- X =< Y,!,Res=pred_true.
check_less_than_equal(_,_,pred_false).



:- use_module(bool_pred).

:- use_module(kernel_objects,[exhaustive_kernel_check_wf/2,
  exhaustive_kernel_check/1, exhaustive_kernel_check/2, exhaustive_kernel_fail_check/1]).

:- assert_must_succeed(exhaustive_kernel_check_wf(b_interpreter_check:conjoin(pred_false,pred_false,pred_false,b(truth,pred,[]),b(truth,pred,[]),WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(b_interpreter_check:conjoin(pred_false,pred_true,pred_false,b(truth,pred,[]),b(truth,pred,[]),WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(b_interpreter_check:conjoin(pred_true,pred_false,pred_false,b(truth,pred,[]),b(truth,pred,[]),WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(b_interpreter_check:conjoin(pred_true,pred_true,pred_true,b(truth,pred,[]),b(truth,pred,[]),WF),WF)).
:- assert_must_fail(b_interpreter_check:conjoin(pred_true,pred_false,pred_true,b(truth,pred,[]),b(truth,pred,[]),_WF)).
:- assert_must_fail(b_interpreter_check:conjoin(pred_true,pred_true,pred_false,b(truth,pred,[]),b(truth,pred,[]),_WF)).


% same as and_equality/3 but for pred_false/pred_true rather than eq_obj/pred_false
:- block conjoin(-,-,-,?,?,?).
conjoin(X,Y,Res,LHS,RHS,WF) :-  %  print(conjoin(X,Y,Res)),nl, translate:print_bexpr(LHS),nl,%
   (   Res==pred_true -> X=pred_true,Y=pred_true
     ; X==pred_true  -> Res=Y
     ; X==pred_false -> Res=pred_false
     ; Y==pred_true  -> Res=X
     ; Y==pred_false -> Res=pred_false
     ; Res==pred_false -> conjoin_false0(X,Y,LHS,RHS,WF)
     ; add_error_fail(conjoin,'Illegal values: ', conjoin(X,Y,Res,LHS,RHS,WF))
   ).
conjoin_false0(X,Y,_LHS,_RHS,_WF) :- X==Y,!,
        %print(conjoin_false_eqeq(X,Y)),nl, translate:print_bexpr(_LHS),nl,
        X=pred_false.
conjoin_false0(X,Y,_LHS,_RHS,_WF) :-
        bool_negate_check(X,Y),!
        . %,print(conjoin_false_neqeq(X,Y)),nl,translate:print_bexpr(_LHS),nl.
conjoin_false0(X,Y,LHS,RHS,WF) :-
        %%Prio=1, %%
        %%(preferences:preference(use_smt_mode,full) -> FullPrio=1.5 ;
            get_priority_of_boolean_expression(LHS,Prio),
            count_number_of_conjuncts(RHS,NrC),  
            FullPrio is Prio+(NrC-1)/10,
        get_wait_flag(FullPrio,conjoin,WF,LWF), %%
        % print('SETUP conjoin_false: '), print(FullPrio), print(' '), print(NrC), print(' '), translate:print_bexpr(LHS),nl, % kernel_waitflags:portray_waitflags(WF),
        conjoin_false(X,Y,LHS,RHS,LWF).
% missing rule: if X==Y -> X=pred_true
:- block conjoin_false(-,-,?,?,-).
conjoin_false(X,Y,_LHS,_RHS,_LWF) :-  %  print(conjoin_false(X,Y)),nl, translate:print_bexpr(_LHS),nl,%
   (   X==pred_true  -> % print('TRUE: '), translate:print_bexpr(_LHS),nl,
                        % print(' FORCING FALSE: '), translate:print_bexpr(_RHS),nl,
                        pred_false=Y
     ; X==pred_false -> true  % , print('FALSE: '), translate:print_bexpr(_LHS),nl
     ; Y==pred_true  -> pred_false=X
     ; Y==pred_false -> true    %, print('FALSE: '), translate:print_bexpr(_RHS),nl
     ; (   % print(forcing_conjoin_lhs_false(X,Y,_LWF)), translate:print_bexpr(_LHS), print(' == FALSE '),nl,
         X=pred_false
         ;
            % print(forcing_conjoin_rhs_false(X,Y,_LWF)), translate:print_bexpr(_LHS), print(' == TRUE ; '), nl,
         X=pred_true,Y=pred_false )
   ).


:- assert_must_succeed(exhaustive_kernel_check(b_interpreter_check:disjoin(pred_false,pred_false,pred_false,_,_,_WF))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],b_interpreter_check:disjoin(pred_false,pred_true,pred_true,_,_,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(b_interpreter_check:disjoin(pred_true,pred_true,pred_true,_,_,_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(b_interpreter_check:disjoin(pred_true,pred_true,pred_false,_,_,_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(b_interpreter_check:disjoin(pred_true,pred_false,pred_false,_,_,_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(b_interpreter_check:disjoin(pred_false,pred_true,pred_false,_,_,_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(b_interpreter_check:disjoin(pred_false,pred_false,pred_true,_,_,_WF))).

:- block disjoin(-,-,-,?,?,?).
disjoin(X,Y,Res,LHS,RHS,WF) :-   %print(disjoin(X,Y,Res,WF)),nl, %% translate:print_bexpr(LHS),print(' or '),translate:print_bexpr(RHS),nl,%%
   (   Res==pred_false -> X=pred_false,Y=pred_false
     ; X==pred_true  -> Res=pred_true
     ; X==pred_false -> Res=Y
     ; Y==pred_true  -> Res=pred_true
     ; Y==pred_false -> Res=X
     ; Res==pred_true  -> disjoin_true0(X,Y,LHS,RHS,WF)
     ; add_error_fail(disjoin,'Illegal values: ',disjoin(X,Y,Res,LHS,RHS,WF))
   ).
disjoin_true0(X,Y,_LHS,_RHS,_WF) :- X==Y,!,
        %print(disjoin_true_eqeq(X,Y)),nl, translate:print_bexpr(_LHS),nl,
        X=pred_true.
%disjoin_true0(X,Y,LHS,_,WF) :- !, disjoin_true(X,Y,_).
disjoin_true0(X,Y,LHS,_RHS,WF) :- %% print('disjoin_true0 '),  translate:print_bexpr(LHS), print(' OR '),  translate:print_bexpr(_RHS),nl, %%
       %%(preferences:preference(use_smt_mode,full) -> FullPrio=2.5 ;
            get_priority_of_boolean_expression(LHS,FullPrio),
            %, count_number_of_disjuncts(RHS,NrC),  
            % FullPrio is Prio+(NrC-1)/10  %  , print(disj_prio(Prio,NrC,RHS,FullPrio)),nl
        get_wait_flag(FullPrio,disjoin,WF,LWF), %%
        disjoin_true(X,Y,LWF).
% missing rule: if X==Y -> X=pred_true ; if X==~Y -> no need to setup choice point
:- block disjoin_true(-,-,-).
disjoin_true(X,Y,_LWF) :-  %%  print(disjoin_true(X,Y,_LWF)),nl, %%
   (   X==pred_true  -> true
     ; X==pred_false -> pred_true=Y
     ; Y==pred_true  -> true
     ; Y==pred_false -> pred_true=X
     ; (  %% print(forcing_disjoin_true(X,Y,_LWF)),nl,  %%
        X=pred_true; 
        X=pred_false,Y=pred_true % these two unifications will happen atomically !
       )
      %add_error_fail(disjoin_true,'Illegal values: ',disjoin_true(X,Y))
   ).

:- assert_must_succeed(exhaustive_kernel_check(b_interpreter_check:impl(pred_false,pred_false,pred_true))).
:- assert_must_succeed(exhaustive_kernel_check(b_interpreter_check:impl(pred_false,pred_true,pred_true))).
:- assert_must_succeed(exhaustive_kernel_check(b_interpreter_check:impl(pred_true,pred_false,pred_false))).
:- assert_must_succeed(exhaustive_kernel_check(b_interpreter_check:impl(pred_true,pred_true,pred_true))).
:- assert_must_succeed(exhaustive_kernel_fail_check(b_interpreter_check:impl(pred_false,pred_false,pred_false))).
:- assert_must_succeed(exhaustive_kernel_fail_check(b_interpreter_check:impl(pred_false,pred_true,pred_false))).
:- assert_must_succeed(exhaustive_kernel_fail_check(b_interpreter_check:impl(pred_true,pred_false,pred_true))).
:- assert_must_succeed(exhaustive_kernel_fail_check(b_interpreter_check:impl(pred_true,pred_true,pred_false))).

impl(X,Y,Res) :- % print(impl(X,Y,Res)),nl,
   ((var(X),var(Y),var(Res)) -> bool_equality(X,Y,EqXY) ; true /* impl4 will not block anyway */),
   impl4(X,Y,EqXY,Res).
:- block impl4(-,-,-,-).
impl4(X,Y,EqXY,Res) :- 
  (  Res==pred_false -> X=pred_true,Y=pred_false
   ; Res==pred_true -> imply_true(X,Y,EqXY)
   ; X==pred_false -> Res=pred_true
   ; X==pred_true -> Res=Y
   ; Y==pred_true -> Res=pred_true
   ; Y==pred_false -> negate(X,Res)
   ; EqXY==pred_true -> Res=pred_true % X => X is always true
   ; EqXY==pred_false -> Y=Res % not(Y) => Y  is true iff Y is true
   ; add_error_fail(impl,'Illegal values: ',impl(X,Y,EqXY,Res))
  ).
imply_true(X,Y) :- %print(imply_true(X,Y)),nl,
   ((var(X),var(Y)) -> bool_equality(X,Y,EqXY) ; true /* imply_true will not block anyway */),
   imply_true(X,Y,EqXY).
:- block imply_true(-,-,-).
imply_true(X,Y,EqXY) :-  %print(imply_true(X,Y,EqXY)),nl,
  (  X==pred_false -> true
   ; X==pred_true -> Y=pred_true
   ; Y==pred_true -> true
   ; Y==pred_false -> X=pred_false
   ; EqXY==pred_true -> true
   ; EqXY==pred_false -> X=pred_false % X => not(X)  ===>  X=pred_false
   ; add_error_fail(imply_true,'Illegal values: ',imply_true(X,Y,EqXY))
  ).


:- assert_must_succeed(exhaustive_kernel_check(b_interpreter_check:equiv(pred_false,pred_false,pred_true))).
:- assert_must_succeed(exhaustive_kernel_check(b_interpreter_check:equiv(pred_false,pred_true,pred_false))).
:- assert_must_succeed(exhaustive_kernel_check(b_interpreter_check:equiv(pred_true,pred_false,pred_false))).
:- assert_must_succeed(exhaustive_kernel_check(b_interpreter_check:equiv(pred_true,pred_true,pred_true))).
:- assert_must_succeed(exhaustive_kernel_fail_check(b_interpreter_check:equiv(pred_false,pred_false,pred_false))).
:- assert_must_succeed(exhaustive_kernel_fail_check(b_interpreter_check:equiv(pred_false,pred_true,pred_true))).
:- assert_must_succeed(exhaustive_kernel_fail_check(b_interpreter_check:equiv(pred_true,pred_false,pred_true))).
:- assert_must_succeed(exhaustive_kernel_fail_check(b_interpreter_check:equiv(pred_true,pred_true,pred_false))).

% b_interpreter_check:equiv(X,Y,Res),X=Y, Res==pred_true
equiv(X,Y,Res) :- %print(equiv(X,Y,Res)),nl,
    bool_equality(X,Y,Res).

% -------------------------------------------------

/* not yet ready for production:
:- block b_propagate_exists(-,-,?,?,?,?).
b_propagate_exists(Res,BodyRes,WaitVars,Parameters,TypedVals,WF) :- % determine that a solution can be found
        print(b_propagate_exists(Res,BodyRes,WaitVars,Parameters,TypedVals,WF)),nl,
        nonvar(BodyRes),!, print(exists_body_determined(BodyRes,Res,Parameters)),nl,
        Res=BodyRes,
        when(ground(WaitVars), % sufficient to find one solution
              (b_enumerate_exists(Parameters,TypedVals,WF) -> true)).
b_propagate_exists(pred_true,BodyRes,WaitVars,Parameters,TypedVals,WF) :- 
        BodyRes=pred_true, % determine that a solution can be found
        when(ground(WaitVars),
              (b_interpreter:b_enumerate_exists(Parameters,TypedVals,WF) -> true)).
b_propagate_exists(pred_false,BodyRes,WaitVars,Parameters,TypedVals,WF) :- 
       kernel_waitflags:portray_waitflags(WF), get_enumeration_finished_wait_flag(WF,EWF),
        when(ground(EWF), % leaves pending goals !!
              ((kernel_waitflags:portray_waitflags(WF),BodyRes=pred_true,
                b_interpreter:b_enumerate_exists(Parameters,TypedVals,WF)) % check for all not exists
               -> fail
               ;  % avoid leaving pending subgoals; maybe we can do this better ??
                 kernel_waitflags:portray_waitflags(WF),
                 BodyRes=pred_false, print(not_exists_ok(TypedVals)),nl,
                 (b_interpreter:b_enumerate_exists(Parameters,TypedVals,WF) -> print(found_not_exists_sol(TypedVals)),nl)
             )).
*/