% (c) 2014-2014 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(clpfd_lists,[
                 lazy_fd_value_check/4,
                 try_in_fd_value_list_check/4,
                 avl_fd_value_check/4,
                 in_fd_value_list_wf/4, in_fd_value_list_wf/3,
                 get_fd_value_list/5,
                 try_get_fd_value_list/4,
                 get_fd_value/3]).

:- use_module(preferences).
:- use_module(error_manager).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides a way to call the CLPFD clpfd_inlist and clpfd:element predicates for lists of FD values.').

% -----------------------------------------------------

:- block complete_type(-).
% check if we have a complete value type or if we extract only part of the value
complete_type(global(_)).
complete_type(integer).
%complete_type(couple(A,B)) :- fail. % THIS IS NOT COMPLETE !! -> the element checks are done for each component in isolation ! (there is no guarantee that the pair is in the list of pairs)

% -----------------------------------------------------

:- block try_in_fd_value_list_check(-,?,?,?).
try_in_fd_value_list_check([],_,_,_) :- fail.
try_in_fd_value_list_check([H|T],El,Type,WF) :-
    try_get_fd_value_type(El,H,Type),  % print(fd_val_type(El,H,T)),nl,
    !,
    get_fd_value_list([H|T],Type,FDList,GroundList,El),
    get_fd_value(Type,El,ElFD),
    in_fd_value_list_wf(GroundList,ElFD,FDList,WF).
try_in_fd_value_list_check(S,El,Type,WF) :-
    (S=[_|_] -> true
      ; add_internal_error('Illegal call: ',try_in_fd_value_list_check(S,El,Type,WF))).

% try get fd_value type from two values; one from element one from set
try_get_fd_value_type(V1,V2,T) :- var(V1),!, try_get_fd_value(T,V2,_).
try_get_fd_value_type((A1,_),(A2,_),T) :- nonvar(T),T=couple_left(TA),!,
   try_get_fd_value_type(A1,A2,TA). % useful for domain checks
try_get_fd_value_type((A1,B1),(A2,B2),Type) :- !,
   (try_get_fd_value_type(A1,A2,TA)
    -> (try_get_fd_value_type(B1,B2,TB) 
         -> Type = couple(TA,TB)  % we have two values
          ; %print(failed_right(B1,B2,TB)),nl,
            Type = couple_left(TA))
    ;  Type = couple_right(TB),
       try_get_fd_value_type(B1,B2,TB)).
try_get_fd_value_type(freeval(ID,Case,Val1),V2, freeval_case(ID,Case,Type)) :- !,
    nonvar(Case), ground(ID),
    ((nonvar(V2),V2  = freeval(ID2,Case2,Val2), (ID,Case) == (ID2,Case2))
      -> try_get_fd_value_type(Val1,Val2,Type)
      ;  % TO DO: if the cases do not match and if Val1 is a variable then we may miss FD values
         %print(case_mismatch(ID,Case,V2)),nl,
         kernel_freetypes:get_freeval_type(ID,Case,CaseType),
         print(freeval_case_type(ID,Case,CaseType,Val1)),nl,
         % TO DO: use CaseType in case Val1 is not instantiated enough
         try_get_fd_value(Type,Val1,_)).
try_get_fd_value_type(V1,V2,T) :- 
    (var(V2) -> try_get_fd_value(T,V1,_) ; try_get_fd_value(T,V2,_)).

% TO DO: merge both versions of try_get_fd_value
try_get_fd_value(T,V,_) :- var(V),var(T),!,fail.
try_get_fd_value(TV,(A,B),FD) :- var(TV),!,
   (try_get_fd_value(TA,A,FDA)
    -> (try_get_fd_value(TB,B,FDB)  % What if this fails because B is still a variable !? --> ideally, typing info would help
         -> TV = couple(TA,TB), FD=fd_pair(FDA,FDB)  % we have two values
         ; TV = couple_left(TA), FD=FDA)
    ;  TV = couple_right(TB),
       try_get_fd_value(TB,B,FD)).
try_get_fd_value(couple(TA,TB),(A,B),fd_pair(FDA,FDB)) :- 
   try_get_fd_value(TA,A,FDA),try_get_fd_value(TB,B,FDB).
try_get_fd_value(couple_left(T),(A,_),FDA) :- try_get_fd_value(T,A,FDA). % useful for domain checks
% to do: try and get both and do multiple in_list/element calls later
try_get_fd_value(couple_right(T),(_,A),FDA) :- try_get_fd_value(T,A,FDA). % useful for range checks
try_get_fd_value(rec_fields(T),rec(Fields),fd_rec(FD)) :- 
   try_get_field_fd_values(T,Fields,FD),
   T \= [].
try_get_fd_value(global(GS),fd(X,GS),X).
try_get_fd_value(integer,int(X),X).

try_get_field_fd_values(_,Fields,_) :- var(Fields),!,fail. % we don't know all the fields; better do nothing
try_get_field_fd_values(Types,[],FDVals) :- !,Types=[],FDVals=[].
try_get_field_fd_values(_,[FN|_],_) :- var(FN),!,fail.
try_get_field_fd_values(_,[field(Name,_)|_],_) :- var(Name),!,fail.
try_get_field_fd_values(Types,[field(Name,Value)|TF],FDVals) :-
   try_get_fd_value(T,Value,FDVal),
   !,
   Types = [field(Name,T)|TTypes],
   FDVals = [FDVal|TFD],
   try_get_field_fd_values(TTypes,TF,TFD).
try_get_field_fd_values([field_no_fd(Name)|Types],[field(Name,_)|TF],FDVals) :-
   try_get_field_fd_values(Types,TF,FDVals).


get_fd_value(T,V,E) :- var(V),var(T),!,
    add_internal_error('Variable arguments: ',get_fd_value(V,T,E)),fail.
get_fd_value(couple(TA,TB),(A,B),fd_pair(FDA,FDB)) :-
    get_fd_value(TA,A,FDA), get_fd_value(TB,B,FDB).
get_fd_value(couple_left(T),(A,_),FDA) :- get_fd_value(T,A,FDA).
get_fd_value(couple_right(T),(_,A),FDA) :- get_fd_value(T,A,FDA).
get_fd_value(rec_fields(T),rec(Fields),fd_rec(FD)) :- get_field_fd_values(T,Fields,FD).
get_fd_value(freeval_case(ID,Case,T),freeval(ID,Case,A),FDA) :-
   get_fd_value(T,A,FDA).
get_fd_value(global(GS),fd(X,GS),X).
get_fd_value(integer,int(X),X).

get_field_fd_values([],[],[]).
get_field_fd_values([field_no_fd(Name)|TTypes],[field(Name,_)|TF],FDVals) :-
   get_field_fd_values(TTypes,TF,FDVals).
get_field_fd_values([field(Name,T)|TTypes],[field(Name,Value)|TF],[FDVal|TFD]) :- 
   get_fd_value(T,Value,FDVal),
   get_field_fd_values(TTypes,TF,TFD).


% -----------------------------------------------------

% extract from a Set (List) representation a list of FDValues which can be used by the clpfd library
try_get_fd_value_list(Set,Type,FDList,GroundList) :-
    try_get_fd_value_list5(Set,Type,FDList,true,GroundList).
try_get_fd_value_list5(V,_,_,_,_) :- var(V),!,fail.
try_get_fd_value_list5([],_,[],G,G).
try_get_fd_value_list5([H|T],Type,[FDH|FDT],GroundIn,GroundOut) :-
   try_get_fd_value(Type,H,Value),!, FDH=Value,
   ((GroundIn=true,number(FDH)) -> Ground1=true ; Ground1=false),
   get_fd_value_list5(T,Type,FDT,Ground1,GroundOut,_NoFilterEl).

% gets a (filtered) FD Value list (El used for filtering)
get_fd_value_list(Set,Type,FDList,GroundList,El) :-
    get_fd_value_list5(Set,Type,FDList,true,GroundList,El).
get_fd_value_list5(V,_,_,_,_,_) :- var(V),!,fail.
get_fd_value_list5([],_,[],G,G,_).
get_fd_value_list5([H|T],Type,FDValList,GroundIn,GroundOut,El) :-
  cannot_match(H,El),!, % we can filter this FD Value out
  get_fd_value_list5(T,Type,FDValList,GroundIn,GroundOut,El).
get_fd_value_list5([H|T],Type,FDValList,GroundIn,GroundOut,El) :-
   (get_fd_value(Type,H,FDH)
    ->  FDValList = [FDH|FDT],
        ((GroundIn=true,number(FDH)) -> Ground1=true ; Ground1=false)
    ;   FDValList=FDT,
        print(get_fd_value_failed(Type,H,_)),nl, % should not happen
        Ground1=GroundIn
   ),
   get_fd_value_list5(T,Type,FDT,Ground1,GroundOut,El).

% -----------------------------------------------------

:- use_module(library(clpfd)).
:- use_module(clpfd_interface).
:- use_module(kernel_waitflags,[get_enumeration_finished_wait_flag/2]).

in_fd_value_list_wf(Element,List,WF) :- (ground(List) -> Gr=true ; Gr=false),
   in_fd_value_list_wf(Gr,Element,List,WF).

% assert that an FD Value is in a list of FD Values
% in_fd_value_list(ListIsGround, FDElement, ListOfFDValues)
in_fd_value_list_wf(true,Element,List,_WF) :- clpfd_inlist(Element,List).
in_fd_value_list_wf(false,Element,List,WF) :- 
   get_enumeration_finished_wait_flag(WF,EWF),
   in_non_ground_fd_value_lists(Element,List,EWF,top).

in_non_ground_fd_value_lists(FDP,List,EWF,_) :- nonvar(FDP), FDP = fd_pair(A,B), !,
   split_fd_pair_list(List,ListA,ListB),
   in_non_ground_fd_value_lists(A,ListA,EWF,left),
   in_non_ground_fd_value_lists(B,ListB,EWF,right).
in_non_ground_fd_value_lists(FDR,List,EWF,_) :- nonvar(FDR), FDR = fd_rec(Fields), !,
   %print(treat_fd_rec_fields(Fields,List)),nl,
   treat_fd_rec_fields(Fields,List,EWF).
in_non_ground_fd_value_lists(Element,List,_EWF,Path) :-
   Path \= top, % otherwise we know List not to be ground
   %print(el(Element,List)),nl,
   ground(List), %ground_number_list(List,Min,Max,Len),
   !,
   clpfd_interface:clpfd_inlist(Element,List). % this is considerably faster; especially in compiled version (see SLOT-PERFORMANCE2)
in_non_ground_fd_value_lists(Element,List,_EWF,_Path) :- 
   tools:exact_member(Element,List), % avoid generating choice point for element position below
   !.
in_non_ground_fd_value_lists(Element,List,EWF,_Path) :-  % print(element(List,Element)),nl, 
   clpfd:element(Nr,List,Element),
   label_el_nr(Nr,Element,List,EWF). % asserts that Element is the _Nr-th element of List; in contrast to clpfd_inlist List do not have to be numbers
   


/* using this + Element in Min..Max
   does not seem to buy anything/ a lot to comparing calling clpfd_inlist for large lists
ground_number_list([],0,0,0).
ground_number_list([H|T],Min,Max,Len) :- number(H),
   ground_number_list_acc(T,H,H,1,Min,Max,Len).
ground_number_list_acc([],Min,Max,Len,Min,Max,Len).
ground_number_list_acc([H|T],Min0,Max0,Len0,Min,Max,Len) :- number(H),
   (H<Min0 -> Min1 = H ; Min1 = Min0),
   (H>Max0 -> Max1 = H ; Max1 = Max0),
   L1 is Len0+1,
   ground_number_list_acc(T,Min1,Max1,L1,Min,Max,Len).
*/

treat_fd_rec_fields([],_,_).
treat_fd_rec_fields([V|TF],List,EWF) :-
   get_next_fd_rec_field(List,FieldFDList,RemainingList),
   %print(check(V,FieldFDList)),nl,
   in_non_ground_fd_value_lists(V,FieldFDList,EWF,rec),
   treat_fd_rec_fields(TF,RemainingList,EWF).

get_next_fd_rec_field([],FieldList,RemList) :- !, FieldList=[], RemList=[].
get_next_fd_rec_field([fd_rec([FDV|Rem]) | TList], [FDV|TFieldList], [ fd_rec(Rem) | TRemList]) :- !,
   get_next_fd_rec_field(TList, TFieldList, TRemList).
get_next_fd_rec_field(List,FieldList,RemainingList) :-
   add_internal_error('Illegal call: ',get_next_fd_rec_field(List,FieldList,RemainingList)),
   FieldList=[], RemainingList=[].

:- block label_el_nr(?,-,?,?), label_el_nr(-,?,?,-).
label_el_nr(ElementNr,Element,List,_) :- %print(label(ElementNr,Element,List,_b)),nl,
  (number(ElementNr) -> true % clpfd:element already has a solution
  ; exact_member_nr(Element,List,1,Nr) -> ElementNr=Nr
  ; %translate:print_clpfd_variable(ElementNr),nl,
    when(ground(List), % we have to wait until List is ground; otherwise labeling ElementNr will instantiate List and we cannot use the cut !
     (clpfd:labeling([],[ElementNr]) -> true)) % avoid pending co-routines in case the Element appears multiple times in the list !
      % Note: ElementNr is a local variable not used outside of in_non_ground_fd_value_lists
   ).

exact_member_nr(X,[Y|_],Acc,R) :- X==Y, !, Acc=R.
exact_member_nr(X,[_|T],Acc,R) :- A1 is Acc+1, exact_member_nr(X,T,A1,R).

split_fd_pair_list([],[],[]).
split_fd_pair_list([fd_pair(A,B)|T],[A|TA],[B|TB]) :- split_fd_pair_list(T,TA,TB).


% -----------------------------------------------------

% a lazy, blocking version which waits for Set to become sufficiently instantiated and then
%  calls in_fd_value_list_wf
% Trys to transform E:Set into calls to clpfd library: ;
% When FullyChecked becomes true, then there is no need to perform other processing (apart from labeling/enumeration of the involved values)
% When FullyChecked becomes false
lazy_fd_value_check(Set,E,_,FC) :- nonvar(Set), Set=[H|T], T==[],
   !, % we have a one-element singleton set 
   %tools_printing:print_term_summary(equal_object(H,E)),nl,
   kernel_objects:equal_object(H,E),
   FC=true.
lazy_fd_value_check(Set,E,WF,FullyChecked) :- preferences:preference(use_clpfd_solver,true),!,
   lazy_in_list_propagation_block(Set,E,WF,FullyChecked).
lazy_fd_value_check(_Set,_E,_WF,false).
   
:- block lazy_in_list_propagation_block(-,?,?,?).
lazy_in_list_propagation_block([],_,_,_) :- fail. % nothing can be element of empty set
lazy_in_list_propagation_block([H|T],E,WF,FullyChecked) :- contains_fd_element(H,Res,WF),
   lazy_in_list_propagation_block_2(Res,H,T,E,WF,FullyChecked).

:- block lazy_in_list_propagation_block_2(-,?,?,?,?,?).
lazy_in_list_propagation_block_2(pred_false,_H,_T,_E,_WF,false). % nothing to propagate
lazy_in_list_propagation_block_2(pred_true,H,T,E,WF,FullyChecked) :- % print(bound_h(H)),nl,
   bound_list(T,Bound),
   lazy_in_list_propagation_block_3(Bound,H,T,E,WF,FullyChecked).

:- block lazy_in_list_propagation_block_3(-,?,?,?,?,?).
lazy_in_list_propagation_block_3(pred_false,_H,_T,_E,_WF,FC) :- !, FC=false. % nothing to propagate
lazy_in_list_propagation_block_3(pred_true,H,T,E,WF,FullyChecked) :-
   try_get_fd_value_type(E,H,Type),   % print(fd_val_type(E,H,Type)),nl,
   get_fd_value_list([H|T],Type,FDVals,GroundList,E), % try and extract values which can be used for CLPFD
   get_fd_value(Type,E,EFD),!,
   FDVals \= [], % if we have the empty list then membership fails
   (complete_type(Type) -> FullyChecked = true ; FullyChecked = false),
   % print(lazy_in_element(FullyChecked,GroundList,E,[H|T],WF)),nl,
   %  print(in_fd_value_list_wf(GroundList,EFD,FDVals,WF)),nl,
   in_fd_value_list_wf(GroundList,EFD,FDVals,WF).
lazy_in_list_propagation_block_3(pred_true,_H,_T,_E,_WF,FullyChecked) :- !,
   % should only happen for freeval with no fd type inside or for freevalues if element is not instantiated (we don't know which freval case we have)
   FullyChecked=false.
lazy_in_list_propagation_block_3(P,H,T,E,WF,false) :-
   add_internal_error('Illegal call: ',lazy_in_list_propagation_block_3(P,H,T,E,WF)).
   
:- block contains_fd_element(-,?,?).
contains_fd_element(int(_),R,_WF) :- !, R=pred_true.
contains_fd_element(fd(_,_),R,_WF) :- !, R=pred_true.
contains_fd_element((A,B),R,WF) :-  !,contains_fd_element(A,RA,WF), contains_fd_element(B,RB,WF), 
   b_interpreter_check:disjoin(RA,RB,R,priority(10000),priority(10000),WF). % TO DO: examine priority to be used
contains_fd_element(rec(Fields),R,WF) :- !, contains_field_fd_el(Fields,R,WF).
contains_fd_element(freeval(_ID,_Case,_Val),R,_WF) :- !,
    % We assume that a freetype can contain FD elements; TO DO: really check !
    % ID is something of the form 'List'(integer) ; we use get_freeval_type(ID,_Case) from kernel_freetypes to get subtypes
    % we could look at the value; but we could be unlucky in that this particular Case has no fd subtype but others do
    % anyway: try_get_fd_value_type will fail if there is no fd type
    R=pred_true.
contains_fd_element(_,pred_false,_).

:- block contains_field_fd_el(-,?,?).
contains_field_fd_el([],R,_WF) :- !, R=pred_false.
contains_field_fd_el([field(_,V)|_],R,WF) :- contains_fd_element(V,R,WF). % TO DO: look at other fields 
% deal with freetypes: freeval(ID,Case,Value); but here we need a different scheme: we need to find the case of the element and only look at elements in the list with the same constructor
% we could also deal with set types {x,y} : { {1,2}, {2,3} }  -->  x/y : 1..3
% TO DO: maybe filter out non-matching elements also here (see filter_non_matching_elements for quick_propagate)

% set output variable as soon as we can determine if we have a bound list
:- block bound_list(-,?).
bound_list(List,BoundList) :-
  (List = [] -> BoundList=pred_true
   ; List = [_|T] -> bound_list(T,BoundList)
   ; BoundList=pred_false). % we have avl_set or global_set or closure; TO DO: also allow avl_set


% -----------------------------------------------------

% generate an in_fd_value_list_wf call for checking that E is a member of an AVL set encoding
:- use_module(library(avl),[avl_domain/2]).
avl_fd_value_check(AVL,E,WF,FullyChecked) :-
   AVL = node(TopEl,true,_,_,_),!,
   contains_fd_element(TopEl,Res,WF),
   (Res==pred_true
     -> avl_domain(AVL,DomainList),
        DomainList = [H|T],
        lazy_in_list_propagation_block_3(pred_true,H,T,E,WF,FullyChecked)
      ; FullyChecked=false).
avl_fd_value_check(AVL,E,WF,FullyChecked) :-
    add_internal_error('Illegal call: ',avl_fd_value_check(AVL,E,WF,FullyChecked)),
    FullyChecked=false.

% -----------------------------------------------------

% remove from a list of FDValues those that belong to objects that definitely cannot match Element
% this improves the precision of a later in_fd_value_list_wf call

% can_match: quick check to see if the (AVL or List) element could match 

cannot_match(A,B) :- \+ can_match(A,B). % Note: probably better to wrap can_match into negation to avoid instantiations

can_match(A,_) :- var(A),!.
can_match(_,B) :- var(B),!.
can_match((A,B),(ElA,ElB)) :- !,can_match(A,ElA), can_match(B,ElB).
can_match(int(A),R) :- !, (\+ ground(R) -> true ; R=int(A)). % unification could trigger other constraint propagations and be expensive
can_match(fd(Nr,GS),R) :- !, (\+ ground(R) -> true ; R=fd(Nr,GS)). % ditto (see, e.g., test 985 - graph iso)
can_match(pred_false,R) :- !, R=pred_false.
can_match(pred_true,R) :- !, R=pred_true.
can_match(string(A),R) :- !, (\+ ground(R) -> true ; R=string(A)).
can_match(term(A),R) :- !, (\+ ground(R) -> true ; R=term(A)).
can_match(freeval(ID,Case,Value),R) :- !,
    (var(R) -> true
     ; R=freeval(IDR,CaseR,VR),
       ((var(IDR) ; var(CaseR)) -> true
        ; (ID,Case)=(IDR,CaseR),
          can_match(Value,VR))).
can_match([],R) :- !, R \= avl_set(_), R\= [_|_].
can_match(avl_set(_),R) :- !, R \= [].
can_match([_|_],R) :- !, R \= [].
can_match(_,_). % catch-all other types: to do: add records,...


  
