% (c) 2009-2012 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- multifile user:term_expansion/6.

:- volatile definitionnr/3, clause_counter_collected/6, call_counter_collected/4.
:- dynamic definitionnr/3, clause_counter_collected/6, call_counter_collected/4.

:- use_module(library(lists)).
:- use_module(module_information).

% use module_info/3 because we are not in a module here
:- module_info(coverage_term_expander,group,coverage_analysis).
:- module_info(coverage_term_expander,description,'This module generates various statistics by expanding and counting terms. Used in the coverage reports.').

%failing_clause(fail). % fail is no longer counted as a failing predicate - usually this is not an error but used on purpose
failing_clause(add_error_fail(_,_,_)).
failing_clause(add_error_and_fail(_,_)).
failing_clause(add_error_and_fail(_,_,_)).
% the following are not failing per se, but can never be covered if there is no internal error in ProB
failing_clause(add_internal_error(_,_)).
failing_clause(add_internal_error(_,_,_,_)).
failing_clause(add_failed_call_error(_)).
failing_clause(add_internal_error_with_span(_,_,_,_)).

reviewed_clause('$NOT_COVERED'(_)).

uncoverable_clause(X) :- failing_clause(X), ! ; reviewed_clause(X).

layout_sub_term([],_,[]).
layout_sub_term([H|T],N,Res) :-
    (N=<1 -> Res=H ; N1 is N-1, layout_sub_term(T,N1,Res)).

% do not use the internal_flatten/2 defined in tools to avoid loading a module before the term expansion is complete
% same goes for internal_assert_once/1
internal_flatten(List,FlatList) :- internal_flatten1(List,[],FlatList).
internal_flatten1([],Acc,FlatList) :- !, Acc = FlatList.
internal_flatten1([H|T],Tail,List) :- !, internal_flatten1(H,FlatList,List), internal_flatten1(T,Tail,FlatList).
internal_flatten1(NonList,Tail,[NonList|Tail]).
internal_assert_once(Predicate) :-
    ( call(Predicate) -> true
    ; otherwise -> assert(Predicate)).

coverage_counter_body(X,_Layout,coverable,_File,_Name,_Counter) :- var(X),!.
% look into not and ~~mnf
coverage_counter_body(\+(X),Layout,Result,File,Name,Counter) :-
    !,
    coverage_counter_body(X,Layout,Result,File,Name,Counter).
coverage_counter_body((_A -> B),Layout,Result,File,Name,Counter) :-
    !,
    layout_sub_term(Layout,3,Layout2),
    coverage_counter_body(B,Layout2,Result,File,Name,Counter).
coverage_counter_body((_A -> B ; C),Layout,Result,File,Name,Counter) :-
    !,
    layout_sub_term(Layout,2,LayoutAB),
    layout_sub_term(LayoutAB,3,LayoutB),
    layout_sub_term(Layout,3,LayoutC),
    coverage_counter_body(B,LayoutB,ResultB,File,Name,Counter),
    coverage_counter_body(C,LayoutC,ResultC,File,Name,Counter),
    merge_results(ResultB,ResultC,Result).
coverage_counter_body(if(_A,B,C),Layout,Result,File,Name,Counter) :-
    !,
    layout_sub_term(Layout,3,LayoutB),
    layout_sub_term(Layout,4,LayoutC),
    coverage_counter_body(B,LayoutB,ResultB,File,Name,Counter),
    coverage_counter_body(C,LayoutC,ResultC,File,Name,Counter),
    merge_results(ResultB,ResultC,Result).
coverage_counter_body((A,B),Layout,Result,File,Name,Counter) :-
    !,
    layout_sub_term(Layout,2,LayoutA),
    layout_sub_term(Layout,3,LayoutB),
    coverage_counter_body(A,LayoutA,ResultA,File,Name,Counter),
    coverage_counter_body(B,LayoutB,ResultB,File,Name,Counter),
    ((ResultA = uncoverable ; ResultB = uncoverable)
     -> Result = uncoverable,
        assert_call_counter(File,Name,Counter,LayoutA),
        assert_call_counter(File,Name,Counter,LayoutB)
     ;  Result = coverable).
coverage_counter_body((A;B),Layout,Result,File,Name,Counter) :-
    !,
    layout_sub_term(Layout,2,LayoutA),
    layout_sub_term(Layout,3,LayoutB),
    coverage_counter_body(A,LayoutA,ResultA,File,Name,Counter),
    coverage_counter_body(B,LayoutB,ResultB,File,Name,Counter),
    merge_results(ResultA,ResultB,Result).
coverage_counter_body(X,Layout,uncoverable,File,Name,Counter) :-
    uncoverable_clause(X),
    !, assert_call_counter(File,Name,Counter,Layout).
coverage_counter_body(_X,_Layout,coverable,_File,_Name,_Counter).

merge_results(uncoverable,uncoverable,uncoverable).
merge_results(coverable,uncoverable,coverable).
merge_results(uncoverable,coverable,coverable).
merge_results(coverable,coverable,coverable).

assert_call_counter(File,Name,Counter,Layout) :-
    internal_flatten(Layout,Flat),
    maplist(assert_call_counter2(File,Name,Counter),Flat).
assert_call_counter2(File,Name,Counter,SingleLine) :-
    internal_assert_once(user:call_counter_collected(File,Name,Counter,SingleLine)).

coverage_counter((:- _),_Layout) :- !.
coverage_counter(end_of_file,_Layout) :- !.

coverage_counter((Head :- Body), [LayoutHead | LayoutSub]) :-
    !,
    prolog_load_context(file,File),
    prolog_load_context(module,Module),
    functor(Head,Name,Arity),
    (retract(definitionnr(Module,Name/Arity,Counter)) -> true ; Counter = 0),
    CNeu is Counter + 1,
    assert(definitionnr(Module,Name/Arity,CNeu)),
    layout_sub_term([LayoutHead|LayoutSub],3,SubLay),
    coverage_counter_body(Body,SubLay,Result,File,Module:Name/Arity,Counter),
    (
      (Result = uncoverable ; uncoverable_clause(Head))
      -> internal_flatten([LayoutHead|LayoutSub],FlatLayout),
         FlatLayout = [FirstLine|_], last(FlatLayout,LastLine),
         internal_assert_once(user:clause_counter_collected(File,Module:Name/Arity,Counter,LayoutHead,FirstLine,LastLine))
      ;  true
    ).
coverage_counter(Fact, _Layout) :-
    !,
    prolog_load_context(module,Module),
    functor(Fact,Name,Arity),
    (retract(definitionnr(Module,Name/Arity,Counter)) -> true ; Counter = 0),
    CNeu is Counter + 1,
    assert(definitionnr(Module,Name/Arity,CNeu)).

% Term Expansion to count intentional fails
user:term_expansion(Term1, Lay1, Tokens1, Term1, Lay1, [coverage_tracing | Tokens1]) :-
    nonmember(coverage_tracing, Tokens1), % do not expand if already expanded
    coverage_counter(Term1, Lay1),
    !.
