% (c) 2014-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(disprover_test_runner, [test_repl/0, run_disprover_on_all_pos/1, load_po_file/1, print_disprover_stats/0]).

:- use_module(pathes,[set_search_pathes/0]). % called first to set_compile_time_search_pathes
%:- set_search_pathes.

portray_message(informational, _).

:- use_module(module_information).
:- module_info(group,cbc).
:- module_info(description,'Rodin Prover / Disprover').


:- use_module(eclipse_interface).
:- use_module(prob2_interface).
:- use_module(disprover).
:- use_module(error_manager).
:- use_module(tools).
:- use_module(solver_interface).
:- use_module(debug).

:- use_module(library(system)).
:- use_module(library(lists)).



test_repl :- set_search_pathes,
   init_eclipse_preferences,
   disprover_load_at_repl_startup,
   disprover_help,
   disprover_repl_loop.


disprover_repl_loop :- print('DISPROVER ==> '),read(T), 
  (quit_command(T) -> true; 
   test_eval(T) -> disprover_repl_loop
   ; print('Command failed !!'),nl, disprover_repl_loop).


disprover_help :- print('Available commands: run, run(PO), timeout(T), stats, list, q, x, FILE, v, vv, write(S,FILE)'),nl.

:- meta_predicate wall(0).
wall(Call) :-
    statistics(walltime,[Start,_]),
    call(Call),
    statistics(walltime,[Stop,_]), WT is Stop-Start,
    format('Walltime: ~w ms~n',[WT]).

quit_command(q).
quit_command(quit).
quit_command(halt).

test_eval(run) :- !,wall(run).
test_eval(run(PO)) :- !,disprover_timeout(TO), wall(run_on_po(PO,TO)).
test_eval(timeout(T)) :- !, set_timeout(T).
test_eval(time_out(T)) :- !, set_timeout(T).
test_eval(stats) :- !,print_disprover_stats.
test_eval(x) :- !,halt.
test_eval(list) :- !,list.
test_eval(help) :- !,disprover_help.
test_eval(h) :- !,disprover_help.
test_eval(v) :- !,tcltk_turn_debugging_on(19).
test_eval(vv) :- !,tcltk_turn_debugging_on(4).
test_eval(load(FILE)) :- !,load_po_file(FILE).
test_eval(wp) :- !, write_pos(true).
test_eval(write(Status)) :- !, write_pos(Status).
test_eval(wp(FILE)) :- !, tell(FILE), write_pos(true), told.
test_eval(write(Status,FILE)) :- !, tell(FILE), write_pos(Status), told.
test_eval(cse) :- !,enable_cse.
test_eval(kodkod) :- !,enable_kodkod.
test_eval(Filename) :- format('Trying to consult file: ~w~n',[Filename]),
    load_po_file(Filename), print_disprover_stats.


print_disprover_stats :- project_name(N), format('Project ~w~n',[N]),
         generated(_,D),
         machine_name(M), format('Machine ~w (~w)~n',[M,D]),
         print_nr_pos('',_,_),
         print_nr_pos('proven (within Rodin)',true,_),
         print_nr_pos('proven (by ProB)',_,true).
         
print_nr_pos(Msg,RodinStatus,ProBStatus) :-
         findall(Name,disprover_po_with_info(Name,RodinStatus,ProBStatus),L),
         length(L,Len),
         ((L = [First|_],last(L,Last))
           -> format('Number of ~w POs: ~w (~w -- ~w)~n',[Msg,Len,First,Last])
           ;  format('Number of ~w POs: ~w~n',[Msg,Len])
          ).

disprover_po_with_info(Name,RodinStatus,ProBStatus) :-
   disprover_po(Name,_Context,_Goal,_All,_Sel,RodinStatus),
   (disprover_result(Name,ProBStatus) -> true ; ProBStatus = 'not_run').

:- dynamic disprover_timeout/1.
disprover_timeout(300).
set_timeout(N) :- retractall(disprover_timeout(_)), assert(disprover_timeout(N)).

run :- run_disprover_on_all_pos(_).

run_disprover_on_all_pos(_) :- retractall(disprover_result(_,_)), disprover_timeout(TO),run_on_po(_,TO), fail.
run_disprover_on_all_pos(Summary) :- findall(R,disprover_result(_,R),L),
       count_occurences(L,Occs),
       format('~nDisprover Run Summary: ~w~n',[Occs]),
       add_info(true,Occs,O1), add_info(false,O1,O2), add_info(unknown,O2,O3),
       add_info(failure,O3,Summary).

add_info(Info,List,Result) :- member(Info-_,List) -> List=Result ; Result = [Info-0|List].


run_on_po(PO_Name,TimeOut) :-
    disprover_po(PO_Name,Context,Goal,AllHypotheses,SelectedHypotheses,Status),
    format('Testing ~w (~w)~n',[PO_Name,Status]),
    load_event_b_project([],[Context],[],_),
    start_animation,
    debug_println(5,loaded_context),
    (my_disprove(Goal,AllHypotheses,SelectedHypotheses,TimeOut,OutResult)
     -> translate_result_for_po(OutResult,ProverResult),
        format('Disprover Result ~w (PO is ~w)~n',[OutResult,ProverResult]),
        assert(disprover_result(PO_Name,ProverResult))
     ;  add_error(disprover,'Disprover failed: ',(PO_Name:Status)),
        assert(disprover_result(PO_Name,'failure'))
     ).

translate_result_for_po(solution(_),R) :- !, R=false.
translate_result_for_po(contradiction_found,R) :- !,R=true.
translate_result_for_po(_,unknown).

:- dynamic extra_opts/1.
extra_opts([]).
update_opts(New) :-
  retract(extra_opts(_)),
  assert(extra_opts(New)).
enable_cse :- XO = [use_common_subexpression_elimination/true,
                     use_common_subexpression_also_for_predicates/true],
        update_opts(XO).
enable_kodkod :- XO = [try_kodkod_on_load/true, kodkod_for_components/true], update_opts(XO).

my_disprove(Goal,AllHypotheses,SelectedHypotheses,Timeout,OutResult) :- 
    Opts = [use_smt_mode/true,use_clpfd_solver/true,use_chr_solver/true],
    extra_opts(X), append(X,Opts,AllOpts),
    % Note: disprover_mode is always set to true
    disprove_with_opts(Goal,AllHypotheses,SelectedHypotheses,Timeout,AllOpts,OutResult).

:- use_module(translate).
:- use_module(bmachine,[type_with_errors/4]).

list :- disprover_po(Name,Context,Goal,_All,_Sel,Status), get_disprover_result(Name,DPRes),
   format('~w (~w : ~w) : ',[Name,Status,DPRes]),
   load_event_b_project([],[Context],[],_),
   start_animation,
   type_check_in_machine_context([Goal],[TypedGoal]),
   translate:print_bexpr(TypedGoal),nl,fail.
list.

:- dynamic project_name/1.
:- dynamic machine_name/1.
:- dynamic generated/2.
:- dynamic disprover_po/6.
:- dynamic disprover_result/2.
project_name(unknown).
machine_name(unknown).
generated(unknown,unknown).
disprover_po(unknown,unknown,unknown,unknown,unknown,unknown).
disprover_result(unknown,unknown).

reset_po_facts :- retractall(project_name(_)),
   retractall(machine_name(_)), retractall(generated(_,_)),
   retractall(disprover_po(_,_,_,_,_,_)), retractall(disprover_result(_,_)).

% write POs which have a certain status; can be used to generate .pl files with only proven or not-proven POs
write_pos(Status) :- 
  format('% Proof Obligations with status = ~w~n',[Status]),
  write_facts(project_name(_)), write_facts(machine_name(_)),
  write_facts(generated(_,_)),
  disprover_result(Name,Status),
  write_facts(disprover_po(Name,_Context,_Goal,_All,_Sel,_RodinStatus)),
  fail.
write_pos(_) :- nl.

write_facts(Fact) :- Fact, 
   write_term(Fact,[quoted(true),numbervars(true)]), write('.'),nl,fail.
write_facts(_).

get_disprover_result(PO_Name,Result) :-
   (disprover_result(PO_Name,R) -> Result=R ; Result = 'not_yet_run').


load_po_file(File) :-  reset_po_facts,
    % consult_without_redefine_warning
    prolog_flag(redefine_warnings, Old, off),
    (% load in dynamic mode: so that we can retract later
     my_po_consult(File)
      -> OK=true ; OK=false),
    prolog_flag(redefine_warnings, _, Old),
    (OK=true -> true ;
      add_error_and_fail(bmachine,'Consulting of File failed: ',File)).

my_po_consult(File) :- on_exception(error(existence_error(_,_),_),
                    load_files(File,[compilation_mode(assert_all)]), 
                    add_error_fail(my_consult,'File does not exist: ',File)).


disprover_load_at_repl_startup :- prolog_flag(argv,ArgV),ArgV = [FILE],
  load_po_file(FILE).
disprover_load_at_repl_startup  :-
  load_po_file('../prob_examples/examples/ProofObligations/ABZ14_Mery/M3_mch.pl').
disprover_load_at_repl_startup.
  
% :- load_po_file('../../prob_examples/examples/ProofObligations/ABZ14_Mery/M2_mch.pl').
% :- load_po_file('../../prob_examples/examples/ProofObligations/ABZ14_Mery/M3_mch.pl').
% :- load_po_file('../../prob_examples/examples/ProofObligations/ABZ14_Mery/M4_mch.pl').
% :- load_po_file('../../prob_examples/public_examples/EventBPrologPackages/Disprover/ProofTests/DefSetCst_ctx.pl').
% '/Users/leuschel/git_root/prob_examples/public_examples/SMT/QF_IDL/queens_bench/toroidal_bench/toroidal_queen_2_QF_IDL_ctx.pl'.

%:- test_repl.

