% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(eclipse_interface,
  [
   clear_loaded_machines/0,
   clear_loaded_machines_wo_errors/0,

   get_rand/4, % DEPRECATED, remove when switching to the new core
   set_rand/4, % DEPRECATED, remove when switching to the new core

   debug_console/2,

   typecheck_b_project/3,

   load_b_file/1, load_prob_file/1,
   load_eventb_file/1,
   load_tla_file/1,
   load_xtl_file/1,

   load_classical_b/1, % deprecated
   load_classical_b/2, % deprecated

   find_shortest_trace_to_current_state/1,
   find_shortest_trace_to_current_state2/2,

   do_ltl_modelcheck/6,

   get_invariants/1,
   get_name/1,

   evaluate/3,

   es_reset/0,  %evalstore: reset, remove all dynamic stores

   load_cspm_spec_from_pl_file/1,

   init_eclipse_preferences/0,

   set_eclipse_card/2,

   getAllOperations/1, % deprecated, remove in 2.0
   getSetElements/1,
   getInternalRep/1,
   get_machine_objects/5,

   get_all_transitions/1,

   getErrorMessages/1,

   ops_with_timeout/1, ops_with_virtual_timeout/1, ops_with_real_timeout/1,

   getCurrentStateID/1,
   getVarCstValue/3, applyVarCstValue/4,
   isRootState/1, isConstantsSetUpState/1,
   getOperationParameterNames/2, % deprecated, remove in 2.0

   get_operation_infos/1,


   % getCurrentStateValues/2, no longer exists ??!
   getStateValues/2,

   test_boolean_expression_in_node/2,
   computeOperations/1,
   executeOperationNr/1, executeOperationNr_on_computed_node/1,

   setCurrentState/1,
   computeOperationsForState/2, % Used in ProB 1.0

   evaluate_expression/2,

   print_states_for_dot/1,

   evaluate_expression_as_tree/4,
   evaluate_properties_as_tree/1,
   evaluate_invariant_as_tree/2,
   evaluate_precondition_as_tree/3,
   evaluate_raw_expression_as_tree/3,

   %in contrast to evaluate_expression, it supports predicates and does typing
   evaluate_raw_expression/3,
   evaluate_raw_expressions/3,

   sap_generate_local_testcases/5,
   sap_find_test_path/4,
   execute_custom_operations/6,

   evaluation_get_top_level/1,
   evaluation_expand_formula/3,
   evaluation_get_values/3,
   evaluation_insert_formula/3,

   deadlock_freedom_check/1,
   deadlock_freedom_check/2,
   invariant_check/2,

   save_as_classical_b/4,

   analyse_predicate/6,
   get_classicalb_machine_objects/2,

   dot_hierarchy/1,

   refinement_check/2,

   load_sets_for_cbc/1,

   cbc_static_assertion_violation_checking/1,
   get_vacuous_invariants/1, get_vacuous_guards/1,

   quick_describe_unsat_properties/1,
   find_state_satisfying_predicate/2,

   prob_construct_trace/6
  ]).

:- use_module(module_information).
:- module_info(group,cli).
:- module_info(description,'This module provides the Prolog interface to Java and other languages (usually called via socket server).').

:- use_module(library(system),[environ/2]).
:- use_module(library(lists)).

:- use_module(error_manager).
:- use_module(self_check).
:- use_module(bsyntaxtree).
:- use_module(version).
:- use_module(eventhandling,[announce_event/1]).

:- use_module(model_checker).

:- use_module(b_global_sets,[enum_global_type/2, add_prob_deferred_set_elements_to_store/3]).
:- use_module(translate).
:- use_module(store).

:- use_module(bvisual2).

:- use_module(library(codesio)).

:- use_module(state_space,[visited_expression/2, visited_expression_id/1]).
:- ensure_loaded(tcltk_interface).

:- use_module(evalstores,[evalstore_reset/0,evalstore_create_store_by_state/2,
                          evalstore_eval_formula/4,evalstore_bindings/2]).

:- use_module(b_state_model_check,[cbc_deadlock_freedom_check/3]).

:- use_module(extension('user_signal/user_signal'), [user_interruptable_call_det/2]).

:- use_module(extension('counter/counter'), [get_counter/2]).
:- use_module(prob2_interface, [get_state/2, is_initialised_state/1]).

% analyse_predicate(+Type,-Desc,-Result,-Total,-False,-Unknown)

analyse_predicate(Type,Desc,Result,Total,False,Unknown) :-
 user:tcltk_analyse_option(Type,Desc),
 user:tcltk_analyse(ascii,Type,Result,Total,False,Unknown).


/* a list of the calls provided by this interface; useful for the whitelist of the socket server */


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       WARNING: This is dangerous in a release. It allows to execute arbitrary code
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


check_key(Key) :-
    environ('PROB_DEBUGGING_KEY',Key).

:- dynamic current_prob_module/1.
current_prob_module(user).

debug_console(Key,Commands) :-
    check_key(Key), debug_console2(Commands).
debug_console2([]).
debug_console2([H|T]) :-
    nonvar(H),
    debug_command(H), debug_console2(T).

debug_command(set_module(X)) :-
    !, current_module(X), retractall(current_prob_module(_)), assert(current_prob_module(X)).
debug_command(Command) :-
    current_prob_module(M),
    write(M:Command),nl,
    call(M:Command).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


/* ------------------------- */
/* Machine Loading Interface */
/* ------------------------- */

:- use_module(pref_definitions,[b_get_preferences_from_machine/0]).
:- use_module(specfile,[b_or_z_mode/0,set_animation_mode/1,set_animation_minor_mode/1,animation_mode/1,
                        set_currently_opened_file/1]).
:- use_module(bmachine_construction,[check_machine/5, get_constructed_machine_name/2]).
:- use_module(probltlsrc(ltl),[ltl_model_check_with_ce/4,ltl_formula_structure/3]).
:- use_module(probltlsrc(ltl_tools),[typecheck_temporal_formula/3]).
%:- use_module(tcltk_interface, [find_shortest_trace_to_node/5]).

:- use_module(state_space).
:- use_module(bmachine).


% DEPRECATED, only here because the old Rodin plugin calls the predicates
get_rand(0,0,0,0).
set_rand(_,_,_,_).

:- use_module(preferences,[reset_all_temporary_preferences/0]).
clear_loaded_machines :-
   clear_loaded_machines_wo_errors,
   reset_errors, 
   % now done via event-handling: reset_all_temporary_preferences,
   garbage_collect_atoms,!. % reclaim atoms if possible
clear_loaded_machines :- add_error(clear_loaded_machines,'clear_loaded_machines Failed').

% new profiler
%:- use_module('../extensions/profiler/profiler.pl').

% old profiler
:- use_module(runtime_profiler,[reset_runtime_profiler/0]).

:- use_module(eval_strings,[reset_eval_strings/0]).
:- use_module(eventhandling, [announce_event/1]).
clear_loaded_machines_wo_errors :-
   announce_event(reset_specification),
   announce_event(clear_specification),
   % Now done via event_handling: reduce_graph_reset reset_runtime_profiler,  reset_refinement_checker state_space_initialise,    b_machine_hierarchy:reset_hierarchy,    b_interpreter:reset_b_interpreter,    reset_model_checker    haskell_csp:clear_cspm_spec, b_machine_reset,
%   reset_dynamics, % new profiler
   reset_eval_strings.

% Errors is a list of error(ErrorDescription,Position)
% Position is the number of the Syntaxtree-Node
typecheck_b_project(Main,Machines,Errors) :-
    check_machine(Main,Machines,[],_,Errors).

load_classical_b(Machines) :- % DEPRECATED
     set_animation_mode(b),
     load_classical_b(Machines,Errors),
     add_errors(Errors).
load_classical_b(Machines,Errors) :-
    [First|_] = Machines,
    get_constructed_machine_name(First,Main),
    b_set_machine(Main,Machines,Errors).
% Don't forget to call update_preferences_from_spec after this and before start_animation


add_errors([]).
add_errors([H|T]) :- error_manager:add_error('eclipse_interface','Error loading machine',H),add_errors(T).

:- use_module(debug,[debug_println/2]).
:- use_module(parsercall,[call_tla2b_parser/1, tla2b_filename/2,  tla2prob_filename/2]).
load_tla_file(Filename) :-
   call_tla2b_parser(Filename),
   tla2prob_filename(Filename,GeneratedProbFile),
   debug_println(10,loading_generated_prob_file(GeneratedProbFile)),
   load_prob_file(GeneratedProbFile),
   %load_b_file(GeneratedBFile),
   debug_println(10,set_animation_minor_mode(tla)),
   set_animation_minor_mode(tla).
load_b_file(Filename) :-
   clear_loaded_machines_wo_errors,
   set_animation_mode(b),
   bmachine:b_load_machine_from_file(Filename),
   set_currently_opened_file(Filename).
load_prob_file(Filename) :-
   clear_loaded_machines_wo_errors,
   set_animation_mode(b),
   bmachine:b_load_machine_probfile(Filename),
   set_currently_opened_file(Filename).

:- use_module(prob2_interface, [load_event_b_project/4]).

save_as_classical_b(File,Machines,Contexts,Errors) :-
    /* first set the number of animated levels to 0*/
    get_preference(number_animated_abstractions,OldNum),
    call_cleanup( ( set_preference(number_animated_abstractions,0),
                    load_event_b_project(Machines,Contexts,[],Errors),!),
                  set_preference(number_animated_abstractions,OldNum)),
    ( Errors==[] ->
        b_show_eventb_as_classicalb(Rep,true),
        tools:write_to_file(File,Rep)
    ; otherwise ->
        /* Do nothing, but return the list of errors*/
        true).


load_eventb_file(Filename) :-
   clear_loaded_machines_wo_errors,
   set_animation_mode(b), set_animation_minor_mode(eventb),
   b_load_eventb_project(Filename),
   set_currently_opened_file(Filename).

:- use_module(xtl_interface,[open_xtl_file/1]).
load_xtl_file(Filename) :-
   clear_loaded_machines_wo_errors,
   set_animation_mode(xtl),
   xtl_interface:open_xtl_file(Filename),
   set_currently_opened_file(Filename).

get_all_transitions(Transitions) :-
    findall(trans(TransId,Src,Event,PP,Dst,Infos),
            ( transition(Src,Event,TransId,Dst),
                translate_event(Event,PP),
                get_event_infos(TransId,Infos)),
            Transitions).

get_event_infos(TransId,Infos) :-
    findall(Info, transition_info(TransId,Info), Infos1),
    filter_transition_infos(Infos1,Infos).

filter_transition_infos([],[]).
filter_transition_infos([I|Irest],[O|Orest]) :-
    filter_transition_info(I,O),
    filter_transition_infos(Irest,Orest).

filter_transition_info(event(StackI),event(StackO)) :- !,
    filter_event_stack(StackI,StackO).
filter_transition_info(I,I).

filter_event_stack([],[]).
filter_event_stack([event(Name,Params)|Irest],[event(Name,Strings)|Orest]) :-
    translate_bvalue_l(Params,Strings),
    filter_event_stack(Irest,Orest).

translate_bvalue_l([],[]).
translate_bvalue_l([V|Vrest],[S|Srest]) :-
    translate_bvalue(V,S),
    translate_bvalue_l(Vrest,Srest).


print_states_for_dot(X) :- visualize_graph:tcltk_print_states_for_dot(X).



do_ltl_modelcheck(Formula,Max,Mode,Atomics,Structure,Result) :-
   typecheck_temporal_formula(Formula,TypeCheckedFormula,Status),
   ( Status=ok ->
       ltl_formula_structure(TypeCheckedFormula,Atomics,Structure),
       ltl_model_check_with_ce(TypeCheckedFormula,Max,Mode,Result1),
       ltl_adapt_operations(Result1,Result)
   ; otherwise ->
       Atomics=[], Structure=[],
       Result=typeerror([])).

ltl_adapt_operations(counterexample(CE1,LoopEntry,PathToCE1), counterexample(CE,LoopEntry,PathToCE)) :- !,
    filter_ops_and_ids(PathToCE1,root,PathToCE),
    ltl_adapt_ce(CE1,CE).
ltl_adapt_operations(Result,Result).

ltl_adapt_ce([],[]).
ltl_adapt_ce([atom(State,Evals,OpTuple)|Irest],[atom(State,Evals,Transition)|Orest]) :-
    ltl_adapt_ce2(OpTuple,State,Transition),
    ltl_adapt_ce(Irest,Orest).
ltl_adapt_ce2(none,_State,none).
ltl_adapt_ce2((TransId,Action,DstId),State,Transition) :-
    create_operation_term(TransId,Action,State,DstId,Transition).


find_shortest_trace_to_current_state(TraceStrings) :-
 find_shortest_trace_to_current_state2(TraceStrings, _).

find_shortest_trace_to_current_state2(TraceStrings, TraceIDs) :-
 getCurrentStateID(ID),
 user:find_shortest_trace_to_node(root,ID,Trace,TraceIDs),
 translate_events(Trace,TraceStrings).


% TODO: Should we put these predicates in another module? They are not called by any Java Interface
:- use_module(state_space, [time_out_for_node/3]).
ops_with_timeout(OpNameList) :-
    getCurrentStateID(StateId),
    ops_with_timeout(StateId,OpNameList,_).
ops_with_virtual_timeout(OpNameList) :-
    getCurrentStateID(StateId),
    ops_with_timeout(StateId,OpNameList,virtual_time_out(_)).
ops_with_real_timeout(OpNameList) :-
    getCurrentStateID(StateId),
    ops_with_timeout(StateId,OpNameList,time_out).
ops_with_timeout(StateId,OpNameList,Type) :-
    findall(OpName, time_out_for_node(StateId,OpName,Type), OpNameList).

/* ------------------------- */
/* CSP Loading Interface */
/* ------------------------- */

:- use_module(probcspsrc(haskell_csp),[parse_and_load_cspm_file/1,load_cspm_pl_file/1]).
load_cspm_spec_from_pl_file(CSPMFile) :-
   clear_loaded_machines_wo_errors,
   specfile:set_animation_mode(cspm),
   load_cspm_pl_file(CSPMFile),
   set_currently_opened_file(CSPMFile).

/* --------------------- */
/* Preferences Interface */
/* --------------------- */

:- use_module(preferences,
  [init_preferences/0,set_preference/2,get_preference/2,
   eclipse_preference/2,
   preference_val_type/2,preference_description/2,preference_default_value/2,
   preference_category/2]).


% will only return normal preferences, not advanced ones


:- use_module(model_checker,[reset_depth_breadth_first_mode/0]).
init_eclipse_preferences :- preferences:init_preferences, model_checker:reset_depth_breadth_first_mode.

% set Cardinality of SET; a bit like scope_SET == Scope DEFINITION
:- use_module(b_global_sets,[set_user_defined_scope/2]).
:- use_module(prob2_interface, [convert_cli_arg/2]).
set_eclipse_card(Set,Scope) :-
    convert_cli_arg(Scope,Value),
    set_user_defined_scope(Set,Value).


/* ------------------ */
/* getErrorMessages/1 */
/* ------------------ */
/* Returns a list of strings containing all error messages (of the error_manager).
   Clears all the errors. Returns the empty list if no errors occurred since
   the last call to getErrorMessages.

  TODO: This is equivalent with the get_error_messages predicate in prob2_interface
  We changed the name in the ProB 2.0 interface for coding convention.
   */

getErrorMessages(ListOfErrMsgs) :-
  (get_all_errors_and_reset(ListOfErrMsgs) ->  true  ; ListOfErrMsgs=[]).



/* ---------------------------- */
/* Syntax Information Interface */
/* ---------------------------- */

getAllOperations(L) :-
   (b_or_z_mode
     -> findall(OperationName, b_top_level_operation(OperationName), L)
     ;  L=[]).  % FIXME: provide list for CSP-M and Promela


get_operation_infos(L) :-
    findall(op(Name,Results,Parameters),b_get_machine_operation(Name,Results,Parameters,_,_),L).




getSetElements(L) :-  /* get a list of all fd(N,S) elements of all sets and precompute their translation */
   findall(fds(Nr,Set,String),
           (enum_global_type(fd(Nr,Set),_), translate_bvalue(fd(Nr,Set),String)), L).


:- use_module(specfile,[get_internal_representation/1]).
getInternalRep(X) :- get_internal_representation(X).

get_classicalb_machine_objects(Constants, Variables) :-
    b_get_machine_constants(C),translate_list(C,Constants),
    b_get_machine_variables(V),translate_list(V,Variables).


translate_list([],[]).
translate_list([H|T],[HT|TT]) :- translate_bexpression(H,HT), translate_list(T,TT).



get_machine_objects(Sections,SetElements,Constants,Variables,Operations) :-
    b_get_animated_sections(Sections),
    findall(TId,
            ( enum_global_type(fd(Nr,Set),_),
                translate_bvalue(fd(Nr,Set),String),
                create_texpr(identifier(String),global(Set),[],TId)),
            SetElements),
    b_get_machine_constants(Constants),
    b_get_machine_variables(Variables),
    b_get_promoted_machine_operations(Operations1),
    maplist(remove_op,Operations1,Operations).
remove_op(TId,NTId) :-
    remove_bt(TId,identifier(Id),identifier(NId),NTId),
    remove_op2(Id,NId).
remove_op2(op(Op),Op) :- !.
remove_op2(Op,Op).

/* ------------------- */
/* Animation Interface */
/* ------------------- */


get_invariants(List) :- %print(start),nl,
   b_get_invariant_from_machine(B),
%   print(got(B)),nl,
   conjunction_to_list(B,L), %print(L),nl,
   maplist(mk_inv,L,List).

mk_inv(C,inv(Text,Pos)) :-  translate_bexpression(C,Text),get_texpr_pos(C,P), (P=pos(Pos,_,_,_,_,_) -> true;Pos = P),!.
mk_inv(C,_) :- add_internal_error(mk_inv,'Failed: ',C, unknown),fail.

get_name(Name) :- b_machine_name(Name).

   /* get the ID of the current state */
getCurrentStateID(CurID) :-
    current_state_id(CurID).
   /* get the ID of the current state */



isRootState(ID) :- ID=root.

:- use_module(specfile,[state_corresponds_to_set_up_constants/2, state_corresponds_to_initialised_b_machine/1,
                        state_corresponds_to_initialised_b_machine/2]).

isConstantsSetUpState(ID) :-
    % true if no constants exists or if all constants have been valued
    visited_expression(ID,State),
    specfile:state_corresponds_to_set_up_constants(State,_).




   /* compute enabled operations, if required, and return a list of
      strings of all possible operations and arguments */
computeOperations(Operations) :-
   user:tcltk_get_options(list(Operations)).


executeOperationNr(Nr) :-
   user:compute_all_transitions_if_necessary, /* make sure options are computed; TO DO: improve */
   user:tcltk_perform_nr(Nr).
executeOperationNr_on_computed_node(Nr) :- % a version which does not recompute transitions if necessary
   user:tcltk_perform_nr(Nr).


setCurrentState(ID) :- /* jumps to the given node; can be backtracked */
   visited_expression_id(ID),
   user:tcltk_goto_state(jump,ID).


test_boolean_expression_in_node(ResID,BoolExpr) :-
   visited_expression(ResID,CurState),
   specfile:state_corresponds_to_initialised_b_machine(CurState,CurBState),
   ~~pp_cll(b_interpreter:b_test_boolean_expression_wf(BoolExpr,[],CurBState)).

/* USED IN PROB 1.0*/
computeOperationsForState(StateID, OpsAndIDs) :- /* compute the enabled operations
   (without the backtrack options) for a state ID which is not necessarily the current one */
   visited_expression_id(StateID),
   user:tcltk_compute_options(StateID,OpsSTAndIDs),
   print('OpsSTAndIDs: '),print(OpsSTAndIDs),nl,
   (filter_ops_and_ids(OpsSTAndIDs,StateID,OpsAndIDs) -> true
     ; print(filter_ops_and_ids_failed(OpsSTAndIDs)),nl,fail).

/*
op(140,io,23,23,[[in(r14_1m)],'SetRoute',multi_span(460,35,460,71,16934,36,src_span(460,31,460,32,16930,1))],['{<< in(r14_1m) >>}','<< SetRoute >>','<< multi_span(460,35,460,71,16934,36,src_span(460,31,460,32,16930,1)) >>'],[],')
*/

:- use_module(bmachine,[b_get_machine_operation_parameter_names/2]).


getOperationParameterNames(Operation,Paras) :-
   b_get_machine_operation_parameter_names(Operation,Paras).

filter_ops_and_ids([],_,[]).
filter_ops_and_ids([OpTuple|T],StateID,[FOp|FT]) :-
   ((OpTuple = (Id,OpTerm,Dst), create_operation_term(Id,OpTerm,StateID,Dst,FOp))
     -> true
     ; print(filter_op_failed(OpTuple,StateID)),nl, FOp = ('?')),
   filter_ops_and_ids(T,StateID,FT).

filter_op(TransId,OpTerm,Src,Dst,JavaOpTerm) :-
   (animation_mode(cspm) ->
       filter_op_csp(TransId,OpTerm,Src,Dst,JavaOpTerm)
   ; otherwise ->
       filter_op1(TransId,OpTerm,Src,Dst,JavaOpTerm)).

filter_op1(TransId,OpTerm,Src,Dst,
           op(TransId,Op,Src,Dst,Arguments,ArgumentsPretty,Infos,State)) :-
   extract_args_and_return_values(OpTerm,Op,Arguments),
   get_state(Dst,BState), translate:translate_bstate(BState,State),
   prettyprint(Arguments,ArgumentsPretty),
   get_event_infos(TransId,Infos).

filter_op_csp(TransId,OpTerm,Src,Dst,op(TransId,Op,Src,Dst,Arguments,ArgumentsPretty,Infos,CSPState)) :-
   extract_csp_args_and_return_values(OpTerm,Op,Arguments),
   translate:translate_event(OpTerm,PPEvent),
   tools: split_atom(PPEvent,['.','!'],[_Op|ArgumentsPretty]),
   get_state(Dst,CSPState),
   get_event_infos(TransId,Infos).

extract_csp_args_and_return_values(io(Args,ChName,_SPAN),ChName,Args).
extract_csp_args_and_return_values(start_cspm(Name),Name,[]).
extract_csp_args_and_return_values(start_cspm_MAIN,'start_cspm_MAIN',[]).
extract_csp_args_and_return_values(tick(_),tick,[]).
extract_csp_args_and_return_values(tau(_),tau,[]).
extract_csp_args_and_return_values(_OP,'?',[]).

prettyprint([],[]).
prettyprint([H|T],[P|R]) :- translate_bvalue(H,P), prettyprint(T,R).

:- dynamic result_warning_emitted/0.

extract_args_and_return_values(Term,OpName,Arguments) :-
    ( Term='-->'(OpTerm,Results) ->
        ( result_warning_emitted -> true
        ; otherwise ->
            print('Warning: results of operations are currently treated like arguments'),nl,
            assert(result_warning_emitted)),
        append(Results,Args,Arguments)
    ; otherwise -> Term=OpTerm,Args=Arguments
    ),
    OpTerm =.. [OpName|Args].



% returns all variables and their values of the given state as a list
% - constants and variables that are defined for the machine but
%   not present in the given state get the value '?'
% - for each variable a term of the form
%     binding(Id,Value,PPValue,Tag)
%   is in the result list, where
%     Id is the variable's name,
%     Value is its value a Prolog term
%     PPValue is its pretty-printed value
% NOTE: Tag no longer generated: TO DO REVIEW:
%     Tag is one of (cst,var,unknown), depending on whether the
%       variable is a constant, a variable or not defined in the
%       machine description but present in the state
getStateValues(CurID,VariableState) :- /* get variable values */
  get_state(CurID,CurBState),
  findall( binding(Id,Value,PPValue),
           (member(bind(Id,Value),CurBState),
               ( translate_bvalue(Value,PPValue) -> true
               ; otherwise                       -> PPValue='**pretty-print failed**')),
           VariableState).

getVarCstValue(ID,VarOrCst,TVals) :-
  /* a function to get the value of a particular Variable or Constant;
     treats relations specially: a list of relation pairs is returned */
  get_state(ID,CurBState),
  member(bind(VarOrCst,Val),CurBState),
  (Val = [ (_,_) | _]  /* check if it is a relation */
    ->  findall(TFact,relation_fact(_F,VarOrCst,Val,TFact),TVals)
    ;   (TVals = [TVal], Val = TVal)
   ).

 relation_fact(_F,_VarOrCst,Val,TFact) :-
   member((L,R), Val), Fact = (L,R),
   Fact=TFact.

applyVarCstValue(ID,VarOrCstFunction,DomValue,RanResult) :-
  /* a function to apply a particular Variable or Constant Function/Relation to a value;
     treats relations specially: a list of relation pairs is returned */
    get_state(ID,CurBState),
    member(bind(VarOrCstFunction,Val),CurBState),
    member((DomValue,R),Val),
    R= RanResult.

/* ------------------------------- */

/* evaluates an expression in Prolog format and computes the result in the current state */
evaluate_expression(Expression,Result) :-
  current_expression(_CurID,CurState),
  b_interpreter:b_compute_expression(Expression,[],CurState,Result).



%%
% Code for invariant debug
%

:-use_module(bvisual).

evaluate_properties_as_tree(Tree):-
    bvisual:get_properties(TExpr),
    evaluate_expression_as_tree(root, [], TExpr, Tree).

evaluate_invariant_as_tree(StateID, Tree):-
    bvisual:get_invariant(TExpr),
    evaluate_expression_as_tree(StateID, [], TExpr, Tree).

evaluate_precondition_as_tree(StateID, OpName, Tree):-
    bvisual:get_operation_pre(OpName, TExpr, _),
    evaluate_expression_as_tree(StateID, [], TExpr, Tree).


evaluate_raw_expression_as_tree(StateID, Raw, Tree):-
    bmachine:b_type_expression(Raw, [variables], _, TExpr, Errors),
    ( var(TExpr) ->
        print(error_while_typing_raw_expression(Errors)),nl,
        flush_output,
        !,
        fail
        ;
        evaluate_expression_as_tree(StateID, [], TExpr, Tree)
    ).



evaluate_expression_as_tree1(StateID, LocalState, TExpr, Tree) :-
    get_state(StateID, CurState),
    get_tree_from_expr(T0, TExpr, LocalState, CurState),
    get_pretty_print_tree(T0, Tree).

evaluate_expression_as_tree(StateID, LocalState, Expr, Tree) :-
    get_state(StateID, State),
    get_tree_from_expr2(State, LocalState, Expr, T),
    get_tree_node_info(T, Children, NewLocalState, Metas),
    translate_bexpression(Expr, Pretty),
    create_tree_node(Tree, Children, NewLocalState, [pretty_long_desc(Pretty) | Metas]).

evaluate_raw_expression(StateID, Raw, Val):-
    bmachine:b_type_expression(Raw, [variables], _, TExpr, Errors),
    ( var(TExpr) ->
        print(error_while_typing_raw_expression(Errors)),nl,
        flush_output,
        !,
        fail
        ;
        get_state(StateID, State),
        bvisual:x_get_value_of_expr(State, [], TExpr, Val)
    ).

evaluate_raw_expressions(StateID, Raws, Vals):-
    maplist(evaluate_raw_expression(StateID), Raws, Vals).



/* ------------------------- */
/* Execute Operation         */
/* ------------------------- */

% MaxNrOfSolutions is the maximum number of solutions ProB returns. Also we should consider existing solutions if the Predicate holds.
%:- use_module(succeed_max,[succeed_max_call/2]).

/* PROB 1.0 */
execute_custom_operations(CurID, OpName, ParsedPredicate, MaxNrOfSolutions, TOperations, Errors) :-
    setCurrentState(CurID),

    ( (OpName = '$initialise_machine';OpName =  '$setup_constants' )  -> Scope = [prob_ids(visible),variables];
      otherwise ->    b_top_level_operation(OpName),
                      b_get_machine_operation(OpName,_,Parameters,_),
                       Scope = [identifier(Parameters),prob_ids(visible),variables] % prob scope allows one to use identifiers for deferred set elements
    ),
    (MaxNrOfSolutions<1
     -> add_message(eclipse_interface,'execute_custom_operations MaxNrOfSolutions < 1: ',MaxNrOfSolutions)
      ; true),
    bmachine:b_type_expression(ParsedPredicate, Scope, _, TypedPred, Errors),
    retractall(bmachine:b_machine_temp_predicate(_)),
    assert(bmachine:b_machine_temp_predicate(TypedPred)),
    %prob_socketserver:writeln_log(parsed_pred(ParsedPredicate)), %%
    %prob_socketserver:writeln_log(succeed_max_call(execute_custom_operation(CurID,OpName,TO),MaxNrOfSolutions)), %%
    findall(TO, execute_custom_operation(CurID,OpName,TO,MaxNrOfSolutions), TOperations),
    %length(TOperations,Len), prob_socketserver:writeln_log(findall_length(Len)),
    retractall(bmachine:b_machine_temp_predicate(_)),
    clear_context_state.

% TO DO: check if we need to recompute the operation effect: if the ParsedPredicate is TRUE
%  (TRUE = equal(none,integer(none,1),integer(none,1)) ) & MaxNrOfSolutions <=
%        what has already been used previously, we can simply reuse transition from the state space

execute_custom_operation(CurID,OpName,Operation_op,Max) :-
    set_context_state(CurID),
    visited_expression(CurID,InState),
    specfile:compute_operation_effect_max(InState,OpName,Operation,NewState,_TransPathInfo,Max),
    % prob_socketserver:writeln_log(sol(OpName,NewState)), %%
    user:add_trans_id(CurID,Operation,NewState,[],NewID,[],TransId),
    % prob_socketserver:writeln_log(ids(OpName,Operation,NewID,TransId)), %%
    create_operation_term(TransId,Operation,CurID,NewID, Operation_op).



% create a term op(...) with infos for Java interface about an operation-transition
create_operation_term(TransId,Op,CurID,NewID, JavaOpTerm) :-
   (filter_op(TransId,Op,CurID,NewID, JavaOpTerm) -> true
     ; add_error(eclipse_interface,'filter_op failed: ', filter_op(TransId,Op,CurID,NewID, JavaOpTerm)),fail).

/* ------------------------- */
/* SAP Testcase Generation   */
/* ------------------------- */
:- use_module(sap).


sap_generate_local_testcases(GFile,LFile,MaxNodes,Sat,Unsat) :-
    sap:tcl_create_local_testsuite(GFile,LFile,MaxNodes,Sat,Unsat).

sap_find_test_path(Events,EndPredicate,Timeout,Operations) :-
    type_predicate(EndPredicate,TypedPredicate),
    ( create_testcase_path(init,Events,TypedPredicate,Timeout,Trace) -> true
    ; otherwise -> Trace = []),
    % Trace can be "timeout" or "interrupt", too
    ( is_list(Trace) -> op_tuples_to_full_operations(Trace,Operations)
    ; otherwise      -> Operations=Trace).

op_tuples_to_full_operations([],[]).
op_tuples_to_full_operations([(Id,OpTerm,Src,Dst)|Trest],[Op|Orest]) :-
    create_operation_term(Id,OpTerm,Src,Dst,Op),
    op_tuples_to_full_operations(Trest,Orest).

type_predicate(Predicate,TypedPredicate) :-
    b_type_expression(Predicate,[variables],pred,TypedPredicate,Errors),
    add_all_perrors(Errors,[],sap_target_predicate_type_error),
    Errors = [].

/* ------------------------- */
/* Formula evaluation        */
/* ------------------------- */
evaluation_get_top_level(Tops) :-
    % Currently, we do not want to see the rodin information positons
    % in the expression viewer
    suppress_rodin_positions,
    bv_get_top_level(TopIds),
    findall( top(Id,Label,Children),
             ( member(Id,TopIds), bv_expand_formula(Id,Label,Children) ),
             Tops).
evaluation_expand_formula(Id,Label,Children) :-
    % Currently, we do not want to see the rodin information positons
    % in the expression viewer
    suppress_rodin_positions,
    bv_expand_formula(Id,Label,Children).
evaluation_get_values(Ids,StateId,Values) :-
    bv_get_values(Ids,StateId,Values).
evaluation_insert_formula(AST,ParentId,Id) :-
    b_type_expression(AST,[variables],_,Typed,Errors),
    ( Errors == [] ->
        bv_insert_formula(Typed,ParentId,Id)
    ; otherwise ->
        add_error_and_fail(eclipse_interface,'Could not type-check AST','')).

/* ------------------------------- */
/* Find state satisfying predicate */
/* ------------------------------- */
:- use_module(b_state_model_check,[b_set_up_valid_state_with_pred/2]).
find_state_satisfying_predicate(Predicate,Result) :-
    b_type_expression(Predicate,[variables],pred,TPredicate,Errors),
    ( Errors == [] ->
        find_state_satisfying_predicate1(TPredicate,Result)
    ; otherwise ->
        Result = errors(Errors)).
find_state_satisfying_predicate1(Predicate,Result) :-
    user_interruptable_call_det(user:catch_clpfd_overflow_call1(b_set_up_valid_state_with_pred(State,Predicate)),
                                InterruptResult),!,
    ( InterruptResult = interrupted ->
        Result = interrupted
    ;  State = time_out ->
        Result = interrupted
    ; otherwise ->
        Result = state_found(Transition,StateId),
        create_helper_transition(root,find_valid_state,State,StateId,Transition)).
find_state_satisfying_predicate1(_Predicate,no_valid_state_found).


/* ------------------------- */
/* Check for deadlocks       */
/* ------------------------- */

:- use_module(solver_interface, [call_with_smt_mode_enabled/1]).

deadlock_freedom_check(Result) :-
    create_texpr(truth,pred,[],Predicate),
    deadlock_freedom_check1(Predicate,Result).
deadlock_freedom_check(Predicate,Result) :-
    b_type_expression(Predicate,[variables],pred,TPredicate,Errors),
    ( Errors == [] ->
        deadlock_freedom_check1(TPredicate,Result)
    ; otherwise ->
        Result = errors(Errors)).
deadlock_freedom_check1(Predicate,Result) :-
    % always do a deadlock check with SMT mode enabled
    call_with_smt_mode_enabled(deadlock_freedom_check2(Predicate,Result)).
deadlock_freedom_check2(Predicate,Result) :-
    user_interruptable_call_det(user:catch_clpfd_overflow_call1(cbc_deadlock_freedom_check(State,Predicate,0)),
                                InterruptResult),!,
    ( InterruptResult = interrupted ->
        Result = interrupted
    ;  State = time_out ->
        Result = interrupted
    ; otherwise ->
        Result = deadlock(Transition,StateId),
        create_helper_transition(root,deadlock_check,State,StateId,Transition)).
deadlock_freedom_check2(_Predicate,no_deadlock_found).

create_helper_transition(SrcId,Op,DstState,DstId,Transition) :-
    user:tcltk_add_new_transition_transid(SrcId,Op,DstId,DstState,[],TransId),
    transition(SrcId,Action,TransId,DstId),!,
    create_operation_term(TransId,Action,SrcId,DstId,Transition).


/* ------------------------------- */
/* Find values                     */
/* ------------------------------- */
%find_values(Commands,Predicate,MaxNumberSolutions,Solutions) :-
%    setup_userdef_state_space(Commands,root,State,Ids,ConstantsId).

%setup_userdef_state_space2(copy_state(Id),CState,State,Ids,ConstantsId) :-
%    visited_expression(Id,S),
%    ( S=concrete_constants(Constants) ->
%        ( CState=root -> ConstantsId = S
%        ; otherwise   -> ConstantsId = CState),


/* ------------------------------- */
/* Check for invariant violations  */
/* ------------------------------- */

invariant_check(all,Result) :-
    findall(OpName,b_get_machine_operation(OpName,_,_,_),Ops),
    invariant_check2(Ops,Result).
invariant_check(ops(Ops),Result) :-
    invariant_check2(Ops,Result).
invariant_check2(Ops,Result) :-
    call_with_smt_mode_enabled(user_interruptable_call_det(invariant_check3(Ops,Res,[]), Status)),
    ( Status = interrupted -> Result = interrupted
    ; otherwise            -> Result = Res).
invariant_check3([]) --> !.
invariant_check3([Op|Rest]) -->
    invariant_check_for_single_op(Op),
    invariant_check3(Rest).
:- use_module(clpfd_interface,[catch_clpfd_overflow_call1/1]).
invariant_check_for_single_op(OpName,In,Out) :-
    ( clpfd_interface:catch_clpfd_overflow_call1(b_state_model_check:state_model_check(OpName,State1,Operation,State2)) ->
        In = [counterexample(OpName,Trans1,Trans2)|Out],
        atom_concat( invariant_check_ , OpName, RootTrans),
        create_helper_transition(root,    RootTrans,State1,StateId1,Trans1),
        create_helper_transition(StateId1,Operation,State2,_StateId2,Trans2)
    ; otherwise ->
        In = Out).

/* ------------------------------- */
/* CBC Refinement Checking         */
/* ------------------------------- */
:- use_module(cbc_refinement_checks, [cbc_refinement_check/2]).
refinement_check(ResultsString,Result) :-
    call_with_smt_mode_enabled(cbc_refinement_check(ResultsString,Result)).

/* ------------------------------- */
/* CBC Sets                        */
/* ------------------------------- */

:- use_module(library(sets), [list_to_set/2]).
:- use_module(prob2_interface, [initialise_specification/0]).
load_sets_for_cbc(Sets) :-
    list_to_set(Sets,SetOfSets),
    load_sets_for_cbc_aux(SetOfSets).
    %initialise_specification. % now has to be called afterwards
load_sets_for_cbc_aux([]) :-
    load_event_b_project([],[event_b_context(none,'CBCSets',[])],[exporter_version(3)],[]), !.
load_sets_for_cbc_aux(SetOfSets) :-
    maplist(create_set_term,SetOfSets,SetTerms),
    load_event_b_project([],[event_b_context(none,'CBCSets',[sets(none,SetTerms)])],[exporter_version(3)],[]).

create_set_term(Name,deferred_set(none,Name)).

/* ----------------------------- */
/* CBC Static Assertion Checking */
/* ----------------------------- */
:- use_module(b_state_model_check, [cbc_static_assertions_check/1]).
cbc_static_assertion_violation_checking(Result) :-
    user_interruptable_call_det(catch_clpfd_overflow_call1(cbc_static_assertions_check(State)),InterruptResult), !,
    ( InterruptResult = interrupted ->
        Result = interrupted
    ;  State = time_out ->
        Result = interrupted
    ;  State = no_counterexample_found ->
        Result = no_counterexample_found
    ;  State = no_counterexample_exists(_,_) ->
        Result = no_counterexample_exists
    ; otherwise ->
        State = counterexample_found(RealState),
        Result = counterexample_found(Transition,StateId),
        create_helper_transition(root,deadlock_check,RealState,StateId,Transition)).

/* ------------------------- */
/* Generate Visualizations   */
/* ------------------------- */

evaluate(In,Result,Warnings) :- tcltk_interface:tcltk_eval(In,Result,Warnings,_Solution).

generate_tree_from_bool_expression(BExpr,Tree,BSVars):-
    (
        current_state_corresponds_to_initialised_b_machine,!,
        specfile:current_b_expression(CurState0),!,

   ~~mnf(b_interpreter:insert_before_substitution_variables(BSVars,[],CurState0,CurState0,CurState))

        ;
        CurState = []
    ),
    ~~mnf(get_tree_from_expr(Tree, BExpr, [], CurState)).


/* ------------------------- */
/* Dot Printing              */
/* ------------------------- */

dot_hierarchy(Filename) :- b_machine_hierarchy:print_hierarchy(Filename).

/* ------------------------- */
/* Evalstore                 */
/* ------------------------- */
es_reset :- evalstore_reset.

get_vacuous_invariants(L) :-
   findall(I,vacuous_invariant(I),L).

vacuous_invariant(Inv) :-
   b_get_invariant_from_machine(Invariant),
   b_interpreter:member_conjunct(Inv,Invariant,_),
   get_texpr_expr(Inv,implication(LHS,_RHS)),
   print('Checking vor vacuity: '), translate:print_bexpr(Inv),nl,
   (test_boolean_expression_in_node(ResID,LHS)
     -> print(lhs_true_in_state(ResID)),nl,nl,  % Implication can be triggered in at least one state
        fail
     ;  print(' *** VACUOUS *** '),nl,nl
    ).

get_vacuous_guards(L) :-
   findall(G,vacuous_guard(_,G),L).

vacuous_guard(OpName,Result) :-
    get_guard_with_one_negated(OpName,Guard,Neg),
    translate:translate_bexpression_with_limit(Guard,GuardTS),
    format('Checking vor vacuity: ~w : ~w~n',[OpName,GuardTS]),
    %translate:print_bexpr(Neg),nl,
    (test_boolean_expression_in_node(ResID,Neg)
      -> print(guard_individually_false_in(ResID)),nl,nl,
         fail
      ;  print(' *** VACUOUS *** '),nl,nl,
         tools:ajoin([OpName,' : ',GuardTS],Result)
     ).


:- use_module(predicate_debugger,[get_operation_enabling_condition/6]).
%:- use_module(b_ast_cleanup, [clean_up/3]).
:- use_module(b_interpreter_components,[construct_optimized_exists/3]).

% get a guard, by translating all parameters into existential quantifiers
get_guard_with_one_negated(OpName,SingleGuard,FullModifiedGuard) :-
  get_operation_enabling_condition(OpName,Parameters,EnablingCondition,_BecomesSuchVars,false,true),
  % TO DO: partition; avoid lifting becomes_such_that conditions
  %   (in EventB the feasibility PO will ensure that Guard => BecomeSuchThat is ok)
  negate_one_guard(EnablingCondition,SingleGuard,ModifiedEnablingCondition),
  construct_optimized_exists(Parameters,ModifiedEnablingCondition,FullModifiedGuard).

:- use_module(probsrc(bsyntaxtree),[conjunction_to_list/2,occurs_in_expr/2]).
negate_one_guard(Conjunct,SingleGuard,ModifiedConjunct) :-
   conjunction_to_list(Conjunct,L),
   append(Rest1,[SingleGuard|Rest2],L),
   create_negation(SingleGuard,NegSingleGuard),
   append(Rest1,[NegSingleGuard|Rest2],NewGuard), % TO DO: better treatment for well-definedness !?
   conjunct_predicates(NewGuard,ModifiedConjunct).

:- use_module(b_interpreter, [tcltk_quick_describe_unsat_properties/1, tcltk_unsatisfiable_components_exist/0]).
quick_describe_unsat_properties(Res) :-
    tcltk_unsatisfiable_components_exist
    -> tcltk_quick_describe_unsat_properties(TRes),  % sometimes this can be 'The properties were satisfiable' if tcltk_unsatisfiable_components_exist faulty
       Res = unsat_properties(TRes)
    ;  Res = no_unsat_properties_found.


% This just bundles a few commands used by BMotionStudio.
% Should avoid communication overhead.
% Replaced by ProB 2.0 Api
:- use_module(prob2_interface,[state_property/3,op_timeout_occurred/2,get_state_errors/2]).
prob_construct_trace([],[],[],[],[],[]).
prob_construct_trace([Name|Names],[Pred|Preds],[NrOfSol|NrOfSols],
                     [state(NewStateId,StateValues,Initialised,InvKO,MaxOpsReached,Timeout,OpTimeout,StateErrors,UnsatProps,OperationsForState)|States],
                     [FirstOp|Ops],Errors) :-
    getCurrentStateID(Id),
    execute_custom_operations(Id,Name,Pred,NrOfSol,Op,LocalErrors),
    Op = [FirstOp|_],
    FirstOp = op(_,_,_,NewStateId,_,_,_,_),
    % the following is essentially ExploreStateCommand
    computeOperationsForState(NewStateId,OperationsForState),
    getStateValues(NewStateId,StateValues),
    state_property(initialised,NewStateId,Initialised),
    state_property(invariantKO,NewStateId,InvKO),
    state_property(max_operations_reached,NewStateId,MaxOpsReached),
    state_property(timeout_occurred,NewStateId,Timeout),
    op_timeout_occurred(NewStateId,OpTimeout),
    get_state_errors(NewStateId,StateErrors),
    quick_describe_unsat_properties(UnsatProps),
    setCurrentState(NewStateId),
    prob_construct_trace(Names,Preds,NrOfSols,States,Ops,FurtherErrors),
    append(LocalErrors,FurtherErrors,Errors).
