% (c) 2014 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(mutation,[mutate/3, mutate_guards/2]).


% with this module you can mutate a predicate into a predicate that implies that the negation of the original predicate is true
% the difference with create_negation is that it sometimes generates multiple solutions
%  not(x<y)  ---> produces two mutants:  x=y   +  x>y
% it is related to MCDC coverage which make the guards false

% use_module(probsrc(mutation)), mutate_guards(Op,MP).

:- use_module(library(lists)).
:- use_module(predicate_debugger,[get_operation_enabling_condition/6]).
%:- use_module(b_ast_cleanup, [clean_up/3]).
:- use_module(b_interpreter_components,[construct_optimized_exists/3]).
:- use_module(probsrc(bsyntaxtree)).
:- use_module(probsrc(error_manager)).
:- use_module(tools,[ajoin/2]).

mutate_guards(OpName,MutatedGuards) :-
   get_operation_enabling_condition(OpName,Parameters,EnablingCondition,_BecomesSuchVars,false,true),
   % TO DO: partition; avoid lifting becomes_such_that conditions 
   %   (in EventB the feasibility PO will ensure that Guard => BecomeSuchThat is ok)
   dif(OpName, '$initialise_machine'),
   nl,nl,print(OpName),nl,
   translate:print_bexpr(EnablingCondition),nl,
   mutate(EnablingCondition,TargetPred,_),
   construct_optimized_exists(Parameters,TargetPred,MutatedGuards),
   nl,print('Mutated : '),
   translate:print_bexpr(MutatedGuards),nl.


mutate(Predicate,TargetPred, conjunct_false(PosNr,SinglePred,InnerTree)) :- 
   is_a_conjunct(Predicate,_,_),
   !,
   % check that every conjunct can be set individually to false:
   % A1 & ... Ai & ... Ak  ---------->  A1 & A2 ... not(Ai) & ... Ak
   conjunction_to_list(Predicate,List),
   select_pred(List, Rest1, SinglePred, Rest2, PosNr),
   mutate(SinglePred, InnerTargetPred, InnerTree),
   append(Rest1,[InnerTargetPred|Rest2],NewConj), % TO DO: better treatment for well-definedness !?
   % if SinglePred is a WD-Guard for Rest2: do not require to analyse Rest2 ?!
   conjunct_predicates(NewConj,TargetPred).
mutate(Predicate,TargetPred, disjunct(ITA,ITB)) :- 
   is_a_disjunct(Predicate,A,B),
   !, 
   mutate(A,MutationA,ITA),
   mutate(B,MutationB,ITB),
   conjunct_predicates([MutationA,MutationB],TargetPred).
mutate(Predicate,TargetPred, implication_rhs(ITR)) :-
   is_an_implication(Predicate,LHS,RHS),
   !,  
   mutate(RHS,MutationRHS,ITR),
   conjunct_predicates([LHS,MutationRHS],TargetPred).
mutate(Predicate,TargetPred, equivalence(Pos,ITR)) :-
   is_an_equivalence(Predicate,LHS,RHS),
   !,  
   (Pos=rhs,mutate(RHS,MutationRHS,ITR),
    conjunct_predicates([LHS,MutationRHS],TargetPred)
    ;
    Pos=lhs,mutate(LHS,MutationLHS,ITR),
    conjunct_predicates([MutationLHS,RHS],TargetPred)
   ).
mutate(Predicate,TargetPred, negation) :-
   is_a_negation(Predicate,NPred),
   !,
   TargetPred = NPred.
mutate(b(E,pred,Info),b(MutationE,pred,Info),atomic(E)) :-
   mutate_atomic_pred(E,MutationE),
   !.
mutate(Predicate,TargetPred,false(Predicate)) :- !, create_negation(Predicate,TargetPred).

% to do: write create_opt_negation
mutate_atomic_pred(equal(A,B),not_equal(A,B)).
mutate_atomic_pred(not_equal(A,B),equal(A,B)).
mutate_atomic_pred(member(A,B),not_member(A,B)).
mutate_atomic_pred(not_member(A,B),member(A,B)).
mutate_atomic_pred(subset(A,B),not_subset(A,B)).
mutate_atomic_pred(not_subset(A,B),subset(A,B)).
mutate_atomic_pred(subset_strict(A,B),not_subset_strict(A,B)).
mutate_atomic_pred(not_subset_strict(A,B),subset_strict(A,B)).


mutate_atomic_pred(less(A,B),equal(A,B)).
mutate_atomic_pred(less(A,B),greater(A,B)).
mutate_atomic_pred(less_equal(A,B),greater(A,B)).
mutate_atomic_pred(greater(A,B),X) :- mutate_atomic_pred(less(B,A),X).
mutate_atomic_pred(greater_equal(A,B),X) :- mutate_atomic_pred(less_equal(B,A),X).


select_pred(List, Rest1, Pred, Rest2, PosNr) :-
   append(Rest1,[Pred|Rest2],List), length(Rest1,L1), PosNr is L1+1.
