% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


portray_message(informational, _).

%% sp_4_1_3_nested_exception_patch.pl BEGIN
%% Compile this to patch a bug in 4.1.3 affecting nested exception handlers.
%% This source code is a no-op on other versions of sicstus than 4.1.3.
:- if(current_prolog_flag(dialect,sicstus)).
:- if(current_prolog_flag(version_data,sicstus(4,1,3,0,_))).

prolog:(matching_exception(internal, Ball, Templ) :- !,
        Templ = Ball).
prolog:(matching_exception(prolog, '$reserved'(_), _) :- !,
        fail).
prolog:(matching_exception(prolog, time_out, _) :- !,
        fail).
prolog:(matching_exception(_, Ball, Template) :-
        compound(Ball),
        sysball_iso_equiv(Ball,Iso1Exc), !,
        Template = error(Iso1Exc, Ball)).
prolog:(matching_exception(_, Ball, Ball)).

:- endif.                       % sicstus 4.1.3
:- endif.                       % sicstus
%% sp_4_1_3_nested_exception_patch.pl END

%:- include('self_check_off.pl').

% use module_info/3 because we are not in a module here
:- use_module(module_information).
:- module_info(prob_cli,group,infrastructure).
:- module_info(prob_cli,description,'ProB start file in cli mode.').

:- use_module(debugging_calls).
:- register_debugging_call pp_mnf(*), pp_cll(*), mnf(*), mnf(-,*), det_call(*).
:- disable_debugging_calls.

:- use_module(pathes,[set_search_pathes/0]). % called first to set_compile_time_search_pathes
:- use_module(tools,[string_concatenate/3,arg_is_number/2]).

:- compile(tcltk_interface).
%:- compile(gui_tcltk).

:- use_module(eclipse_interface).
:- use_module(prob2_interface,[start_animation/0, is_initialised_state/1,
                               set_eclipse_preference/2, update_preferences_from_spec/0,
                               load_cspm_spec_from_cspm_file/1]).


:- use_module(prob_socketserver).

:- use_module(coverage_statistics,[compute_the_coverage/4]).
:- use_module(value_persistance, [set_storage_directory/1,unset_storage_directory/0]).

do_one_gui_event.
tk_do_all_events.

:- dynamic junit_mode/1.

:- use_module(b_trace_checking,[check_default_trace_for_specfile/1, tcltk_check_state_sequence_from_file/1]).
%:- use_module(cspParserHs,'cia/haskell/java_api/cspParserHs.pl',
%               [initCspParser/0]).

:- use_module(library(file_systems),[file_exists/1]).
:- use_module(debug).
:- use_module(preferences).
:- use_module(probltlsrc(ltl),[ltl_check_assertions/2,ltl_model_check/4]).
:- use_module(probltlsrc(ctl),[ctl_model_check/4]).
:- use_module(test_typechecker,[run_typecheck_testcase/2]).

:- use_module(self_check,[disable_interaction_on_errors/0]).
:- use_module(error_manager).

:- use_module(extension('zmq/master/master')).
:- use_module(extension('zmq/worker/worker')).

:- use_module(self_check,[perform_self_check/1,turn_off_run_time_type_checks/0,turn_on_run_time_type_checks/0]).

:- use_module(library(system)).
:- use_module(library(codesio)).

:- use_module(log_analyser).
:- use_module(bmachine,[b_write_machine_representation_to_file/2,full_b_machine/1]).

:- use_module(xtl_interface,[set_cspm_main_process/1]).
:- use_module(visualize_graph,[tcltk_print_states_for_dot/1]).
:- use_module(state_space_reduction,[write_transition_diagram_for_expr_to_dotfile/2,
                                     write_signature_merge_to_dotfile/1]).
:- use_module(reduce_graph_state_space,[print_dot_for_dfa_from_nfa/1]).
:- use_module('kodkod/kodkod_test').
:- use_module(predicate_analysis,[test_predicate_analysis/0]).
:- use_module(b_show_history,[write_history_to_file/2,write_values_to_file/1,write_history_to_user_output/1]).
:- use_module(external_functions,[check_if_predicates_were_evaluated/1]).
:- use_module(sap, [write_all_deadlocking_paths_to_xml/1]).

:- use_module(smtlib(smtlib2_cli)).
:- use_module(disprover_test_runner, [run_disprover_on_all_pos/1, load_po_file/1,print_disprover_stats/0]).

:- meta_predicate if_option_set(-,0).
:- meta_predicate if_option_set(-,0,0).
:- meta_predicate if_option_set_loaded(-,-,0).

printsilent(S) :- option(silent) -> true ; print(S).
nls :- option(silent) -> true ; nl.
formatsilent(FS,Args) :- (option(silent) -> true ; format(FS,Args)).

start_probcli_timer(timer(T1,WT1)) :-
                   statistics(runtime,[T1,_]),
                   statistics(walltime,[WT1,_]).
stop_probcli_debug_timer(Timer,Msg) :- (debug_mode(on) -> stop_probcli_timer(Timer,Msg) ; true).
stop_probcli_timer(timer(T1,WT1),Msg) :-
        statistics(runtime,[T2,_]), TotTime is T2-T1,
        statistics(walltime,[WT2,_]), WTotTime is WT2-WT1,
        format('~w ~w ms walltime (~w ms runtime), since start: ~w ms~n',[Msg,WTotTime,TotTime,WT2]),
        !.
stop_probcli_timer(Timer,Msg) :- add_internal_error('Illegal timer call: ',stop_probcli_timer(Timer,Msg)).
print_total_probcli_timer :-
        statistics(runtime,[T2,_]),
        statistics(walltime,[WT2,_]),
        format('Since start of probcli: ~w ms walltime (~w ms runtime)~n',[WT2,T2]).


:- use_module(library(timeout)).
:- meta_predicate timeout_call(0,-).
timeout_call(Call,NOW) :- option(timeout(TO)),!,
   statistics(runtime,[T1,_]),
   time_out(Call,TO,Res),
   statistics(runtime,[T2,_]), Runtime is T2-T1,
   printsilent('Runtime: '), printsilent(Runtime), printsilent(' ms'),nls,
   (Res=time_out -> print('*** Timeout occurred: '), print(TO),nl,
                    print('*** Call: '), print(Call),nl,
                    nl,
                    writeln_log(timeout_occurred(NOW,Call))
                 ;  true).
timeout_call(Call,_NOW) :-
   statistics(runtime,[T1,_]),
   call(Call),
   statistics(runtime,[T2,_]), Runtime is T2-T1,
   printsilent('Runtime: '), printsilent(Runtime), printsilent(' ms'),nls.

:- use_module(junit_tests,[set_junit_dir/1, print_junit/2, create_junit_result/6]).
set_junit_mode(X) :- set_junit_dir(X), statistics(runtime,[Start,_]), assert(junit_mode(Start)).


go_proxy :-
    catch( go_cli(['-s','8888']), halt(ExitCode),
           ( nl,write('CLI halt prevented, exit code '),write(ExitCode),nl) ).

go_cli :-
    prolog_flag(argv,ArgV),
    go_cli_with_junit_check(ArgV).


reset_cli :-
  reset_errors,
  reset_expected_error_occurred,
  preferences:reset_to_defaults,
  unset_storage_directory,
  % now done via event handling: clear_dynamic_predicates_for_POR,
  %     retractall(refinement_checker: generated_predeterministic_refinement(_,_,_)),
  %     retractall(refinement_checker: determinism_check_div_found(_)),
  %     reset_cspm_main_process,
  (file_loaded(_)
   -> clear_loaded_machines,
      retractall(file_loaded(_,_)),
      retractall(loaded_main_file(_,_))
   ;  true).

go_cli_with_junit_check(ArgV) :-
    catch( go_cli(ArgV),
           halt(ExitCode),
           ( ExitCode = 0 ->
               true
           ; junit_mode(S) ->
               statistics(runtime,[E,_]), T is E - S,
	       create_junit_result(ArgV,'Integration Tests','Integration_Test',T,error([ExitCode]),Result),
               print_junit([Result],'Integration_Test')
           ; throw(halt(ExitCode)))).

go_cli(ArgV) :-
     (go_cli2(ArgV) -> true
     ; print_error('INTERNAL ERROR OCCURRED !'),nl,
       print_error(' (go_cli failed)'),nl, halt_exception(1)).

no_command_issued :- \+ command_option(_).
command_option(X) :- option(X), \+ not_command_option(X).
not_command_option(verbose).
not_command_option(very_verbose).
not_command_option(profiling_on).
not_command_option(set_pref(_,_)).
not_command_option(set_card(_,_)).
not_command_option(set_argv(_)).
not_command_option(silent).

%:- use_module(extension('myheap/myheap')).
go_cli2(ArgV) :-
    %print_total_probcli_timer,
    set_search_pathes,
    %myheap_init,
    init_eclipse_preferences,
    external_functions:reset_argv,
    cli_init_options(ArgV,RemArgV),
    %% cli_init_options(['-ll'|ArgV],RemArgV),  %% comment in to have an automatically logging probcli
    %% cli_init_options(['-vv','-version'|ArgV],RemArgV),  %% comment in to have an automatically verbose probcli
    !,
    ((RemArgV=[File],no_command_issued,
      get_filename_extension(File,Ext), \+ do_not_execute_automatically(Ext))
       % then assume we want to do -execute_all:
      -> assert(option(execute(2147483647,false)))
       ; true),
    if_option_set(test_mode,set_random_seed_to_prologs_current_random_state, set_new_random_seed),
    if_option_set(verbose,
                  verbose, %tcltk_turn_debugging_on(19),
                  tcltk_turn_debugging_off),
    if_option_set(very_verbose,
                  very_verbose), %tcltk_turn_debugging_on(5)),
    if_option_set(profiling_on,profiling_on),
    debug_print(9,'Command Line Arguments: '),debug_println(9,ArgV),
    if_option_set(very_verbose,
                  print_options),
    debug_print(6,'Command Line File Args: '),debug_println(6,RemArgV),
    if_option_set(cache_storage(StorageDir), set_storage_directory(StorageDir)),
    if_option_set(parsercp(ParserLoc),
                  assert(parsercall:parser_location(ParserLoc))),
    cli_show_help(ArgV,RemArgV),
    generate_time_stamp(Datime,NOW),
    if_option_set(log(F),
                  cli_start_logging(F,NOW,Datime,RemArgV)),
    if_option_set(evaldot(F),
                  set_eval_dot_file(F)),
    if_option_set(runtimechecking,
                  turn_on_run_time_type_checks,
                  turn_off_run_time_type_checks),
    if_option_set(junit(X),
                  set_junit_mode(X)),
    set_prefs,
    if_option_set(set_argv(F),
                  set_argv(F)),
    if_option_set(selfcheck(_),
                  cli_start_selfcheck),
    if_option_set(typechecker_test(Filename),
                  (run_typecheck_testcase(Filename,typesok) -> halt_prob(NOW); halt1_prob(NOW))),
    if_option_set(print_version, print_version),
    if_option_set(print_short_version, print_short_version),
    if_option_set(very_verbose,
                  preferences:print_preferences),
    if_option_set(zmq_worker, zmq_start_worker(localhost,5000,0,NOW)),
    if_option_set(zmq_worker2(MasterIP, Port, Recover), zmq_start_worker(MasterIP,Port,Recover,NOW)),
    debug_println(6,loading(RemArgV)),
    cli_load_files(RemArgV,NOW), % all CLI arguments which are not understood are supposed to be files to be treated
    debug_println(6,finished_loading(RemArgV)),
    if_options_set(analyse_log(LF,RowID,ForGnu),
                  analyse_log(LF,RowID,ForGnu)),
    if_options_set(check_log(LF),
                  check_log(LF)),
    if_option_set(socket(Port),
                  cli_start_socketserver(Port)),
    % check_all_expected_errors_occurred(NOW),  % is now checked for each file; socket_server should not generate errors ?
    debug_println(20,'% ProB Finished'),
    cli_print_junit_results(ArgV),
    writeln_log_time(prob_finished(NOW)).

cli_init_options(ArgV,RemArgV) :- %print(argv(ArgV)),nl,
    append(ProBArgV,['--'|BArgV],ArgV),!, % pass arguments after -- to B via external_functions
    cli_init_options2(ProBArgV,RemArgV),
    debug_println(20,set_argv_from_list(BArgV)),
    external_functions:set_argv_from_list(BArgV).
cli_init_options(ArgV,RemArgV) :- cli_init_options2(ArgV,RemArgV).
cli_init_options2(ArgV,RemArgV) :-
    retractall(option(_)),
    %%assert(option(log('/tmp/ProBLog.log'))), print('LOGGING'),nl, %% coment in to build a version of probcli that automatically logs
    ( get_options(ArgV,recognised_option,Options,RemArgV) ->
        assert_all_options(Options)
    ; otherwise ->
        print_error(get_options_failed(ArgV)),definite_error_occurred).
cli_show_help(ArgV,RemArgV) :-
    ( (option(help) ; ArgV=[]) ->
        print_help, (RemArgV=[] -> halt_exception ; true)
    ; true).
cli_start_logging(F,NOW,Datime,RemArgV) :-
    debug_print(20,'%logging to: '), debug_println(20,F), set_log_file(F),
    writeln_log(start_logging(NOW,F)),
    version(V1,V2,V3,Suffix), revision(Rev),  lastchangeddate(LCD),
    writeln_log(version(NOW,V1,V2,V3,Suffix,Rev,LCD)),
    prolog_flag(version,PV),
    writeln_log(prolog_version(NOW,PV)),
    findall(Opt, option(Opt), Options),
    writeln_log(options(NOW,Options,RemArgV)),
    datime(Datime,DateRecord),
    writeln_log(date(NOW,DateRecord)).

:- use_module(eventhandling,[store_virtual_event/1]).
cli_start_selfcheck :-
      %clear_loaded_machines_wo_errors,
      b_set_initial_machine,
      store_virtual_event(clear_specification), % TO DO: try and get rid of the need for this
      start_animation,
      option(selfcheck(Module)),
      (perform_self_check(Module) -> true ; true),
      (test_error_occurred(_) -> error_occurred(selfcheck) ; true),
      fail.
cli_start_selfcheck.


:- dynamic file_loaded/2.
file_loaded(Status) :- file_loaded(Status,_File).

%cli_load_files(RemArgV,NOW) :-
%    if_options_set(l_cspm_pl_file(PlFile),cli_l_cspm_pl_file(PlFile)),!,
%    assert(file_loaded(true,PlFile)),
%    cli_process_loaded_file(NOW,PlFile),
%(RemArgV=[] -> true ; cli_load_files2(RemArgV,NOW)).
cli_load_files([],NOW) :- % no files are provided
   !,
   ( options_allow_start_without_file -> cli_set_empty_machine
     ; we_did_something -> true
     ; otherwise -> print('No file to process'),nl),
   writeln_log_time(start_processing_empty_machine(NOW)),
   cli_process_loaded_file(NOW,'$EMPTY_MACHINE'),
   check_all_expected_errors_occurred(NOW). % check that all expected errors occurred; below they will be checked for each file
cli_load_files(RemArgV,NOW) :- cli_load_files2(RemArgV,NOW).


cli_set_empty_machine :- % TO DO: do this more properly here and for initialise_required
%       announce_event(clear_specification),
       set_animation_mode(b),
%       announce_event(start_initialising_specification),
       bmachine:b_set_empty_machine,
       assert(file_loaded(true,empty_machine)).
       %announce_event(specification_initialised).


options_allow_start_without_file :- option(eval_repl).
options_allow_start_without_file :- option(eval_string_or_file(_,_,_,_)).
options_allow_start_without_file :- option(check_log(_)).

we_did_something :- option(print_version).
we_did_something :- option(print_short_version).

option_only_works_for_single_file(zmq_assertion(_PortStart,_Logfile)).
option_only_works_for_single_file(zmq_master2(_N,_Max,_PortStart,_Workers,_Logfile)).
option_only_works_for_single_file(zmq_master(_Workers,_Logfile)).

clear_loaded_files :- (file_loaded(_) -> clear_loaded_machines_wo_errors ; true).
cli_load_files2([],_) :- !.
cli_load_files2([F1,F2|T],NOW) :-
    option(Option),
    option_only_works_for_single_file(Option),!,
    add_error(probcli,'The following option can only be used for a single file: ',Option),
    add_error(probcli,'Multiple files provided: ',[F1,F2|T]),
    halt_prob(NOW).
cli_load_files2(RemArgV,NOW) :-
    %print_total_probcli_timer,
    clear_loaded_files,
    retractall(file_loaded(_,_)),
    RemArgV = [MainFile0|Rest],!,
    cli_load_files3(MainFile0,Rest,NOW).
cli_load_files3(MainFile0,Rest,NOW) :-
    safe_absolute_file_name(MainFile0,MainFile,[access(none)]), % converts Windows slash into Unix slash,...
    if_option_set(file_info,print_file_info(MainFile)),
    ( file_exists(MainFile) ->
        debug_println(6,file_exists(MainFile)),
        ( load_main_file(MainFile,NOW,Already_FullyProcessed) ->
          (Already_FullyProcessed==true -> true
           ; assert(file_loaded(true,MainFile)),
             debug_println(20,'% Processing File'),
             writeln_log_time(start_processing(NOW)),
             start_probcli_timer(Timer),
             (cli_process_loaded_file(NOW,MainFile)
               -> stop_probcli_debug_timer(Timer,'% Finished Processing File after')
               ;  print_error('### Processing file failed: '), print_error(MainFile)
             ),
             writeln_log_time(finished_processing(NOW))
           ),
		  get_errors,
		  get_state_space_errors
        ; otherwise ->
            assert(file_loaded(error,MainFile)),
            print_error('### Loading Machine Failed'),
            writeln_log_time(loading_failed(NOW,MainFile)),
            error_occurred(load_main_file),
            get_errors
        ),nl
    ; otherwise ->
        nl, assert(file_loaded(error,MainFile)),
        (number(MainFile0)
            -> print_error('### Specified number not associated with command and does not exist as file: '),
               print_error(MainFile0)
         ; atom_codes(MainFile0,[45|_]) % starts with a dash - : probably an illegal command-line option
             -> print_error('### Specified Option or File does not exist: '),print_error(MainFile0)
          ; print_error('### Specified File does not exist: '),print_error(MainFile)),
        definite_error_occurred
    ),
    check_all_expected_errors_occurred(NOW),
    reset_expected_error_occurred, % reset for next file
    NOW1 is NOW+1, update_time_stamp(NOW1),
    cli_load_files2(Rest,NOW1).

print_file_info(F) :-
   print('Specification_File('), print(F), print(')'),nl.

% ---------------------

% process all the commands for a loaded file:
cli_process_loaded_file(NOW,MainFile) :-
    (real_error_occurred -> print_error('% *** Errors occurred while loading ! ***'),nl,nl ; true),
    get_errors,
    if_option_set(kodkod_performance(KPFile,Iterations),
                  compare_kodkod_performance1(KPFile,Iterations,NOW)),
    if_option_set(kodkod_comparision(MaxResiduePreds),
                  test_kodkod_and_exit(MaxResiduePreds,NOW)),
    % if_option_set(add_csp_guide(CspGuide), tcltk_add_csp_file(CspGuide)), %% moved to later to ensure B machine is precompiled; allows e.g. type_check_csp_and_b to run

    if_option_set(csp_main(MAINPROC),
                  set_cspm_main_process(MAINPROC)),

    if_option_set(zmq_master2(N,Max,PortStart,Recover,Logfile), zmq_start_master(invariant,N,Max,PortStart,Recover,Logfile)),
    if_option_set(zmq_master(Logfile), zmq_start_master(invariant,200,-1,5000,0,Logfile)),

    % STARTING ANIMATION/MODEL CHECKING
    cli_start_animation(NOW),
    if_option_set_loaded(trace_check,trace_check,
                         cli_start_trace_check(MainFile)),

    cli_process_loaded_file_afer_start_animation(NOW).

cli_process_loaded_file_afer_start_animation(NOW) :-
    ifm_option_set(cli_print_machine_statistics,
                  cli_print_machine_statistics),
    ifm_option_set(pretty_print_internal_rep(PPFILE),
                  b_write_machine_representation_to_file(needed,PPFILE)),
    ifm_option_set(pretty_print_all_types(PPFILE),
                  b_write_machine_representation_to_file(all,PPFILE)),

    if_option_set_loaded(state_trace(File),state_trace,
                         cli_start_trace_state_check(File)),

    (initialise_required
      -> check_loaded(initialise),
         cli_start_initialisation(NOW),
         writeln_log_time(initialised(NOW))
      ;  true),

    if_option_set(zmq_assertion(PortStart, Logfile),
                  zmq_start_master(assertion,PortStart, Logfile)),

    if_option_set(cli_start_mc_with_tlc,cli_start_mc_with_tlc),
    if_option_set(cli_symbolic_model_check,cli_symbolic_model_check),

    if_option_set_loaded(cli_check_properties,check_properties,
                         cli_check_properties(NOW)),
    ifm_option_set_loaded(cli_check_assertions(ALL,ReqInfos),check_assertions,
                         cli_check_assertions(ALL,ReqInfos,NOW)),
    if_option_set(set_goal(GOAL),
                  cli_set_goal(GOAL)),
    if_option_set(set_searchscope(GOAL),
                  cli_set_searchscope(GOAL)),
    ifm_option_set_loaded(cli_mc(Nr),model_check,
                         cli_start_model_check(Nr,NOW)),
    ifm_option_set_loaded(cli_random_animate(Steps,ErrOnDeadlock),animate,
                         cli_random_animate(NOW,Steps,ErrOnDeadlock)),
    ifm_option_set_loaded(execute(Steps,ErrOnDeadlock),animate,
                         cli_execute(Steps,ErrOnDeadlock)),
    if_option_set_loaded(pa_check,predicate_analysis,
                         test_predicate_analysis),

    cbc_check(NOW),
    if_options_set(feasibility_analysis_csv(EnablingCsvFile),
                   do_feasibility_analysis(EnablingCsvFile)),
    ifm_option_set_loaded(mcm_tests(ADepth,AMaxS,ATarget,Output),mcm_test_cases,
                         mcm_test_case_generation(ADepth,AMaxS,ATarget,Output)),
    ifm_option_set_loaded(all_deadlocking_paths(File),all_deadlocking_paths,
                         write_all_deadlocking_paths_to_xml(File)),
    ifm_option_set_loaded(cbc_tests(ADepth,ATarget,Output),cb_test_cases,
                         cbc_test_case_generation(ADepth,ATarget,Output)),
    ifm_option_set_loaded(test_description(TestDescFile),cb_test_cases,
                         test_generation_by_xml_description(TestDescFile)),
    if_options_set(csp_in_situ_refinement_check(P,Type,Q),
                  cli_csp_in_situ_refinement_check(P,Type,Q,NOW)),
    if_options_set(csp_checkAssertion(Proc,Model,AssertionType),
                   cli_checkAssertion(Proc,Model,AssertionType,NOW)),
    if_options_set(refinement_check(RefFile,PerformSingleFailures,RefNrNodes),
                  cli_start_refinement_check(RefFile,PerformSingleFailures,RefNrNodes,NOW)),
    if_options_set(ltl_forumla_model_check(Formula,Expected),cli_ltl_model_check(Formula,Expected,_)),
    if_options_set(ctl_forumla_model_check(Formula,Expected),cli_ctl_model_check(Formula,Expected,_)),
    % TO DO print ctl/ltl statistics
    if_options_set(csp_get_assertions,cli_csp_get_assertions),
    if_options_set(set_csp_event_as_visible,cli_set_csp_event_as_visible),
    if_options_set(cspm_evaluate_expression(E),cli_cspm_evaluate_expression(E)),
    if_options_set(eval_csp_expression(E),cli_eval_csp_expression(E)),
    if_options_set(csp_add_unicode(File), cli_csp_add_unicode(File)),
    if_options_set(csp_remove_unicode(File), cli_csp_remove_unicode(File)),
    if_options_set(csp_translate_to_file(PlFile),cli_csp_translate_to_file(PlFile)),
    if_options_set(get_min_max_coverage(FileName),cli_get_min_max_coverage(FileName)),
    if_options_set(get_coverage_information(FileName),cli_get_coverage_information(FileName)),
    if_options_set(vacuity_check,cli_vacuity_check),
    if_option_set_loaded(check_goal,check_goal,cli_check_goal),
    if_option_set_loaded(animate,animate,
                         (animate_machine -> true ; true)),
    evaluate_from_commandline,
    if_option_set_loaded(ltl_assertions,check_ltl_assertions,
                         (timeout_call(ltl_check_assertions,NOW) -> true; true)),
    ifm_option_set_loaded(ltl_file(LtlFilename),check_ltl_file,
                         (ltl_check_file(LtlFilename) -> true; true)),
    ifm_option_set_loaded(history(HistoryFilename),history,
                         cli_print_history(HistoryFilename)),
    ifm_option_set_loaded(print_values(ValuesFilename),values,
                         cli_print_values(ValuesFilename)),
    if_options_set(save_state_for_refinement(SaveRefF),
                  refinement_checker:tcltk_save_specification_state_for_refinement(SaveRefF)),
    if_options_set(state_space_dot(DotFile),
                  tcltk_print_states_for_dot(DotFile)),
    if_options_set(state_space_merge_dot(DotFile),
                  write_signature_merge_to_dotfile(DotFile)),
    if_options_set(state_space_dfa_dot(DotFile),
                  print_dot_for_dfa_from_nfa(DotFile)),
    if_options_set(state_as_graph_dot(DotFile),
                  graph_canon:print_cstate_graph(DotFile)),
    if_options_set(transition_diagram_dot(Expr,DotFile),
                  write_transition_diagram_for_expr_to_dotfile(Expr,DotFile)),
    if_options_set(generate_dot_from_invariant(DotFile),
                  generate_dot_from_invariant(DotFile)),
    if_options_set(generate_dot_from_assertions(DotFile),
                  generate_dot_from_assertions(DotFile)),
    if_options_set(generate_dot_from_properties(DotFile),
                  generate_dot_from_properties(DotFile)),
    if_options_set(generate_dot_from_deadlock_po(DotFile),
                  generate_dot_from_deadlock_po(DotFile)),
    if_options_set(evaluate_expression_over_history_to_csv_file(Expr,DotFile),
                  user:evaluate_expression_over_history_to_csv_file(Expr,DotFile)),
    if_options_set(enabling_analysis_csv(EnablingCsvFile),
                   do_enabling_analysis_csv(EnablingCsvFile)),
    if_options_set(expect_evaluation(EEvals),
                   check_if_predicates_were_evaluated(EEvals)),
    ifm_option_set(coverage(Nodes,Operations,ShowEnabledInfo),
                  cli_show_coverage(Nodes,Operations,ShowEnabledInfo,NOW)),
    ifm_option_set(check_statespace_hash(ExpectedHash,Kind),
                  cli_check_statespace_hash(ExpectedHash,Kind)),
    ifm_option_set(variable_coverage,
                  cli_show_variable_coverage(NOW)),
    ifm_option_set(coverage(ShowEnabledInfo),
                  cli_show_coverage(ShowEnabledInfo,NOW)),
    if_option_set(save_state(File),
                  (debug_println(20,'% Saving state space to file'),
                   state_space:tcltk_save_state(File))),
    if_option_set(prob_profile,
                  prob_profile),
    if_option_set(statistics,
                  cli_print_statistics),
    if_option_set(profile_statistics,
                  profile_statistics),
    if_option_set(show_cache,
                  show_cache),
    if_option_set(check_complete, check_complete),
    if_option_set(check_complete_operation_coverage, check_complete_operation_coverage),
    if_option_set(check_scc_for_ltl_formula(LtlFormula,SCC),cli_check_scc_for_ltl_formula(LtlFormula,SCC)).

:- use_module(value_persistance,[show_cache_file_contents/0,show_cache_file_constants/0]).
show_cache :- option(verbose),!,show_cache_file_contents,nl.
show_cache :- show_cache_file_constants,nl.

% new profiler
%:- use_module('../extensions/profiler/profiler.pl').
%cli_print_statistics :- pen,nl,garbage_collect,statistics,nl,state_space:state_space_initialise_with_stats.

% old profiler
:- use_module(hit_profiler).
cli_print_statistics :- profile_statistics,
    nl,garbage_collect,statistics,nl,
    state_space:state_space_initialise_with_stats.

check_complete :-
  (tcltk_find_max_reached_node(Node1) ->
     add_error(check_complete,'Maximum number of transitions reached for at least one state: ',Node1) ; true),
  (not_all_transitions_added(Node2) ->
     add_error(check_complete,'At least one state was not examined: ',Node2) ; true),
  (not_invariant_checked(Node3) ->
     add_error(check_complete,'The invariant was not checked for at least one state: ',Node3) ; true).

check_complete_operation_coverage :-
  (state_space: operation_name_not_yet_covered(OpName) ->
     add_error(check_complete_operation_coverage,'At least one operation is not covered: ', OpName)
  ; true).

show_operation_coverage_summary :- findall(OpName, state_space: operation_name_not_yet_covered(OpName), Uncovered),
   (Uncovered=[] -> print(' All operations have been covered'),nl
     ; format(' The following operations were not covered:~n   ~w~n',[Uncovered])).

% ---------------------

animation_mode_does_not_support_animation(File) :-
    loaded_main_file(smt2,File).
cli_start_animation(NOW) :-
    file_loaded(true,LoadedFile),
    \+ animation_mode_does_not_support_animation(LoadedFile),
    !,
    debug_println(20,'% Starting Animation'),
    writeln_log_time(start_animation(NOW)),
    start_probcli_timer(Timer1),
    start_animation_without_computing,
    stop_probcli_debug_timer(Timer1,'% Finished Starting Animation'),
    if_option_set(add_csp_guide(CspGuide), tcltk_add_csp_file(CspGuide)),

    getAllOperations(Ops),
    debug_print(20,'Operations: '), debug_println(20,Ops),

    (we_need_only_static_assertions(ALL)
      -> debug_println(20,'% Projecting on static ASSERTIONS'),
         print('% PROJECTING ON STATIC ASSERTIONS'),nl,
         b_interpreter:set_projection_on_static_assertions(ALL) ; true),

    (option(load_state(File))
      -> debug_println(20,'% Loading stored state from file'),
         state_space:tcltk_load_state(File)
       ; computeOperations_for_root_required ->
         debug_println(20,'% Searching for valid initial states'),
         start_probcli_timer(Timer2),
         cli_computeOperations(EO),
         stop_probcli_debug_timer(Timer2,'% Finished searching for valid initial states'),
         debug_println(10,EO)
       ; debug_println(20,'% No initialisation required')
    ).
cli_start_animation(_NOW).

start_animation_without_computing :-
    update_preferences_from_spec,
    set_prefs, % override SET_PREF in DEFINITIONS with values from command-line;
    start_animation,
    if_option_set(plugin(Plugin),
            (plugins:initialise_plugins,plugins:start_plugin_for_current_mode(Plugin))),
    get_errors.

% an execution engine with minimal overhead: states are not stored in visited_expression database, only first enabled operation is taken
cli_execute(Steps,ErrorOnDeadlock) :-
    current_expression(ID,CurState),
    debug_println(9,starting_execution_from(ID)),
    start_ms_timer(Start),
    (b_top_level_operation(_) -> temporary_set_preference(filter_unused_constants,true,CHNG) ; true), % we can filter out unused constants, unless there are no operations in which case the user probably wants to see the constant values
    cli_execute_aux(1,Steps,CurState,ErrorOnDeadlock,'$NO_OPERATION'),
    (b_top_level_operation(_) -> reset_temporary_preference(filter_unused_constants,CHNG) ; true),
    stop_ms_timer(Start).

:- use_module(external_functions,[reset_side_effect_occurred/0, side_effect_occurred/1]).
cli_execute_aux(Nr,Steps,CurState,_ErrorOnDeadlock,_LastActionName) :- Nr>Steps,!,
    formatsilent('Stopping execution after ~w steps~n',[Steps]),
    print_state_silent(CurState).
cli_execute_aux(Nr,Steps,CurState,ErrorOnDeadlock,LastActionName) :-
    (\+ cli_invariant_ok(CurState,LastActionName) ->
           N1 is Nr-1,
           format('INVARIANT VIOLATED after ~w steps (after ~w).~n',[N1,LastActionName]),!,
           print_state_silent(CurState)
    % TO DO: the same for assertions
     ; cli_goal_found(CurState) ->
           N1 is Nr-1,
           format('GOAL FOUND after ~w steps (after ~w).~n',[N1,LastActionName]),!,
           print_state_silent(CurState)
    ;      reset_side_effect_occurred,
           specfile_trans(CurState,ActionName,Act,NewState,_TransInfo,Residue),!, % no time-out !
           (Residue=[] -> true ; error_occurred(cli_execute_residue(ActionName,Residue,Act))),
           (option(animate_stats)
             -> translate:translate_event(Act,Str),
                truncate_animate_action(Str,TStr),
                format('~w~5|: ~w~n',[Nr,TStr]) ; true),
           N1 is Nr+1,
           ((NewState=CurState, \+ side_effect_occurred(file))
             -> formatsilent('Infinite loop reached after ~w steps (looping on ~w).~n',[N1,ActionName])
              ; cli_execute_aux(N1,Steps,NewState,ErrorOnDeadlock,ActionName))
    ).
cli_execute_aux(Nr,_Steps,CurState,ErrorOnDeadlock,LastActionName) :- N1 is Nr-1,
    formatsilent('Deadlock reached after ~w steps (after ~w).~n',[N1,LastActionName]), print(ok),nl,
    (ErrorOnDeadlock=true -> error_occurred(deadlock) ; true),
    print_state_silent(CurState).

print_state_silent(CurState) :- (option(silent) -> true ; translate:print_state(CurState),nl).

cli_invariant_ok(_,_) :- option(no_invariant_violations),!. % user asks not to check it
cli_invariant_ok(_,_) :- get_preference(do_invariant_checking,false),!. % user asks not to check it via preference
cli_invariant_ok(S,_) :- \+ state_corresponds_to_initialised_b_machine(S),!. % not yet initialised
cli_invariant_ok(CurState,LastActionName) :-
    (b_specialized_invariant_for_op(LastActionName,Invariant) -> true
       %, print('Specialized invariant: '),translate:print_bexpr(Invariant),nl
      ; b_get_invariant_from_machine(Invariant)),
    b_interpreter:b_test_boolean_expression_for_ground_state(Invariant,[],CurState ).

cli_goal_found(_):- option(no_goal),!,fail.
cli_goal_found(S) :- \+ state_corresponds_to_initialised_b_machine(S),!,fail. % not yet initialised
cli_goal_found(CurState) :-
   b_get_machine_goal(Goal),
   b_interpreter:b_test_boolean_expression_for_ground_state(Goal,[],CurState ).

% random animation
cli_random_animate(Steps,Err) :- probcli_time_stamp(NOW),
    cli_random_animate(NOW,Steps,Err).
cli_random_animate(NOW,Steps,ErrorOnDeadlock) :-
    writeln_log_time(starting_random_animate(NOW,Steps)),
    start_ms_timer(S),
    perform_random_steps(Steps,ErrorOnDeadlock),
    stop_ms_timer(S,Res),
    %tcltk_save_history_as_trace_file(user),
    debug_println(20,finished_random_animate(Steps,Res)),
    printsilent(finished_random_animate(Steps,Res)),nls,
    writeln_log_time(finished_random_animate(NOW,Steps,ErrorOnDeadlock,Res)).

we_need_only_static_assertions(ALL) :- specfile:b_or_z_mode,
  (option(cli_check_assertions(main,_)) -> ALL=main
     ; option(cli_check_assertions(ALL,_))),
   % we do an assertion check
   \+ ((option(A), option_requires_all_properties(A))),
   b_get_assertions(ALL,dynamic,[]). % the assertions do not reference variables

% do we need all properties/constants of the machine, or only certain ones (e.g., static)
option_requires_all_properties(cli_mc(_)).
option_requires_all_properties(cli_check_properties(_)). % we may want to check all properties
option_requires_all_properties(cli_random_animate(_)).
option_requires_all_properties(trace_check).
option_requires_all_properties(state_trace(_)).
option_requires_all_properties(mcm_tests(_,_,_,_)).
option_requires_all_properties(cbc_tests(_,_,_)).
option_requires_all_properties(animate).
option_requires_all_properties(initialise). % INITIALISATION may access constants
option_requires_all_properties(eval_repl).
option_requires_all_properties(eval_string_or_file(_,_,_,_)).
option_requires_all_properties(ltl_assertions).
option_requires_all_properties(ltl_file(_)).
option_requires_all_properties(refinement_check(_,_,_)).
option_requires_all_properties(cli_random_animate(_)).
option_requires_all_properties(cli_start_mc_with_tlc).
option_requires_all_properties(cli_symbolic_model_check).

cli_start_trace_check(MainFile) :-
     debug_println(20,'% Starting Trace Check: '),
     ( check_default_trace_for_specfile(MainFile) -> true
     ; otherwise -> error_occurred(trace_check)).
cli_start_trace_state_check(File) :-
     debug_println(20,'% Starting Trace Check: '),
     ( tcltk_check_state_sequence_from_file(File) -> true
     ; otherwise -> error_occurred(state_trace)).

% is it necessary to compute enabled operations for root state
computeOperations_for_root_required :- initialise_required.
computeOperations_for_root_required :- option(trace_check).
computeOperations_for_root_required :- option(state_trace(_)).
computeOperations_for_root_required :- option(ltl_assertions).
computeOperations_for_root_required :- option(cli_random_animate(_,_)).
computeOperations_for_root_required :- option(socket(_)).
computeOperations_for_root_required :- option(cli_mc(_)).
computeOperations_for_root_required :- option(ltl_file(_)).
computeOperations_for_root_required :- option(ltl_forumla_model_check(_,_)).
computeOperations_for_root_required :- option(ctl_forumla_model_check(_,_)).
computeOperations_for_root_required :- option(refinement_check(_,_,_)).
computeOperations_for_root_required :- option(csp_in_situ_refinement_check(_,_)).
computeOperations_for_root_required :- option(csp_checkAssertion(_,_)).
computeOperations_for_root_required :- option(mcm_tests(_,_,_,_)).
computeOperations_for_root_required :- option(mcm_cover(_)).

% is an initialisation mandatory:
initialise_required :- option(initialise), \+ file_loaded(true,empty_machine).
initialise_required :- \+ option(trace_check), \+ option(state_trace(_)),
                       \+ option(load_state(_)),
                       \+ file_loaded(true,empty_machine), init_req2.
init_req2 :- option(cli_check_properties).
init_req2 :- option(zmq_assertion(_,_)).
init_req2 :- option(cli_check_assertions(_,_)).
init_req2 :- option(eval_string_or_file(_,_,_,_)). % ensure that we initialise/precompile empty machine in case no main file specified; currently no longer required

initialise :- probcli_time_stamp(NOW),cli_start_initialisation(NOW). % for REPL

cli_start_initialisation(NOW) :-
    debug_println(20,'% Performing INITIALISATION: '),
    (perform_random_initialisation -> true ;
      writeln_log_time(cli_start_initialisation_failed(NOW)),
      fail).
cli_check_properties :- probcli_time_stamp(NOW),
    cli_check_properties(NOW).
cli_check_properties(NOW) :-
    print('% Checking PROPERTIES: '),nl,
    writeln_log_time(starting_check_properties(NOW)),
    ( current_state_corresponds_to_setup_constants_b_machine ->
        set_analyse_hook('_P'),
        predicate_evaluator:tcltk_analyse_properties(_PROPRES,PROPInfos),
        unset_analyse_hook,
        print(PROPInfos),nl,   % ex: [total/33,true/29,false/0,unknown/4,timeout/4,runtime/49950]
        (predicate_evaluator:check_summary_all_true(PROPInfos) -> true
          ; print_error('### Not all PROPERTIES true'), error_occurred(check_properties))
    ; otherwise ->
        (tcltk_debug_properties(PROPRES,false,Satisfiable) ->
         print(PROPRES),nl,
         print(Satisfiable),nl
         ; error_occurred(debug_properties_failed))
    ),
    writeln_log_time(finished_check_properties(NOW,PROPInfos)),
    print('% Finished checking PROPERTIES'),nl.
cli_check_assertions(ALL,RI) :-
    probcli_time_stamp(NOW),
    cli_check_assertions(ALL,RI,NOW).
cli_check_assertions(ALL,ReqInfos,NOW) :-
    print('% Checking ASSERTIONS: '),nl,
    writeln_log_time(starting_check_assertions(NOW)),
    set_analyse_hook('_A'), % for dot output, in case users wants to generate dot files for assertions
    predicate_evaluator:tcltk_analyse_assertions(ALL,_ASSRES,Infos),
    unset_analyse_hook,
    print(Infos),nl,
    check_required_infos(ReqInfos,Infos,check_assertions),
    writeln_log_time(finished_check_assertions(NOW,Infos)),
    print('% Finished checking ASSERTIONS'),nl,!.
cli_check_assertions(ALL,ReqInfos,NOW) :-
    add_internal_error('Analyse ASSERTIONS unexpectedly failed',cli_check_assertions(ALL,ReqInfos,NOW)),
    error_occurred(internal_error).
cli_set_goal(GOAL) :-
    debug_println(20,set_goal(GOAL)), %print(set_goal(GOAL)), nl,
    bmachine:b_set_machine_goal(GOAL).
cli_set_searchscope(GOAL) :-
    debug_println(20,set_searchscope(GOAL)), %print(set_goal(GOAL)), nl,
    format('Setting SCOPE for verification: ~w~n (Only states satsifying this predicate will be examined)~n',[GOAL]),
    bmachine:b_set_machine_searchscope(GOAL).
cli_check_goal :-
    print('% Checking GOAL predicate: '),nl,
    tcltk_analyse_goal(List,Summary),
    debug_println(20,List),
    debug_println(20,Summary),
    check_required_infos([false/0,unknown/0],Summary,check_goal).
cli_mc(Nr) :- probcli_time_stamp(NOW), cli_start_model_check(Nr,NOW).
cli_start_model_check(Nr,NOW) :-
    debug_print(20,'% Starting Model Check, Bound: '),
    debug_println(20,Nr),
    writeln_log_time(starting_model_check(NOW,Nr)),
    model_check(Nr,Time,MCRes,NOW),
    nl,
    writeln_log_time(model_check(NOW,Nr,Time,MCRes)).
cli_start_mc_with_tlc :-
    debug_println(20,'% Starting Model Check using TLC.'),
    ((animation_mode(b), \+ animation_minor_mode(eventb)) -> true; error_manager: add_error_and_fail(mc_with_tlc,'TLC4B tool can be used only for classical B models.')),
    on_exception(error(E,_),
                 safe_absolute_file_name(prob_lib('TLC4B.jar'),TLC4BTool),
                 error_manager: add_error_fail(get_tlc_command,'Could not find TLC4B.jar file.',E)),
    construct_and_execute_tlc_command(TLC4BTool).
construct_and_execute_tlc_command(TLC4BTool) :-
    parsercall: get_java_command_path(JavaCmd),
    loaded_main_file(File),
    system_call: system_call(JavaCmd, ['-jar', TLC4BTool, File], Text,JExit),
    print(exit(JExit)),nl,
    (JExit=exit(0)
     ->  true
     ;   error_manager: add_error(construct_and_execute_tlc_command,'Error while model checking with TLC: ',TLC4BTool/File),
         name(T,Text),
         error_manager: add_error_fail(construct_and_execute_tlc_command,'Std error: ',T)
    ).

:- use_module(symbolic_model_checker(symbolic_model_checker)).
cli_symbolic_model_check :-
    debug_println(20,'% Starting Model Check using TLC.'),
    (animation_mode(b)
     -> true
     ;  error_manager:add_error_and_fail(cli_symbolic_model_check,'Symbolic Model Checking is currently only available for B and Event-B.')),
    symbolic_model_checker:symbolic_model_check(Result),
    handle_symbolic_model_check_result(Result).
handle_symbolic_model_check_result(counterexample_found) :- !, error_occurred(invariant_violation).
handle_symbolic_model_check_result(property_holds) :- !.
handle_symbolic_model_check_result(solver_and_provers_too_weak) :- !, error_occurred(model_check_incomplete).

zmq_start_master(invariant,N,Max,PortStart,Workers,Logfile) :-
    start_animation_without_computing,
    zmq_get_initialisation_term(InitTerm),
    (option(strict_raise_error) -> Strict = 1 ; Strict = 0),
    start_master(InitTerm,N,Max,PortStart,Workers,Strict,Logfile),
    halt.
zmq_start_master(assertion,PortStart,Logfile) :-
    current_state_corresponds_to_setup_constants_b_machine,
    animation_mode(b),
    full_b_machine(Machine),
    b_get_assertions(_,static,SAss),
    b_get_assertions(_,dynamic,DAss),
    append(SAss,DAss,Ass),
    count_assertions(Ass,0,N),
    assert(master:assertion_count(N)),
    current_expression(_,State1),
    specfile:state_corresponds_to_set_up_constants(State1,State),
    zmq_get_important_options(Options),
    (option(strict_raise_error) -> Strict = 1 ; Strict = 0),
    start_master(assertions(classical_b(Machine,Options),State,Ass),2,-1,PortStart,0,Strict,Logfile),
    halt.

zmq_get_initialisation_term(Term) :-
    (animation_mode(b) ; animation_mode(csp_and_b)), % CSP file not yet added when ZMQ master starts working
    \+ animation_minor_mode(eventb),
    option(add_csp_guide(CspGuide)),
    !, % Classical B + CSP
    debug_println(20,'ZMQ: Transferring CSP || B model'),
    full_b_machine(Machine),
    % TO DO: extract CSP Term rather than file name: will not work for distribution on other file-systems
    zmq_get_important_options(Options),
    Term = classical_b_with_csp(Machine,CspGuide,Options).
zmq_get_initialisation_term(Term) :-
    debug_println(20,'Generating ZMQ Worker Initialisation'),
    animation_mode(b), \+ animation_minor_mode(eventb), !, % Classical B
    debug_println(20,'ZMQ: Transferring Classical-B model'),
    full_b_machine(Machine),
    zmq_get_important_options(Options),
    Term = classical_b(Machine,Options).
zmq_get_initialisation_term(Term) :-
    animation_mode(b), animation_minor_mode(eventb), !, % Event-B
    debug_println(20,'ZMQ: Transferring Event-B model'),
    full_b_machine(Machine),
    zmq_get_important_options(Options),
    Term = eventb(Machine,Options).
zmq_get_initialisation_term(Term) :-
    animation_mode(cspm),
    loaded_main_file(MainCSPFile),!, % TO DO: pass CSP Prolog term rather than file name (for distribution)
    zmq_get_important_options(Options),
    debug_println(20,'ZMQ: Transferring CSP specification'),
    Term = csp_specification(MainCSPFile,Options).
zmq_get_initialisation_term(_Term) :-
    \+ real_error_occurred, % otherwise error occured while loading
    animation_mode(Mode),
    add_internal_error('Unsupported formalism for ZMQ', Mode),
    fail.

zmq_get_initialisation_term(filename(FN)) :-
    loaded_main_file(FN).

% get important command-line options to be transmitted to probcli worker
zmq_get_important_options(Options) :- findall(O, (option(O), zmq_important_option(O)), Options),
   debug_println(20,transferring_zmq_options_to_workers(Options)).
zmq_important_option(coverage(_)).
zmq_important_option(expect_error(_,_)).
zmq_important_option(file_info).
zmq_important_option(log(_)).
zmq_important_option(print_version).
zmq_important_option(print_short_version).
zmq_important_option(profiling_on).
zmq_important_option(set_card(_,_)).
zmq_important_option(set_pref(_,_)).
zmq_important_option(set_preference_group(_,_)).
zmq_important_option(verbose).
zmq_important_option(statistics).
zmq_important_option(variable_coverage).
zmq_important_option(very_verbose).
% we could consider also supporting: -argv, -cache, -prefs FILE  csp_main(ProcessName) profiling_on prob_profile runtimechecking

% set options received by a zmq worker
zmq_set_important_options(Options) :- debug_println(20,setting_zmq_options(Options)),
     maplist(zmq_set_option,Options).
zmq_set_option(file_info) :- !, file_info.
zmq_set_option(log(F)) :- !,
    generate_time_stamp(Datime,NOW),
    cli_start_logging(F,NOW,Datime,[zmq_worker]).
zmq_set_option(print_version) :- !, print_version.
zmq_set_option(print_short_version) :- !, print_short_version.
zmq_set_option(profiling_on) :- !, profiling_on.
zmq_set_option(set_card(P,V)) :- !, set_eclipse_card(P,V).
zmq_set_option(set_pref(P,V)) :- !, set_pref(P,V).
zmq_set_option(set_preference_group(P,V)) :- !, set_preference_group(P,V).
zmq_set_option(verbose) :- !, verbose.
zmq_set_option(very_verbose) :- !, very_verbose.
zmq_set_option(O) :- zmq_delayed_option(O),!, assert(option(O)). % DO IT LATER
zmq_set_option(O) :- add_internal_error('Unsupported option for ZMQ worker: ',zmq_set_option(O)).

zmq_delayed_option(coverage(_)).
zmq_delayed_option(expect_error(_,_)).
zmq_delayed_option(statistics).
zmq_delayed_option(variable_coverage).

zmq_start_worker(Master, Port, Recover, NOW) :-
    start_worker(Master,Port,Recover,zmq_worker_load_model),
    formatsilent('ZMQ worker finished (Master: ~w, Port:~w)~n',[Master,Port]),
    cli_process_loaded_file_afer_start_animation(NOW),
    printsilent('Exiting probcli worker'),nls,
    halt_prob(NOW).

zmq_worker_load_model(classical_b(Machine,Options)) :-
    debug_println(20,'ZMQ WORKER: Loading classical B model'),
    zmq_set_important_options(Options),
    bmachine:(b_machine_reset, assert_machine(Machine)),
    set_animation_mode(b),
    prob2_interface:start_animation, cli_computeOperations(_)
    .
zmq_worker_load_model(classical_b_with_csp(Machine,CspGuide,Options)) :-
    debug_println(20,'ZMQ WORKER: Loading CSP || B model'),
    zmq_set_important_options(Options),
    bmachine:(b_machine_reset, assert_machine(Machine)),
    set_animation_mode(b),
    prob2_interface:start_animation,
    tcltk_add_csp_file(CspGuide), % TO DO: use CSP Prolog term rather than filename <----------------
    prob2_interface:start_animation,
    cli_computeOperations(_)
    .
zmq_worker_load_model(eventb(Machine,Options)) :-
    zmq_set_important_options(Options),
    bmachine:(b_machine_reset, assert_machine(Machine)),
    set_animation_mode(b), set_animation_minor_mode(eventb),
    prob2_interface:start_animation, cli_computeOperations(_).
zmq_worker_load_model(csp_specification(CSPFile,Options)) :-
    zmq_set_important_options(Options),
    load_cspm_spec_from_cspm_file(CSPFile), % TO DO: pass CSP Prolog term rather than filename
    prob2_interface:start_animation, cli_computeOperations(_).
zmq_worker_load_model(filename(FN)) :-
    printsilent('loading file by filename\n'),flush_output,
    ( is_eventb_b(FN) ->
        eclipse_interface:load_eventb_file(FN)
    ; otherwise ->
        bmachine:b_load_machine_probfile(FN)),
    prob2_interface:start_animation.
zmq_worker_load_model(assertions(Machine,State,Assertions)) :-
   assert(assertion_counter(-1)),
   printsilent(loaded_model_for_assertion_checking),nl,
   zmq_worker_load_model(Machine),
   assert(worker:assertion_state(State)),
   make_assertionwps(Assertions).
   % assert current state
zmq_worker_load_model(Other) :-
    add_internal_error('ZMQ worker: Unexpected machine description', zmq_worker_load_model(Other)),
    fail.


:-dynamic assertion_counter/1.

count_assertions([],A,A).
count_assertions([H|T],A,R) :- bsyntaxtree:conjunction_to_list(H,HL),
                               length(HL,N1),
                               NN is A + N1,
                               count_assertions(T,NN,R).

make_assertionwps([]).
make_assertionwps([H|T]) :- bsyntaxtree:conjunction_to_list(H,HL),
                            sort_assertions(HL,SL),
                            append_assertion(SL),
                            make_assertionwps(T).

append_assertion([]).
append_assertion([H|T]) :- assertion_counter(N),
                           retractall(assertion_counter(_)),
                           N1 is N + 1,
                           assert(assertion_counter(N1)),
                           assert(worker:assertion_task(N1,H)),
                           append_assertion(T).

%assertions_order(A,B) :- term_size(A,NA),term_size(B,NB), NA > NB.
sort_assertions(X,X).
% :- samsort(assertions_order,X,Y).



is_eventb_b(FN) :- append(_,FN,".eventb").
 % load_model(Initialisation)


load_model(classical_b(M)) :- !,eclipse_interface:load_classical_b(M).
load_model(load_event_b_project(A,B,C,D)) :- !, prob2_interface:load_event_b_project(A,B,C,D), start_animation.
load_model(X) :- print('*** Unknown model: '), print(X),nl.


:- use_module(predicate_evaluator).
:- use_module(bmachine,[b_machine_name/1]).
set_analyse_hook(AddPrefix) :- % set a hook to write false/unknown expressions into a dot file
   reset_dot_file_number,
   if_options_set(dot_analyse_output_prefix(_Path),
                  (set_dot_file_prefix_if_option_set(AddPrefix),
                   register_conjunct_error_hook(user:pred_eval_hook))).
unset_analyse_hook :- predicate_evaluator:reset_conjunct_error_hook.

set_dot_file_prefix_if_option_set(AddPrefix) :-
   if_options_set(dot_analyse_output_prefix(Path),
                  (loaded_main_file(MainFile),
                   get_modulename_filename(MainFile,RootName),
                   % we could also use b_machine_hierarchy:main_machine_name(RootName)
                   string_concatenate(Path,RootName,P1),
                   string_concatenate(P1,AddPrefix,FullPath),
                   set_dot_file_prefix(FullPath),
                   debug_println(9,dot_file_prefix(FullPath)))).

% Status: true, false, unknown
pred_eval_hook(_Conjunct,true,_EnumWarning,_IsExpanded, _CS) :-
  \+ option(dot_generate_for_all_formulas),!. % don't generate .dot for true formulas, unless explicitly requested
pred_eval_hook(Conjunct,Status,_EnumWarning,_IsExpanded, CS) :-
   printsilent('Generating dotfile for: '),printsilent(CS),nls,
   (write_dot_graph_to_new_file(Status,Conjunct) -> true
     ; add_error(dot_output,'Writing dot to file failed: ',CS)).


:- dynamic dot_file_prefix/1.
:- dynamic dot_file_number/1.

dot_file_prefix('~/Desktop/dot').
set_dot_file_prefix(F) :- retractall(dot_file_prefix(_)), assert(dot_file_prefix(F)).
dot_file_number(0).
reset_dot_file_number :- retractall(dot_file_number(_)), assert(dot_file_number(0)).
get_next_nr(GNr) :- retract(dot_file_number(Nr)), N1 is Nr+1,
  assert(dot_file_number(N1)), GNr = Nr.
write_dot_graph_to_new_file(Status,BExpr) :-
    dot_file_prefix(Dir),get_next_nr(Nr),
    string_concatenate('_',Status,Str1),
    string_concatenate(Nr,Str1,NS),
    string_concatenate(Dir,NS,F1),
    atom_concat(F1,'.dot',FileName),
    user:write_dot_file_for_pred(BExpr,FileName).

% get dot file name if dot_output has been set
get_dot_file(Type,FileName) :- option(dot_analyse_output_prefix(_)),
    set_dot_file_prefix_if_option_set(Type),
    dot_file_prefix(Dir),
    string_concatenate('_',Type,Str1),
    string_concatenate(Dir,Str1,F1),
    atom_concat(F1,'.dot',FileName).

:- use_module(refinement_checker).
cli_csp_in_situ_refinement_check(P,Type,Q,NOW) :-
    debug_println(20,'% Starting CSP Refinement Check'),
    writeln_log_time(starting_csp_refinement_check(NOW)),
    loaded_main_file(CSPFile),
    ajoin_with_sep(['assert',P,Type,Q], ' ',Assertion),
    ( timeout_call(user:tcltk_check_csp_assertion(Assertion,CSPFile,'False',RefTrace),NOW)
      -> check_ref_result(RefTrace)
      ; otherwise -> true),
    writeln_log_time(finished_csp_refinement_check(Assertion,NOW)).
cli_start_refinement_check(RefFile,PerformSingleFailures,RefNrNodes,NOW) :-
    debug_println(20,'% Starting Refinement Check'),
    writeln_log_time(starting_refinement_check(NOW)),
    refinement_checker:tcltk_load_refine_spec_file(RefFile),
    ( timeout_call(tcltk_refinement_search(RefTrace,PerformSingleFailures,RefNrNodes),NOW)
      -> check_ref_result(RefTrace)
      ; otherwise -> true),
    writeln_log_time(finished_refinement_check(NOW)).
check_ref_result(RefTrace) :-
   ( RefTrace==no_counter_example ->
            print('==> Refinement Check Successful'),nl
        ; otherwise ->
            print('*** Refinement Check Counter-Example: ***'),nl, print(RefTrace),nl,
            print('*** Refinement Check Failed ***'),nl,
            error_occurred(refinement_check_fails)).
cli_checkAssertion(Proc,Model,AssertionType,NOW) :-
    debug_println(20,'% Starting CSP Deadlock free Check'),
    writeln_log_time(starting_csp_assertion_check(Assertion,NOW)),
    loaded_main_file(CSPFile),
    ajoin(['assert ',Proc,' :[ ',AssertionType,'[',Model,']',' ]'],Assertion),
    ( timeout_call(user:tcltk_check_csp_assertion(Assertion,CSPFile,'False',ResTrace),NOW)
      -> check_model_result(ResTrace)
      ; otherwise -> true),
    writeln_log_time(finished_csp_assertion_check(Assertion,NOW)).
check_model_result(ResTrace) :-
    ( ResTrace==no_counter_example ->
             print('==> Model Check Successful'),nl
         ;otherwise ->

             print('*** Model Check Counter-Example: ***'),nl,print(ResTrace),nl,
             print('*** Model Check Failed ***'),nl,
             error_occurred(model_check_fails)).
cli_csp_get_assertions :-
    loaded_main_file(CSPFile),
    get_csp_assertions_as_string(CSPFile,String),
    print('*** Assertions in File (separated by $) ***'),nl,print(String),nl.
cli_set_csp_event_as_visible :-
    set_preference(csp_event_visible_to_user,true).
cli_cspm_evaluate_expression(E) :-
    (loaded_main_file(CSPFile) ->
	      haskell_csp: cspm_eval_expression_from_csp_file(CSPFile, '--src=', E, _ResText, _ErrText)
	    ; haskell_csp: cspm_eval_expression(E,_ResText,_ErrText)
    ).
cli_eval_csp_expression(E) :-
    (loaded_main_file(CSPFile) ->
	      haskell_csp: evaluate_csp_expression(E, CSPFile, Res)
	    ; haskell_csp: evaluate_csp_expression(E,Res)
    ), print('Evaluated Expression: '),nl,print(Res),nl.
cli_csp_add_unicode(File) :-
    loaded_main_file(CSPFile),
    cspm_file_add_remove_unicode(CSPFile, File,'--addUnicode=').
cli_csp_remove_unicode(File) :-
    loaded_main_file(CSPFile),
    cspm_file_add_remove_unicode(CSPFile, File,'--removeUnicode=').
cli_csp_translate_to_file(PlFile) :-
    loaded_main_file(CSPFile),parse_and_load_cspm_file_into_specific_pl_file(CSPFile,PlFile).
cli_check_scc_for_ltl_formula(LtlFormula,SCC) :-
    ltl_fairness: check_scc_ce(LtlFormula,SCC).
cli_get_min_max_coverage(Filename) :-
    coverage_statistics:tcltk_compute_min_max(list(R)),
    tools:filter(tools: is_a_comment('/*','*/'),R,_Comments,Res),
    tools:write_lines_to_file(Filename,Res).
%    pretty_print_min_max_coverage_to_file(FileName).
:- use_module(coverage_statistics,[pretty_print_coverage_information_to_file/1]).
cli_get_coverage_information(FileName) :-
    pretty_print_coverage_information_to_file(FileName).
cli_vacuity_check :-
    eclipse_interface:get_vacuous_invariants(L),
    (L=[] -> print('No vacuous invariants'),nl
     ; maplist(add_vacuous_invariant,L)).
add_vacuous_invariant(Inv) :-
    translate:translate_bexpression(Inv,TI),
    add_error(vacuity_check,'Vacuous invariant: ',TI).
cli_start_socketserver(Port) :-
    printsilent('Starting Socket Server'),nls,
    safe_absolute_file_name(prob_home('.'),AppDir),
    printsilent('Application Path: '),printsilent(AppDir),nls,
    disable_interaction_on_errors,
    ( start(Port) -> true
    ; otherwise ->
        print('Starting socket server failed, Port: '), print(Port),nl),
    printsilent('Finished Socket Server'),nls.
cli_check_statespace_hash(Expected,Kind) :-
    print('Computing hash of entire statespace: '),
    state_space:compute_full_state_space_hash(Hash),
    print(Hash),nl, % TO DO: maybe also compute hash for transitions and check that
    (Hash=Expected -> true
      ; Kind=='64bit', \+ platform_is_64_bit -> format('Hash does not match ~w (but was computed on 64-bit system)~n',[Expected])
      ; Kind=='32bit', platform_is_64_bit -> format('Hash does not match ~w (but was computed on 32-bit system)~n',[Expected])
      ; add_error(hash,'Expected Statespace Hash to be: ',Expected)).
cli_show_coverage(ShowEnabledInfo,NOW) :-
    cli_show_coverage(_Nodes,_Operations,ShowEnabledInfo,NOW).
cli_show_coverage(Nodes,Operations,ShowEnabledInfo,NOW) :-
    ShowEnabledInfo==just_check_stats,!, % no printing of individual transition coverage
    state_space:get_state_space_stats(TotalNodeSum,TotalTransSum,_ProcessedTotal), % no computation overhead
    writeln_log(computed_coverage(NOW,TotalNodeSum,TotalTransSum)),
    check_totals(Nodes,Operations,TotalNodeSum,TotalTransSum).
cli_show_coverage(Nodes,Operations,ShowEnabledInfo,NOW) :-
    ShowEnabledInfo== just_summary,
    !, % no printing of detailed transition coverage (avoid traversing state space)
    state_space:get_state_space_stats(TotalNodeSum,TotalTransSum,_ProcessedTotal), % no computation overhead
    writeln_log(computed_coverage(NOW,TotalNodeSum,TotalTransSum)),
    format('Coverage:~n States: ~w~n Transitions: ~w~n',[TotalNodeSum,TotalTransSum]),
    show_operation_coverage_summary,
    check_totals(Nodes,Operations,TotalNodeSum,TotalTransSum).
cli_show_coverage(Nodes,Operations,ShowEnabledInfo,NOW) :-
    print('Coverage:'),nl,
    compute_the_coverage(Res,TotalNodeSum,TotalTransSum,ShowEnabledInfo),
    writeln_log(computed_coverage(NOW,TotalNodeSum,TotalTransSum)),
    print(Res),nl,
    check_totals(Nodes,Operations,TotalNodeSum,TotalTransSum).
check_totals(Nodes,Operations,TotalNodeSum,TotalTransSum) :-
    ( Nodes=TotalNodeSum -> true
    ; otherwise ->
        add_error(probcli,'Unexpected number of nodes: ',TotalNodeSum),
        add_error(probcli,'Expected: ',Nodes),error_occurred(coverage)),
    ( Operations=TotalTransSum -> true
    ; otherwise ->
        add_error(probcli,'Unexpected number of transitions: ',TotalTransSum),
        add_error(probcli,'Expected: ',Operations),error_occurred(coverage)).
cli_show_variable_coverage(NOW) :-
    print('Variable (min/max) coverage: '),nl,
    coverage_statistics:tcltk_compute_min_max(list(R)),
    print_list(R),nl,
    writeln_log(computed_variable_coverage(NOW,R)).

cli_print_machine_statistics :-
    b_machine_name(Name),
    %(b_get_main_filename(File) -> true ; File=unknown),
    format('Machine statistics for ~w~n',[Name]),
    findall(Key/Nr,b_machine_statistics(Key,Nr),L),
    maplist(print_keynr,L),!.
cli_print_machine_statistics :- add_error(machine_stats,'Could not obtain machine statistics').
print_keynr(Key/Nr) :- format(' ~w : ~w~n',[Key,Nr]).

cli_print_junit_results(ArgV) :-
    junit_mode(S),!,
    statistics(runtime,[E,_]),
    T is E - S,
    create_junit_result(ArgV,'Integration Tests','Integration_Test',T,pass,Result),
    print_junit([Result],'Integration_Test').
cli_print_junit_results(_).

cli_print_history(HistFile) :-
    findall( O, option(history_option(O)), Options),
    debug_println(9,writing_history_to_file(HistFile)),
    (Options=[trace_file] -> tcltk_save_history_as_trace_file(HistFile)  % save as Prolog trace file for replay with -t
     ; write_history_to_file(HistFile,Options) -> true
     ; add_error(history,'Writing history to file failed: ',HistFile)).

cli_print_values(ValuesFilename) :-
    (write_values_to_file(ValuesFilename) -> true ; add_error(sptxt,'Writing values to file failed: ',ValuesFilename)).

:- use_module(library(terms),[term_hash/2]).
:- dynamic probcli_time_stamp/1.
generate_time_stamp(NOW,TS) :- retractall(probcli_time_stamp(_)),
  now(NOW),
  prolog_flag(argv,ArgV),term_hash(ArgV,Hash),
  Rnd is Hash mod 1000,
  % random(0,1000,Rnd),  always returns 216 % TO DO: try to get milliseconds from some library function
  TS is (NOW*1000)+Rnd,
  assert(probcli_time_stamp(TS)).
update_time_stamp(NOW1) :- retractall(probcli_time_stamp(_)),
  assert(probcli_time_stamp(NOW1)).

get_errors :- \+ real_error_occurred,!,
              (get_error(_Source,_Msg) -> print('*** Warnings occurred'),nl ; true),
              reset_errors.
get_errors :- get_error_sources.

get_error_sources :- get_error(ErrSource,_), !, error_occurred(ErrSource),
      findall(1,get_error(ErrSource,_),L), length(L,Nr),
      (Nr>0 -> N1 is Nr+1, print_error('*** Occurences of this error: '),print_error(N1) ; true),
      get_error_sources.
get_error_sources.

get_state_space_errors :- option(strict_raise_error),!,
              (invariant_violated(ID) -> print(invariant_violation(ID)),nl,
                                         error_occurred(invariant_violation) ; true),
              (state_error(_,_,abort_error(TYPE,_,_,_)) -> error_occurred(TYPE) ; true),
              get_state_errors.
get_state_space_errors.

get_state_errors :- state_error(_,_,X), X\=invariant_violated, X\=abort_error(_,_,_,_),
  create_state_error_description(X,Msg),error_occurred(Msg),fail.
get_state_errors.

create_state_error_description(eventerror(Event,Error,_),Description) :- !,
    functor(Error,Functor,_),
    ajoin(['event_error:',Event,':',Functor],Description).
create_state_error_description(StateError,Description) :-
    functor(StateError,Functor,_),
    atom_concat('state_error:',Functor,Description).

check_loaded(Action) :-
    ( file_loaded(true)  -> true
    ; file_loaded(error) -> fail /* we have already generated error message */
    ; otherwise ->
        add_error(probcli,'No file specified; cannot perform action: ',Action),
        error_occurred(loading),fail).

:- dynamic loaded_main_file/2.
loaded_main_file(File) :- loaded_main_file(_Ext,File).

load_main_file(MainFile,NOW,Already_FullyProcessed) :- retractall(loaded_main_file(_,_)),
               debug_print(20,'% Loading: '), debug_println(20,MainFile),
               writeln_log_time(loading(NOW,MainFile)),
               get_filename_extension(MainFile,Ext),
               debug_println(6,file_extension(Ext)),
               start_probcli_timer(Timer),
               load_spec_file(Ext,MainFile,Already_FullyProcessed),
               stop_probcli_debug_timer(Timer,'% Finished loading'),
               (Already_FullyProcessed==true -> true
                ; assert(loaded_main_file(Ext,MainFile))).
%load_spec_file('pl',MainFile) :- !, load_cspm_spec_from_pl_file(MainFile). % no longer needed ?
load_spec_file('csp',MainFile) :- !, load_cspm_spec_from_cspm_file(MainFile).
load_spec_file('cspm',MainFile) :- !, load_cspm_spec_from_cspm_file(MainFile).
load_spec_file('P',MainFile) :- !, load_xtl_file(MainFile).
load_spec_file('p',MainFile) :- !, load_xtl_file(MainFile). % sometimes windows is confused about the upper case letter....
load_spec_file('eventb',MainFile) :- !, load_eventb_file(MainFile).
load_spec_file('v',MainFile) :- !,
    print('---------------------------------'),nl,
    print('Loading: '),print(MainFile),nl,
    load_b_file(MainFile). % Siemens Rule File; maybe in future use -eval_rule_file
load_spec_file('prob',MainFile) :- !,load_prob_file(MainFile).
load_spec_file('mch',MainFile) :- !,load_b_file(MainFile).
load_spec_file('sys',MainFile) :- !,load_b_file(MainFile).
load_spec_file('ref',MainFile) :- !,load_b_file(MainFile).
load_spec_file('imp',MainFile) :- !,load_b_file(MainFile).
load_spec_file('fuzz',MainFile) :- !,tcltk_open_z_file(MainFile).
load_spec_file('tla',MainFile) :- !, load_tla_file(MainFile).
load_spec_file('pml',MainFile) :- !,parsercall:call_promela_parser(MainFile),
   parsercall:promela_prolog_filename(MainFile,PrologFile),
   printsilent(consulting(PrologFile)),nls,
   tcltk_open_promela_file(PrologFile).
load_spec_file(EXT,MainFile) :- print_error('### Unknown file extension, assuming B machine:'),
                                 print_error(EXT),
                                 load_b_file(MainFile).

load_spec_file('pl',MainFile, Already_FullyProcessed) :- !, Already_FullyProcessed=true,
   printsilent('Processing PO file: '),printsilent(MainFile),nls,
   load_po_file(MainFile),
   printsilent('Running ProB Disprover'), nls,
   run_disprover_on_all_pos(Summary),
   print_disprover_stats,
   (option(cli_check_disprover_result(Infos)) -> check_required_infos(Infos,Summary,load_po_file)
    ; option(strict_raise_error) -> check_required_infos([false-0,unknown-0,failure-0],Summary,load_po_file)
    % TO DO: provide way for user to specify expected info
     ; true),
     clear_loaded_machines.
load_spec_file('smt2',MainFile,Already_FullyProcessed) :- !, Already_FullyProcessed=true,
   printsilent('Processing SMT file: '),printsilent(MainFile),nls,
   smtlib2_file(MainFile).
load_spec_file(EXT,F,false) :- load_spec_file(EXT,F).


% do not perform -execute_all if no parameters provided
do_not_execute_automatically('pl').
do_not_execute_automatically('smt2').

test_kodkod_and_exit(MaxResiduePreds,NOW) :-
    start_animation_without_computing,
    test_kodkod(MaxResiduePreds),
    halt_prob(NOW).

compare_kodkod_performance1(KPFile,Iterations,NOW) :-
    start_animation_without_computing,
    compare_kodkod_performance(KPFile,Iterations),
    halt_prob(NOW).

:- use_module(version).

print_short_version :-
    version(V1,V2,V3,Suffix),revision(Rev),
    format('VERSION ~p.~p.~p-~p (~p)~N',[V1,V2,V3,Suffix,Rev]).

:- use_module(parsercall,[get_parser_version/1]).
print_version :-
    print('ProB Command Line Interface'),nl,
    version(V1,V2,V3,Suffix),
    revision(Rev),  lastchangeddate(LCD), prolog_flag(version,PV),
    ( option_verbose ->
        safe_absolute_file_name(prob_home('.'),AppDir),
        get_parser_version(ParserVersion),
        format('  VERSION ~p.~p.~p-~p (~p)~N  ~p~N  Prolog: ~p~N  Application Path: ~p~N  Java Parser: ~p~N',
           [V1,V2,V3,Suffix,Rev,LCD,PV,AppDir,ParserVersion])
     ;
        format('  VERSION ~p.~p.~p-~p (~p)~N  ~p~N  Prolog: ~p~N~N',
           [V1,V2,V3,Suffix,Rev,LCD,PV])
    ), print_compile_time_flags.

print_compile_time_flags :-
   findall(Flag, compile_time_flag(Flag), Flags),
   (Flags=[] -> true ; format('  COMPILE TIME FLAGS: ~w~N',[Flags])).

:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(prob_safe_mode,true)).
compile_time_flag(prob_safe_mode).
:- endif.
:- if(environ(prob_data_validation_mode,true)).
compile_time_flag(prob_data_validation_mode).
:- endif.
:- if(environ(partially_evaluate,true)).
compile_time_flag(partially_evaluate).
:- endif.
:- if(environ(partially_evaluate_compile,true)).
compile_time_flag(partially_evaluate_compile).
:- endif.
:- if(environ(no_wd_checking,true)).
compile_time_flag(no_wd_checking).
:- endif.
:- if(environ(use_pe_cache,true)).
compile_time_flag(use_pe_cache).
:- endif.
:- if(environ('SP_JIT',disabled)).
compile_time_flag('SP_JIT'(disabled)).
:- endif.
:- if(environ('SP_JIT',yes)).
compile_time_flag('SP_JIT'(yes)). % this is the default
:- endif.
:- if(environ('SP_JIT',no)).
compile_time_flag('SP_JIT'(no)).
:- endif.
:- if(environ('SP_JIT_COUNTER_LIMIT',_)).
compile_time_flag('SP_JIT_COUNTER_LIMIT_SET'). % default is 0
:- endif.
:- if(environ('SP_JIT_CLAUSE_LIMIT',_)).
compile_time_flag('SP_JIT_CLAUSE_LIMIT_SET'). % default is 1024
:- endif.
:- if(environ('SP_SPTI_PATH',verbose)).
compile_time_flag('SP_SPTI_PATH'(verbose)). % can be set to verbose
:- endif.

compile_time_flag(_) :- fail.

print_help :- print_version,
          print('Usage: probcli FILE [OPTIONS]'),nl,
          print('  OPTIONS are: '),nl,
          print('     -mc Nr       model check; checking at most Nr states'),nl,
          print('     -model_check model check without limit on states explored'),nl,
          print('     -noXXX       XXX=dead,inv,goal,ass (for model check)'),nl,
          print('     -bf          proceed breadth-first'),nl,
          print('     -df          proceed depth-first'),nl,
          print('     -timeout N   total timeout in ms for model and refinement checking'),nl,
          print('     -t           trace check (associated .trace file must exist)'),nl,
          print('     -init        initialise specification'),nl,
          print('     -cbc OPNAME  constraint-based invariant checking for an operation'),nl,
          print('                  (you can also use OPNAME=all)'),nl,
          print('     -cbc_deadlock  constraint-based deadlock checking'),nl,
          print('                  (you can also use -cbc_deadlock_pred PRED)'),nl,
          print('     -cbc_assertions  constraint-based static assertion checking'),nl,
          print('     -cbc_refinement  constraint-based static refinement checking'),nl,
          print('     -cbc_sequence S  constraint-based search for sequence of operations'),nl,
          print('     -strict      raise error if model-checking finds counter example'),nl,
          print('                  or trace checking fails or any error state found'),nl,
          print('     -expcterr ERR  expect error to occur (ERR=cbc,mc,ltl,...)'),nl,
          print('     -animate Nr  random animation (max. Nr steps)'),nl,
          print('     -animate_all random animation until a deadlock is reached'),nl,
          print('     -animate_stats  provide feedback which operations are executed'),nl,
          print('     -his File    write history to File'),nl,
          print('     -his_option O  additional option when writing a history (show_init or show_states)'),nl,
          print('     -sptxt File  save constants and variable values to File'),nl,
          print('     -cache Directory  automatically save constants to files and avoid recomputation'),nl,
          print('     -det_check   check if animation steps are deterministic'),nl,
          print('     -det_constants  only check if SETUP_CONSTANTS step is deterministic'),nl,
          print('     -i           interactive animation. Only for interactive sessions,'),nl,
          print('                  the output can arbitrarily change in fututre versions. '),nl,
          print('                  Do not build automatic tools using the interactive mode'),nl,
          print('     -repl        start interactive read-eval-loop'),nl,
          print('     -eval "E"    evaluate expression or predicate'),nl,
          print('     -c           print coverage statistics'),nl,
          print('     -cc Nr Nr    print and check coverage statistics'),nl,
          print('     -vacuity_check look for vacuous implications in invariant'),nl,
          print('     -statistics  print memory and other statistics at the end'),nl,
          print('     -p PREF Val  set Preference to Value'),nl,
          print('     -prefs FILE  set Preferences from Prolog File'),nl,
          print('     -card GS Val set scope of B deferred set'),nl,
          print('     -goal "PRED" set GOAL predicate for model checker'),nl,
          print('     -check_goal  check GOAL (after -mc, -t, or -animate)'),nl,
          print('     -scope "PRED"set scope predicate for model checker'),nl,
          print('     -s Port      start socket server on given port'),nl,
          print('     -ss          start socket server on port 9000'),nl,
          print('     -sf          start socket server on some free port'),nl,
          print('     -l LogFile   log activities in LogFile'),nl,
          print('     -ll          log activities in /tmp/prob_cli_debug.log'),nl,
          print('     -lg LogFile  analyse logfile using gnuplot'),nl,
          print('     -pp FILE     pretty-print internal representation to file'), nl,
          print('     -ppf FILE    like -pp, but force printing of all type infos'),nl,
          print('     -v           verbose'),nl,
          print('     -mc_with_tlc model check using the model checker TLC'),nl,
 %%         print('     -vv          very verbose'),nl,  % no effect when debugging calls disabled
          print('     -version     print version information (-svers for short info)'),nl,
          print('     -assertions  check ASSERTIONS'),nl,
          print('     -main_assertions  check ASSERTIONS from main file only'),nl,
          print('     -properties  check PROPERTIES'),nl,
          print('     -cache Dir   use directory "Dir" to cache constants and variables'),nl,
          print('     -ltlfile F   check LTL formulas in file F'),nl,
          print('     -ltlassertions  check LTL assertions (in DEFINITIONS)'),nl,
          print('     -ltllimit L  explore at most L states when model-checking LTL or CTL'),nl,
          print('     -ltlformula \"F\" check the LTL formula F'),nl,
          print('     -ctlformula \"F\" check the CTL formula F'),nl,
          print('     -save File   save state space for later refinement check'),nl,
          print('     -refchk File refinement check against previous saved state space'),nl,
          print('     -mcm_tests Depth MaxStates EndPredicate File'),nl,
          print('                  generate test cases with maximum length Depth, explore'),nl,
          print('                  maximally MaxStates, the last state satisfies EndPredicate'),nl,
          print('                  and the test cases are written to File'),nl,
          print('     -mcm_cover Operation'),nl,
          print('                  when generating MCM test cases, Operation should be covered'),nl,
          print('     -cbc_tests Depth EndPredicate File'),nl,
          print('                  generate test cases by constraint solving with maximum'),nl,
          print('                  length Depth, the last state satisfies EndPredicate'),nl,
          print('                  and the test cases are written to File'),nl,
          print('     -cbc_cover Operation'),nl,
          print('                  when generating CBC test cases, Operation should be covered'),nl,
%          print('     -cbc_cover_all   try and cover all operations'),nl, % is now default if no cbc_cover provided
          print('     -test_description File'),nl,
          print('                      read information for test generation from File'),nl,
          print('     -spdot File      write graph of the state space to a dot file.'),nl,
          print('     -spmdot File     write graph signature merge of the state space to a dot file.'),nl,
          print('     -spddot File     write graph of the state space as a DFA to a dot file.'),nl,
          print('     -sptdot E File   write transition diagram for expression E to a dot file.'),nl,
          print('     -dot_output      Path generate dot files for false assertions/properties'),nl,
          print('     -dot_all         also generate dot files for true assertions/properties'),nl,
          print('     -csvhist E File  evaluate expression over history and generate CSV file.'),nl,
          print('     -load_state File load state from a saved state (generated by PrB Tcl/Tk or -save_state)'),nl,
       % For Eclipse Version only
       %%  print('     -parsercp  CP class path of the B Parser, this has to be a valid Java class path'),nl,
       %%   print('     -cspm         load CSP-M .csp file rather than B Machine .mch/.ref/.imp File'),nl,
       %%   print('     -csp         load CSP-M .pl file rather than B Machine File'),nl,
          print('     -cspref Spec [m= Impl File'),nl,
          print('                  checks a refinement statement,'),nl,
          print('                  where Spec and Impl are processes from File, and \'m\' the type of the refinement:'),nl,
          print('                  \'T\' for traces, \'F\' for failures, or \'FD\' for failures-divergences.'),nl,
          print('     -cspdeadlock P m File'),nl,
          print('                  checks a process for deadlock,'),nl,
          print('                  where \'P\' is a process from File, and \'m\' the type of the model:'),nl,
          print('                  \'F\' for failures and \'FD\' for failures-divergences.'),nl,
          print('     -cspdeterministic P m File'),nl,
          print('                  checks a process for determinism,'),nl,
          print('                  where \'P\' is a process from File, and \'m\' the type of the model:'),nl,
          print('                  \'F\' for failures and \'FD\' for failures-divergences.'),nl,
          print('     -csplivelock P File'),nl,
          print('                  checks a process for divergence,'),nl,
          print('                  where \'P\' is a process from File.'),nl,
	        print('     -csp_eval "E" evaluate CSP-M expression.'),nl,
          print('     -csp-guide File   CSP||B: Use the CSP File to control the B machine'),nl,
          print('     '),nl,
          ( \+ option_verbose -> true
           ;
	      print('     -cspm_eval "E" evaluate CSP-M expression with the CSPM tool.'),nl,
	      print('     -test_mode       set random seed to the Prolog\'s current random state'),nl,
          print('     -expcteval Evals expect that the given predicates have been evaluated or not (see external predicate IS_EVALUATED)'),nl,
          print('     -rc          runtime checking of types/pre-/post-conditions'),nl,
          print('     -state_trace File      read a file of B predicates (one per line) and try find a matching trace.'),nl

          ),
          print(' FILE extensions are: '),nl,
          print('      .mch         for B abstract machines'),nl,
          print('      .ref         for B refinement machines'),nl,
          print('      .imp         for B implementation machines'),nl,
          print('      .csp, .cspm  for CSP-M files, same format as FDR'),nl,
          print('      .eventb      for Event-B packages exported from Rodin ProB Plugin'),nl,
          print('      .tla         for TLA+ models'),nl,
          ( option_verbose ->
              print(' Preferences PREF are: '),nl,
              print_eclipse_prefs
          ; otherwise ->
              print(' Use --help -v to print avalaible preferences PREF'),nl
          ),
          print(' More info at: http://www.stups.uni-duesseldorf.de/ProB/'),nl,
          print('       and at: http://www.prob2.de/probcli/'),nl,
          % http://www.stups.uni-duesseldorf.de/ProB/index.php5/Using_the_Command-Line_Version_of_ProB
          nl.


set_argv(V) :-
   debug_println(20,set_argv(V)),
   external_functions:set_argv_from_atom(V).

set_prefs :-
   if_option_set(socket(_), % then we may need the event(.) transition_info for the Java API
       preferences:set_preference(store_event_transinfo,true)),
   option(set_prefs_from_file(File)),
   debug_println(20,load_preferences(File)),
   preferences:load_preferences(File),
   fail.
set_prefs :-
   option(set_preference_group(P,V)),
   debug_println(20,set_preference_group(P,V)),
   set_preference_group(P,V),
   fail.
% eclipse preference or 'normal preference'
set_prefs :-
   option(set_pref(P,V)),
   set_pref(P,V),
   fail.
set_prefs :- option(set_card(P,V)),
   debug_println(20,set_card(P,V)),
   set_eclipse_card(P,V),
   fail.
set_prefs :-
    ( option(breadth_first) -> user:set_depth_breadth_first_mode(breadth_first)
    ; option(depth_first)   -> user:set_depth_breadth_first_mode(depth_first)
    ; otherwise -> true
    ).
set_pref(P,V) :-
   debug_println(20,set_pref(P,V)),
   ( eclipse_preference(P,_)
     -> set_eclipse_preference(P,V)
    ; obsolete_eclipse_preference(P) -> add_warning(probcli,'Obsolete preference: ',P)
     ; % might be a term if its a plugin preference
     atom_codes(P,Codes),
     append(Codes,".",Codes2), % to make term readable by read_from_codes
     read_from_codes(Codes2,Preference),
     ((nonvar(Preference),preference_val_type(Preference,_))
        -> prob2_interface: convert_cli_arg(V,Value),
           set_preference(Preference,Value)
        ;  add_error(probcli,'Illegal preference: ',P)
     )
   ).


model_check(Nr,Time,ErrRes,NOW) :-
   option(timeout(TO)),!,
   time_out(model_check(Nr,Time,ErrRes),TO,Res),
   (Res=time_out -> print('*** Timeout occurred while Model Checking: '), print(TO),nl,nl,
                    writeln_log(timeout_occurred(NOW,model_check(Nr,Time,ErrRes))),
                    Time=time_out(TO), ErrRes=time_out
                 ; true).
model_check(Nr,Time,ErrRes,_NOW) :- model_check(Nr,Time,ErrRes).

:- use_module(model_checker,[model_checking_is_incomplete/6]).

add_model_checking_warnings(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations) :-
   %print(check(model_checking_is_incomplete(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations,Msg,Term))),nl,
   model_checking_is_incomplete(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations,Msg,Term),
   nl,add_warning(model_check_incomplete,Msg,Term),
   fail.
add_model_checking_warnings(_,_,_,_).

model_check(Nr,Time,ErrRes) :-
   statistics(runtime,[T1,_]),
   statistics(walltime,[W1,_]),
   (option(no_deadlocks) -> FindDeadlocks=0 ; FindDeadlocks=1),
   (option(no_invariant_violations) -> FindInvViolations=0 ; FindInvViolations=1),
   (option(no_goal) -> FindGoal=0 ; FindGoal=1),
   (option(no_assertion_violations)
     -> FindAssViolations=0
      ; FindAssViolations=1
    ),
   get_preference(por,POR),
   (get_preference(use_pge,true) -> PartialGuardEvaluation=1; PartialGuardEvaluation=0),
   StopAtFullCoverage=0,
   %STOPMCAFTER =  86400000, /* 86400000 = 1 day timeout */
   STOPMCAFTER = 1152921504606846975, /* equals 13,343,998,895 days */
   InspectExistingNodes = 1,
   (user:do_model_check(Nr,NodesAnalysed,STOPMCAFTER,ErrRes,FindDeadlocks,FindInvViolations,FindGoal,FindAssViolations,StopAtFullCoverage,POR,PartialGuardEvaluation, InspectExistingNodes)
      -> (statistics(runtime,[T2,_]), statistics(walltime,[W2,_]),
           Time1 is T2-T1, WTime is W2-W1,
           (model_checker: expired_static_analysis_time(AnalysisTime) ->
               Time is Time1 - AnalysisTime
           ;   Time = Time1),
           format('Model Checking Time: ~w ms (~w ms walltime)~n',[Time,WTime]),
           format('States analysed: ~w~n',[NodesAnalysed]),
           state_space: get_state_space_stats(_,NrTransitions,_),
           print('Transitions fired: '),print(NrTransitions),nl,
          (ErrRes = no
           -> print('No Counter Example Found, not all nodes visited')
           ;  (ErrRes=all
                -> (tcltk_find_max_reached_node
                     -> print('No Counter Example found. However, not all transitions were computed !')
                     ;  print('No Counter Example found. ALL nodes visited.')
                   ),
                   add_model_checking_warnings(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations)
                ;   print('*** COUNTER EXAMPLE FOUND ***'),nl,
                    debug_println(20,ErrRes),nl,
                    user:translate_error_for_tclk(ErrRes,TclTkRes),
                    print(TclTkRes),nl,
                    (option(silent) -> true
                      ; print('*** TRACE: '),nl,
                        user:tcltk_get_history(list(Hist)),
                        reverse(Hist,Trace),
                        print_list(Trace,1)
                    ),
                    error_occurred(TclTkRes)
              )
          )
         )
      ;  ( statistics(runtime,[T2,_]),
           Time1 is T2-T1,
           (model_checker: expired_static_analysis_time(AnalysisTime) ->
               Time is Time1 - AnalysisTime
           ;   Time = Time1),
           print('Model Checking Time: '), print(Time), print(' ms'),nl,
           print_error('*** Model Checking FAILED '),nl,
           ErrRes=fail,
           definite_error_occurred)
   ).
cbc_check(NOW) :-
     option(cbc_deadlock_check(DeadlockGoalPred)),
     print('% Starting Constraint-Based DEADLOCK check '),nl,
     writeln_log_time(cbc_deadlock_check(NOW)),
     (tcltk_constraint_find_deadlock_state_with_goal(DeadlockGoalPred,false,Res)
       -> write_result_to_file(Res),
         (Res=time_out ->
            print('*** TIME_OUT occurred ***'),nl,
            error_occurred(cbc_deadlock_check_time_out)
          ; print('*** DEADLOCK state found ***'),nl,
            error_occurred(cbc_deadlock_check),
            print('*** STATE = '),nl,
            current_b_expression(DBState), translate:print_bstate(DBState),nl,
            print('*** END DEADLOCKING STATE '),nl
         )
       ;  write_result_to_file(no_deadlock_found),
          print('No DEADLOCK state found'),nl
     ),fail.
cbc_check(NOW) :-
     option(constraint_based_check(OpName)),OpName\=all,
     check_loaded(constraint_based_check),
     print('% Starting Constraint-Based Check for Operation: '), print(OpName),nl,
     writeln_log_time(cbc_operation_check(NOW,OpName)),
	  (tcltk_constraint_based_check_with_timeout(OpName,Result)
		->  print('% Constraint-Based Check Result: '),nl,
			print(Result),nl,
			write_result_to_file(Result),
			(Result=time_out
			  ->  print('*** TIMEOUT OCCURRED ***'),nl, error_occurred(cbc)
			  ;  (Result=ok -> print('NO ERRORS FOUND'),nl)
				   ; print('*** CONSTRAINT-BASED CHECK FOUND ERRORS ***'),nl,error_occurred(cbc) )
		;  write_result_to_file(constraint_based_check_failed),
		   add_error(probcli,'ConstraintBasedCheck unexpectedly failed'),
		   (b_get_machine_operation(OpName,_,_,_) -> true
		     ; add_error(probcli,'Unknown Operation Name: ',OpName)),
		   definite_error_occurred
	  ),nl,
	  fail.
cbc_check(NOW) :-
     option(constraint_based_check(all)),
     check_loaded(constraint_based_check),
     print('% Starting Constraint-Based Check for all Operations: '),nl,
     writeln_log_time(cbc_all_check(NOW)),
          (tcltk_constraint_based_check(list(Result),ErrorsWereFound)
            ->  print('% Constraint-Based Check Result: '),nl,
                print(Result),nl,
                write_result_to_file(Result),
                (ErrorsWereFound=true
                  ->  print('*** CONSTRAINT-BASED CHECK FOUND ERRORS ***'),nl, error_occurred(cbc)
                  ;  (ErrorsWereFound=false -> print('NO ERRORS FOUND'),nl)
                       ; print('*** TIMEOUT OCCURRED ***'),nl,error_occurred(cbc)
                )
            ;  write_result_to_file(cbc_check_failed),
               add_internal_error('ConstraintBasedCheck unexpectedly failed. ',cbc_check(NOW)),definite_error_occurred
          ),nl,fail.
cbc_check(NOW) :-
	 option(cbc_assertions(AllowEnumWarning,Options)),
     check_loaded(cbc_assertions),
     print('% Starting Constraint-Based static ASSERTIONS check '),nl,
     writeln_log_time(cbc_assertions_start(NOW)),
     (tcltk_constraint_find_static_assertion_violation(Res,Options)
       -> process_cbc_assertion_result(Res,AllowEnumWarning)
       ;  write_result_to_file(cbc_assertions_failed),
          print('CBC Check failed'),nl,
          error_occurred(cbc_assertions_failure)
     ),writeln_log_time(cbc_assertions_finished(NOW,Res)),fail.
%cbc_check(NOW) :-
%     option(cbc_pred(TargetPredString)),
%     check_loaded(cbc_sequence),
%     print('% Starting Constraint-Based Check for Predicate: '), print(TargetPredString),nl,
%     b_set_up_valid_state_with_pred(...)  TO DO: add this feature
cbc_check(NOW) :-
     option(cbc_sequence(Sequence,TargetPredString,Findall)),
     check_loaded(cbc_sequence),
     print('% Starting Constraint-Based Check for Sequence: '), print(Sequence),
     (TargetPredString='' -> true ; print(' with target: '), print(TargetPredString)),nl,
     writeln_log_time(cbc_sequence(NOW)),
     (tcltk_cbc_find_trace(Sequence,TargetPredString,Findall,Res)
       -> (Res=ok -> print('Sequence found and executed'),nl
           ; Res=time_out -> error_occurred(cbc_sequence_time_out)
           ; Res=no_solution_found -> print('*** NO SOLUTION FOUND '),error_occurred(cbc_sequence_no_solution_found)
           ; Res=nr_cbc_sols(NrSols) -> print('*** # SOLUTIONS FOUND: '),print(NrSols),nl
           ; print('*** Unknown result: '), print(Res),nl,
             error_occurred(cbc_sequence)
          )
       ;  print('*** Internal error: Check failed '), error_occurred(cbc_sequence), Res=internal_error
     ),writeln_log_time(cbc_sequence(NOW,Res)),fail.
cbc_check(NOW) :-
     option(cbc_refinement),
     check_loaded(cbc_refinement),
     print('% Starting Constraint-Based static refinement check '),nl,
     writeln_log_time(cbc_refinement_start(NOW)),
     tcltk_cbc_refinement_check(list(Result),ErrorsWereFound),
     print('% Constraint-Based Refinement Check Result: '),nl,print(Result),nl,
     (ErrorsWereFound = time_out -> print('*** TIME_OUT occurred ***'),nl,error_occurred(cbc_refinement_time_out) ;
      ErrorsWereFound = true -> print('*** Refinement Violation found ***'),nl,error_occurred(cbc_refinement) ;
      otherwise -> print('No static Refinement Violation found'),nl
     ),writeln_log_time(cbc_refinement_finished(NOW,ErrorsWereFound)),fail.
cbc_check(_).

:- use_module(solver_interface,[predicate_uses_unfixed_deferred_set/2, unfixed_typed_id_in_list/3]).
:- use_module(bsyntaxtree,[get_texpr_id/2]).
process_cbc_assertion_result(time_out,_) :- !,
        write_result_to_file(no_counterexample_found('"TIME_OUT"')),
	    print('*** TIME_OUT occurred ***'),nl,
	    error_occurred(cbc_assertions_time_out).
process_cbc_assertion_result(no_counterexample_exists(Constants,TotPredicate),AllowEnumWarning) :- !,
		print('No counter-example to ASSERTION exists'),nl,
		(unfixed_typed_id_in_list(TID,CType,Constants) % TO DO: look only at component
		  -> write_deferred_set_used(AllowEnumWarning),
			 get_texpr_id(TID,CID),pretty_type(CType,CTypeS),
		     format('Warning: Some constants use deferred sets (e.g., ~w:~w) which have only been checked for a single cardinality!~n',[CID,CTypeS])
		  ; predicate_uses_unfixed_deferred_set(TotPredicate,CType)
		    -> write_deferred_set_used(AllowEnumWarning),pretty_type(CType,CTypeS),
		       format('Warning: Some quantified variables use deferred sets (e.g., ~w) which have only been checked for a single cardinality!~n',[CTypeS]) % happens for tests 1173, 1174
		  ; write_result_to_file(no_counterexample_exists)
		    %,print('Computing unsat core: '),nl,unsat_cores:unsat_core(TotPredicate,Core),print('CORE: '),translate:print_bexpr(Core),nl
		). % WE HAVE A PROOF
process_cbc_assertion_result(no_counterexample_found,AllowEnumWarning) :- !,
		write_result_to_file(no_counterexample_found('"Enumeration Warning"')),
		print('No counter-example for ASSERTION found (*enumeration warning occured*)'),nl,
		(AllowEnumWarning=true -> true ; error_occurred(cbc_assertions_enumeration_warning)).
process_cbc_assertion_result(counterexample_found,_) :- !,
		write_result_to_file(counterexample_found),
        print('*** Counter-example for ASSERTION found ***'),nl,
		error_occurred(cbc_assertions),
		print('*** STATE = '),nl,
		current_b_expression(DBState), translate:print_bstate(DBState),nl,
		print('*** END ASSERTION counter-example STATE '),nl,
		(get_dot_file('cbc_assertions',DFile) -> generate_dot_from_assertions(DFile) ; true).
process_cbc_assertion_result(Res,A) :-
		write_result_to_file(Res),
		add_internal_error('Unknown: ',process_cbc_assertion_result(Res,A)).


write_deferred_set_used(AllowEnumWarning) :-
             write_result_to_file(no_counterexample_found('"Deferred Sets Used"')),
			 (AllowEnumWarning=true -> true ; error_occurred(cbc_assertions_enumeration_warning)).

:- use_module(tools_io,[safe_open_file/4]).
write_result_to_file(Result) :- option(cbc_result_file(FILE)),
   safe_open_file(FILE,write,Stream,[]),
   !,
   write(Stream,Result),
   close(Stream).
write_result_to_file(_).



if_option_set(Option,Call) :-
    if_option_set(Option,Call,true).
if_option_set(Option,Then,Else) :-
    ( option(Option) -> call(Then)
    ; otherwise      -> call(Else)).
ifm_option_set(Option,Call) :-
    ifm_option_set(Option,Call,true).
ifm_option_set(Option,Then,Else) :- % can perform multiple options
   findall(Then,option(Option),As),
   (As=[] -> call(Else) ; perform(As)).
perform([]).
perform([A|T]) :- call(A), perform(T).
if_option_set_loaded(Option,Action,Call) :-
    ( (option(Option),check_loaded(Action)) ->
        call(Call)
    ; otherwise -> true).
ifm_option_set_loaded(Option,Action,Call) :- % can perform multiple options
   findall(Call,(option(Option),check_loaded(Action)),As), perform(As).



if_options_set(Option,Call) :- % allow multiple solutions for Option
   option(Option),call(Call),fail.
if_options_set(_,_).

print_options :- print('CLI OPTIONS: '),nl,
                 option(Option), print(Option), nl, fail.
print_options :- nl.

:- use_module(enabling_analysis,[infeasible_operation_cache/1]).

mcm_test_case_generation(ADepth,AMaxStates,ATarget,Output) :-
    arg_is_number(ADepth,MaxDepth),
    arg_is_number(AMaxStates,MaxStates),
    bmachine:b_parse_machine_predicate(ATarget,Target),!,
    get_comma_separated_options(mcm_cover,Events),
    print('mcm test case generation, maximum search depth: '),print(MaxDepth),nl,
    print('mcm test case generation, maximum number of states: '),print(MaxStates),nl,
    print('mcm test case generation, target state predicate: '),print_bexpr(Target),nl,
    print('mcm test case generation, output file: '),print(Output),nl,
    print('mcm test case generation, events to cover: '),print_list(Events),nl,
    explore_and_generate_testcases(Events,Target,MaxDepth,MaxStates,Output,NumTests,Uncovered),
    print('mcm test case generation, generated test cases: '),print(NumTests),nl,
    print_uncovered('mcm test case generation, ',Uncovered).
mcm_test_case_generation(_ADepth,_AMaxStates,_ATarget,_Output) :-
    print_error('### MCM Test Case Generation failed'),
    error_occurred(mcm_tests).

cbc_test_case_generation(ADepth,TargetString,Output) :-
    arg_is_number(ADepth,MaxDepth),
    ( option(cbc_cover_all) -> Events=all
       ; (get_comma_separated_options(cbc_cover,Events), Events \= []) -> true
       ; Events=all ),
    (\+ option(cbc_cover_final) -> FEvents = Events
     ; Events=all -> FEvents=all,
         add_error(cbc_cover_final,'Option cbc_cover_final not compatible with trying to cover all events')
     ; FEvents = final(Events),
       print('constraint based test case generation, target events considered final'),nl),
    print('constraint based test case generation, maximum search depth: '),print(MaxDepth),nl,
    print('constraint based test case generation, target state predicate: '),print(TargetString),nl,
    print('constraint based test case generation, output file: '),print(Output),nl,
    (TargetString = '#not_invariant' ->
       print('constraint based test case generation, performing bounded model checking'),nl
    ;  print('constraint based test case generation, events to cover: '),print_list(Events),nl),
    cbc_gen_test_cases_from_string(FEvents,TargetString,MaxDepth,Output,Uncovered),
    !,
    %print('mcm test case generation, generated test cases: '),print(NumTests),nl,
    (TargetString = '#not_invariant'
      -> tcl_get_stored_test_cases(Tests),
        (Tests=[] -> print('No counterexample found'),nl ; add_error(invariant_violation,'Invariant violation found by bmc'))
      ; (Uncovered=[_|_],option(strict_raise_error))
       -> add_error(cbc_tests,'Uncovered events: ',Uncovered)
      ; print_uncovered('constraint based test case generation, ',Uncovered)
    ).
cbc_test_case_generation(_ADepth,_ATarget,_Output) :-
    print_error('### CBC Test Case Generation failed'),
    error_occurred(cbc_tests).

print_uncovered(Msg,Uncovered) :-
   include(infeasible_operation_cache,Uncovered,Infeasible),
    (Infeasible=[]
      -> format('~wuncovered events: ',[Msg]),print_list(Uncovered),nl
      ;  format('~winfeasible uncovered events: ',[Msg]),print_list(Infeasible),nl,
         exclude(infeasible_operation_cache,Uncovered,Feasible),
         format('~wuncovered events: ',[Msg]),print_list(Feasible),nl
    ).

print_list(all) :- print('** all **').
print_list([]) :- print('** none **').
print_list([H|T]) :- print(H),print(' '),print_list2(T).
print_list2([]).
print_list2([H|T]) :- print(H),print(' '),print_list2(T).

get_comma_separated_options(Option,Selection) :-
    functor(O,Option,1),
    findall(E, (option(O),arg(1,O,CommaSep),
                   split_by_comma(CommaSep,Es),
                   member(E,Es)),
            Selection).

split_by_comma(NonAtomic,Res) :- \+ atomic(NonAtomic),!, Res=[NonAtomic].
split_by_comma(String,Strings) :-
    atom_chars(String,Chars),
    split_by_comma2(Chars,Strings).
split_by_comma2(Chars,[A|Rest]) :-
    append(AChars,[','|B],Chars),!,
    atom_chars(A,AChars),
    split_by_comma2(B,Rest).
split_by_comma2(Chars,[String]) :- atom_chars(String,Chars).


ltl_check_assertions :-
    (option(ltl_limit(Limit)) -> true; Limit= -1), % -1 means no limit
    print('LTL Model checking assertions'),nl,
    ltl_check_assertions(Limit,Outcome),
    ( Outcome = pass -> print('LTL check passed'),nl
    ; Outcome = fail -> print('*** LTL check failed'),nl,error_occurred(ltl)
    ; Outcome = no_tests -> print('*** No LTL assertions found, test failed'),nl,definite_error_occurred
    ; otherwise -> print('*** An error occurred in the LTL assertion test'),nl,
	definite_error_occurred).

:- use_module(probltlsrc(ltl),[parse_ltlfile/2]).
ltl_check_file(Filename) :-
    (option(ltl_limit(Limit)) -> true; Limit= -1), % -1 means no limit
    print('LTL Model checking assertions'),nl,
    parse_ltlfile(Filename, Formulas),
    ( real_error_occurred ->
        print('An error occurred while parsing the LTL file.\n'), fail
    ; otherwise ->
        ltl_check_formulas(Formulas,Limit)).

:- use_module(probltlsrc(ltl),[ltl_model_check2/4]).
ltl_check_formulas([],_) :-
    print('All LTL formulas checked.\n').
ltl_check_formulas([formula(Name,F)|Rest],Limit) :-
    print('Checking formula '),print(Name),print(':\n'),
    ltl_model_check2(F,Limit,init,Status),
    ( Status == no ->
        print('Counter-example found for formula \"'),print(Name),
        print('\", saving trace file.\n'),
        ajoin(['ltlce_', Name, '.trace'], Tracefile),
        tcltk_save_history_as_trace_file(Tracefile),
        add_error(ltl_counterexample,'Counter-example was found')
    ; Status == ok ->
        ltl_check_formulas(Rest,Limit)
    ; Status == incomplete ->
        ajoin(['Model was not completly model-checked, aborted after ',Limit,' new states'],
              Msg),
        add_error(ltl,Msg)
    ; otherwise ->
        ajoin(['Model checker returns unexpected result (',Status,')'],Msg),
        add_error(ltl,Msg)).

cli_ltl_model_check(Formula,ExpectedStatus,Status) :-
    (option(ltl_limit(Max)) -> true; Max = -1), % -1 means no limit
    print('LTL model checking formula'),nl,
    ltl_model_check(Formula,Max,init,Status),
    check_status(Status,ExpectedStatus,Formula,ltl).

cli_ctl_model_check(Formula,ExpectedStatus,Status) :-
    (option(ltl_limit(Max)) -> true; Max = -1), % -1 means no limit
    print('CTL model checking formula'),nl,
    ctl_model_check(Formula,Max,init,Status),
    check_status(Status,ExpectedStatus,Formula,ctl).

check_expected(St,Exp,Mode) :-
   (St=Exp -> true ; add_error(Mode,'Unexpected model checking result: ',Mode:St)).

check_status(ok,Expected,Formula,ltl) :- !, % TO DO: make uniform ? CTL returns true; LTL returns ok
    format('LTL Formula TRUE.~nNo counter example found for ~w.~n',[Formula]),
    check_expected(true,Expected,ltl).
check_status(true,Expected,Formula,ctl) :- !,
    format('CTL Formula TRUE.~nNo counter example found for ~w.~n',[Formula]),
    check_expected(true,Expected,ctl).
check_status(incomplete,Expected,Formula,LTLorCTL) :- !,
    incomplete_warning(LTLorCTL,Warning),
    add_warning(Warning, 'Warning: Model Check incomplete for: ', Formula),nl,
    format('No counter example found so far for ~w.~n',[Formula]),
    check_expected(incomplete,Expected,LTLorCTL).
check_status(NO,Expected,Formula,LTLorCTL) :- (NO=no ; NO=false),!, % TO DO: make uniform
    (Expected==false -> format('Model Check Counter-Example found for: ~w~n',[Formula])
      ; add_error(LTLorCTL, 'Model Check Counter-Example found for: ', Formula),nl
    ),
    print('Formula '), print('FALSE.'),nl,
        %print('Counter example found for '),print(Formula),print('.'),nl,
    nl,
    check_expected(false,Expected,LTLorCTL).
check_status(Status,Expected,Formula,LTLorCTL) :-
    add_internal_error('Unknown status: ', check_status(Status,Expected,Formula,LTLorCTL)).

incomplete_warning(ltl,ltl_incomplete) :- !.
incomplete_warning(ctl,ctl_incomplete) :- !.
incomplete_warning(X,X).

evaluate_from_commandline :-
    retractall(eval_result(_,_)),
    option(eval_string_or_file(A,B,Q,E)),   %print(eval(A,B,Q,E)),nl,
    % treat eval_string and eval_file together to ensure proper order of evaluation
    %  (only possible side-effect at the moment: formula can adde new machine_string facts)
    evaluate_string_or_file(A,B,Q,E),
    fail.
evaluate_from_commandline :- print_eval_results,
    % treat -repl option
    if_option_set(eval_repl,
                  (evaluate_expressions -> true ; true)).

:- dynamic eval_result/2.
add_eval_result(R) :- retract(eval_result(R,N)),!,
   N1 is N+1, assert(eval_result(R,N1)).
add_eval_result(R) :- assert(eval_result(R,1)).
print_eval_results :- findall(R/N, eval_result(R,N), L), sort(L,SL),
   (SL=[] -> true ; format('Evaluation results: ~w~n',[SL])).

evaluate_string_or_file(string,String,_,Expected) :-
   print('eval: '), print(String),nl,
    eval_string_with_time_out(String,StringResult,EnumWarning,_LS),
    % print(enumwarning(EnumWarning,Expected)),nl,
    add_eval_result(StringResult),
    (Expected\=StringResult
      -> print_error('### Error, expected result to be: '), print_error(Expected),
         error_occurred(eval_string)
      ; EnumWarning = false -> true
      ; gen_enum_warning(Expected) % now also generate enum warnings when Expected not set (eval instead of evalt/evalf)
    ).
evaluate_string_or_file(file,File,Quantifier,Expected) :-
    turn_show_error_source_off, % reduce clutter in user feedback; eval_file used in ProB Logic Calculator for example
    debug_println(10,eval_file(File,Quantifier,Expected)),nl,
    error_manager:reset_error_scopes, % TO DO: avoid that exceptions mess up error scopes in eval_string/file
    statistics(runtime,[Start,_]),nl,
    (eval_file(File,Quantifier,Result,EnumWarning)
      -> add_eval_result(Result),
         (EnumWarning = false -> true ;  gen_enum_warning(EnumWarning)),
         (Result\=Expected -> print_error('### Error, expected result to be: '),
                              print_error(Expected),
                              print_error('### File: '), print_error(File),
                              error_occurred(eval_file)
           ; true)
       ;  add_eval_result(eval_file_failed),
          print_error('### Eval from file failed: '), print_error(File),
          error_occurred(eval_file)
    ),
   (debug_mode(on) -> statistics(runtime,[Stop,_]), Time is Stop - Start,
                      print(File),print(' time: '), print(Time), print(' ms'),nl
      ; true),
  turn_show_error_source_on.

gen_enum_warning(time_out) :- !,error_occurred(eval_string_time_out).
gen_enum_warning(_) :- print_error('### Enumeration warning occurred'),
                       error_occurred(eval_string_enum_warning,warning).
repl :- evaluate_expressions.
:- use_module(parsercall,[ensure_console_parser_launched/0]).
evaluate_expressions :- nl,print('ProB Interactive Expression and Predicate Evaluator '), nl,
                        print('Type ":help" for more information.'),nl,
                        turn_show_error_source_off, % reduce clutter in user feedback
                        (option(evaldot(File))
                          -> print('Solutions written to dot file: '), print(File),nl
                          ;  true
                        ),
                        print('>>> '),
                        (ensure_console_parser_launched
                           -> top_level_eval
                            ; print('>>> ABORTING REPL'),nl),
                        turn_show_error_source_on.

:- use_module(user_interrupts,[interruptable_call/1]).
top_level_eval :-
  catch(top_level_eval1, halt(0), (format('~s', ["Bye."]), nl)).

top_level_eval1 :-
 (interruptable_call(eval1) -> true
  ; print('Evaluation failed or interrupted'),nl,
    print('Use :q to quit REPL'),nl),
  print('>>> '),
  top_level_eval1.
eval :- reset_errors, % get_errors prints errors again and quits in -strict mode
        print('>>> '), garbage_collect, eval1.
eval1 :- repl_multi_read_line(Expr), eval_probcli_repl_line(Expr).

repl_multi_read_line(Line) :- repl_multi_read_line_aux([],Line).
repl_multi_read_line_aux(SoFar,Line) :- read_line(L),
  (append(LFront,[92],L)  % line ends with slash
   -> append(SoFar,LFront,NewSoFar), print('... '),
      repl_multi_read_line_aux(NewSoFar,Line)
    ; append(SoFar,L,Line)).

:- use_module(eval_strings).
:- dynamic trace_eval/0.

generate_atom_list([],[],R) :- !, R=[].
generate_atom_list([],Last,[NewAtom]) :-
   reverse(Last,RL),
   atom_codes(NewAtom,RL).
generate_atom_list([32|X],Last,[NewAtom|T]) :- !,
   reverse(Last,RL),
   atom_codes(NewAtom,RL),
   generate_atom_list(X,[],T).
generate_atom_list([H|X],Last,Res) :- generate_atom_list(X,[H|Last],Res).

:- meta_predicate call_probcli_option(0).
call_probcli_option(Option) :-
   on_exception(error(existence_error(_,_),E),
       call(Option),
       (print('probcli Command not yet supported in REPL: '), print(Option),nl,
        print('error: '), print(E),nl)).

reload_mainfile :- file_loaded(_,MainFile),
      print('Reloading and initialising file: '), print(MainFile),nl,
      clear_loaded_files,
      load_main_file(MainFile,0,_),
      get_errors,
      cli_start_animation(0),
      cli_start_initialisation(0).

% REPL EVAL LOOP:
eval_probcli_repl_line(end_of_file) :- !, eval_line(end_of_file).
eval_probcli_repl_line(Line) :- strip_ws(Line,SLine), eval_line(SLine).

strip_ws([32|T],Res) :- !, strip_ws(T,Res).
strip_ws(C,Res) :- reverse(C,CR), strip_ws2(CR,SCR), reverse(SCR,Res).
strip_ws2([32|T],Res) :- !, strip_ws2(T,Res).
strip_ws2(R,R).

eval_line([]) :- !, print('Type :q or :quit to quit.'),nl,eval.
eval_line(end_of_file) :- !, halt_exception(0).
% Haskell GHCI like syntax
eval_line(":q") :- !, halt_exception(0).
eval_line(":quit") :- !, halt_exception(0).
eval_line(":x") :- !,halt.
eval_line(":b") :- !, browse,eval.
eval_line(":browse") :- !, browse,eval.
eval_line(":r") :- !, eval_line("--reload").
eval_line(":reload") :- !, eval_line("--reload").
eval_line(":h") :- !, eval_help, eval.
eval_line(":help") :- !, eval_help, eval.
eval_line(":?") :- !, eval_help, eval.
eval_line("--reload") :- !,
      (reload_mainfile -> true ; get_errors,print('### Error(s) occured during reload'),nl),
      eval.
% TO DO: other Haskell commands :info E   :l FILE ,  let pattern = expression
eval_line("--prefs") :- !,print_eclipse_prefs, eval.
eval_line([45|Command]) :- generate_atom_list([45|Command],[],ArgV),
   % try and parse like commands passed to probcli
   get_options(ArgV,recognised_option,Options,[],fail),
   print('Executing probcli command: '),print(Options),nl,!,
   (maplist(call_probcli_option,Options) -> true
     ; print('Failed to execute probcli arguments'),nl),
   eval.
eval_line("+") :- !, add_last_expression_to_unit_tests, eval.
eval_line("$+") :- !, preferences:temporary_set_preference(expand_avl_upto,-1,CHNG),
    print_last_value,preferences:reset_temporary_preference(expand_avl_upto,CHNG),
    eval.
eval_line("$$") :- !, print_last_expression, eval.
eval_line(":show") :- !, show_last_expression_as_table, eval.
eval_line([36,36,Num]) :- Num>47, Num<58,!, % $$0 - $$9 commands to print last expression with indentation
   N1 is Num-48,indent_print_last_expression(N1), eval.
eval_line("$") :- !, print_last_info, eval.
eval_line(":statistics") :- !, print_last_info, eval.
eval_line(":stats") :- !, print_last_info, eval.
eval_line("!trace") :- !, eval_line("^").
eval_line("^") :- !, (retract(trace_eval) -> print('TRACING OFF'),nl ; assert(trace_eval), print('TRACING ON'),nl), eval.
eval_line("!observe") :- !, toggle_observe_evaluation.
eval_line("%") :- !, print('PROFILING : '), %spy([avl:avl_size/2]),
   (current_prolog_flag(profiling,on)
    -> set_prolog_flag(profiling,off), print('OFF') ;
       set_prolog_flag(profiling,on), print('ON')),
    nl,print('USE %% to print profile info'),nl,eval.
eval_line("%%") :- !, nl,print('PROFILE INFORMATION:'), nl,
   on_exception(error(existence_error(_,_),_),print_profile, print('CAN ONLY BE USED WHEN RUNNING PROB FROM SOURCE')),
   nl, debug:timer_statistics,
   eval.
eval_line("%%%") :- !, nl,print('PROFILE INFORMATION (Starting TK Viewer):'), nl,
   on_exception(error(existence_error(_,_),_),(use_module(library(gauge)),view), print('CAN ONLY BE USED WHEN RUNNING PROB FROM SOURCE')),
   nl,eval.
eval_line("!debug") :- !, nl,print('ENTERING PROLOG DEBUG MODE:'),
   on_exception(error(existence_error(_,_),_),debug, print('CAN ONLY BE USED WHEN RUNNING PROB FROM SOURCE')),
   nl,
   eval.
eval_line([D,C,T,L|FORMULA]) :- [D,C,T,L] = "$ctl",!,
    atom_codes(F,FORMULA),
    (cli_ctl_model_check(F,_,Status)
     -> (Status=false -> write_history_to_user_output([show_init,show_states]) ; true)
     ; print('CTL Syntax: ExUy,EXx,AXx,EFx,AGx,EX[Op]x,e(Op),{B-Pred}'),nl),
    eval.
eval_line([D,L,T,L|FORMULA]) :- [D,L,T] = "$lt",!,
    atom_codes(F,FORMULA),
    (cli_ltl_model_check(F,_,Status)
     -> (Status=no
     -> write_history_to_user_output([show_init,show_states]) ; true)
     ; print('LTL Operators: G,F,X,U,W,R,not,&,or,=>'),nl,
       print('LTL Propositions: {B-Pred}, e(Op), [Op], true, false, sink'),nl,
       print('Past-LTL Operators: Y,H,O,S,T (dual to X,G,F,U,R)'),nl
     ),
    eval.
eval_line("@") :- !, get_preference(find_abort_values,OldVal),
  print('Try more aggressively to detect ill-defined expressions: '),
  (OldVal=true -> Val=false ; Val=true), print(Val),nl,
  temporary_set_preference(find_abort_values,Val) , eval.
eval_line("!") :- !, toggle_eval_det,eval.
eval_line("!norm") :- !, toggle_normalising,eval.
eval_line([104,101,108,112]) :- !, /* help */
       eval_help,
       eval.
eval_line(ExpressionOrPredicate) :- (trace_eval -> trace ; true),
      (eval_codes(ExpressionOrPredicate,exists,_,_,_,_)
        -> eval ; print('Evaluation failed'),nl,eval).

profiling_on :- set_prolog_flag(profiling,on), print('% PROFILING ON'),nl.


eval_help :-
       print('ProB Interactive Expression and Predicate Evaluator '), nl,
       print('Type a valid B expressions or predicates, followed by RETURN or ENTER.'),nl,
       print('You can spread input over multiple lines by ending lines with "\\".'),nl,
       browse_machine,
       print('You can also type one of the following commands: '),nl,
       (option_verbose ->
          print('    + to save last expression to ProB unit tests.'),nl,
          print('    ! to go to deterministic propagation only mode.'),nl
         ; true),
       print('    let x = E to define a new local variable x and :b to browse the variables'),nl,
       print('    :t E to get the type of an expression and :r to reload the machine'),nl,
       print('    :show to display the last result as a table (if possible)'),nl,
       print('    :stats to print the type and evaluation time for last query'),nl,
       print('    -PROBCLIARGS to pass command-line probcli arguments to the REPL'),nl,
       print('    $ctl F or $ltl F to check a CTL or LTL formula.'),nl,
       print('    :q to exit.'),nl.

browse :- browse_machine, browse_repl_lets.

browse_machine :-
       print_sets,
       print('Available CONSTANTS: '),
       bmachine:b_get_machine_constants(Constants),
       bsyntaxtree:get_texpr_ids(Constants,CN),print(CN),nl,
       print('Available VARIABLES: '),
       bmachine:b_get_machine_variables(Variables),
       bsyntaxtree:get_texpr_ids(Variables,VN),print(VN),nl.

print_sets :- print('Available SETS: '), b_global_set(GS), print_set(GS),fail.
print_sets :- nl.

print_set(GS) :- print(GS), \+ is_b_global_constant(GS,_,_),!, print(' ').
print_set(GS) :- print(' = {'), is_b_global_constant(GS,_,Cst), print(Cst), print(' '),fail.
print_set(_) :- print(' } ').

% showing relations as tables:

:- use_module(table_tools).
show_last_expression_as_table :- \+ last_expression(_,_Expr,_Value),!,
   print('Please evaluate an expression or predicate first.'),nl.
show_last_expression_as_table :-
   get_last_result_value(Expr,_,Value),
   expand_and_translate_to_table(Value,Header,Table),
   ((get_header_from_expression(Expr,ExprHeader),
     same_length(Header,[_|ExprHeader]))
     -> print_table(['Nr'|ExprHeader],Table)
     ;  print_table(Header,Table)).

:- use_module(bsyntaxtree,[get_texpr_ids/2]).
get_header_from_expression(b(comprehension_set(IDs,_),_,_),Header) :-
   get_texpr_ids(IDs,Header).


% Simple Animator

animate_machine :- nl,print('IMPORTANT: Do not use this mode for automatic tools.'),nl,
                   print('The output format can change arbitrarily in future versions.'),nl,
                   print('Please terminate your input with a dot (.) and then type return.'),nl,nl,
                   animate_machine2.

animate_machine2 :-
  print_current_state,
  cli_computeOperations(Ops),
  length(Ops,Max),
  print('Enabled Operations: '),nl,
  print_options(Ops,1),
  print(' ==> '),!,
  read(Nr),
  ((number(Nr),Nr>0,Nr=<Max)
    -> cli_executeOperationNr(Nr,Ops,0)
    ;  fail
  ),!,
  animate_machine2.
animate_machine2.

print_current_state :- getCurrentStateID(CurID), print('ID ==> '), print(CurID),nl,
  getStateValues(CurID,State),
  print_bindings(State),
  ((specfile:b_or_z_mode,\+is_initialised_state(CurID))
  -> print(' Not yet initialised.'),print_mode_info, debug_println(10,state(State)) ; nl).

print_mode_info :- animation_mode(M), (animation_minor_mode(MM) -> true ; MM=''),
  format('Animation Mode = ~w [~w]~n',[M,MM]).

cli_computeOperations(Ops) :- option(animate_stats),!, % provide statistics about the animation
   nl,
   start_probcli_timer(Timer),
   computeOperations(Ops),
   stop_probcli_timer(Timer,'Time: ').
cli_computeOperations(Ops) :- computeOperations(Ops).

cli_executeOperationNr(Nr,Options,StepNr) :-
   (option(animate_stats)
     -> nth1(Nr,Options,Action),
        truncate_animate_action(Action,TA),
        (StepNr>1 -> format('performing step ~w : ~w~n',[StepNr,TA])
          ; format('performing ~w~n',[TA]))
     ; true),
   executeOperationNr_on_computed_node(Nr).

% optionally truncate animation action atom for printing:
truncate_animate_action(Action,TA) :-
   (option_verbose -> TA = Action
	  ; \+ atom(Action) -> TA = Action
	  ; truncate_atom(Action,100,TA)).

perform_random_step(StepNr) :- perform_random_step(_Ops,_Len,_RanChoice,StepNr).
perform_random_step(Ops,Len,RanChoice,StepNr) :-
  cli_computeOperations(Ops),
  length(Ops,Len), Len>0,
  debug_println(20,perform_random_step(Len,StepNr)),
  L1 is Len+1,
  ((do_det_checking, Len>1)
    ->  print_error('### Non-deterministic step in animate or init'),
        print_error('### State:'),
        print_current_state,  print_error('### Enabled Operations: '), print_options(Ops,1),
        error_occurred(det_check)
    ; true),
  random(1,L1,RanChoice),
  debug_println(20,random(L1,RanChoice)),
  cli_executeOperationNr(RanChoice,Ops,StepNr).

do_det_checking :- option(det_check),!.
do_det_checking :- option(det_constants_check),getCurrentStateID(root),
   b_mode, b_machine_has_constants_or_properties.

perform_random_steps(Nr,_) :- \+ number(Nr),!,
   print_error('### Argument to animate not a number'), print_error(Nr),error_occurred(animate).
perform_random_steps(Nr,_) :- Nr<0, !,
   print_error('### Argument to animate is a negative number'), print_error(Nr),error_occurred(animate).
perform_random_steps(Nr,ErrorOnDeadlock) :- perform_random_steps_aux(0,Nr,ErrorOnDeadlock).

perform_random_steps_aux(Nr,Nr,_) :- !, debug_println(9,performed_random_steps(Nr)).
perform_random_steps_aux(Nr,Max,ErrorOnDeadlock) :-
   N1 is Nr+1,
   (perform_random_step(N1)
     -> perform_random_steps_aux(N1,Max,ErrorOnDeadlock)
     ;  /* deadlock */
      ErrorOnDeadlock=true ->
       print_error('### Deadlock occurred during -animate, at step number:'), print_error(Nr),
       error_occurred(animate)
     ; print('% Deadlock occurred during -animate, at step number:'), print(Nr),nl
    ).

perform_random_initialisation :- getCurrentStateID(State), State \= root, !,
   print_error('### init can only be used in initial state'), print_error(State),error_occurred(initialisation).
perform_random_initialisation :- b_mode, b_machine_has_constants_or_properties,!,
   (perform_random_step(Ops,_Len,RanChoice,1)
    -> nth1(RanChoice,Ops,Choice), %print(Choice),nl,
       (Choice = 'PARTIAL_SETUP_CONSTANTS'
         -> error_occurred(setup_constants_fails)
         ; true)
    ; error_occurred(setup_constants_fails),fail), % $setup_constants
   perform_random_init_after_setup_constants. % $initialise_machine
perform_random_initialisation :- (perform_random_step(1) -> true ; error_occurred(initialisation_fails),fail).


perform_random_init_after_setup_constants :- \+ option(initialise), we_need_only_static_assertions(_),!,
         print('% NOT INITIALISING MACHINE (not required)'),nl.
 %  debug_println(20,'% NOT INITIALISING MACHINE (not required)').
perform_random_init_after_setup_constants :-
   (perform_random_step(2) -> true ; error_occurred(initialisation_fails),fail).

:- use_module(enabling_analysis,[tcltk_cbc_enabling_analysis/1, print_enable_table/1]).
do_enabling_analysis_csv(EnablingCsvFile) :-
    debug_println(9,peforming_enabling_analysis),
    tcltk_cbc_enabling_analysis(list(R)),
    debug_println(9,writing_to_file(EnablingCsvFile)),
    my_tell(EnablingCsvFile),
    print_enable_table(R),
    told,!, debug_println(9,finished_writing_to_file(EnablingCsvFile)).
do_enabling_analysis_csv(EnablingCsvFile) :- add_error(enabling_analysis,'Enabling analysis failed',EnablingCsvFile).

:- use_module(enabling_analysis,[feasible_operation/2]).
do_feasibility_analysis(EnablingCsvFile) :-
    debug_println(9,peforming_feasibility_analysis),
    findall(list([Op,Res]),feasible_operation(Op,Res),R),
    debug_println(9,writing_to_file(EnablingCsvFile)),
    my_tell(EnablingCsvFile),
    print_enable_table([list(['Event','Feasibility'])|R]),
    told,!, debug_println(9,finished_writing_to_file(EnablingCsvFile)).
do_feasibility_analysis(EnablingCsvFile) :- add_error(feasibility_analysis,'Feasibility analysis failed',EnablingCsvFile).

my_tell(File) :- on_exception(error(_E,_), % existence_error(_,_)
                   tell(File), add_error_fail(tell,'File cannot be written to: ',File)).

print_options([],_).
print_options([H|T],N) :-
  print(' '), print(N), print(':'), print(H),nl,
  N1 is N+1,
  print_options(T,N1).

print_list([],_) :- !.
print_list([H|T],N) :- !,
  print(' '), print(N), print(': '),print(H), print(': '),
  (visited_expression(H,State) -> print(State) ; true),nl,
  N1 is N+1,
  print_list(T,N1).
print_list(X,_) :- print('### not a list: '), print(X),nl.

print_bindings([]) :- !.
print_bindings([binding(Var,_,PPV)|T]) :- !, print(Var),print('='),print(PPV),
  (T=[] -> true ; print(', '), print_bindings(T)).
print_bindings([binding(Var,_,PPV,_Tag)|T]) :- !, print(Var),print('='),print(PPV),
  (T=[] -> true ; print(', '), print_bindings(T)).
print_bindings(X) :- print('### Internal Error: illegal binding list: '), print(X),nl.

:- dynamic expected_error_occurred/1.
:- dynamic error_did_not_occur/1.
reset_expected_error_occurred :- retractall(expected_error_occurred(_)).
check_all_expected_errors_occurred(_NOW) :-  get_errors, get_state_space_errors,
  % first assert time_out errors; not caught by other mechanisms at the moment
  ((  state_space:time_out_for_node(_ID,_,time_out)
    ; state_space:time_out_for_invariant(_ID)
    ; state_space:time_out_for_assertions(_ID) )
   -> error_occurred(time_out),fail ; fail).
check_all_expected_errors_occurred(_NOW) :- % now assert virtual time_out error
   (state_space:time_out_for_node(_ID,_,virtual_time_out(_))
    -> error_occurred(virtual_time_out)).
check_all_expected_errors_occurred(NOW) :- retractall(error_did_not_occur(_)),
  option(expect_error(Type)),
  \+ expected_error_occurred(Type),
  print('*** Expected Error of following type to occur: '), print(Type),nl,
  writeln_log_time(expected_error_did_not_occur(NOW,Type)),
  assert(error_did_not_occur(Type)),
  (option(strict_raise_error) -> definite_error_occurred ; fail).
check_all_expected_errors_occurred(_NOW) :-
   (option(expect_error(_))
    -> (error_did_not_occur(_) -> print('*** Some expected errors did NOT occur !')
         ; print('All expected errors occurred.')),nl
    ;  true).

error_occurred(warning(Type)) :- !, error_occurred(Type,warning).
error_occurred(Type) :- error_occurred(Type,error).

error_occurred(Type,_) :- option(expect_error(Type)),!,
    assert(expected_error_occurred(Type)).
error_occurred(Type,ErrOrWarning) :-
    (probcli_time_stamp(NOW) -> true ; NOW=unknown),
    writeln_log(error_occurred(NOW,Type)),
    (option(strict_raise_error)
     -> print_error('*** Unexpected error occurred ***'),
        print_error(Type),
        definite_error_occurred
     ; (ErrOrWarning=error,serious_error(Type))
      -> print_error('*** Serious error occurred ***'),
         print_error(Type),
         definite_error_occurred
     ;  print_probcli_error_non_strict(Type,ErrOrWarning)).

% a list of serious errors: if these occur; then return code different from 0 even in non-strict mode
serious_error(get_java_command_path).
serious_error(internal_error(_)).

print_probcli_error_non_strict(parse_machine_predicate_error,_) :-
    !. % have already been reported
print_probcli_error_non_strict(Type,ErrOrWarning) :-
  (ErrOrWarning=warning -> print_error('*** warning occurred ***')
     ; print_error('*** error occurred ***')),
  print_error(Type).

definite_error_occurred :- print_error('*** Abnormal termination of probcli !'),
  (file_loaded(_,File) -> print_error('*** for_file'(File)) ; true),
  (probcli_time_stamp(NOW) -> halt1_prob(NOW)
   ; writeln_log(halt(1)),
     halt_exception(1)
  ).

halt_prob(NOW) :-
  writeln_log(prob_halted_prematurely(NOW)),
  check_all_expected_errors_occurred(NOW),
  halt_exception.
halt1_prob(NOW) :-
  writeln_log(prob_halted_prematurely(NOW)),
  halt_exception(1).


check_required_infos([],_,_).
check_required_infos([H|T],Infos,ErrType) :-
  (member(H,Infos) -> check_required_infos(T,Infos,ErrType)
     ; print('*** Unexpected result while checking: '),
       translate_err_type(ErrType,ES), print(ES),nl,
       print('***    expected : '), print(H),nl,
       print('***          in : '), print(Infos),nl,
       error_occurred(ErrType)).
translate_err_type(check_assertions,'ASSERTIONS') :- !.
translate_err_type(cli_check_assertions,'ASSERTIONS') :- !.
translate_err_type(check_goal,'GOAL') :- !.
translate_err_type(load_po_file,'PROOF OBLIGATIONS') :- !.
translate_err_type(X,X).


mc_ok_arg(Arg,X) :- Arg==all,!,prolog_flag(max_tagged_integer,X).
mc_ok_arg(Arg,N) :- tools:arg_is_number(Arg,N).


:- dynamic option/1.
assert_all_options([]).
assert_all_options([Opt|T]) :- assert(option(Opt)), assert_all_options(T).

% replace a leading double-dash -- by a single dash
remove_leading_dash(X,RX) :- atom(X), atom_codes(X,[45,45,H|T]),
    H \= 45, atom_codes(RX,[45,H|T]).
recognised_option(X,Opt,Args,Condition) :- remove_leading_dash(X,RX),!,
   recognised_option(RX,Opt,Args,Condition).
recognised_option(X,Opt,[],true) :- recognised_option(X,Opt).
recognised_option(X,Opt,Args,true) :- recognised_option(X,Opt,Args).

recognised_option('-mc',cli_mc(N),[Arg],user:mc_ok_arg(Arg,N)).
recognised_option('-model_check',cli_mc(LimitNr),[],true) :- max_tagged_integer(LimitNr).
recognised_option('-model-check',cli_mc(LimitNr),[],true) :- max_tagged_integer(LimitNr).
recognised_option('-timeout',timeout(N),[Arg],tools:arg_is_number(Arg,N)).
recognised_option('-s',socket(S),[Arg],tools:arg_is_number(Arg,S)).
recognised_option('-cc',coverage(N,N2,just_check_stats),[Arg,Arg2],
                   (arg_is_number_or_wildcard(Arg,N),arg_is_number_or_wildcard(Arg2,N2))).
recognised_option('-hash',check_statespace_hash(H,_),[Arg],tools:arg_is_number(Arg,H)).
recognised_option('-hash64',check_statespace_hash(H,'64bit'),[Arg],tools:arg_is_number(Arg,H)).
recognised_option('-hash32',check_statespace_hash(H,'32bit'),[Arg],tools:arg_is_number(Arg,H)).
recognised_option('-ltllimit',ltl_limit(Nr),[Arg], tools:arg_is_number(Arg,Nr)).
recognised_option('-check_disprover_result',cli_check_disprover_result([true-TNr,false-FNr,unknown-UNr,failure-0]),[T,F,U],
   (arg_is_number_or_wildcard(T,TNr),arg_is_number_or_wildcard(F,FNr),arg_is_number_or_wildcard(U,UNr))).
recognised_option('-aa',cli_check_assertions(all,[true/TNr,false/FNr,unknown/UNr]),[T,F,U],
   (arg_is_number_or_wildcard(T,TNr),arg_is_number_or_wildcard(F,FNr),arg_is_number_or_wildcard(U,UNr))).
recognised_option('-ma',cli_check_assertions(main,[true/TNr,false/FNr,unknown/UNr]),[T,F,U],
   (arg_is_number_or_wildcard(T,TNr),arg_is_number_or_wildcard(F,FNr),arg_is_number_or_wildcard(U,UNr))).
recognised_option('-kodkod_comparision',kodkod_comparision(Nr),[Arg],tools:arg_is_number(Arg,Nr)).
recognised_option('-kodkod_performance',kodkod_performance(File,Nr),[File,Arg],tools:arg_is_number(Arg,Nr)).
recognised_option('-animate',cli_random_animate(N,true),[Steps],tools:arg_is_number(Steps,N)).
recognised_option('-execute',execute(N,true),[Steps],tools:arg_is_number(Steps,N)).
recognised_option('-expcteval',expect_evaluation(List),[Arg],user:evalchecklist(Arg,List)).
recognised_option('-zmq_master2',zmq_master2(N,Max,PortStart,Recover,Logfile),[SN,SMax,SPortStart,SRecover,Logfile],
                  tools:(arg_is_number(SN,N),arg_is_integer(SMax,Max),arg_is_number(SPortStart,PortStart),arg_is_number(SRecover,Recover))).
recognised_option('-zmq_assertions',zmq_assertion(PortStart, Logfile),[SPortStart,Logfile],
                  tools:(arg_is_number(SPortStart, PortStart))).
recognised_option('-zmq_master',zmq_master(Logfile),[Logfile], true).
recognised_option('-zmq_worker2',zmq_worker2(MasterIP, Port, Recover),[MasterIP, SPort, SRecover],
                  tools:(arg_is_number(SPort,Port), arg_is_number(SRecover, Recover))).
recognised_option('-plugin',plugin(Plugin), [Plugin]).
recognised_option('-p',set_pref(PREF,PREFVAL),[PREF,PREFVAL]).
recognised_option('-pref',set_pref(PREF,PREFVAL),[PREF,PREFVAL]).
recognised_option('-pref_group',set_preference_group(Group,Val),[Group,Val]).
recognised_option('-pref',set_pref(PREF,PREFVAL),[PREF,PREFVAL]).
recognised_option('-prefs',set_prefs_from_file(PREFFILE),[PREFFILE]).
recognised_option('-card',set_card(SET,SCOPE),[SET,SCOPE]).
recognised_option('-argv',set_argv(ARGV),[ARGV]).
recognised_option('-goal',set_goal(GOAL),[GOAL]).
recognised_option('-scope',set_searchscope(GOAL),[GOAL]).
recognised_option('-searchscope',set_searchscope(GOAL),[GOAL]).
recognised_option('-eval',eval_string_or_file(string,E,exists,_),[E]).
recognised_option('-evalt',eval_string_or_file(string,E,exists,'TRUE'),[E]).
recognised_option('-evalf',eval_string_or_file(string,E,exists,'FALSE'),[E]).
recognised_option('-eval_file',eval_string_or_file(file,F,exists,_),[F]).
recognised_option('-evalt_file',eval_string_or_file(file,F,exists,'TRUE'),[F]).
recognised_option('-eval_rule_file',eval_string_or_file(file,F,forall,_),[F]).
recognised_option('-parsercp',(L),[L]).
recognised_option('-l',log(Log),[Log]).
recognised_option('-log',log(Log),[Log]).
recognised_option('-lg',analyse_log(Log,timestamp,true),[Log]).
recognised_option('-lgr',analyse_log(Log,revision,false),[Log]).
recognised_option('-check_log',check_log(Log),[Log]).
recognised_option('-expcterr',expect_error(Type),[Type]).
recognised_option('-expecterr',expect_error(Type),[Type]).
recognised_option('-expect',expect_error(Type),[Type]).
recognised_option('-pp',pretty_print_internal_rep(File),[File]).
recognised_option('-ppf',pretty_print_all_types(File),[File]).
recognised_option('-his',history(File),[File]).
recognised_option('-his_option',history_option(Option),[Option]).
recognised_option('-sptxt',print_values(Filename),[Filename]).
recognised_option('-cache',cache_storage(D),[D]).
recognised_option('-show_cache',show_cache,[]).

recognised_option('-MAIN',csp_main(ProcessName),[ProcessName]).

recognised_option('-ltlfile',ltl_file(Filename),[Filename]).
recognised_option('-ltlformula',ltl_forumla_model_check(Formula,_),[Formula]).
recognised_option('-ltlformulat',ltl_forumla_model_check(Formula,true),[Formula]).
recognised_option('-ltlformulaf',ltl_forumla_model_check(Formula,false),[Formula]).
recognised_option('-ctlformula',ctl_forumla_model_check(Formula,_),[Formula]).
recognised_option('-ctlformulat',ctl_forumla_model_check(Formula,true),[Formula]).
recognised_option('-ctlformulaf',ctl_forumla_model_check(Formula,false),[Formula]).
recognised_option('-load_state',load_state(Filename),[Filename]).
recognised_option('-save_state',save_state(Filename),[Filename]).
recognised_option('-save',save_state_for_refinement(Filename),[Filename]).
recognised_option('-refchk',refinement_check(Filename,0,100000),[Filename]).

%recognised_option('-cspref',csp_in_situ_refinement_check(assertRef('False',val_of(AbsP1,no_loc_info_available),Type,val_of(ImplP2,no_loc_info_available),no_loc_info_available),'False'),[AbsP1,Type,ImplP2]).
recognised_option('-cspref',csp_in_situ_refinement_check(AbsP1,Type,ImplP2),[AbsP1,Type,ImplP2]).
% -cspref R [F= Q
recognised_option('-cspdeadlock',csp_checkAssertion(Proc,Model,'deadlock free'),[Proc,Model]).
% -cspdeadlock R F
recognised_option('-cspdeterministic',csp_checkAssertion(Proc,Model,'deterministic'),[Proc,Model]).
% -cspdeterministic R F
recognised_option('-csplivelock',csp_checkAssertion(Proc,'FD','livelock free'),[Proc]).
% -csplivelock R
recognised_option('-cspm_eval', cspm_evaluate_expression(Expr),[Expr]).
recognised_option('-csp_eval', eval_csp_expression(Expr),[Expr]).
recognised_option('-csp_add_unicode', csp_add_unicode(File), [File]).
recognised_option('-csp_remove_unicode', csp_remove_unicode(File), [File]).
recognised_option('-get_csp_assertions_as_string',csp_get_assertions,[]).
recognised_option('-prologOut',csp_translate_to_file(PlFile),[PlFile]).
recognised_option('-csp-guide',add_csp_guide(Filename),[Filename]).
recognised_option('-csp_guide',add_csp_guide(Filename),[Filename]).
recognised_option('-csp_event',set_csp_event_as_visible,[]). % used only with the -csp-guide option
recognised_option('-get_min_max_coverage',get_min_max_coverage(FileName),[FileName]).
recognised_option('-get_coverage_information',get_coverage_information(FileName),[FileName]).
recognised_option('-vacuity_check',vacuity_check,[]).
recognised_option('-cbc',constraint_based_check(OpName),[OpName]).
recognised_option('-cbc_deadlock',cbc_deadlock_check(true),[]).
recognised_option('-cbc_assertions',cbc_assertions(true,[]),[]).
recognised_option('-cbc_assertions_proof',cbc_assertions(false,[]),[]). % do not allow enumeration warnings
recognised_option('-cbc_assertions_tautology_proof',cbc_assertions(false,[tautology_check]),[]). % do not allow enumeration warnings + disregard PROPERTIES
recognised_option('-cbc_result_file',cbc_result_file(FILE),[FILE]). % write result to FILE
recognised_option('-cbc_refinement',cbc_refinement,[]).
recognised_option('-cbc_deadlock_pred',cbc_deadlock_check(GoalPred),[GoalPred]).
recognised_option('-cbc_sequence',cbc_sequence(OpSequence,'',single_solution),[OpSequence]).
recognised_option('-cbc_sequence_all',cbc_sequence(OpSequence,'',findall),[OpSequence]).
recognised_option('-cbc_sequence_with_target',cbc_sequence(OpSequence,TargetPredString,single_solution),[OpSequence,TargetPredString]).
recognised_option('-cbc_sequence_with_target_all',cbc_sequence(OpSequence,TargetPredString,findall),[OpSequence,TargetPredString]).
recognised_option('-comment',comment(UserComment),[UserComment]). % not processed by tool, but will be stored in log-file and used by log_analyser
recognised_option('-junit',junit(Dir),[Dir]).
recognised_option('-mcm_tests', mcm_tests(Depth,MaxStates,EndPred,Output),[Depth,MaxStates,EndPred,Output]).
recognised_option('-mcm_cover', mcm_cover(Event),[Event]).
recognised_option('-cbc_tests', cbc_tests(Depth,EndPred,Output),[Depth,EndPred,Output]).
recognised_option('-cbc_cover', cbc_cover(Event),[Event]).
recognised_option('-cbc_cover_match', cbc_cover(match_event(Event)),[Event]). % find events which have Event String occuring somewhere in name
recognised_option('-cbc_cover_all', cbc_cover_all,[]).  % is now default if no cbc_cover provided
recognised_option('-cbc_cover_final', cbc_cover_final,[]).
recognised_option('-bmc', cbc_tests(Depth,'#not_invariant',''),[Depth]).
recognised_option('-test_description', test_description(File), [File]).
recognised_option('-all_paths', all_deadlocking_paths(File), [File]).
recognised_option('-spdot',state_space_dot(File),[File]).
recognised_option('-spmdot',state_space_merge_dot(File),[File]).
recognised_option('-spddot',state_space_dfa_dot(File),[File]).
recognised_option('-sgdot',state_as_graph_dot(File),[File]).
recognised_option('-sptdot',transition_diagram_dot(Expr,File),[Expr,File]).
recognised_option('-invdot',generate_dot_from_invariant(File),[File]).
recognised_option('-propdot',generate_dot_from_properties(File),[File]).
recognised_option('-assdot',generate_dot_from_assertions(File),[File]).
recognised_option('-deaddot',generate_dot_from_deadlock_po(File),[File]).
recognised_option('-dot_output',dot_analyse_output_prefix(Path),[Path]).
recognised_option('-csvhist',evaluate_expression_over_history_to_csv_file(Expr,File),[Expr,File]).
recognised_option('-evaldot',evaldot(File),[File]).
recognised_option('-enabling_analysis',enabling_analysis_csv(user_output),[]).
recognised_option('-enabling_analysis_csv',enabling_analysis_csv(EnablingCsvFile),[EnablingCsvFile]).
recognised_option('-feasibility_analysis',feasibility_analysis_csv(user_output),[]).
recognised_option('-feasibility_analysis_csv',feasibility_analysis_csv(EnablingCsvFile),[EnablingCsvFile]).
recognised_option('-state_trace',state_trace(File),[File]).
recognised_option('-typecheckertest',typechecker_test(File),[File]).
recognised_option('-scc_trace',check_scc_for_ltl_formula(LtlFormula,SCC),[LtlFormula,SCC]).
recognised_option('-selfcheck_module',selfcheck(M),[M]).

recognised_option('-dot_all',dot_generate_for_all_formulas). % generate dot also for true formulas
recognised_option('-animate_all',cli_random_animate(2147483647,false)).
recognised_option('-execute_all',execute(2147483647,false)).
recognised_option('-animate_stats',animate_stats).
recognised_option('-check_goal',check_goal).
recognised_option('-check_goal',check_goal).
recognised_option('-ltlassertions',ltl_assertions).
recognised_option('-assertions',cli_check_assertions(all,[false/0,unknown/0])).
recognised_option('-main_assertions',cli_check_assertions(main,[false/0,unknown/0])).
recognised_option('-properties',cli_check_properties).
recognised_option('-selfcheck',selfcheck(_)).
recognised_option('-pacheck',pa_check).
recognised_option('-det_check',det_check). % check if animation is deterministic
recognised_option('-det_constants',det_constants_check). % check if animation for setup_constants is deterministic
recognised_option('-bf',breadth_first).
recognised_option('-breadth',breadth_first).
recognised_option('-df',depth_first).
recognised_option('-depth',depth_first).
recognised_option('-strict',strict_raise_error).
recognised_option('-silent',silent).
recognised_option('-statistics',statistics).
recognised_option('-profile_statistics',profile_statistics).
recognised_option('-nodead',no_deadlocks).
recognised_option('-noinv',no_invariant_violations).
recognised_option('-nogoal',no_goal).
recognised_option('-noass',no_assertion_violations).
%recognised_option('-POR',with_reduction).
recognised_option('-i',animate).
recognised_option('-repl',eval_repl). % used to be -eval
recognised_option('-c',coverage(false)).
recognised_option('-cs',coverage(just_summary)).
recognised_option('-coverage',coverage(false)).
recognised_option('-coverage_summary',coverage(just_summary)).
recognised_option('-machine_stats',cli_print_machine_statistics).
recognised_option('-vc',variable_coverage).
recognised_option('-variable_coverage',variable_coverage).
recognised_option('-cv',coverage(true)).
recognised_option('-v',verbose).
recognised_option('-vv',very_verbose).
recognised_option('-verbose',verbose).
recognised_option('-very_verbose',very_verbose).
recognised_option('-profiling_on',profiling_on). % Prolog profiling
recognised_option('-profile',prob_profile). % ProB Operation profiling
recognised_option('-version',print_version).
recognised_option('-svers',print_short_version).
recognised_option('-short_version',print_short_version).
recognised_option('-file_info',file_info).
recognised_option('-t',trace_check).
recognised_option('-init',initialise).
recognised_option('-initialise',initialise).
recognised_option('-ll',log('/tmp/prob_cli_debug.log')).
recognised_option('-ss',socket(9000)).
recognised_option('-sf',socket(_)).
recognised_option('-help',help).
recognised_option('-h',help).
recognised_option('-rc',runtimechecking).
recognised_option('-test_mode',test_mode).
recognised_option('-check_complete',check_complete).
recognised_option('-check_complete_operation_coverage', check_complete_operation_coverage).
recognised_option('-mc_with_tlc', cli_start_mc_with_tlc).
recognised_option('-symbolic_model_check', cli_symbolic_model_check).
%recognised_option('-csp',csp_pl_mode).
%recognised_option('-cspm',csp_mode).

recognised_option('-zmq_worker',zmq_worker).

% some utilities to be able to call the above options directly from repl:
verbose :- tcltk_turn_debugging_on(19).
very_verbose :- tcltk_turn_debugging_on(5).
file_info :- file_loaded(true,MainFile), print_file_info(MainFile).
coverage(ShowEnabledInfo) :- probcli_time_stamp(NOW), cli_show_coverage(ShowEnabledInfo,NOW).
variable_coverage :- probcli_time_stamp(NOW), cli_show_variable_coverage(NOW).
silent :- (option(silent) -> true ; assert(option(silent))).

option_verbose :- (option(verbose) -> true ; option(very_verbose)).

% new profiler
%:- use_module(('../extensions/profiler/profiler.pl').
%prob_profiler :- pen.

% old profiler
:- use_module(runtime_profiler,[print_runtime_profile/0]).
prob_profile :- print_runtime_profile.

% convert a string of the form "+x,-y,z" to a list [evaluated(x),not_evaluated(y),evaluated(z)]
evalchecklist(Arg,List) :-
    split_atom(Arg,[',',' '],AList),
    evalchecklist2(AList,List).
evalchecklist2([],[]).
evalchecklist2([Atom|Arest],[Term|Trest]) :-
    atom_chars(Atom,Chars),Chars=[C|Crest],
    evalchecklist3(C,Crest,Term),
    evalchecklist2(Arest,Trest).
evalchecklist3(+,Crest,evaluated(Pred)) :- !,atom_chars(Pred,Crest).
evalchecklist3(-,Crest,not_evaluated(Pred)) :- !,atom_chars(Pred,Crest).
evalchecklist3(C,Crest,evaluated(Pred)) :- atom_chars(Pred,[C|Crest]).

set_random_seed_to_prologs_current_random_state :-
     getrand(CurrState),setrand(CurrState).

set_new_random_seed :-
   now(TimeStamp), % getting the unix time
   setrand(TimeStamp). % setting new random seed by every execution of probcli

halt_exception :- halt_exception(0).
halt_exception(Code) :- throw(halt(Code)).



runtime_entry(start) :- go_cli.

save :- save_program('probcli.sav').
