% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 



:- module(prob_socketserver, [start/1,set_log_file/1,
                              writeln_log/1, writelnstr_log/1,
                              writeln_log_time/1]).

:- use_module(version).
:- use_module(eclipse_interface).
:- use_module(prob2_interface).
:- use_module(extension('user_signal/user_signal'),[get_user_signal_ref/1]).
:- use_module(library(sockets)).
:- use_module(library(lists)).

:- op(1150, fx, type).

prolog_socket :- start(_).

:- dynamic socket_opened/1.

start(Port) :-
	on_exception(E, start2(Port),
	    ( E=halt(C) -> throw(halt(C))
        ; otherwise ->
            my_format_error("Exception in Socket Server: ~w~n",[E]),fail)),
	!.
start(Port) :- my_format_error("Listening to Port failed: ~w~n",[Port]),
   retract(socket_opened(Socket)),
   my_format("Attempting to close socket: ~w~n",[Socket]),
   socket_server_close_sp4(Socket),
   my_format("Socket closed: ~w~n",[Socket]).
    

socket_server_open_sp4(Port,Socket) :-
     socket_server_open(Port,Socket,[loopback(true)])
      -> true
	  ; (my_format_error("Unable to bind sp4 socket: ~w to localhost:~w.~n",[Socket,Port]),fail).
	
	
socket_server_close_sp4(Socket) :- socket_server_close(Socket).

socket_accept_sp4(Socket, Stream) :- socket_server_accept(Socket, Client, Stream, [type(text)]),
   my_format("Connected: ~w~n",[Client]).

start2(Port) :-
	% print(socket('AF_INET', Socket)),nl,
	socket_server_open_sp4(Port,Socket),
    assert(socket_opened(Socket)),
	%open(prologport, write,Stream),
	%format(Stream,'~w~n',[Port]),
	%close(Stream),
	prolog_flag(user_output, Stdout, Stdout),
	format(Stdout,'Port: ~w~n', [Port]),
    writeln_log(port(Port)),
    revision(Revision),
    format(Stdout,'probcli revision: ~w~n',[Revision]),
    get_user_signal_ref(Ref),
    format(Stdout,'user interrupt reference id: ~w~n',[Ref]),
    format(Stdout,'-- starting command loop --~n', []),
	flush_output(Stdout),
	wait(Socket),
	socket_server_close_sp4(Socket),!,
    my_format("Socket closed: ~w~n",[Socket]),
    writeln_log(socket_closed(Socket)).
  

%% Wait for a connection on the given socket
wait(Socket) :-
	socket_accept_sp4(Socket, Stream),
	read_socket(Socket, Stream).

%% We have a connection so just wait for valid terms, and return either
%% OK X=val\000Y=val1
%% FAIL\000
%% ERROR SHORT long message here\000
read_socket(Socket,Stream) :-	
   %print(listening_on_socket(Socket)),nl, flush_output(user_output),
	on_exception(E,
		     read_term(Stream, Term, [variable_names(Vars)]),
		     (portray_clause(user_output,goal_parse_error),Parse=fail,exception(E,Stream))),
	
	%print(received_term(Term)),nl, flush_output(user_output),
	(Parse=ok ->
	    (
	      ((Term == end_of_file ; Term==halt) ->
		  (
		    socket_server_close_sp4(Socket),
		    throw(halt(0))
		  )
	      ;
   		  %portray_clause(user_output,calling(Term)),
		  make_call(Stream, Term, Vars)
		  %portray_clause(user_output,exit(Term))
	      )
	    )	    
	;
	    (
	      %% Parse failed, but we have already written error message
	      %print(an_exception_occurred(E)),nl,
	      portray_clause(user_output, an_exception_occurred(E))
	    )
	),
	%print(done(Term)),nl,
	%flush_output(user_output),
	%flush_output,
	%put_code(Stream, 1),
	
	flush_output(Stream),
	flush_output(user_output),
	read_socket(Socket, Stream).
	

	


%% We have a valid term so make the call
make_call(Stream,Term, Vars) :-
	% write(make_call(Term)),nl,flush_output(user_output),
	on_exception(CallEx,
		       do_call(Term,Vars,Result),
	           (portray_clause(user_output, exception(make_call/3, CallEx)),
		    Result=exception(CallEx)
	           )),
	write_canonical(Stream,Result),
	nl(Stream), put_code(Stream,1),
	flush_output(Stream).

do_call(Call,Vars,yes(Vars)) :- %print(do_call(Call)),nl,flush_output(user_output),
     writeln_log(do_call(Call)),
     (white_list(Call)
       -> call(Call)
       ;  my_format_error('Call not on whitelist: ~w.\n',[Call]),fail
     ),
     ((ground(Vars);Call=debug_console(_,_)) ->
	 % print(success(Call)),nl,flush_output(user_output),
	 writeln_log(success(Call));

	 raise_exception(result_is_not_ground)),
%     \+ next_char_is_semicolon(Stream), /*Otherwise force backtracking */
     !. 
do_call(_Call,_Vars,no) :-  
     writeln_log(no).

:- dynamic exported_by_eclipse_interface/2.
white_list(halt). % needed to stop ProB
white_list(true). % to prevent fails because of empty queries
white_list((A,B)) :- white_list(A),white_list(B). % concatenation used to send several commands at once
white_list(C) :- nonvar(C),functor(C,Functor,Arity), % all public predicates of the module 
    is_exported_by_eclipse_interface(Functor,Arity). % eclipse_interface can be used
is_exported_by_eclipse_interface(Functor,Arity) :-
    exported_by_eclipse_interface(Functor,Arity),!.
is_exported_by_eclipse_interface(Functor,Arity) :-
    % check that the list has not already been generated
    \+ exported_by_eclipse_interface(_,_),
    % generate the list
    create_exported_by_eclipse_interface_list,
    % do the check again
    exported_by_eclipse_interface(Functor,Arity).
create_exported_by_eclipse_interface_list :-
    % fail-driven loop over the predicates imported from eclipse_interface
    predicate_property(prob_socketserver:P,imported_from(eclipse_interface)),
    functor(P,Functor,Arity),
    assert(exported_by_eclipse_interface(Functor,Arity)),
    fail.
create_exported_by_eclipse_interface_list :-
    % fail-driven loop over the predicates imported from prob2_interface
    predicate_property(prob_socketserver:P,imported_from(prob2_interface)),
    functor(P,Functor,Arity),
    assert(exported_by_eclipse_interface(Functor,Arity)),
    fail.
create_exported_by_eclipse_interface_list.


next_char_is_semicolon(Stream) :- %print(peeking(Stream)),nl,
    peek_code(Stream,X),!,
    ((X=10;X=13)
      -> (get_code(Stream,_),next_char_is_semicolon(Stream))
      ; (X=59,get_code(Stream,_))).
    
	    
get_stream_info(S,Info) :-
	on_exception(E,get_stream_info1(S,Info),(portray_clause(user_output, exception(E)))).

get_stream_info1(S,blank):-
	portray_clause(user_output, stream_is(S)),
	%true.
	fail.
get_stream_info1(S, Info) :-
	stream_property(S,file_name(Info)),!.

get_stream_info1(S, Info) :-
	
	stream_property(S,alias(Info)),
	!.
get_stream_info1(_S, unknown).


tokens_to_string([],[]).
tokens_to_string([A|As], [B|Bs]) :-
	token_to_string1(A,B),
	tokens_to_string(As,Bs).


token_to_string1(atom(A)-_, A) :- !.
token_to_string1(var(_,S,_)-_, A) :-
	!,
	name(A,S).


token_to_string1(A, A).


%% Exception handlers
%% First Argument is Sicstus exception, second is verbose message for user
exception(E,_Stream) :-
 portray_clause(user_output,E),fail.


exception(syntax_error(Call, Pos, Msg, Tokens,Lineend),Stream) :-
	!,
	portray_clause(user_output,syntax_stream(Call)),
	(Call = read_term(Err,_,_) ->
	    get_stream_info(Err,Info)
	;
	    Info = unknown_stream
	),
	tokens_to_string(Tokens,Where),

	my_format_error(Stream,'exception(ERR-~@PARSE~@~w in ~w\nlines::~w-~w\nCheck ~w)\n', [put_code(1),put_code(1),Msg,Info,Pos,Lineend,Where]),
	put_code(Stream, 1).

exception(existence_error(_, _, _, _, past_end_of_stream),Stream) :-
	!,
	my_format_error(Stream,'exception(ERR-~@EXISTENCE~@Attempted to read past end of stream)', [put_code(1),put_code(1)]),
	put_code(Stream, 1)	.
	
exception(existence_error(_, _,procedure,PredSpec,_),Stream) :-
	!,	
	my_format_error(Stream,'exception(ERR-~@EXISTENCE~@Unable to call predicate ~w)', [put_code(1),put_code(1),PredSpec]),
	put_code(Stream, 1)	.

exception(existence_error(_, _,file,PredSpec,_),Stream) :-
	!,	
	my_format_error(Stream,'exception(ERR-~@EXISTENCE~@Unable to open file ~w)', [put_code(1),put_code(1),PredSpec]), 
	put_code(Stream, 1)	.

exception(E,Stream) :-
	my_format_error(Stream,'exception(ERR-~@UNKNOWN~@~w)', [put_code(1),put_code(1),E]),
	put_code(Stream, 1),
	portray_clause(user_output,unknown_exception(E)).


:- dynamic logfile/1.
set_log_file(F) :- retractall(logfile(_)), assert(logfile(F)).

my_format(FS,Args) :- my_format(user_output,FS,Args).
my_format_error(FS,Args) :- my_format(user_error,FS,Args).
my_format_error(Stream,FS,Args) :- my_format(Stream,FS,Args).    
    
my_format(Stream,FS,Args) :- format(Stream,FS,Args),
    (logfile(F)
      -> (open(F,append,S),
          format(S,FS,Args),
          close(S))
     ; true
    ).


writelnstr_log(Str) :- 
    (logfile(F)
      -> (open(F,append,S),
          write(S,'% '),write(S,Str),nl(S),
          close(S))
     ; true
    ).
writeln_log(Term) :- 
    (logfile(F)
      -> (open(F,append,S),
          write_term(S,Term,[quoted(true)]), write(S,'.'),nl(S),
          close(S))
     ; true
    ).

writeln_log_time(Term) :- Term=..[H|Args],
    (logfile(_) ->
      statistics(runtime,[Time,_]),
      statistics(walltime,[WTime,_]),
      append(Args,[Time,WTime],NArgs),
      NT =.. [H|NArgs],
      writeln_log(NT)
    ; true).

%runtime_entry(start) :- go_cli.

%save :- save_program('probcli.sav').


     
