% (c) 2014 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(prologTasks,[prologTaskStart/3, prologTaskStep/3, prologTaskAbort/1, prologTaskFinish/1]).

:- use_module(error_manager).
:- use_module(tools).
:- use_module(library(lists)).



:- use_module(module_information,[module_info/2]).
:- module_info(group,tcltk).
:- module_info(description,'This module provides the interface to the Tcl/Tk tree_inspector.').

:- dynamic task_step/3.

prologTaskStart(test,'Test Task',1000) :- !.
prologTaskStart(Category,CatMsg,Length) :-
    get_category_name(Category,CatMsg),
    retractall(task_step(Category,_,_)),
    % collect all task steps:
    findall(task_step(Category,Prolog,Msg),get_task_steps(Category,Prolog,Msg),Tasks),
    %print(task_steps(Category,Tasks)),nl,
    length(Tasks,Length),
    maplist(assert,Tasks),nl.


prologTaskStep(test,Nr,NewMsg) :- !,
   tools:ajoin(['Running ', Nr],NewMsg).
prologTaskStep(Category,Nr,NewMsg) :-  %print(step(Nr)),nl,
    retract(task_step(Category,PrologCode,Msg)),
    % print(calling(PrologCode)),nl,
    statistics(walltime,[Start,_]),
    call(PrologCode),
    statistics(walltime,[Stop,_]),
    T is Stop-Start,
    % print(done(Msg)),nl,
    (get_task_info(Category,Nr,Info)
     -> tools:ajoin(['Step ', Nr, ': ', Msg, ' [', Info, '] (', T, ' ms)'],NewMsg)
     ;  tools:ajoin(['Step ', Nr, ': ', Msg, ' (', T, ' ms)'],NewMsg)
    ).
% TO DO: optionally return other stats : nr of test-cases, nr of feasible/infeasbile ops


prologTaskAbort(Category) :- retract(task_step(Category,_PrologCode,Msg)),
    %add_internal_error('Pending task: ',task_step(Category,PrologCode,Msg)),
    print(aborting(Msg)),nl,
    fail.
prologTaskAbort(_).

% prologTaskFinish will also be called in case of Abort

prologTaskFinish(cbc_tests(_EventSel,_TargetPredicateStr,_MaxLength,_Filename,_Final)) :- !,
    sap:cbc_gen_test_cases_finish_tasks.
prologTaskFinish(_).

:- use_module(sap).
:- use_module(enabling_analysis).
:- use_module(bmachine, [b_top_level_operation/1]).
:- use_module(state_space, [operation_name_not_yet_covered/1]).
:- use_module(tools, [ajoin/2]).


get_category_name(feasibility,'Feasibility Analysis').
get_category_name(enabling(_),'Enabling Analysis').
get_category_name(cbc_tests(_,_,_,_,_),'CBC Testcase Generation').

get_task_info(cbc_tests(_,_,_,_,_),Nr,Info) :-
    N1 is Nr+1,
    sap:get_counter_info(N1,Paths,Tests,Timeouts),
    sap:cbc_get_nr_uncovered_events(Unc),
    tools:ajoin(['Uncovered: ',Unc, ', Paths: ',Paths,', Tests: ',Tests,', Timeouts: ',Timeouts],Info).

% maybe we should move this to the corresponding modules, which register their tasks
get_task_steps(feasibility,PrologCode,Msg) :-
   b_top_level_operation(OpName),
   operation_name_not_yet_covered(OpName),
   \+ feasible_operation_cache(OpName,_), % not yet computed
   PrologCode = enabling_analysis:feasible_operation(OpName,_), % will compute the cache
   tools:ajoin(['feasibility of ',OpName],Msg).
get_task_steps(enabling(ExtraTimeOut),PrologCode,Msg) :- %print(get_task(ExtraTimeOut)),nl,
   enabling_analysis:init_or_op(OpName), %print(op(OpName)),nl,
   \+ enabling_analysis:cbc_enable_analysis_cache(OpName,_,_,ExtraTimeOut), % not yet computed
   PrologCode = findall(1,enabling_analysis:cbc_enable_analysis(OpName,_,_,ExtraTimeOut),_),
   tools:ajoin(['enabling of ',OpName],Msg).
get_task_steps(cbc_tests(EventSel,TargetPredicateStr,MaxLength,Filename,Final),PrologCode,Msg) :-
   sap:cbc_gen_test_cases_task_step(EventSel,TargetPredicateStr,MaxLength,Filename, Final, Depth, PrologCode),
   tools:ajoin(['cbc_tests for depth ',Depth],Msg).

