/* 
 * created by dewin dennis.winter@uni-duessldorf.de 2008
 */

:- module(h_int, [promela_transition/3, promela_property/2, open_promela_file/1,eval/4]).
:- use_module(library(codesio)).
:- use_module(library(samsort)).
:- use_module(library(avl)).

:- use_module(probsrc(module_information)).
:- module_info(group,promela).
:- module_info(description,'This is the Promela (the Spin model checker language) extension.').

:- dynamic inst/2, line/5.

:- volatile current_promela_file_opened/1.
:- dynamic current_promela_file_opened/1.

open_promela_file(File) :- 
    print(open_promela_file(File)),nl,
    consult_without_redefine_warning(File),
    retractall(current_promela_file_opened(_)),
    assert(current_promela_file_opened(File)),
    print(new_xtl_file(File)),nl.

consult_without_redefine_warning(File) :-
    prolog_flag(redefine_warnings, Old, off),
    prolog_flag(single_var_warnings, Old2, off),
    (consult(File)
    -> OK=true ; OK=false),
    prolog_flag(redefine_warnings, _, Old),
    prolog_flag(single_var_warnings, _, Old2),
    OK=true.

promela_transition(root, start_Promela, X) :- start(X).
promela_transition(S, L, S2) :- trans(L,S,S2).
		
% --------------------------------------------------------

start(Out) :-
    print('promela interpreter version 1.5, created by dennis winter'),nl,
    empty_avl(Vars),
    empty_avl(Procs),
    empty_avl(Chans),
    define(env(Vars,Procs,Chans), var(sys(pid),0), global, In2),
    define(In2, var(sys(chan_id),0), global, In3),
    define(In3, var(sys(unsafe),false), global, In4),
    define(In4, var(sys(mode),normal), global, In5),
    init(1,PC2,In5,In6),
    init2(PC2,In6,Out).

init(PC, PC2, In, Out) :-
    inst(PC,S),
    %print(inst(PC,S)),nl,
    handle(S,In,In2),
    NewPC is PC+1,
    init(NewPC,PC2,In2,Out),!.

init(PC,PC,In,In).
	

% define array variable
get_init_value(def(Type,array(Name,K)), Value, In, Out) :-
    ( Type=bool 
      -> generate_false(K,Value), In=Out;  
      Type=chan
      -> generate_chanid_array(K, Value, In, Out);
      (Type=bit; Type=byte;
       Type=pid; Type=short; Type=int;
       Type=unsigned; Type=mtype)
      -> generate_zero(K,Value), In=Out;
      %else: user-defined type
      generate_usertype_array(K, Type, Value, In, In2),
      define(In2, var(type(Name),Type), global, Out) ),!.

% define variable
get_init_value(def(Type,Name), Value, In, Out) :-
    ( Type=bool
      -> Value=false, In=Out;
      Type=chan
      -> generate_chanid(Value, In, Out);
      (Type=bit; Type=byte;
       Type=pid; Type=short; Type=int;
       Type=unsigned; Type=mtype)
      -> Value=0, In=Out;
      %else: user-defined type
      generate_usertype(Type, 0, [], Value, In, In2),
      define(In2, var(type(Name),Type), global, Out) ).

% generate channel-id array
generate_chanid_array(0, [], In, In) :- !.
generate_chanid_array(K, [CID|Tail], In, Out) :-
    generate_chanid(CID, In, In2),
    K2 is K-1,
    generate_chanid_array(K2, Tail, In2, Out).
    
% generate channel-id
generate_chanid(CID, In, Out) :-
    lookup(var(sys(chan_id)), global, In, CID),
    CID2 is CID+1,
    define(In, var(sys(chan_id),CID2), global, In2),
    store(In2, chan(CID,empty), global, Out),!.

% generate usertype array
generate_usertype_array(0, _, [], In, In) :-!.
generate_usertype_array(K, Type, [Value|Tail], In, Out) :-
    generate_usertype(Type, 0, [], Value, In, In2),
    K2 is K-1,
    generate_usertype_array(K2, Type, Tail, In2, Out).

% generate usertype
generate_usertype(Type, N, ValueIn, ValueOut, In, Out) :-
    inst(_,typedef(Type,N,def(EType,EName))),
    get_init_value(def(EType,EName), Value, In, In2),
    append(ValueIn, [Value], ValueIn2),
    N2 is N+1, 
    generate_usertype(Type, N2, ValueIn2, ValueOut ,In2, Out),!.

generate_usertype(_, _, Value, Value, In, In).

% define globals
handle(def(Type,Name), In, Out) :-
    trans3(def(Type,Name), global, In, Out).

% typedef
handle(typedef(_,_,_),In,In).

% assign
handle(assign(mtype,Value), In, Out) :-
    define(In, var(mtype,Value), global, Out),!.

handle(assign(Name,Value), In, Out) :-
    trans3(assign(Name,Value), global, In, Out).

% stop global initialisation here.
handle(proctype(_,_,_), _, _) :- fail.


set_record_element(Type, [array(EName,expr(K))], Value, V, PID, In, NewValue) :-
    inst(_, typedef(Type,X,def(_,array(EName,_)))),
    get_element(X, Value, Value2),
    eval(K, PID, In, K2),
    set_element(K2, Value2, V, Value3),
    set_element(X, Value, Value3, NewValue),!.
set_record_element(Type, [EName], Value, V, _, _, NewValue) :-
    inst(_, typedef(Type,X,def(_,EName))),
    set_element(X, Value, V, NewValue),!.
set_record_element(Type, [array(EName,expr(K))|T], Value, V, PID, In, NewValue) :-
    inst(_, typedef(Type,X,def(EType,array(EName,_)))),
    get_element(X, Value, Value2),
    eval(K, PID, In, K2),
    get_element(K2, Value2, V2),
    set_record_element(EType, T, V2, V, PID, In, V3),
    set_element(K2, Value2, V3, Value3),
    set_element(X, Value, Value3, NewValue),!.
set_record_element(Type, [EName|T], Value, V, PID, In, NewValue) :-
    inst(_, typedef(Type,X,def(EType,EName))),
    get_element(X, Value, V2),
    set_record_element(EType, T, V2, V, PID, In, V3),
    set_element(X, Value, V3, NewValue).


init2(PC, In, Out) :-
    inst(PC,S),
    %print(inst(PC,S)),nl,
    handle2(S,PC,In,In2),
    NewPC is PC+1,
    init2(NewPC,In2,Out),!.

init2(_,In,In).

    
handle2(proctype(active(X),Proc,_), PC, In, Out) :-
    NewPC is PC+1,
    lookup(var(sys(pid)), global, In, PID),
    store(In, proc(active(PID,Proc),NewPC), global, In2),
    NewPID is PID+1,
    define(In2, var(sys(pid),NewPID), global, In3),
    (X =< 1
     -> In3=Out
     ;  NewX is X-1,
        handle2(proctype(active(NewX),Proc,_), PC, In3, Out)
    ),!.

handle2(proctype(inactive,_,_), _, In, In).

handle2(init, PC, In, Out) :-
    NewPC is PC+1,
    lookup(var(sys(pid)), global, In, PID),
    store(In, proc(active(PID,init),NewPC), global, In2),
    NewPID is PID+1,
    define(In2, var(sys(pid),NewPID), global, Out).
    
handle2(_,_,In,In).

/********************************************************************/

%para_func(def(_,Name),Name).

%map([], _, []).
%map([H|T], P, [H2|T2]) :-
%    Call =.. [P,H,H2], call(Call), map(T,P,T2).


% for type bool or mtype hone_value() must be called every time.
% for all other types, hone_value() must only be called in case of an overflow.
check_type(Name,Type,Value,NewValue,In) :-
	( check_type2(Type,Value,Value2,In) 
	  -> ( (Type == bool; Type == mtype)
	       -> hone_value(Type,Value2,NewValue,In)
		   ;  NewValue = Value2 )
	  ;  print('Warning: Overflow of '),print(variable(Name,Type,Value)),
	     hone_value(Type,Value,NewValue,In),
	     print(', '),print(new_value(NewValue)),print('.'),nl ).

check_type2(bit, Value, NewValue, _) :-
	number(Value),
	NewValue is round(Value),
	(NewValue == 0; NewValue == 1),!.

check_type2(bool, Value, NewValue, _) :-
	number(Value),
	NewValue is round(Value),
	(NewValue == 0; NewValue == 1),!.

check_type2(byte, Value, NewValue, _) :-
	number(Value),
	NewValue is round(Value),
	NewValue >= 0, 
	NewValue =< 255,!.

check_type2(pid, Value, NewValue, _) :-
	number(Value),
	NewValue is round(Value),
	NewValue >= 0, 
	NewValue =< 255,!.

check_type2(short, Value, NewValue, _) :-
	number(Value),
	NewValue is round(Value),
	NewValue >= -32768, 
	NewValue =< 32767,!.

check_type2(int, Value, NewValue, _) :-
	number(Value),
	NewValue is round(Value),
	NewValue >= -2147483648, 
	NewValue =< 2147483647,!.

check_type2(mtype, Value, Value2, In) :-
	number(Value),
	Value2 is round(Value),
	lookup(var(mtype), global, In, MList),
	length(MList, N),
	Value2 >= 1,
	Value2 =< N,!.

check_type2(chan, Value, _) :-
	% what's the maximum chan_id?
	number(Value),
	print('chan typecheck. Use define(In var(chan(..)..) instead. chan_id: '),print(Value),nl.

hone_value(bit, Value, NewValue, _) :-
	NewValue is Value mod 2,!.

hone_value(bool, Value, NewValue, _) :-
	Value2 is Value mod 2,
	( false(Value2)
	  -> NewValue = false
	  ;  NewValue = true ),!.

hone_value(byte, Value, NewValue, _) :-
	NewValue is Value mod 256,!.

hone_value(pid, Value, NewValue, _) :-
	NewValue is Value mod 256,!.

hone_value(short, Value, NewValue, _) :-
	( Value > 32767
	  -> Value2 is Value mod 32768,
	     NewValue is Value2 - 32768 ),
    ( Value < -32768
	  -> Value2 is Value mod 32769,
	     NewValue is Value2 - 1 ),!.

hone_value(int, Value, NewValue, _) :-
	( Value > 2147483647
	  -> Value2 is Value mod 2147483648,
	     NewValue is Value2 - 2147483648 ),
    ( Value < -2147483648
	  -> Value2 is Value mod 2147483649,
	     NewValue is Value2 - 1 ),!.

hone_value(mtype, Value, ctype(NewValue), In) :-
	lookup(var(mtype), global, In, MList),
	length(MList, N),
	Value2 is Value -1,
	Value3 is Value2 mod N,
	Value4 is Value3 + 1,
	K is N - Value4,
	get_element(K, MList, NewValue),!.

hone_value(chan, Value, Value, _).

/********************************************************************/

% store variable
% (stores singles, arrays, records; all of them as global or as local variable)
store(env(Vars,Procs,Chans), var(Name,Type,expr(V)), PID, env(NewVars,Procs,Chans)) :-
    % set Name4
    ( Name = record([Name2|Tail]) -> true; Name2 = Name ),
    ( Name2 = array(Name3,expr(K)) -> true; Name3 = Name2 ),
    ( (PID \= global, lookup2(key(Name3,PID),Vars,_)) 
      -> Name4 = key(Name3,PID)
      ;  Name4 = Name3 ),
    lookup2(Name4, Vars, Value),
    eval(V, PID, env(Vars,Procs,Chans), V1),
	check_type(Name,Type,V1,V2,env(Vars,_,_)),
    % set Value4
    ( Name2 = array(_,_) 
      -> eval(K,PID,env(Vars,Procs,Chans),K2), get_element(K2,Value,Value2)
      ;  Value2 = Value ),
    ( Name = record(_) 
      -> lookup2(type(Name3),Vars,DType), set_record_element(DType,Tail,Value2,V2,PID,env(Vars,_,_),Value3)
      ;  Value3 = V2 ),
    ( Name2 = array(_,_)
      -> set_element(K2,Value,Value3,Value4)
	% Value = [_|_]: for example short c[3] = 3, then set all elements of c to 3.
	  ;  Name \= record(_), Value = [_|_]
	     -> set_all_elements(Value,V2,Value4)
		 ;  Value4 = Value3 ),
	%print(end_store(Name,Type,expr(V))),nl,
    store2(Vars,Name4,Value4,NewVars).

% store global proctype
store(env(Vars,Procs,Chans), proc(PID,PC), global, env(Vars,NewProcs,Chans)) :-
    store2(Procs, PID, PC, NewProcs).

% store global channel
store(env(Vars,Procs,Chans), chan(Name,Value), _, env(Vars,Procs,NewChans)) :-
    store2(Chans, Name, Value, NewChans).


% define local variable
define(env(Vars,Procs,Chans), var(Name,Value), PID, env(NewVars,Procs,Chans)) :-
    PID \= global,
    store2(Vars, key(Name,PID), Value, NewVars),!.

% define global variable
define(env(Vars,Procs,Chans), var(Name,Value), global, env(NewVars,Procs,Chans)) :-
    store2(Vars, Name, Value, NewVars).


%store2([],Key,Value,[Key/Value]) :-
%    print(store_new_var(Key,Value)),nl.
%store2([Key/_|T],Key,Value,[Key/Value|T]).
%store2([Key2/Value2|T],Key,Value,[Key2/Value2|BT]) :-
%   Key \= Key2, store2(T,Key,Value,BT).

store2(AVL,Key,Value,NewAVL) :-
    avl_store(Key, AVL, Value, NewAVL).

% lookup variable
% (looks up singles, arrays, records; all of them as global or as local variable)
lookup(var(Name), PID, env(Vars,Procs,Chans), Value3) :-
    % set Name4
    %print(lookup(var(Name))),nl,
    ( Name = record([Name2|Tail]) -> true; Name2 = Name ),
    ( Name2 = array(Name3,expr(K)) -> true; Name3 = Name2 ),
    ( (PID \= global, lookup2(key(Name3,PID),Vars,_)) 
      -> Name4 = key(Name3,PID)
      ;  Name4 = Name3 ),
    lookup2(Name4, Vars, Value),
    % get Value3
    ( Name2 = array(_,_) 
      -> (eval(K,PID,env(Vars,Procs,Chans),K2), get_element(K2,Value,Value2))
      ;  Value2 = Value ),
    ( Name = record(_) 
      -> (lookup2(type(Name3),Vars,Type), get_record_element(Type,Tail,Value2,PID,env(Vars,_,_),Value3))
      ;  Value3 = Value2 ).

% lookup global proctype
lookup(proc(PID), global, env(_,Procs,_), PC) :-
    lookup3(PID, Procs, PC).

% lookup global channel
lookup(chan(Name), global, env(_,_,Chans), Value) :-
    lookup2(Name, Chans, Value).


%lookup2(Key,[],_) :- 
%    %ground(Key), 
%    print(lookup_var_not_found(Key)),nl,fail.
%lookup2(Key,[Key/Value|_T],Value).
%lookup2(Key,[Key2/_|T],Value) :-
%    Key \= Key2, lookup2(Key,T,Value).

% look up, given a key
lookup2(Key,AVL,Value) :-
	avl_fetch(Key,AVL,Value).

% look up, if there are possibly multiple matchings for key or value
lookup3(Key,AVL,Value) :-
    avl_member(Key,AVL,Value).


%remove(Key,[],_) :-
%    %print(remove_var_not_found(Key)),nl.
%    print(remove_var_not_found_error(Key)),nl,fail.
%remove(Key,[Key/_|T],T).
%remove(Key,[Key2/Value|T],[Key2/Value|T2]) :-
%    Key \= Key2, remove(Key,T,T2).

remove(Key,AVL,NewAVL) :-
    avl_delete(Key, AVL, _, NewAVL).

% remove local vars
remove_vars(PID, env(Vars,Procs,Chans), env(NewVars,Procs,Chans)) :-
    %findall(key(Name,PID), lookup2(key(Name,PID),Vars,_), List),
    findall(key(Name,PID), lookup3(key(Name,PID),Vars,_), List),
    remove_func(List,Vars,NewVars).

% remove pointers to channels and channels itself
remove_chans(PID, env(Vars,Procs,Chans), env(NewVars,Procs,NewChans)) :-
    findall(key(chan(Name),PID), lookup3(key(chan(Name),PID),Vars,_), List),
    remove_func(List,Vars,NewVars),
    findall(CID, lookup3(key(chan(_),PID),Vars,CID), List2),
    remove_channels(List2,NewVars,Chans,NewChans).

remove_children(PID2, env(Vars,Procs,Chans), env(NewVars,Procs,Chans)) :-
    findall(child(PID1, PID2), lookup3(child(PID1,PID2), Vars, true), List),
    remove_func(List,Vars,NewVars).

remove_proc(active(PID,Name), env(Vars,Procs,Chans), env(Vars,NewProcs,Chans)) :-
    remove(active(PID,Name), Procs, NewProcs).

remove_func([], Vars, Vars).
remove_func([H|T], Vars, NewVars) :-
    remove(H, Vars, Vars2),
    remove_func(T, Vars2, NewVars).

% remove global channels (only when no pointer is pointing to channel)
remove_channels([], _, Chans, Chans).
remove_channels([ID|Tail], Vars, Chans, NewChans) :-
    number(ID),
    ( (lookup3(chan(_),Vars,ID); lookup3(key(chan(_),_),Vars,ID))
      -> Chans2=Chans
      ;  remove(ID, Chans, Chans2) ),
    %print(remove_channel(ID)),nl,
    remove_channels(Tail, Vars, Chans2, NewChans).
% list of CID's (array)
remove_channels([[]|Tail], Vars, Chans, NewChans) :-
    remove_channels(Tail, Vars, Chans, NewChans).
% list of CID's (array)
remove_channels([[H|T]|Tail], Vars, Chans, NewChans) :-
    number(H),
    ( (lookup3(chan(_),Vars,H); lookup3(key(chan(_),_),Vars,H))
      -> Chans2=Chans
      ;  remove(H, Chans, Chans2) ),
    %print(remove_channels(H)),nl,
    remove_channels([T|Tail], Vars, Chans2, NewChans).

    
/************************************************************/

eval_list([], _, _, []).
eval_list([expr(H)|T], PID, In, [H2|T2]) :-
    eval(H, PID, In, H2),
    eval_list(T, PID, In, T2),!.

eval(cond(X,Y,Z), PID, In, R) :-
    eval(X, PID, In, X1),
    ( true(X1)
      -> eval(Y, PID, In, R)
      ;  eval(Z, PID, In, R) ).
eval(or(X,Y), PID, In, R) :-
    eval(X, PID, In, R2),
    ( true(R2)
      -> R=R2
      ;  eval(Y, PID, In, R) ).
eval(and(X,Y), PID, In, R) :-
    eval(X, PID, In, R2),
    ( true(R2)
      -> eval(Y, PID, In, R)
      ;  R=R2 ).
eval(eq(X,Y), PID, In, R) :-
    eval(X, PID, In, X1), eval(Y, PID, In, Y1),
    ( X1==Y1
      -> R=1
      ;  R=0 ).
eval(uneq(X,Y), PID, In, R) :-
    eval(X, PID, In, X1), eval(Y, PID, In, Y1),
    ( X1==Y1
      -> R=0
      ;  R=1 ).
eval(lt(X,Y), PID, In, R) :-
    eval(X, PID, In, X1), eval(Y, PID, In, Y1),
    ( X1<Y1 -> R=1; R=0 ).
eval(gt(X,Y), PID, In, R) :-
    eval(X, PID, In, X1), eval(Y, PID, In, Y1),
    ( X1>Y1 -> R=1; R=0 ).
eval(leqt(X,Y), PID, In, R) :-
    eval(X, PID, In, X1), eval(Y, PID, In, Y1),
    ( X1=<Y1 -> R=1; R=0 ).
eval(geqt(X,Y), PID, In, R) :-
    eval(X, PID, In, X1), eval(Y, PID, In, Y1),
    ( X1>=Y1 -> R=1; R=0 ).
eval('+'(X,Y), PID, In, R) :-
    eval(X, PID, In, X1), eval(Y, PID, In, Y1), R is X1+Y1.
eval('-'(X,Y), PID, In, R) :-
    eval(X, PID, In, X1), eval(Y, PID, In, Y1), R is X1-Y1.
eval('*'(X,Y), PID, In, R) :-
    eval(X, PID, In, X1), eval(Y, PID, In, Y1), R is X1*Y1.
eval('/'(X,Y), PID, In, R) :-
    eval(X, PID, In, X1), eval(Y, PID, In, Y1), R is X1/Y1.
eval('mod'(X,Y), PID, In, R) :-
    eval(X, PID, In, X1), eval(Y, PID, In, Y1), R is X1 mod Y1.
eval(not(X), PID, In, R) :-
    eval(X, PID, In, X1),
    ( true(X1) -> R=0; R=1 ).
% bit complement
eval(compl(X), PID, In, R) :-
    eval(X, PID, In, X1), 
	R is \ X1.
eval(bitor(X,Y), PID, In, R) :-
    eval(X, PID, In, X1), eval(Y, PID, In, Y1),
	R is X1 \/ Y1.
eval(bitxor(X,Y), PID, In, R) :-
    eval(X, PID, In, X1), eval(Y, PID, In, Y1),
	R is (X1 \/ Y1) /\ \(X1 /\ Y1).
eval(bitand(X,Y), PID, In, R) :-
    eval(X, PID, In, X1), eval(Y, PID, In, Y1),
	R is X1 /\ Y1.
eval(lshift(X,Y), PID, In, R) :-
    eval(X, PID, In, X1), eval(Y, PID, In, Y1), 
	R is X1 << Y1.
eval(rshift(X,Y), PID, In, R) :-
    eval(X, PID, In, X1), eval(Y, PID, In, Y1),
	R is X1 >> Y1.
eval((expr(X)), PID, In, R) :-
    eval(X, PID, In, R),!.
eval(true, _, _, 1) :- !.
eval(false, _, _, 0) :- !.
eval(skip, _, _, 1) :- !.
eval(X, _, _, X) :-
    (number(X); X=[_|_]),!.
eval(ctype(X), _, In, R) :-
    lookup(var(mtype), global, In, Value),
	get_element(K, Value, X),
	length(Value, N),
	R is N - K.
eval(string(X), _, _, R) :-
	atom_codes(X,R).
eval(timeout, _, In, R) :-
    ( find_exec(In) -> R=0; R=1 ),!.
eval(pid, PID, _, PID) :- !.
eval(nr_pr, _, env(_,Procs,_), R) :-
    %length(Procs,R),!.
    avl_size(Procs,R),!.
eval(Name, PID, In, R) :-
    % looks up singles, chan(Name), array(Name,K), record(Name).
    (atomic(Name); Name=chan(_); Name=array(_,_); Name=record(_)),
    lookup(var(Name), PID, In, R2),
	eval(R2, PID, In, R).
eval(nfull(chan(Name)), PID, In, R) :-
    eval(empty(chan(Name)), PID, In, R). % fixed by leuschel
eval(empty(chan(Name)), PID, In, R) :-
    lookup(var(chan(Name)), PID, In, CID),
    lookup(chan(CID), global, In, Value),
    ( Value = [[empty]|_] -> R=1; R=0 ).
eval(full(chan(Name)), PID, In, R) :-
    eval(nempty(chan(Name)), PID, In, R).
eval(nempty(chan(Name)), PID, In, R) :-
    lookup(var(chan(Name)), PID, In, CID),
    lookup(chan(CID), global, In, Value),
    ( Value = [[empty]|_] -> R=0; R=1 ).
eval(len(chan(Name)), PID, In, R) :-
    lookup(var(chan(Name)), PID, In, CID),
    lookup(chan(CID), global, In, Value),
    chan_length(Value,R).
    %print(chan_length(Value,R)),nl.
eval(expr_poll(Name,Msg), PID, In, R) :-
    ( trans3(poll(Name,Msg),PID,In,_) -> R=1; R=0 ).
eval(expr_random_poll(Name,Msg), PID, In, R) :-
    ( trans3(random_poll(Name,Msg),PID,In,_) -> R=1; R=0 ).

true(X) :-
    \+false(X).
false(X) :-
    X==0.


chan_length(empty,0).
chan_length([],0).
chan_length([[empty]|_],0).
chan_length([H|T],R) :-
    H \= [empty], chan_length(T,X), R is X+1.
    

get_record_element(Type, [array(EName,expr(K))], Value, PID, In, R) :-
    inst(_, typedef(Type,X,def(_,array(EName,_)))),
    get_element(X, Value, Value2),
    eval(K, PID, In, K2),
    get_element(K2, Value2, R),!.
get_record_element(Type, [EName], Value, _, _, R) :-
    inst(_, typedef(Type,X,def(_,EName))),
    get_element(X,Value,R),!.
get_record_element(Type, [array(EName,expr(K))|T], Value, PID, In, R) :-
    inst(_, typedef(Type,X,def(EType,array(EName,_)))),
    get_element(X, Value, Value2),
    eval(K, PID, In, K2),
    get_element(K2, Value2, V2),
    get_record_element(EType, T, V2, PID, In, R),!.
get_record_element(Type, [EName|T], Value, PID, In, R) :-
    inst(_, typedef(Type,X,def(EType,EName))),
    get_element(X, Value, V2),
    get_record_element(EType, T, V2, PID, In, R).


/************************************************************/

% e.g.: get_element(4,[1,2,3,4,5],5), means 4th element of L is 5.
get_element(R,List,E) :- get_element2(0,List,E,R).
get_element2(X,[E|_],E,X).
get_element2(X,[_|T],E,R) :- Y is X+1, get_element2(Y,T,E,R).

% e.g.: set_element(1,[1,2,3,4],8,[1,8,3,4])
set_element(0,[_|T1],V,[V|T1]).
set_element(X,[H1|T1],V,[H1|T2]) :- Y is X-1, set_element(Y,T1,V,T2).

% e.g.: set_all_elements([1,2,3,4],6,[6,6,6,6])
set_all_elements([],_,[]).
set_all_elements([_|T1],E,[E|T2]) :-
	set_all_elements(T1,E,T2).


% e.g.: remove_element(2,[1,2,3,4],[1,2,4])
remove_element(0,[_|T1],T1).
remove_element(X,[H1|T1],[H1|T2]) :- Y is X-1, remove_element(Y,T1,T2).

% e.g.: X=4 -> L=[0,0,0,0]
generate_zero(0,[]) :- !.
generate_zero(X,[0|R]) :- Y is X-1, generate_zero(Y,R).

generate_false(0,[]) :- !.
generate_false(X,[false|R]) :- Y is X-1, generate_false(Y,R).

generate_chan(0,_,[]) :- !.
generate_chan(X,Msg,[Msg|R]) :- Y is X-1, generate_chan(Y,Msg,R).

/************************************************************/

trans(io([Statement], proc(PID,Proc), Span), In, Out) :-
    lookup(proc(active(PID,Proc)), global, In, PC),
    inst(PC, S),
    lookup(var(sys(mode)), global, In, MValue),
    %print(try_execute(PC,S)),nl,
	check_mode(MValue, S),
	trans2(S, Statement, PC, PC2, PID, Span, In, In2),
    %print(executed(PC,S)),nl,
	check_jump(PC2, NewPC),
	%print(jump(PC2,NewPC)),nl,
    ( NewPC \== dead 
      -> store(In2, proc(active(PID,Proc),NewPC), global, Out)
      ;  remove_proc(active(PID,Proc), In2, Out) ).

check_mode(normal, _).
check_mode(p_atomic, p_atomic(_)).
check_mode(d_step, d_step(_)).

% executes igoto or ibreak when necessary.
check_jump(PC, NewPC) :-
	( inst(PC,S), check_jump2(S, PC2)
	  -> check_jump(PC2, NewPC)
	  ;  NewPC = PC ).

check_jump2(igoto(X), NewPC) :-
	( number(X) 
	  -> NewPC is X
	  ; inst(Y,label(X)), NewPC is Y ).
check_jump2(ibreak(X), NewPC) :-
	number(X),
	NewPC is X.
check_jump2(p_atomic(igoto(X)), NewPC) :-
	check_jump2(igoto(X), NewPC).
check_jump2(p_atomic(ibreak(X)), NewPC) :-
	check_jump2(ibreak(X), NewPC).
check_jump2(d_step(igoto(X)), NewPC) :-
	check_jump2(igoto(X), NewPC).
check_jump2(d_step(ibreak(X)), NewPC) :-
	check_jump2(ibreak(X), NewPC).


trans2(do(Alt), Statement, _, NewPC, PID, Span, In, Out) :-
    member(A,Alt),
    inst(A,S),
    ( (S=else; S=d_step(else); S=p_atomic(else)) 
	  -> delete(Alt,A,Alt2), 
	     ( find_exec(Alt2,PID,In) -> fail; true )
	  ;  true ),
    trans2(S,Statement,A,NewPC,PID,Span,In,Out).

trans2(if(Alt), Statement, _, NewPC, PID, Span, In, Out) :-
    trans2(do(Alt), Statement, _, NewPC, PID, Span, In, Out).

trans2(unless(M,E), Statement, _, NewPC, PID, Span, In, Out) :-
    inst(E,Escape),
    ( trans2(Escape,Statement,E,NewPC,PID,Span,In,Out)
      -> true
      ;  (inst(M,Main), trans2(Main,Statement,M,NewPC,PID,Span,In,Out)) ).

% provided (constraint on proctypes)
trans2(provided(expr(Expr)), Statement, PC, NewPC, PID, Span, In, Out) :-
    eval(Expr, PID, In, R),
    ( true(R)
      -> (PC2 is PC+1, inst(PC2,S), trans2(S,Statement,PC2,NewPC,PID,Span,In,Out))
      ;  fail ).

% Distinction between d_step_start and d_step is important for to not recognize
% a new starting d_step statement as a continuing d_step.
trans2(d_step_start(S), Statement, PC, NewPC, PID, Span, In, Out) :-
	trans2(d_step(S), Statement, PC, NewPC, PID, Span, In, Out).

trans2(p_atomic_start(S), Statement, PC, NewPC, PID, Span, In, Out) :-
	trans2(p_atomic(S), Statement, PC, NewPC, PID, Span, In, Out).

trans2(d_step(S), Statement, PC, NewPC, PID, Span, In, Out) :-
    trans2(S, Statement, PC, PC2, PID, Span, In, In2),
	check_jump(PC2, NewPC),
    inst(NewPC, S2),
    % look, if next step is d_step
    ( S2 = d_step(_)
      -> define(In2, var(sys(mode),d_step), global, Out)
      ;  define(In2, var(sys(mode),normal), global, Out) ),!.

trans2(p_atomic(S), Statement, PC, NewPC, PID, Span, In, Out) :-
    trans2(S, Statement, PC, PC2, PID, Span, In, In2),
	check_jump(PC2, NewPC),
    inst(NewPC, S2),
	%print(p_atomic_inst(NewPC,S2)),nl,
    % look, if next step is p_atomic and is executable.
    ( S2 = p_atomic(S3), trans2(S3, _, NewPC, _, PID, _, In2, _)
      -> define(In2, var(sys(mode),p_atomic), global, Out)%, print(define(p_atomic)),nl
      ;  define(In2, var(sys(mode),normal), global, Out) ).%, print(define(normal)),nl ).
    
% destructor
trans2(destructor, destructor, PC, NewPC, PID, src_span(SRow,SCol,ERow,ECol,_,_), In, Out) :-
    % ensure, PID has no child.
    In = env(Vars,_,_),
    \+ lookup3(child(PID,_), Vars, true),
    % remove channels that are not being referenced
    remove_chans(PID, In, In2),
    remove_vars(PID, In2, In3),
    remove_children(PID, In3, Out),
    NewPC = dead,
    line(PC,SRow,SCol,ERow,ECol),!.

trans2(goto(X), goto(X), PC, NewPC, _, src_span(SRow,SCol,ERow,ECol,_,_), In, In) :-
    ( number(X) ->  NewPC is X; inst(Y,label(X)), NewPC is Y ),
    inst(NewPC,S),
    % goto into a d_step is not allowed
    S \= d_step(_),
    line(PC,SRow,SCol,ERow,ECol).

trans2(break(X), break(X), PC, NewPC, _, src_span(SRow,SCol,ERow,ECol,_,_), In, In) :-
    NewPC is X,
    line(PC,SRow,SCol,ERow,ECol).

trans2(label(_), Statement, PC, NewPC, PID, Span, In, Out) :-
    PC2 is PC+1,
    inst(PC2,S),
    trans2(S, Statement, PC2, NewPC, PID, Span, In, Out).

trans2(printf(TextList), printf(Text), PC, NewPC, _, src_span(SRow,SCol,ERow,ECol,_,_), In, In) :-
    % ex.: atom_codes(X,[97]).
    %      X = a
    atom_codes(Text,TextList),
    NewPC is PC+1,
    line(PC,SRow,SCol,ERow,ECol).

trans2(printf(TextList,VarList), printf(Result), PC, NewPC, PID, src_span(SRow,SCol,ERow,ECol,_,_), In, In) :-
    % ex.: format("~s~2f",[a,0.3]).
    %      a0.30
	% format_to_codes("a ~c",[13],R).
	% R = [97,32,13] ? 
	%
    eval_list(VarList, PID, In, ValueList),
    atom_codes(Text,TextList),
    format_to_codes(Text,ValueList,ResultList),
    atom_codes(Result,ResultList),
    NewPC is PC+1,
    line(PC,SRow,SCol,ERow,ECol).

trans2(printm(expr(Expr)), printm(Result), PC, NewPC, PID, src_span(SRow,SCol,ERow,ECol,_,_), In, In) :-
    eval(Expr, PID, In, Result),
    NewPC is PC+1,
    line(PC,SRow,SCol,ERow,ECol).

trans2(Statement, Statement, PC, NewPC, PID, src_span(SRow,SCol,ERow,ECol,_,_), In, Out) :-
    trans3(Statement, PID, In, Out),
    NewPC is PC+1,
    line(PC,SRow,SCol,ERow,ECol).

trans2(send(Name,Msg), Statement, PC, NewPC, PID, SpanOut, In, Out) :-
    lookup(var(Name), PID, In, CID),
  %%  print(chan_id(CID)),nl,
    ( lookup(chan(CID), global, In, [])
      -> r_send(CID, Name, Msg, Statement, PID, Span2, In, Out), SpanOut = multi_span(SRow,SCol,ERow,ECol,_,_,Span2)
      ;  n_send(CID, Name, Msg, Statement, PID, In, Out), SpanOut = src_span(SRow,SCol,ERow,ECol,_,_) ),
    NewPC is PC+1,
    line(PC,SRow,SCol,ERow,ECol).

trans2(sorted_send(Name,Msg), Statement, PC, NewPC, PID, SpanOut, In, Out) :-
    lookup(var(Name), PID, In, CID),
   %% print(chan_id(CID)),nl,
    ( lookup(chan(CID), global, In, [])
      -> r_send(CID, Name, Msg, Statement, PID, Span2, In, Out), SpanOut = multi_span(SRow,SCol,ERow,ECol,_,_,Span2)
      ;  s_send(CID, Name, Msg, Statement, PID, In, Out), SpanOut = src_span(SRow,SCol,ERow,ECol,_,_) ),
    NewPC is PC+1,
    line(PC,SRow,SCol,ERow,ECol).

% normal send
n_send(CID, Name, Msg, Statement, PID, In, Out) :-
    lookup(chan(CID), global, In, Value),
    get_element(K, Value, [empty]), % Backtracking!
    %print(map_send(Msg)),nl,
    eval_list(Msg, PID, In, Msg2),
    %print(map2_send(Msg2)),nl,
    set_element(K, Value, Msg2, NewValue),
    store(In, chan(CID,NewValue), global, Out),
    Statement = send(Name,Msg),!.

% rendezvous-send
r_send(CID, Name, Msg, Statement, PID, Span, In, Out) :-
    eval_list(Msg, PID, In, Msg2),
    %print(msg3(Msg2)),nl,
    store(In, chan(CID,[Msg2]), global, In2),
    % search matching receive
    lookup(proc(active(RPID,Proc)), global, In, PC),
    inst(PC, S),
    % S can be recv(), igoto(), ibreak(), etc.
    trans2(S,recv(RName,RMsg), PC, NewPC, RPID, Span, In2, In3),
    lookup(chan(CID), global, In3, [[empty]]),
    % reset channel
    store(In3, chan(CID,[]), global, In4),
    store(In4, proc(active(RPID,Proc), NewPC), global, Out),
    Statement = '.'(r_send(Name,Msg),r_recv(RName,RMsg)).
    % Permit Backtracking in this clause!

% sorted send
s_send(CID, Name, Msg, Statement, PID, In, Out) :-
    lookup(chan(CID), global, In, Value),
    get_element(K, Value, [empty]), % Backtracking!
    %print(map_send(Msg)),nl,
    eval_list(Msg, PID, In, Msg2),
    %print(map2_send(Msg2)),nl,
    set_element(K, Value, Msg2, NewValue),
    %print(newValue(NewValue)),nl,
    samsort(NewValue, SortedValue),
    %print(sortedValue(SortedValue)),nl,
    store(In, chan(CID,SortedValue), global, Out),
    Statement = sorted_send(Name,Msg),!.

% normal receive
trans3(recv(Name,Msg), PID, In, Out) :-
    lookup(var(Name), PID, In, Id),
    lookup(chan(Id), global, In, Value),
    get_element(0, Value, Msg2),
    Msg2 \== [empty],
    %print(map_recv(Msg)),nl,
    recv_eval(Msg, PID, In, In2, Msg2),
    %print(map2_recv(Msg2)),nl,
    remove_element(0, Value, Value2),
    append(Value2, [[empty]], NewValue),
    store(In2, chan(Id,NewValue), global, Out).

% random receive
trans3(random_recv(Name,Msg), PID, In, Out) :-
    lookup(var(Name), PID, In, Id),
    lookup(chan(Id), global, In, Value),
    get_element(K, Value, Msg2), % Backtracking!
    Msg2 \== [empty],
    %print(map_recv(Msg)),nl,
    recv_eval(Msg, PID, In, In2, Msg2),
    %print(map2_recv(Msg2)),nl,
    remove_element(K, Value, Value2),
    append(Value2, [[empty]], NewValue),
    store(In2, chan(Id,NewValue), global, Out),!.

% normal poll
trans3(poll(Name,Msg), PID, In, Out) :-
    lookup(var(Name), PID, In, Id),
    lookup(chan(Id), global, In, Value),
    get_element(0, Value, Msg2),
    Msg2 \== [empty],
    %print(map_recv(Msg)),nl,
    recv_eval(Msg, PID, In, Out, Msg2).
    %print(map2_recv(Msg2)),nl.

% random poll
trans3(random_poll(Name,Msg), PID, In, Out) :-
    lookup(var(Name), PID, In, Id),
    lookup(chan(Id), global, In, Value),
    get_element(_, Value, Msg2), % Backtracking!
    Msg2 \== [empty],
    %print(map_recv(Msg)),nl,
    recv_eval(Msg, PID, In, Out, Msg2),!.
    %print(map2_recv(Msg2)),nl,!.

trans3(expr(run(Proc,ValueList)), PID, In, Out) :-
    inst(PC,proctype(_,Proc,PList)),
    PC2 is PC + 1,
    %-----------------%
    lookup(var(sys(pid)), global, In, PID2),
    store(In, proc(active(PID2,Proc),PC2), global, In2),
    set_local_vars(PList, ValueList, PID, PID2, In2, In3),
    PID3 is PID2+1,
    define(In3, var(sys(pid),PID3), global, In4),
    define(In4, var(child(PID,PID2),true), global, Out),!.

trans3(expr(Name), PID, In, In) :-
    eval(Name, PID, In, R),
    %print(evaluated_expr(PID,Name,R)),nl,
    ( false(R) -> fail; true ).
    
trans3(assert(expr(Name)), PID, In, Out) :-
    eval(Name, PID, In, R),
   %% print(evaluated_expr(Name,R)),nl,
    ( false(R)
      -> define(In, var(sys(unsafe),true), global, Out)
      ; In=Out ).
    
trans3(else, _, In, In).
trans3(skip, _, In, In).

% define locals
trans3(def(Type,Name), PID, In, Out) :-
    get_init_value(def(Type,Name), Value, In, In2),
    ( Name = array(Name2,_) -> true; Name2 = Name ),
    ( Type = chan
      -> define(In2, var(chan(Name2),Value), PID, Out)
      ;  define(In2, var(Name2,Value), PID, Out) ).

% assign rendezvous-channel
trans3(assign(array(Name,expr(K)),chan(0,_)), PID, In, Out) :-
    lookup(var(chan(Name)), PID, In, CIDList),
    eval(K, PID, In, K2),
    get_element(K2, CIDList, CID),
    store(In, chan(CID,[]), global, Out),!.
trans3(assign(Name,chan(0,_)), PID, In, Out) :-
    lookup(var(chan(Name)), PID, In, CID),
    assign_channel_value(CID, [], In, Out),!.
% assign channel
trans3(assign(array(Name,expr(K)),chan(N,_)), PID, In, Out) :-
    lookup(var(chan(Name)), PID, In, CIDList),
    eval(K, PID, In, K2),
    get_element(K2, CIDList, CID),
    generate_chan(N, [empty], Value),
    store(In, chan(CID,Value), global, Out),!.
trans3(assign(Name,chan(N,_)), PID, In, Out) :-
    lookup(var(chan(Name)), PID, In, CID),
    generate_chan(N, [empty], Value),
    assign_channel_value(CID, Value, In, Out),!.

% assign channel to an existing CID
trans3(assign(array(chan(Name),expr(K)),expr(chan(Name2))), PID, In, Out) :-
	lookup(var(chan(Name)), PID, In, CIDList),
	eval(K, PID, In, K2),
	lookup(var(chan(Name2)), PID, In, CID),
	set_element(K2, CIDList, CID, NewCIDList),
	define(In, var(chan(Name),NewCIDList), PID, Out),!.
trans3(assign(chan(Name),expr(chan(Name2))), PID, In, Out) :-
	lookup(var(chan(Name2)), PID, In, CID),
    define(In, var(chan(Name),CID), PID, Out),!.

trans3(assign(vt(Name,Type),Value), PID, In, Out) :-
    store(In, var(Name,Type,Value), PID, Out).

trans3(inc(vt(Name,Type)), PID, In, Out) :-
    lookup(var(Name), PID, In, Value),
    R is Value+1,
    store(In, var(Name,Type,expr(R)), PID, Out).

trans3(dec(vt(Name,Type)), PID, In, Out) :-
    lookup(var(Name), PID, In, Value),
    R is Value-1,
    store(In, var(Name,Type,expr(R)), PID, Out).

/**************************************************************/
append_span(SpanIn, SpanAdd, SpanOut) :-
    ( SpanAdd = src_span(SRow,SCol,ERow,ECol,_,_)
      -> SpanOut = multi_span(SRow,SCol,ERow,ECol,_,_,SpanIn)
      ;  true ),
    ( SpanAdd = multi_span(SRow,SCol,ERow,ECol,_,_,Span2) 
      -> append_span(multi_span(SRow,SCol,ERow,ECol,_,_,SpanIn), Span2, SpanOut)
      ; true ).

% assign Value to CID
assign_channel_value(CID, Value, In, Out) :-
    number(CID),
    store(In, chan(CID,Value), global, Out),!.
% assign Value to a list of CID's
assign_channel_value([], _, In, In).
assign_channel_value([H|Tail], Value, In, Out) :-
    store(In, chan(H,Value), global, In2),
    assign_channel_value(Tail,Value, In2, Out).

% set_local_vars(NameList, ValueList, PID, PID2, In, Out)
set_local_vars([], [], _, _, In, In).
% H2 = single-chan or array-chan
set_local_vars([def(chan,H1)|T1], [expr(H2)|T2], PID, PID2, In, Out) :-
    lookup(var(H2), PID, In, Value),
    define(In, var(chan(H1),Value), PID2, In2),
    set_local_vars(T1, T2, PID, PID2, In2, Out),!.
set_local_vars([def(Type,H1)|T1], [expr(H2)|T2], PID, PID2, In, Out) :-
    eval(H2,PID,In,V),
	check_type(H1,Type,V,V2,In),
    define(In, var(H1,V2), PID2, In2),
    set_local_vars(T1, T2, PID, PID2, In2, Out).


% recv_eval(NameList, PID, In, Out, ValueList)
recv_eval([], _, In, In, []).
recv_eval([ctype(A)|T], PID, In, Out, [R|T2]) :-
	check_type(recv(ctype(A)),mtype,R,ctype(A),In),
    recv_eval(T, PID, In, Out, T2),!.
recv_eval([true|T], PID, In, Out, [R|T2]) :-
	check_type(recv(true),bool,R,true,In),
	recv_eval(T, PID, In, Out, T2),!.
recv_eval([false|T], PID, In, Out, [R|T2]) :-
	check_type(recv(false),bool,R,false,In),
	recv_eval(T, PID, In, Out, T2),!.
recv_eval([skip|T], PID, In, Out, [skip|T2]) :-
	recv_eval(T, PID, In, Out, T2),!.
recv_eval([chan(A)|T], PID, In, Out, [CID|T2]) :-
    define(In, var(chan(A),CID), PID, In2),
    recv_eval(T, PID, In2, Out, T2),!.
recv_eval([A|T], PID, In, Out, [A|T2]) :-
    number(A),
    recv_eval(T, PID, In, Out, T2),!.
recv_eval([underscore|T], PID, In, Out, [_|T2]) :-
    recv_eval(T, PID, In, Out, T2),!.
recv_eval([eval(Expr)|T], PID, In, Out, [R|T2]) :-
    eval(Expr, PID, In, R),
  %%  print(evaluated_expr(Expr,R)),nl,
    recv_eval(T, PID, In, Out, T2),!.
recv_eval([vt(A,AType)|T], PID, In, Out, [R|T2]) :-
    atomic(A),
    store(In, var(A,AType,expr(R)), PID, In2),
    recv_eval(T, PID, In2, Out, T2).
    
% else check
find_exec(Alt, PID, In) :-
    member(A, Alt),
    inst(A,Statement),
    trans2(Statement, _, A, _, PID, _, In, _).

% timeout check
find_exec(In) :-
    lookup(proc(active(PID,_)), global, In, PC),
    inst(PC,Statement),
    Statement \= expr(timeout),
    trans2(Statement, _, PC, _, PID, _, In, _).
    %print(find_exec(PC,Statement)),nl.
    

delete([X|T],X,T).
delete([Y|T],X,[Y|DT]) :- delete(T,X,DT).

/***************************************************************/

promela_property(env(Vars,_,_), '='('OrderedPair'(PID,Name),Value)) :- %var(key(Name,PID),Value)) :- 
    lookup3(key(Name,PID), Vars, Value).
promela_property(env(Vars,_,_), var(Name,Value)) :- 
    lookup3(Name, Vars, Value), Name \= key(_,_).
promela_property(env(_,Procs,_), proc(active(PID,Name),PC)) :- 
    lookup3(active(PID,Name), Procs, PC).
promela_property(env(_,_,Chans), chan(Id,Value)) :- 
    Chans \= [], 
    lookup3(Id, Chans, Value).

promela_property(env(_,Procs,_), active_proctypes(N)) :- 
    avl_size(Procs,N).
promela_property(env(Vars,_,_), unsafe) :- 
    lookup2(sys(unsafe), Vars, Value),
	( Value == true
	  -> true
	  ;  fail ).

% debugging:
%promela_property(env(Vars,_,_), true) :-
%	avl_to_list(Vars,List),
%	print(vars(List)),nl.

%promela_property(env(_,Procs,_), true) :-
%	avl_to_list(Procs,List),
%	print(procs(List)),nl.

%promela_property(env(_,_,Chans), true) :-
%	avl_to_list(Chans,List),
%	print(chans(List)),nl.

%promela_property(env(Vars,Procs,Chans), 0) :-
%    print('Vars: '),portray_avl(Vars),nl,
%    print('Procs: '),portray_avl(Procs),nl,
%    print('Chans: '),portray_avl(Chans),nl,fail.
