% (c) 2009 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

% LTL model checking with promela

:- module(promela_tools,[parse_promela_strings/4,eval_promela/4]).

:- use_module(library(lists)).
:- use_module(library(file_systems)).
:- use_module(library(process)).

:- use_module(h_int, [eval/4]).
:- use_module(probsrc(tools),[host_platform/1]).
:- use_module(probsrc(error_manager)).

:- use_module(probsrc(module_information)).
:- module_info(group,promela).
:- module_info(description,'This module contains helper predicates for the Promela extension.').

eval_promela(Expr, PID, Env, Value) :- eval(Expr,PID,Env,Value).

parse_promela_strings(ExprStrings,PredStrings,[],PromPreds) :-
	( ExprStrings = [_|_] ->
	    throw(parse_error('Unsupported expressions in Promela mode!'))
	; true),
	write_string_to_file(MMachine,PredStrings),
	call_promela_parser(MMachine,PMachine),
	load_promela(PMachine,PromPreds),
	delete_file(PMachine).

write_string_to_file(Filename,Machine) :-
	open(temp('ltl_prob.mch'),write,S,[if_exists(generate_unique_name)]),
	stream_property(S, file_name(Filename)),
	print_string_list(Machine,S),
	close(S).

print_string([],_).
print_string([C|Rest],S) :- put_code(S,C),print_string(Rest,S).
print_string_list([],_).
print_string_list([H|T],S) :-
	% ascii-nr 44 is ','.
	append(H,[44],H2),
	print_string(H2,S),
	print_string_list(T,S).

load_promela(Filename,Term) :-
	open(Filename, read, S),
	read(S,Term),
	close(S).

call_promela_parser(MCH,PL) :-
	(call_promela_parser2(MCH,PL) -> Res=ok; Res=fail),
	delete_file(MCH),
	Res=ok.
call_promela_parser2(MCH,PL) :-
	absolute_file_name(prob_lib('.'), LibDir),
        atom_concat('-Djava.library.path=', LibDir, JLibpath),

	absolute_file_name(prob_lib('Promela.jar'),Jar),
	once(absolute_file_name(path(java),
				JavaCmd,
				[access(exist),extensions(['.exe','']),solutions(all),file_errors(fail)])),

	replace_windows_path_backslashes(JavaCmd,JavaCmdW),
	replace_windows_path_backslashes(Jar,JarW),
	replace_windows_path_backslashes(MCH,MCHW),

	process_create(JavaCmdW,
		       %['-jar',JarW,MCHW],
		       [JLibpath,'-jar',JarW,'-ltl',MCHW],
		       [process(Java),stdout(pipe(JStdout))]),
	
	read_all(JStdout,Text),
	process_wait(Java,JExit),
	(JExit=exit(0) -> true;
	    (name(T,Text),JExit=exit(1) -> 
		raise_exception(parse_error(T));
		add_error(ltl,'Error while parsing predicates',Text),fail)),
	atom_chars(MCH,FilenameM),append(Core,['.','m','c','h'],FilenameM),
	append(Core,['.','p','l'],FilenameP),atom_chars(PL,FilenameP).

/* replace backslashes by forward slashes in atoms, only under windows */
replace_windows_path_backslashes(Old,New) :-
	( host_platform(windows) ->
		name(Old,OldStr),
		replace_string_backslashes(OldStr,NewStr),
		name(New,NewStr)
	; otherwise ->
		Old=New).

% copy from parsercall:
replace_string_backslashes([],[]).
replace_string_backslashes([C|OldRest],[N|NewRest]) :-
        ( C=92 /* backslash */ ->
            N=47 /* forward slash */
        ; otherwise ->
            N=C ),
        replace_string_backslashes(OldRest,NewRest).

read_all(S,Text) :- read_line(S,Line),
	(Line=end_of_file ->  Text="",close(S);
	    append(Line,"\n",Line1),
	    append(Line1,RText,Text),
	    read_all(S,RText)).
