% (c) 2009-2014 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


:- module(state_custom_dot_graph,[tcltk_generate_state_custom_dot_graph/1]).

:- use_module(module_information).
:- module_info(group,dot).
:- module_info(description,'This module provides a way to generate custom state graphs (using info from DEFINITIONS).').

:- use_module(dot_graph_generator).

:- use_module(translate,[    print_bexpr/1 ]).

:- use_module(self_check).
:- use_module(error_manager).

:- use_module(state_space,[current_expression/2]).
:- use_module(bsyntaxtree, [get_texpr_id/2]).

tcltk_generate_state_custom_dot_graph(File) :-
       current_expression(_CurID,State),
       specfile:state_corresponds_to_initialised_b_machine(State,BState),!,
      ( bmachine:b_get_machine_custom_nodes_function(NodesFunction)
        -> true
        ; add_error_and_fail(tcltk_generate_state_custom_dot_graph,'No CUSTOM_GRAPH_NODES Function in DEFINITIONS')),
       (bmachine:b_get_machine_custom_edges_function(EdgesFunction)
        -> true
        ; add_error_and_fail(tcltk_generate_state_custom_dot_graph,'No CUSTOM_GRAPH_EDGES Function in DEFINITIONS')),
       eval_definition_fun(NodesFunction, BState, NodesRes),
       print(custom_nodes(NodesRes)),nl,
       retractall(custom_nodes(_)), assert(custom_nodes(NodesRes)),
       eval_definition_fun(EdgesFunction, BState, EdgesRes),
       retractall(custom_edges_default_label(_)),
       (get_texpr_id(EdgesFunction,EFID)
         -> assert(custom_edges_default_label(EFID)) ; assert(custom_edges_default_label('custom'))),
       print(custom_edges(EdgesRes)),nl,
       retractall(custom_edges(_)), assert(custom_edges(EdgesRes)),
       retractall(custom_node_id(_,_)), retractall(nodectr(_)), 
       retractall(custom_trans_label(_,_)), retractall(trans_ctr(_)), assert(trans_ctr(1)),
       gen_dot_graph(File,state_custom_dot_graph,node_predicate,trans_predicate,none,none),
       print(finished),nl.
tcltk_generate_state_custom_dot_graph(_) :-
       add_error_and_fail(tcltk_generate_state_custom_dot_graph,'Only possible for initialised B machine.').

:- dynamic custom_nodes/1, custom_edges/1, custom_edges_default_label/1.
custom_edges_default_label('custom').

node_predicate(NodeID,SubGraph,NodeDesc,Shape,Style,Color) :- SubGraph=none,
     Shape=box, Style=filled,
     custom_nodes(Nodes),
     (var(Nodes) -> add_error(node_predicate,'Variable custom_nodes: ',Nodes),fail ; true),
     get_col(NodeVal,Nodes,ColVal),
     (translate:translate_bvalue(NodeVal,NodeDesc) -> true ; NodeDesc = '???'),
     translate_colour(ColVal,Color),
     %tools:print_message(node(NodeVal,ColVal,Color,NodeID)),
     gen_id(NodeVal,NodeID). %tools:print_message(nodeid(NodeID)).

get_col(NodeVal,Nodes,ColVal) :- 
   if(member((NodeVal,ColVal),Nodes),true,
      (member(NodeVal,Nodes),ColVal = NodeVal)). % we have a set of nodes maybe
 
:- dynamic custom_node_id/2, nodectr/1.
nodectr(0).

get_ctr(Res) :- retract(nodectr(C)),!, C1 is C+1, assert(nodectr(C1)), Res=C.
get_ctr(0) :- assert(nodectr(1)).

gen_id(NodeVal,ID) :-
 (custom_node_id(NodeVal,ID) -> true
   ; get_ctr(C),assert(custom_node_id(NodeVal,C)),ID=C).

:- dynamic custom_trans_label/2, trans_ctr/1.
trans_ctr(1).
gen_trans(Label,Col) :-  custom_trans_label(Label,C),!,Col=C.
gen_trans(Label,Col) :- translate_colour2(Label,C),!,Col=C.
gen_trans(Label,Col) :-  retract(trans_ctr(Ctr)), C1 is Ctr+1, assert(trans_ctr(C1)),
   translate_colour(int(Ctr),C),
   assert(custom_trans_label(Label,C)), Col=C.

translate_colour(Val,Col) :- (translate_colour2(Val,C) -> Col=C; Col=lightgray).
translate_colour2(int(X),Res) :- !, (translate_int_col(X,Colour) -> Res=Colour ; Res=black).
translate_colour2(string(X),Colour) :- preferences:is_of_type(X,rgb_color),!, X=Colour.
translate_colour2(fd(X,GS),Colour) :- !, translate_fd_col(X,GS,Colour).

% TO DO : add string conversions, rgb values , ...

translate_fd_col(X,GS,Res) :-
   b_global_sets:is_b_global_constant(GS,X,Colour),
   preferences:is_of_type(Colour,rgb_color),!,
   Res=Colour.
translate_fd_col(X,_,Res) :- translate_int_col(X,Colour),!, Res=Colour.
translate_fd_col(_,_,lightgray). 
   
translate_int_col(0,white).
translate_int_col(1,blue).
translate_int_col(2,red).
translate_int_col(3,green).
translate_int_col(4,lightgray).
translate_int_col(5,orange).
translate_int_col(6,yellow).
translate_int_col(7,brown).
translate_int_col(8,violet).
translate_int_col(9,tomato).
translate_int_col(10,darkslateblue).
translate_int_col(11,maroon2).
translate_int_col(12,olivedrab2).
translate_int_col(13,chartreuse3).
translate_int_col(14,darkgray).
% red,green,blue,yellow,orange,black,white,gray,brown,violet,darkred,tomato,darkblue,
%             darkgray,darkviolet,darkslateblue,lightblue,lightgray,maroon2,olivedrab2,
%             steelblue,chartreuse3,chartreuse4

trans_predicate(NodeID,Label,SuccID,Color,Style) :-
   custom_edges(Edges),
     (var(Edges) -> add_error(trans_predicate,'Variable custom_edges: ',Edges),fail ; true),
   member((From,To),Edges),
   trans_pred_aux(From,To,NodeID,SuccID,Label,Color,Style).

% we have a transition where the colour is specified (e.g., CUSTOM_GRAPH_EDGES == {n1,col,n2 | ... }
trans_pred_aux((From,LabelVal),To,NodeID,SuccID,Label,Color,solid) :-
   translate:translate_bvalue(LabelVal,Label),
   gen_trans(LabelVal,Color),
   !,
   gen_id(From,NodeID),
   gen_id(To,SuccID).
trans_pred_aux(From,To,NodeID,SuccID,Label,blue,solid) :- 
   custom_edges_default_label(Label),
   gen_id(From,NodeID),
   gen_id(To,SuccID).

:- use_module(custom_explicit_sets,[try_expand_custom_set/2]).

eval_definition_fun(AnimFunction,BState, FunctionRes) :- %trace,
       b_interpreter:b_compute_expression(AnimFunction,[],BState,FunctionResCl),
       nl, print('FunctionResult'(FunctionResCl)),nl,
       ~~mnf( custom_explicit_sets:try_expand_custom_set(FunctionResCl, FunctionRes) ).
