% (c) 2009-2014 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(state_packing,[reset_stored_values/0,
                         precompile_state_packing/0, % to be called after b_global_sets precompiled
                         pack_state/2, unpack_state/2,
                         % pack_values/2, unpack_values/2,
                         pack_value/2, unpack_value/2,
                         print_stored_values/0, retract_stored_values_with_statistics/0]).

% a module to pack/compress states before asserting them in state_space
:- use_module(specfile,[b_or_z_mode/0, csp_mode/0]).
:- use_module(tools).
:- use_module(b_global_sets).

:- use_module(module_information).
:- module_info(group,animator).
:- module_info(description,'This module packs and upacks states and values.').

%pack_state(concrete_constants(S),R) :- b_or_z_mode,!, R=concrete_constants(PS), pack_state(S,PS). % leads to further reduction in memory consumption, but to a slowdown
pack_state(root,R) :- !, R=root.
pack_state(const_and_vars(ID,S),R) :- % b_or_z_mode, % here for efficiency we assume that we are in B mode; any plug-in should not use this state constructor !!
    !,
    R=const_and_vars(ID,PS),
   % debug:time(state_packing:pack_bind_list(S,PS)).
    pack_bind_list(S,PS).
pack_state([bind(V,Val)|T],R) :- % b_or_z_mode, % again, for effiency we do not check whether we are in B mode here
    !,
    R='$bind_var'(V,PV,PT),
    pack_value(Val,PV),
    pack_bind_list(T,PT).
pack_state(csp_and_b(CSP,B),csp_and_b(PCSP,PB)) :- !,
    pack_state(CSP,PCSP),
    pack_state(B,PB).
pack_state(CSP,R) :- preferences:preference(use_state_packing,true),csp_mode,!,
    pack_csp_expression(CSP,PCSP), R=PCSP.
pack_state(V,V).

:- use_module(error_manager,[add_internal_error/2]).
pack_bind_list([],[]) :- !.
pack_bind_list([bind(V,Val)|T],R) :- !,R='$bind_var'(V,PV,PT),
    pack_value(Val,PV),
    pack_bind_list(T,PT).
pack_bind_list(V,R) :- add_internal_error('Illegal value: ',pack_bind_list(V,R)),R=V.

:- use_module(library(avl)).
unpack_state(root,State) :- !, State=root.
unpack_state(concrete_constants(C),State) :- !, State=concrete_constants(C). % comment out in case concrete_constants are packed after all
unpack_state(PackedState,State) :- empty_avl(E), %tools_printing:print_term_summary(unpack(PackedState,State)),nl, % trace,
     unpack_state(PackedState,State,E,_).

%unpack_state(concrete_constants(S),R) --> {b_or_z_mode},!, {R=concrete_constants(UPS)}, unpack_state(S,UPS). % leads to further reduction
unpack_state(const_and_vars(ID,S),R) --> % {b_or_z_mode}, we assume const_and_vars is only used in B or Z mode or is compatible
   !,
   {R=const_and_vars(ID,UPS)}, unpack_state(S,UPS).
unpack_state('$bind_var'(Var,Val,T),R) --> % {b_or_z_mode}, we assume $bind_var not used by any plug-in
   !,
   {R=[bind(Var,UPV)|UPT]},
   unpack_value(Val,UPV),
   unpack_bind_list(T,UPT).
unpack_state(csp_and_b(CSP,B),csp_and_b(UPCSP,UPB)) --> !, unpack_state(CSP,UPCSP),
     unpack_state(B,UPB).
unpack_state(PCSP,R) -->
    {csp_mode,preferences:preference(use_state_packing,true)},!,
    {unpack_csp_expression(PCSP,R)}.
unpack_state(R,R) --> [].

unpack_bind_list([],[]) --> [].
unpack_bind_list('$bind_var'(Var,Val,T),R) -->  {R=[bind(Var,UPV)|UPT]},
   unpack_value(Val,UPV), %!,
   unpack_bind_list(T,UPT).
%unpack_bind_list('$bind_var'(Var,Val,_),_) --> {add_internal_error('Unpacking value failed: ',Var/Val),fail}.


/* currently commented out in store_transition in state_space.pl module
pack_values([],[]).
pack_values([H|T],[PH|PT]) :- 
  (fail %b_or_z_mode
   -> pack_value(H,PH), pack_values2(T,PT)
    ; H=PH,T=PT).
pack_values2([],[]).
pack_values2([H|T],[PH|PT]) :- pack_value(H,PH), pack_values2(T,PT).

unpack_values([],[]).
unpack_values([H|T],[PH|PT]) :- 
  (fail %b_or_z_mode
   -> unpack_value(H,PH,empty,_), unpack_values2(T,PT)
    ; H=PH,T=PT).
unpack_values2([],[]).
unpack_values2([H|T],[PH|PT]) :- unpack_value(H,PH,empty,_), unpack_values2(T,PT).
*/

:- use_module(library(avl)).
:- use_module(preferences,[preference/2]).
:- use_module(custom_explicit_sets,[is_interval_closure/5, construct_interval_closure/3]).
:- use_module(error_manager,[add_internal_error/2]).
pack_value(Var,R) :- var(Var),!,
  add_internal_error('Illegal variable value: ', pack_value(Var,R)),
   R='$variable'(Var).
pack_value(avl_set(A),R) :- !, %R\==[],
  avl_domain(A,AD),
  %avl_packed_domain(A,ADL,[]),
  (pack_fd_set(AD,Type,PackedInteger) -> R = '$avl_bv'(PackedInteger,Type) % avl bit vector %, print(packed(PackedInteger,AD)),nl
   ; l_pack_value(AD,ADL),
    (preferences:preference(use_state_packing,true) %ADL=[_,_|_]
       -> R='$stored_avl'(ID),store_value(ADL,ID)
         % , print(stored(ID,ADL)),nl
       ; R='$avl_exp'(ADL)) % ,print(packed(A,R)),nl
   ).
pack_value((A,B),(PA,PB)) :- !, pack_value(A,PA), pack_value(B,PB).
pack_value(fd(Nr,T),R) :- pack_basic_fd_value(T,Nr,Atom),!,Atom=R.
pack_value(int(Nr),R) :- !, R=Nr.
pack_value(closure(P,T,B),R) :-  custom_explicit_sets:is_interval_closure(P,T,B,Low,Up),!, 
  %print(pack_interval(Low,Up,B)),nl,
  R='$interval'(Low,Up).
%pack_value(closure(P,T,B),R) :- is_cartesian_product_closure(CPA,A1,A2),!,
%   R='$cartesian_product'(PA1,PA2),
%   pack_value(A1,PA1), pack_value(A2,PA2).
pack_value(V,V).
% to do: records + maybe closures (interval-closures,... ?)

/* not worth it : 
% compute avl_domain and apply pack_value in one go:
avl_packed_domain(empty, Domain, Domain).
avl_packed_domain(node(Key,_,_,L,R), Domain0, Domain) :-
    pack_value(Key,PKey),
    avl_packed_domain(L, Domain0, [PKey|Domain1]),
    avl_packed_domain(R, Domain1, Domain).
*/

l_pack_value([],[]).
l_pack_value([H|T],[PH|PT]) :- pack_value(H,PH), l_pack_value(T,PT).

% TO DO  or [(fd,fd) as bit-vectors if <xx bits

% Pack set of FD values as bit-vector representation
pack_fd_set([fd(X,Type)|T],Type,Result) :- 
   b_global_sets:b_fd_type(Type,1,UpBound),
   (UpBound < 28 -> true /* we are ok on 32-bit systems */
     ; UpBound < 60, platform_is_64_bit),  /* prolog_flag(max_tagged_integer,X), Y is 1<<59, Y<X. */
   Acc is 1<<X,
   pack_fd_set_aux(T,Acc,Result).

pack_fd_set_aux([],A,A).
pack_fd_set_aux([fd(X,_)|T],Acc,Res) :- 
   NewAcc is Acc \/ (1<<X),
   pack_fd_set_aux(T,NewAcc,Res).

% unpack a bit-vector set representation
unpack_fd_set(PackedInteger,Type,Value) :-
   unpack_fd_set_aux(PackedInteger,0,Type,ResList), %print(unpacked(PackedInteger,ResList)),nl,
   custom_explicit_sets:ord_list_to_avlset_direct(ResList,Value,unpack_fd_set).
   
unpack_fd_set_aux(0,_,_,Res) :- !, Res=[].
unpack_fd_set_aux(Nr,X,Type,Res) :-
   (Nr mod 2 =:= 1  ->  Res = [fd(X,Type)-true|T] ; Res=T),
   X1 is X+1, Nr1 is (Nr>>1),
   unpack_fd_set_aux(Nr1,X1,Type,T).

% first rough attempt at packing CSP expressions
pack_csp_expression(esharing(Set,B,C,Span),'$esharing'(PSet,PB,PC,PSpan)) :- !,
   pack_csp_expression(Set,PSet), pack_csp_expression(B,PB),
   pack_csp_expression(C,PC), pack_csp_expression(Span,PSpan).
pack_csp_expression('|||'(A,B,Span),'|||'(PA,PB,PSpan)) :- !,
   pack_csp_expression(A,PA), pack_csp_expression(B,PB), pack_csp_expression(Span,PSpan).
%pack_csp_expression('&'(A,B),'&'(PA,PB)) :- !,
%   pack_csp_expression(A,PA), pack_csp_expression(B,PB).
pack_csp_expression('[]'(A,B,Span),'[]'(PA,PB,PSpan)) :- !,
   pack_csp_expression(A,PA), pack_csp_expression(B,PB), pack_csp_expression(Span,PSpan).
pack_csp_expression(eaParallel(C,A,D,B,Span),eaParallel(PC,PA,PD,PB,PSpan)) :- !,
   pack_csp_expression(A,PA), pack_csp_expression(B,PB), pack_csp_expression(C,PC),
   pack_csp_expression(D,PD), pack_csp_expression(Span,PSpan).
%pack_csp_expression(prefix(C,A,D,B,Span),prefix(PC,PA,PD,PB,PSpan)) :- !,
%   pack_csp_expression(A,PA), pack_csp_expression(B,PB), pack_csp_expression(C,PC),
%   pack_csp_expression(D,PD), pack_csp_expression(Span,PSpan).
%pack_csp_expression(ifte(C,A,D,B,Span,S3),ifte(PC,PA,PD,PB,PSpan,S3)) :- !,
%   pack_csp_expression(A,PA), pack_csp_expression(B,PB), pack_csp_expression(C,PC),
%   pack_csp_expression(D,PD), pack_csp_expression(Span,PSpan).
%pack_csp_expression(V,'$stored_csp'(ID)) :- store_value(V,ID).
pack_csp_expression(V,ID) :- store_value(V,ID).

unpack_csp_expression('$esharing'(PSet,PB,PC,PSpan),esharing(Set,B,C,Span)) :- !,
   unpack_csp_expression(PSet,Set), unpack_csp_expression(PB,B),
   unpack_csp_expression(PC,C), unpack_csp_expression(PSpan,Span).
unpack_csp_expression('|||'(A,B,Span),'|||'(PA,PB,PSpan)) :- !,
   unpack_csp_expression(A,PA), unpack_csp_expression(B,PB), unpack_csp_expression(Span,PSpan).
%unpack_csp_expression('&'(A,B),'&'(PA,PB)) :- !,
%   unpack_csp_expression(A,PA), unpack_csp_expression(B,PB).
unpack_csp_expression('[]'(A,B,Span),'[]'(PA,PB,PSpan)) :- !,
   unpack_csp_expression(A,PA), unpack_csp_expression(B,PB), unpack_csp_expression(Span,PSpan).
unpack_csp_expression(eaParallel(C,A,D,B,Span),eaParallel(PC,PA,PD,PB,PSpan)) :- !,
   unpack_csp_expression(A,PA), unpack_csp_expression(B,PB), unpack_csp_expression(C,PC),
   unpack_csp_expression(D,PD), unpack_csp_expression(Span,PSpan).
%unpack_csp_expression(prefix(C,A,D,B,Span),prefix(PC,PA,PD,PB,PSpan)) :- !,
%   unpack_csp_expression(A,PA), unpack_csp_expression(B,PB), unpack_csp_expression(C,PC),
%   unpack_csp_expression(D,PD), unpack_csp_expression(Span,PSpan).
%unpack_csp_expression(ifte(C,A,D,B,Span,S3),ifte(PC,PA,PD,PB,PSpan,S3)) :- !,
%   unpack_csp_expression(A,PA), unpack_csp_expression(B,PB), unpack_csp_expression(C,PC),
%   unpack_csp_expression(D,PD), unpack_csp_expression(Span,PSpan).
%unpack_csp_expression('$stored_csp'(ID),V) :- !, stored_value(ID,V).
unpack_csp_expression(ID,V) :- !, stored_value(ID,V).
unpack_csp_expression(V,V).

unpack_value(Val,UPVal) :- unpack_value(Val,UPVal,empty,_).

unpack_value(V,R) --> {var(V)},!,
    {add_internal_error('Illegal variable value: ', unpack_value(V,R))}, {R=V}.
unpack_value('$avl_exp'(AD),R) --> !, unpack_avl_expanded(AD,R).
unpack_value('$stored_avl'(ID),R) --> !, 
     {(stored_value(ID,Val) -> true ; add_internal_error('No stored AVL: ',ID),fail)},
     unpack_avl_expanded(Val,R).
unpack_value('$avl_bv'(PackedInteger,Type),R) --> !, {unpack_fd_set(PackedInteger,Type,R)}.
/* comment in to try and re-construct sharing; does seem to increase memory usage for vital_gradient_values
 unpack_value('$stored_avl'(ID),R,InEnv,OutEnv) :- !, 
  (avl_fetch(ID,InEnv,ER) %(expanded(ID,ER),InEnv)
    -> R = ER, OutEnv=InEnv % reuse the already expanded value: will also decrease memory consumption due to sharing
     ; stored_value(ID,Val),
       avl_store(ID,InEnv,R,InEnv2), % OutEnv=[expanded(ID,R)|OutEnv2], % note: ID cannot occur inside Val
       unpack_avl_expanded(Val,R,InEnv2,OutEnv)).
*/
%unpack_value('$cartesian_product'(A,B),R) :- unpack_value(A,PA), unpack_value(B,PB),
%   construct_cartesian_product_closure(PA,PB,R).
unpack_value('$interval'(A,B),R) --> !, {custom_explicit_sets:construct_interval_closure(A,B,R)}.
unpack_value((A,B),(PA,PB)) --> !, unpack_value(A,PA), unpack_value(B,PB).
unpack_value(Nr,R) --> {number(Nr)},!, {R=int(Nr)}.
unpack_value(V,UPV) --> {unpack_basic_fd_value(V,FD)},!,{UPV=FD}.
unpack_value(V,V) --> [].

unpack_avl_expanded(AD,R) -->
  l_unpack_value(AD,ADT),
  {custom_explicit_sets:ord_list_to_avlset_direct(ADT,R,unpack_value)}.
   
l_unpack_value([],[]) --> [].
l_unpack_value([H|T],[PH-true|PT]) --> unpack_value(H,PH), l_unpack_value(T,PT).


:- dynamic pack_basic_fd_value/3, unpack_basic_fd_value/2.
% associate precompiled atoms with all enumerated & deferred set elements; TO DO: only do for "smaller" SETS
precompile_state_packing :- 
   retractall(pack_basic_fd_value(_,_,_)),
   retractall(unpack_basic_fd_value(_,_)),
   b_global_sets:enum_global_type(fd(Nr,T),_),
   string_concatenate('$fd_',T,FDT),
   string_concatenate(FDT,Nr,NewAtom),
   %debug:debug_println(9,pack(T,Nr,NewAtom)),
   assert(pack_basic_fd_value(T,Nr,NewAtom)),
   assert(unpack_basic_fd_value(NewAtom,fd(Nr,T))),fail.
precompile_state_packing.

% ------------------------

:- dynamic stored_value/2, stored_value_hash_to_id/2, next_value_id/1.
next_value_id(0).

reset_stored_values :- retractall(stored_value(_,_)),
   retractall(stored_value_hash_to_id(_,_)),
   retractall(next_value_id(_)), assert(next_value_id(0)).

:- use_module(library(terms),[term_hash/3]).
store_value(Value,ID) :-
    term_hash(Value,[range(smallint),algorithm(sdbm), depth(infinite),if_var(ignore)],Hash),
    %hashing:my_term_hash(Value,Hash),
    store_value_aux(Value,Hash,ID).

store_value_aux(Value,Hash,ID) :-
   stored_value_hash_to_id(Hash,ID),
   stored_value(ID,Value),!. % we have already stored the value
store_value_aux(Value,Hash,ID) :-
   retract(next_value_id(NID)), NID1 is NID+1,
   assert(next_value_id(NID1)),
   assert(stored_value(NID,Value)), assert(stored_value_hash_to_id(Hash,NID)),
   ID=NID.

% state_packing:print_stored_values
print_stored_values :-  % for saving state_space to file
  print_dynamic_pred(state_packing,stored_value,2),
  print_dynamic_pred(state_packing,stored_value_hash_to_id,2),
  print_dynamic_pred(state_packing,next_value_id,1).


retract_stored_values_with_statistics :-
    retract_with_statistics(state_packing,[stored_value(_,_),
                                           stored_value_hash_to_id(_,_)]),
    reset_stored_values.
