% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(state_space_open_nodes,[not_all_transitions_added/1,
          reset_open_ids/0,
          add_id_at_front/1,
          add_id_at_end/1,
          add_id_random/1,
          add_id_to_process/3, % only relevant in heuristic mode, does add_id_at_front
          add_id_with_weight/2,  % only relevant in heuristic mode, does add_id_at_front
          
          pop_id_from_front_direct/1,
          pop_id_from_end_direct/1,
          retract_open_node_direct/1,
          
          open_ids_empty/0,
          
          retract_open_ids_with_statistics/0,
          print_state_space_open_nodes/0]).

:- use_module(tools).
:- use_module(module_information).
:- module_info(group,animator).
:- module_info(description,'This module keeps track of the open (not yet visited) nodes for the animator/model checker.').

% Open Node Queue Interface
% The interface for adding new ids and popping them from the list of open nodes:

:- dynamic not_all_transitions_added/1.
%not_all_transitions_added(v_0).
:- dynamic not_all_z/1.  /* a reverse order (partial) list of the above */
%not_all_z(v_0).

open_ids_empty :- not_all_transitions_added(_),!,fail.
open_ids_empty.

reset_open_ids :- retractall(not_all_transitions_added(_)),
    retractall(not_all_z(_)).
retract_open_ids_with_statistics :-
     retract_with_statistics(state_space_open_nodes,[not_all_transitions_added(_),
                                                     not_all_z(_)]).

add_id_at_front(NodeID) :-
   asserta(not_all_transitions_added(NodeID)),
   assertz(not_all_z(NodeID)). /* same as not_all_transitions_added but in reverse order */
add_id_at_end(NodeID) :-
   assertz(not_all_transitions_added(NodeID)),
   assertz(not_all_z(NodeID)).
add_id_with_weight(NodeID,W) :- print(ignoring_weight_for(NodeID,W)),nl,
   add_id_at_front(NodeID).

add_id_random(NodeID) :- print('add_id_random only implemented in state_space_open_nodes_c'),nl,
   add_id_at_front(NodeID).

add_id_to_process(NodeID,_Hash,_From) :- % only relevant in heuristic mode
   add_id_at_front(NodeID).

pop_id_from_front_direct(NodeID) :-
   retract(not_all_transitions_added(NodeID)),!, %print(df(NodeID)),nl,
   (retract(not_all_z(NodeID)) -> true ; print('*** no not_all_z entry: '), print(NodeID),nl).
pop_id_from_end_direct(NodeID) :- % breadth-first step
   retract(not_all_z(NodeID)), !,/* take from end */  %print(bf(NodeID)),nl,
   (retract(not_all_transitions_added(NodeID)) -> true ; print('*** no not_all_transitions_added entry: '), print(NodeID),nl).
   

retract_open_node_direct(NodeID) :- %print_message(retract(NodeID)),
  retract(not_all_transitions_added(NodeID)),
  retractall(not_all_z(NodeID)).
 
not_all_z_saved(X) :- not_all_z(X). % to avoid name clashes; as consulting occurs in state_space module
not_all_transitions_added_saved(X) :- not_all_transitions_added(X).
print_state_space_open_nodes :-
  print_dynamic_pred(state_space_open_nodes,not_all_z_saved,1),
  print_dynamic_pred(state_space_open_nodes,not_all_transitions_added_saved,1).
  
assert_not_all_transitions_added(NodeID) :- assert(not_all_transitions_added(NodeID)).
assert_not_all_z(NodeID) :- assert(not_all_z(NodeID)).