% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(succeed_max,[reset_max_reached/0,
                       max_reached/0, max_reached/1,
                       succeed_max_call/2, succeed_max_call_id/3]).

:- use_module(tools).

:- use_module(module_information).
:- module_info(group,infrastructure).
:- module_info(description,'This module provides a meta call to retrieve a specified maximum number of solutions.').

:- dynamic max_reached/1.

max_reached :- max_reached(_),!.

assert_max_reached(Id) :- 
   (max_reached(Id) -> true ; assert(max_reached(Id))).
reset_max_reached :- retractall(max_reached(_)).

:- meta_predicate succeed_max_call(0,-).
:- meta_predicate succeed_max_call_id(-,0,-).

succeed_max_call(Module:Call,MaxNrOfSols) :-
    succeed_max_call_id('$unknown',Module:Call,MaxNrOfSols).

:- volatile sols_found/1.
:- dynamic sols_found/1.
/* A version which uses the counter of myheap; still need to ensure it compiles on all platforms */ 
/* :- use_module(extension('myheap/myheap')). %
succeed_max_call(C,Max) :- !, mycounter_set(Max),call(C),
  mycounter_dec(R),
  (R=<0 -> (!, assert_max_reached) ; true).
  */
succeed_max_call_id(Id,_Call,MaxNrOfSols) :- retractall(sols_found(_)),
   MaxNrOfSols==0,!,
   assert_max_reached(Id),fail.
succeed_max_call_id(Id,Call,MaxNrOfSols) :- %print_message(max_call(Call,MaxnrOfSols)), 
    call(Call),
    (retract(sols_found(X)) -> X1 is X+1 ; X1 = 1),
    assert(sols_found(X1)),
    % print_message(sol(X1)), %
    ((X1>=MaxNrOfSols) 
      -> (!, %print_message(max_sol_found(Call,X1,MaxNrOfSols)),
          assert_max_reached(Id)
         )
      ;  true).

% not used    
% succeed_max_reached(MaxNrOfSols) :- sols_found(SF), SF >= MaxNrOfSols.