% (c) 2014-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(mic_generation,[down/4,ctgDown/6]).

:- use_module(probsrc(module_information)).
:- module_info(group,symbolic_model_checker).
:- module_info(description,'Generation of minimal inductive (sub-)clauses.').

:- use_module(library(ordsets)).
:- use_module(library(lists)).

:- use_module(probsrc(bsyntaxtree), [conjunct_predicates/2,
                                     conjunction_to_list/2,
                                     create_negation/2,
                                     create_implication/3]).

:- use_module(symbolic_model_checker(solver_handling)).
:- use_module(symbolic_model_checker(predicate_handling)).

% down algorithm. see "better generalization in ic3" listing 1
down(Cube,I,Frames,CubeOut) :-
    down(Cube,[],I,Frames,CubeOut).

down([],Fixed,_,_,Fixed2) :- reverse(Fixed,Fixed2). % TODO: why do i need to reverse?!
down([C|Cs],Fixed,I,Frames,COut) :-
    append(Cs,Fixed,FC),
    conjunct_predicates(FC,Q),
    (down_aux(Q,I,Frames)
    -> down(Cs,Fixed,I,Frames,COut)
    ;  down(Cs,[C|Fixed],I,Frames,COut)).

down_aux(Q,_I,_Frames) :-
    % return false if I does not imply not q
    % i.e. if there is a solution for I & q
    initial_state(Q), !, fail.
down_aux(Q,I,Frames) :-
    % F_i & not q & T => not q'
    create_negation(Q,NegQ),
    prime_predicate(Q,PrimedQ),
    in_solver_on_level(I,Frames,IS),
    conjunct_predicates([NegQ,PrimedQ|IS],Pred),
    solve(Pred,Result),
    Result = contradiction_found(_), !.
down_aux(Q,I,Frames) :-
    % with F_i ^ not q state
    in_solver_on_level(I,Frames,IS),
    create_negation(Q,NegQ),
    conjunct_predicates([NegQ|IS],Pred),
    solve(Pred,Result),
    Result = solution(F_iQState), !,
    create_state_predicate(F_iQState,F_iQStatePred),
    common_literals(Q,F_iQStatePred,NewQ),
    down_aux(NewQ,I,Frames).

% the ctg down algorithm taken from "better generalization in IC3"
ctgDown(Cube,I,K,FramesIn,FramesOut,CubeOut) :-
    % set current depth to 1
    ctgDown(Cube,[],1,I,K,FramesIn,FramesOut,CubeOut).

ctgDown([],Fixed,_,_,_,Frames,Frames,Fixed2) :- reverse(Fixed,Fixed2). % TODO: why do i need to reverse?!
ctgDown([C|Cs],Fixed,CurDepth,I,K,FramesIn,FramesOut,CubeOut) :-
    append(Cs,Fixed,FC),
    conjunct_predicates(FC,Q),
    (ctgDown_aux(Q,CurDepth,I,K,FramesIn,FramesT)
    -> ctgDown(Cs,Fixed,CurDepth,I,K,FramesT,FramesOut,CubeOut)
    ;  ctgDown(Cs,[C|Fixed],CurDepth,I,K,FramesIn,FramesOut,CubeOut)).

ctgDown_aux(Q,CurDepth,I,K,FramesIn,FramesOut) :-
    % set current counter examples to generalization counter to 0
    ctgDown_aux(Q,CurDepth,0,I,K,FramesIn,FramesOut).

ctgDown_aux(Q,_Depth,_CTGs,_I,_K,Frames,Frames) :-
    % as in the simple down case
    % return false if I does not imply not q
    % i.e. if there is a solution for I & q
    initial_state(Q), !, fail.
ctgDown_aux(Q,_Depth,_CTGs,I,_K,Frames,Frames) :-
    % again this is equivalent to the simple case
    % return true if
    % F_i & not q & T => not q'
    create_negation(Q,NegQ),
    prime_predicate(Q,PrimedQ),
    in_solver_on_level(I,Frames,IS),
    conjunct_predicates([NegQ,PrimedQ|IS],Pred),
    solve(Pred,Result),
    Result = contradiction_found(_), !.
ctgDown_aux(_,Depth,_,_,_,Frames,Frames) :-
    % fail if Depth is too large (currently 2 - needs experiments)
    Depth > 2, !, fail.
ctgDown_aux(Q,Depth,CTGs,I,K,FramesIn,FramesOut) :-
    % with F_i ^ not q state
    in_solver_on_level(I,FramesIn,IS),
    create_negation(Q,NegQ),
    conjunct_predicates([NegQ|IS],Pred),
    solve(Pred,Result),
    Result = solution(F_iQState), !,
    create_state_predicate(F_iQState,F_iQStatePred),
    ctgDown_aux2(F_iQStatePred,Q,Depth,CTGs,I,K,FramesIn,FramesOut).

% the if / else in line 32ff of "better generalization in IC3"
ctgDown_aux2(S,Q,Depth,CTGs,I,K,FramesIn,FramesOut) :-
    % max ctgs currently 3 - needs experiments
    CTGs < 3, I > 0,

    % Init => not S
    get_initial_state_predicate(Init),
    conjunct_predicates([Init,S],IAndS),
    solve(IAndS,ResultIAndS),
    ResultIAndS = contradiction_found(_), !,

    % F_i-1 & not S & T => not s'
    create_negation(S,NegS),
    prime_predicate(S,SPrimed),
    IMinus1 is I - 1,
    in_solver_on_level(IMinus1,FramesIn,F_I),
    conjunct_predicates([NegS,SPrimed|F_I],Consecution),
    solve(Consecution,ResultConsecution),
    ResultConsecution = contradiction_found(_), !,

    % we found a new counterexample to generalization
    NCTGs is CTGs + 1,
    % TODO: for loop to find J
    create_negation(SPrimed,NegSPrimed),
    ctgDown_aux_find_level(I,K,NegS,NegSPrimed,FramesIn,J),
    JMinus1 is J - 1,
    Depth2 is Depth + 1,
    conjunction_to_list(S,SList),
    ctgDown(SList,[],Depth2,JMinus1,K,FramesIn,FramesT,SOut),
    % add modified not s to frames on level J and recur (while loop in paper)
    % instead of adding not s, we add the former conjuncts and use them as disjuncts
    add_clauses_to_level(J,FramesT,SOut,FramesTT),
    ctgDown_aux(Q,Depth,NCTGs,I,FramesTT,FramesOut).

ctgDown_aux2(S,Q,Depth,_CTGs,I,FramesIn,FramesOut) :-
    % the else branch
    common_literals(Q,S,NewQ),
    ctgDown_aux(NewQ,Depth,I,FramesIn,FramesOut).

ctgDown_aux_find_level(K,K,_,_,_,K).
ctgDown_aux_find_level(I,_K,NegS,NegSPrimed,Frames,J) :-
    in_solver_on_level(I,Frames,F_I),
    conjunct_predicates([NegS|F_I],LHS),
    create_implication(LHS,NegSPrimed,Implies),
    solve(Implies,Result),
    Result = contradiction_found(_), !,
    J = I.
ctgDown_aux_find_level(I,K,NegSPrimed,SPrimed,Frames,J) :-
    I2 is I + 1,
    ctgDown_aux_find_level(I2,K,NegSPrimed,SPrimed,Frames,J).

common_literals(Q1,Q2,CQ) :-
    conjunction_to_list(Q1,Q1L),
    conjunction_to_list(Q2,Q2L),
    list_to_ord_set(Q1L,Q1S),
    list_to_ord_set(Q2L,Q2S),
    ord_intersection(Q1S,Q2S,CQS),
    conjunct_predicates(CQS,CQ).
