% (c) 2014-2014 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(table_tools, [expand_and_translate_to_table/3, print_table/2]).

:- use_module(translate).
:- use_module(tools,[ajoin/2]).
:- use_module(library(lists)).

expand_and_translate_to_table(Value,Header,Table) :-  
    expand(Value, ExpVal),
    translate_to_table(ExpVal,1,Table,Header). %, print(table(Res)),nl.

:- use_module(custom_explicit_sets,[is_custom_explicit_set/1]).
expand((A,B),[(A,B)]) :- !. % wrap into a set
expand(Value,ExpVal) :- is_custom_explicit_set(Value),!,
      custom_explicit_sets:try_expand_custom_set(Value, ExpVal).
expand([H|T],V) :- !, V=[H|T].
expand([],V) :- !, V=[].
expand(Val,[Val]).


translate_to_table([],_RowNr,[],['Nr','Elements']) :- !.
translate_to_table([A|T],RowNr,[list([RowNr|LA])|TT],['Nr'|Header]) :- !,
    get_table_entry(A,LA,Header),
    R1 is RowNr+1,
    translate_to_table(T,R1,TT,_).
translate_to_table(A,RowNr,[list([RowNr|AT])],['Nr','Value']) :- !,
    translate:translate:translate_bvalue(A,AT).

get_table_entry(A,[AT],['Elements']) :- A \= (_,_),!,
    translate:translate:translate_bvalue(A,AT).
get_table_entry(A,AT,HA) :- %A=(A1,A2), (A1 = (_,_) ; A2=(_,_)),
    !,
    get_table_entry2(A,[],AT,HA).
%get_table_entry((A,B),[AT,BT],['prj1','prj2']) :- !,
%    translate:translate:translate_bvalue(A,AT),
%    translate:translate:translate_bvalue(B,BT).
    
get_table_entry2((A,B),Path,ABT,HAB) :- !,
    get_table_entry2(A,[1|Path],AT,HA),
    get_table_entry2(B,[2|Path],BT,HB),
    append(AT,BT,ABT), append(HA,HB,HAB).
%get_table_entry2(rec([field(Name,Val)|Rest]),Path,ABT,HAB) :- !,
get_table_entry2(A,Path,[AT],[Header]) :- !,
    reverse(Path,RP), ajoin(['prj'|RP],Header), % to do: do all this more efficiently
    translate:translate:translate_bvalue(A,AT).
    

print_table(Header,Table) :-
   maplist(print_entry,Header),
   nl,
   maplist(print_row,Table).


print_row(list(Row)) :- maplist(print_entry,Row),nl.
% TO DO: determine column length dynamically
print_entry(E) :- print(E), my_atom_length(E,Len), WS is 10-Len, print_ws(WS).
print_ws(X) :- X>1, !, print(' '), X1 is X-1, print_ws(X1).
print_ws(_) :- print(' ').

my_atom_length(AtomOrNumber,Len) :- atom(AtomOrNumber),!, atom_length(AtomOrNumber,Len).
my_atom_length(A,Len) :- name(A,List), length(List,Len).