% (c) 2009-2014 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(tools_printing,[
          print_term_summary/1, print_term_summary_hash/1,
          print_term_summary_wo_nl/1,
          print_functor/1, 
          watch/3,
          trace_unify/2, trace_print/1, trace_print/2]).


:- use_module(module_information).
:- module_info(group,infrastructure).
:- module_info(description,'This module contains printing/debug helper predicates.').

:- use_module(library(terms),[term_hash/2]).
print_term_summary_hash(T) :- print_term_summary(T),
   term_hash(T,H),  %% (H=179896960 -> trace ; true), %%
   format('Hash = ~w~n',H).

print_term_summary(T) :- print_term_summary_wo_nl(T),nl.

print_term_summary_wo_nl(N) :- var(N),!,print('VARIABLE: '),print(N).
print_term_summary_wo_nl((A,B)) :- print('('), print_term_summary_wo_nl(A),
    print(','), print_term_summary_wo_nl(B),print(')').
print_term_summary_wo_nl(clpfd:Call) :- !, print(clpfd:Call).
print_term_summary_wo_nl(Module:Call) :- !,
    print(Module), print(':'),print_term_summary_wo_nl(Call).
print_term_summary_wo_nl(bind(Var,Val)) :- !,print_arg(bind(Var,Val)).
print_term_summary_wo_nl(N) :- functor(N,F,Arity), print(F), print('/'),print(Arity), 
                     N=..[F|Args], l_print_arg(Args).

l_print_arg(V) :- var(V),!, print('...').
l_print_arg([]).
l_print_arg([H|T]) :- print(': '), print_arg(H), l_print_arg(T).



:- use_module(library(avl)).
% USE MODULE NOT PERFORMED TO AVOID LOADING translate and all dependent modules
%:- use_module(translate,[print_bexpr_with_limit/2, print_bstate/1]).

print_arg(N) :- var(N),!,print('VARIABLE: '),print(N).
print_arg(int(N)) :- !, (var(N) -> print_var_integer(N) ; print(int(N))).
print_arg(float(N)) :- !, print(float(N)).
print_arg(fd(N,GS)) :- !, 
  ((number(N),b_global_sets:is_b_global_constant(GS,N,Res)) -> print(Res)
    ; print('fd('),print_integer(N),format(',~w)',GS)).
print_arg(rec(Fields)) :- !, print('rec('), print_arg(Fields),print(')').
print_arg(bind(Var,Val)) :- atomic(Var), !, print(Var),print('/'),print_arg(Val).
print_arg(field(Var,Val)) :- atomic(Var), !, print('field('),print(Var),print('/'),print_arg(Val),print(')').
print_arg(typedvalc(Val,Type,VarID,EnumWarning,Card)) :- 
   atomic(VarID), !, format('typedvalc for ~w :(',[VarID]),print_arg(Val), 
   format(',~w,~w,~w) ',[Type,EnumWarning,Card]).
print_arg(wfx(A,B,C)) :- !, print(wfx(A,B,C)).
print_arg(Module:Call) :- atomic(Module),!,
    print(Module), print(':'),print_arg(Call).
print_arg(A) :- atomic(A),!, print(A).
print_arg(A) :- A=node(_,_,_,_,_),ground(A), !, print('AVLnode.size='),avl_size(A,Sz), print(Sz).
print_arg(avl_set(A)) :- ground(A), !,
  (custom_explicit_sets:singleton_set(avl_set(A),El) -> print('AVL('),print_arg(El),print(')')
   ;  print('AVL.size='),avl_size(A,Sz), print(Sz)).
print_arg((A,B)) :- !, print('('),print_arg(A), print(','), print_arg(B), print(')').
print_arg(string(A)) :- !, print('string('),print_arg(A), print(')').
print_arg([]) :- !, print('[]').
print_arg([H|T]) :- T==[], !, print('['),print_arg(H),print(']').
print_arg([H|T]) :- nonvar(T),T=[H2|T2],T2==[], !,
     print('['),print_arg(H),print(','),print_arg(H2),print(']').
print_arg([H|T]) :- nonvar(H), H=bind(_,_),!, get_list_up_to([H|T],10,S,ALL),
   translate:print_bstate(S),
   (ALL=all -> true ; print(', ... ')).
print_arg([H|T]) :- list_skeleton_size([H|T],Sz,Type), !,
   print(Type), 
   ((Type=closed,ground([H|T])) -> print('&ground') ; true),
   print(' LIST.size='), print(Sz),
   (Sz<6 -> print(' ['), l_print_arg([H|T]), print(']') ; true).
print_arg(closure(P,T,B)) :- !, print('CLOSURE '), 
  (custom_explicit_sets:is_interval_closure(P,T,B,Low,Up) -> print([Low:Up])
    ; print(P)). %, print(' | '), print(_B).
print_arg(mnf(Call)) :- !, print('mnf('),print_term_summary_wo_nl(Call), print(')').
print_arg(pp_mnf(Call)) :- !, print('mnf('),print_term_summary_wo_nl(Call), print(')').
print_arg(pp_cll(Call)) :- !, print('mnf('),print_term_summary_wo_nl(Call), print(')').
print_arg(b(Expr,T,I)) :- !, 
       (T==pred -> print('PRED(') ; print('BEXPR(')),
       print_functor(Expr), print(','),
       translate:print_bexpr_with_limit(b(Expr,T,I),200), print(')').
print_arg(N) :- print_functor(N).

:- use_module(library(clpfd),[fd_dom/2]).
print_var_integer(X) :- print('int(?:'),fd_dom(X,Dom),print(Dom),
    %print(':('),frozen(X,Goal),print(Goal),print(')')
    print(')').

print_integer(I) :- (var(I) -> fd_dom(I,Dom), print('?:'),print(Dom) ; print(I)).
 
get_list_up_to([],_,[],all).
get_list_up_to([H|T],N,R,ALL) :-
  (N<1 -> R=[],ALL=no ; N1 is N-1, R=[H|TR], get_list_up_to(T,N1,TR,ALL)).

list_skeleton_size(X,Sz,Type) :- var(X),!,Sz=0,Type=open.
list_skeleton_size([],0,closed).
list_skeleton_size([_|T],N,Type) :- list_skeleton_size(T,NT,Type), N is NT+1.


print_functor(N) :- var(N),!,print('VARIABLE: '),print(N).
print_functor(N) :- functor(N,F,Arity), print(F), print('/'),print(Arity).

% ---------------

:- block watch(-,?,?).
watch([],N,Pos) :- !,print_data([],N,Pos).
watch([H|T],N,Pos) :- !,print_data(H,N,Pos), P1 is Pos+1, watch(H,N,Pos), watch(T,N,P1).
watch((P1,P2),N,Pos) :- print_data((P1,P2),N,Pos), watch(P2,N,Pos).
watch(Other,N,Pos) :- print_data(Other,N,Pos).
print_data(Data,N,Pos) :- print(N), print(' @ '), print(Pos),print(' : '), print(Data),nl.

% ---------------


trace_unify(A,B) :- trace_unify_aux(A,B,0).
trace_unify_aux(A,B,Lvl) :- (var(A);var(B)),!,indent(Lvl),print(A),print(' <-> '), print(B),
   if(A=B,true,(print(' FAILS'),nl,fail)),nl.
trace_unify_aux(A,B,Lvl) :- functor(A,FA,FAN), functor(B,FB,FBN),
   indent(Lvl),print(FA/FAN),
   ((FA,FAN)=(FB,FBN)
     -> A=..[_|As], B=..[_|Bs], 
        (As=[] -> nl
         ; print('('),nl,l_trace_unify(As,Bs,Lvl))
     ;  print(' FUNCTOR MISMATCH : '), print(FA/FAN), print(' <-> '), print(FB/FBN),nl,fail
   ).
indent(0) :- !. indent(X) :- X>0, print('+-'), X1 is X-1, indent(X1).

l_trace_unify([],[],L) :- indent(L), print(')'),nl.
l_trace_unify([A|TA],[B|TB],Lvl) :- L1 is Lvl+1, trace_unify_aux(A,B,L1),
   l_trace_unify(TA,TB,Lvl).

% ---------------

trace_print(A) :- trace_print(A,0,4).
trace_print(A,Max) :- trace_print(A,0,Max).

:- use_module(library(lists),[maplist/2]).
trace_print(A,Lvl,MaxLvl) :- Lvl >= MaxLvl,!, indent(Lvl), print(A),nl.
trace_print(A,Lvl,_MaxLvl) :- treat_as_atomic(A),!, indent(Lvl), print(A),nl.
trace_print([H|T],Lvl,MaxLvl) :- !, indent(Lvl), print('['),nl,
    l_trace_print([H|T],Lvl,MaxLvl),
    indent(Lvl), print(']'),nl.
%trace_print(A,Lvl,_MaxLvl) :- functor(A,_,1), arg(1,A,AN), atomic(AN),!, indent(Lvl), print(A),nl.
trace_print(A,Lvl,_MaxLvl) :- A=..[_|Args], maplist(atomic,Args),!,
    indent(Lvl), print(A),nl.
trace_print(A,Lvl,MaxLvl) :- functor(A,FA,FAN),
   indent(Lvl),print(FA/FAN),print('('),nl,
   A=..[_|As], l_trace_print(As,Lvl,MaxLvl).

treat_as_atomic(A) :- var(A).
treat_as_atomic(A) :- atomic(A).
treat_as_atomic(nodeid(pos(_,_,_,_,_,_))).
treat_as_atomic(sharing(_,_,_,_)).

l_trace_print([],L,_MaxLvl) :- indent(L), print(')'),nl.
l_trace_print([A|TA],Lvl,MaxLvl) :- L1 is Lvl+1, trace_print(A,L1,MaxLvl),
   l_trace_print(TA,Lvl,MaxLvl).

