% (c) 2009-2014 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(typing_tools,[
     is_infinite_type/1, % check if a type is infinite
     valid_ground_type/1, % check if a type is a valid ground type
     contains_infinite_type/1
]).

:- use_module(bsyntaxtree, [get_texpr_types/2,
                            syntaxtraversion/6,
                            find_typed_identifier_uses/3]).
:- use_module(library(lists), [some/2]).

% check if a type is infinite for ProB animation; deferred sets are not counted as infinite here!
is_infinite_type(integer).
is_infinite_type(string).
is_infinite_type(freetype(_)).
is_infinite_type(set(X)) :- is_infinite_type(X).
is_infinite_type(seq(_)).
is_infinite_type(couple(X,Y)) :- (is_infinite_type(X) -> true ; is_infinite_type(Y)).
is_infinite_type(record(Fields)) :- 
   ((member(field(_,Type),Fields), is_infinite_type(Type)) -> true).

contains_infinite_type([H|T]) :- 
  (is_infinite_type(H) -> true ; contains_infinite_type(T)).

:- use_module(error_manager).

valid_ground_type(X) :- var(X),!, add_error(valid_ground_type,'Variable as type: ',X),fail.
valid_ground_type(X) :- valid_ground_type_aux(X),!.
valid_ground_type(X) :- add_error(valid_ground_type,'Illegal type term: ',X),fail.

%:- use_module(b_global_sets,[b_global_set/1]).

valid_ground_type_aux(any) :- format(user_error,'! Type contains any type~n',[]).
valid_ground_type_aux(integer).
valid_ground_type_aux(string).
valid_ground_type_aux(boolean).
valid_ground_type_aux(global(G)) :- atom(G).
 %bmachine:b_get_machine_set(G).
 %b_global_sets:b_global_set(G).
valid_ground_type_aux(freetype(FT)) :- ground(FT).
valid_ground_type_aux(set(X)) :- valid_ground_type(X).
valid_ground_type_aux(seq(X)) :- valid_ground_type(X).
valid_ground_type_aux(couple(X,Y)) :- valid_ground_type(X), valid_ground_type(Y).
valid_ground_type_aux(record(Fields)) :- valid_fields(Fields).

valid_fields([]).
valid_fields([field(Name,_)|_]) :- \+ atom(Name),!,
   add_error(valid_ground_type,'Illegal record field name: ',Name),fail.
valid_fields([field(_Name,Type)|T]) :- 
   valid_ground_type(Type),
   valid_fields(T).